/*
 * Copyright (c) 2004-2020 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <dlfcn.h>
#include <stdarg.h>
#include <unordered_set>

#include <infiniband/ibdiag/ibdiag.h>
#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h"
#include "infiniband/ibdm/cable/CombinedCable.h"


/****************************************************/
#include <infiniband/ibdiag/ibdiag_version.h>

string get_ibdiag_version()
{
    string version = "\"";
    #ifdef IBDIAG_VERSION
        version += IBDIAG_VERSION;
    #else
        version += "unknown";
    #endif

    //add dependency libraries
    version += "\",\"";
    version += get_ibdm_version();
    version += "\",\"";
    version += get_ibis_version();
    version += "\"";
    return version;
}

u_int8_t get_operational_vl_num(u_int8_t opvl_code)
{
    if ((opvl_code < 1) | (opvl_code > 5))
        return 0;
    if (opvl_code == 5)
        return 15;
    u_int8_t operational_vl = 1;
    opvl_code--;
    operational_vl = (u_int8_t)(operational_vl << opvl_code);
    return operational_vl;
}
/****************************************************/
/******************private methods*******************/
/****************************************************/
void IBDiag::SetLastError(const char *fmt, ...)
{
    IBDIAG_ENTER;
    char buffer[IBDIAG_ERR_SIZE] = {0};
    va_list args;

    va_start(args, fmt);
    vsnprintf(buffer, sizeof(buffer), fmt, args);
    va_end(args);
    this->last_error.assign(buffer);
    IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "-E- %s\n", this->last_error.c_str());
    IBDIAG_RETURN_VOID;
}


void IBDiag::CleanUpInternalDB()
{
    IBDIAG_ENTER;

    std::unordered_set<direct_route_t*> DRs;

    //clean bad direct routes
    for (auto x : this->bad_direct_routes) {
        DRs.insert(x->direct_route);
        delete x;
    }

    //clean loop direct routes
    for (auto x : this->loop_direct_routes)
        DRs.insert(x);

    //clean bfs list
    for (auto x : this->bfs_list)
        DRs.insert(x);

    //clean known node direct routes, because good nodes was contain only parts of routes other can be in bad path
    for (auto & l : this->bfs_known_node_guids)
        for (auto x : l.second)
            DRs.insert(x);

    //release dr objects
    for (auto x : DRs)
        delete x;

    this->ibdiag_discovery_status = DISCOVERY_NOT_DONE;
    this->root_node = NULL;
    this->root_port_num = 0;
    this->fabric_extended_info.CleanUpInternalDB();

    //make sure all lists + maps are empty
    this->bfs_list.clear();
    this->good_direct_routes.clear();
    this->bad_direct_routes.clear();
    this->loop_direct_routes.clear();
    this->duplicated_guids_detection_errs.clear();
    this->bfs_known_node_guids.clear();
    this->bfs_known_port_guids.clear();
    this->errors.clear();

    IBDIAG_RETURN_VOID;
}

void IBDiag::PrintFileTimestamp(const string &file_path, const string &file_type)
{
    IBDIAG_ENTER;

    char timestamp_buf[100] = {0};
    if (!IBFabric::GetFileTimestamp(timestamp_buf, sizeof(timestamp_buf), file_path))
         INFO_PRINT("Load %s file: %s, Timestamp: %s\n", file_type.c_str(), file_path.c_str(),
                    timestamp_buf);
    else
         WARN_PRINT("Load %s file: %s - file is not accessible\n", file_type.c_str(),
                    file_path.c_str());

    IBDIAG_RETURN_VOID;
}

//Returns non zero on error
int IBDiag::OpenFile(const string& name, const OutputControl::Identity& identity, ofstream& sout, bool to_append, bool add_header)
{
    IBDIAG_ENTER;

    int                       rc = 0;
    string                    message;
    OutputControl::Properties properties(identity);
    bool                      isSkipped = false;


    if(properties.is_valid()) {
        if(properties.enabled())
            if(!OutputControl::CreateFolder(properties.path()))
            {
            	string err_message = "Cannot create directory for '" + properties.path() + "'.";
                rc = -1;
            }
            else
            {
			    rc = IBFabric::OpenFile(properties.path().c_str(), sout,
			    		                to_append, message, add_header);

			    if(rc)
				    if(message.empty())
				        string err_message = "Cannot open file '" + properties.path() + "'.";
            }
        else
            isSkipped = true;
	}
    else
    {
        string err_message = "Cannot retrieve output properties for '" + identity.text() + "'";
        rc = -1;
    }

    if (!isSkipped) {
        if (!message.empty())
            this->SetLastError(message.c_str());
        else
            message = properties.path();
    }

    if (properties.in_summary() && !isSkipped)
        AddGeneratedFile(name, message);

    IBDIAG_RETURN(rc);
}

void IBDiag::CloseFile(ofstream & sout) {
    if (sout.is_open())
        sout << endl << endl << "# File closed at : " << IBFabric::GetNowTimestamp() << endl;

    sout.close();
}

void IBDiag::AddGeneratedFile(const string &name, const string &file_name)
{
    if(!name.empty())
    {
        stringstream ss;
        ss << "-I- " << left << setw(35) << name << " : " << file_name << endl;
        this->generated_files_list += ss.str();
    }
}

// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::GetReverseDirectRoute(INOUT direct_route_t *p_reverse_direct_route,
        IN direct_route_t *p_direct_route)
{
    IBDIAG_ENTER;
    CLEAR_STRUCT(*p_reverse_direct_route);
    IBNode *p_curr_node = this->root_node;
    if (!p_curr_node) {         //don't have root node
        this->SetLastError("DB error - can't find reverse direct route for direct route=%s - null root node",
                        Ibis::ConvertDirPathToStr(p_direct_route).c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    int i, j;
    //TODO - validate p_direct_route->length
    p_reverse_direct_route->length = (u_int8_t)(p_direct_route->length - 1);    //without 0 at first byte
    for (i = 1, j = p_direct_route->length - 2; j >= 0; ++i, --j) {
        u_int8_t curr_port = p_direct_route->path.BYTE[i];
        if ((curr_port <= 0) || (curr_port > p_curr_node->numPorts)) {
            this->SetLastError("DB error - can't find reverse direct route for direct route=%s - original direct route with port out of range",
                    Ibis::ConvertDirPathToStr(p_direct_route).c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        if (!p_curr_node->getPort(curr_port) || !p_curr_node->getPort(curr_port)->p_remotePort) {
            this->SetLastError("DB error - can't find reverse direct route for direct route=%s - reached null port",
                    Ibis::ConvertDirPathToStr(p_direct_route).c_str());
            if (!p_curr_node->getPort(curr_port)) {
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Node %s port:%d is null\n",
                           p_curr_node->name.c_str(), curr_port);
            } else {
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Node %s the peer of port:%d is null\n",
                           p_curr_node->name.c_str(), curr_port);
            }
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        p_reverse_direct_route->path.BYTE[j] = p_curr_node->getPort(curr_port)->p_remotePort->num;
        p_curr_node = p_curr_node->getPort(curr_port)->p_remotePort->p_node;
        if (!p_curr_node) {
            this->SetLastError("DB error - can't find reverse direct route for direct route=%s - reached null node",
                    Ibis::ConvertDirPathToStr(p_direct_route).c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_EXCEEDS_MAX_HOPS
int IBDiag::ConcatDirectRoutes(IN direct_route_t *p_direct_route1,
        IN direct_route_t *p_direct_route2,
        OUT direct_route_t *p_direct_route_result)
{
    IBDIAG_ENTER;
    CLEAR_STRUCT(*p_direct_route_result);
    int i;
    if ((p_direct_route1->length + p_direct_route2->length) > IBDIAG_MAX_HOPS) {
        this->SetLastError("Concatenate of direct route1=%s with direct route2=%s exceeded max hops",
                Ibis::ConvertDirPathToStr(p_direct_route1).c_str(),
                Ibis::ConvertDirPathToStr(p_direct_route2).c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_EXCEEDS_MAX_HOPS);
    }
    //insert first direct route elements at the beginning
    for (i = 0; i < p_direct_route1->length; ++i)
        p_direct_route_result->path.BYTE[i] = p_direct_route1->path.BYTE[i];
    //insert second direct route elements at the end
    for (i = 0; i < p_direct_route2->length; ++i)
        p_direct_route_result->path.BYTE[i + p_direct_route1->length] = p_direct_route2->path.BYTE[i];
    //update length
    p_direct_route_result->length = (u_int8_t)(p_direct_route1->length + p_direct_route2->length);
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::ResetAppData(bool force)
{
    static bool done = false;
    if (!force && done) {
        return;
    }
    //mark the node as not checked (xmit wait, port counters extended capability, ext speed counters)
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end();
            ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node)
            continue;
        p_node->appData1.val = 0;
    }
    done = true;
}

void IBDiag::DumpIbisStat(ofstream &sout)
{
    GetIbisPtr()->m_mads_stat.output_all_records(sout, false);
}

/****************************************************/
/*******************public methods*******************/
/****************************************************/
const char* IBDiag::GetLastError()
{
    IBDIAG_ENTER;
    if (this->last_error != "")
        IBDIAG_RETURN(this->last_error.c_str());
    IBDIAG_RETURN("Unknown");
}

IBDiag::IBDiag():
                curr_iteration(0), ibdiag_status(NOT_INITILIAZED),
                ibdiag_discovery_status(DISCOVERY_NOT_DONE), check_duplicated_guids(true),
                check_switch_duplicated_guids(false), root_node(NULL), root_port_num(0),
                no_mepi(false), enable_spst(true), llr_active_cell_size(RETRANS_NO_RETRANS),
                ber_threshold(DEFAULT_BER_THRESHOLD),
                send_node_desc(false), send_port_info(false),
                max_node_info_mads_in_pack(DEAFULT_NODES_INFO_MADS_IN_PACK),
                rail_flow(0), p_rail_regex(NULL), rail_data_collected(false), cable_exported(false),
                export_lib_handle(NULL), pf_export_get_api_version(NULL), pf_export_open_session(NULL),
                pf_export_close_session(NULL), pf_export_data_node(NULL), pf_export_data_port(NULL)
{
    IBDIAG_ENTER;
    IBDIAG_RETURN_VOID;
}


IBDiag::NodeInfoSendData::NodeInfoSendData(list_route_and_node_info &in_list):
                        position_itr(in_list.begin()), end_itr(in_list.end())
{
    IBDIAG_ENTER;
    IBDIAG_RETURN_VOID;
}

IBDiag::DirectRouteAndNodeInfo::DirectRouteAndNodeInfo(): p_direct_route(NULL),
                                                is_filled(false)
{
    IBDIAG_ENTER;
    CLEAR_STRUCT(node_info);
    IBDIAG_RETURN_VOID;
}

IBDiag::~IBDiag()
{
    IBDIAG_ENTER;
    //Make sure that there are no pending and on the wire MADs.
    this->ibis_obj.MadRecAll();

    this->CleanUpInternalDB();

    delete p_rail_regex;
    p_rail_regex = NULL;

    if (this->export_lib_handle)
        dlclose(this->export_lib_handle);

    IBDIAG_RETURN_VOID;
}


// Returns: SUCCESS_CODE / ERR_CODE_INIT_FAILED
int IBDiag::Init()
{
    IBDIAG_ENTER;
    if (this->IsInit() == false) {
        IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Making IBDM using internal log\n");
        ibdmUseInternalLog();

        IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Initialize IBIS\n");
        if (this->ibis_obj.Init()) {
            this->SetLastError("Failed to init ibis object, err=%s", this->ibis_obj.GetLastError());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
        }

        if (this->capability_module.Init()) {
            this->SetLastError("Failed to init capability_module object");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
        }
        this->ibdiag_status = NOT_SET_PORT;
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_INIT_FAILED
int IBDiag::SetPort(const char* device_name, phys_port_t port_num)
{
    IBDIAG_ENTER;
    if (this->IsInit() == false) {
        this->SetLastError("IBDiag initialize wasn't done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    if (this->IsReady() == true) {
        this->SetLastError("IBDiag set_port was already done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Setting IBDiag with (Device name=%s, Port number=%u)\n",
            device_name, port_num);
    if (this->ibis_obj.SetPort(device_name, port_num, true)) {
        this->SetLastError("Failed to set port of ibis object, err=%s", this->ibis_obj.GetLastError());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    this->ibdiag_status = READY;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_INIT_FAILED
int IBDiag::SetPort(u_int64_t port_guid)
{
    IBDIAG_ENTER;
    if (this->IsInit() == false) {
        this->SetLastError("IBDiag initialize wasn't done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    if (this->IsReady() == true) {
        this->SetLastError("IBDiag set_port was already done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    IBDIAG_LOG(TT_LOG_LEVEL_INFO, "Setting IBDiag with (GUID=" U64H_FMT ")\n",
            port_guid);
    if (this->ibis_obj.SetPort(ADB2C_CPU_TO_BE64(port_guid))) {
        this->SetLastError("Failed to set port of ibis object, err=%s", this->ibis_obj.GetLastError());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INIT_FAILED);
    }
    this->ibdiag_status = READY;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


IBNode * IBDiag::GetNodeByDirectRoute(IN const direct_route_t *p_direct_route)
{
    IBDIAG_ENTER;
    //set root node
    IBNode *p_curr_node = this->root_node;
    if (!p_curr_node)
        IBDIAG_RETURN(NULL);

    for (int length = 1; length < p_direct_route->length; ++length) {
        u_int8_t curr_port = p_direct_route->path.BYTE[length];
        if ((curr_port <= 0) || (curr_port > p_curr_node->numPorts))
            IBDIAG_RETURN(NULL);
        if (!p_curr_node->getPort(curr_port) || !p_curr_node->getPort(curr_port)->p_remotePort)
            IBDIAG_RETURN(NULL);

        p_curr_node = p_curr_node->getPort(curr_port)->p_remotePort->p_node;
        if (!p_curr_node)
            IBDIAG_RETURN(NULL);
    }
    IBDIAG_RETURN(p_curr_node);
}


direct_route_t * IBDiag::GetDirectRouteByNodeGuid(IN u_int64_t guid)
{
    IBDIAG_ENTER;
    list_p_direct_route list_direct_routes = this->GetDirectRoutesByNodeGuid(guid);
    if (list_direct_routes.empty())
        IBDIAG_RETURN(NULL);
    IBDIAG_RETURN(list_direct_routes.front());
}


direct_route_t * IBDiag::GetDirectRouteByPortGuid(IN u_int64_t guid)
{
    IBDIAG_ENTER;
    list_p_direct_route list_direct_routes = this->GetDirectRoutesByPortGuid(guid);
    if (list_direct_routes.empty()) {
        IBDIAG_RETURN(NULL);
    }
    IBDIAG_RETURN(list_direct_routes.front());
}


IBPort * IBDiag::GetRootPort()
{
    IBDIAG_ENTER;
    IBPort *p_port = NULL;

    //get root port
    if (!this->root_node) {
        this->SetLastError("DB error - root node is null");
        IBDIAG_RETURN(NULL);
    }
    p_port = this->root_node->getPort(this->root_port_num);
    if (!p_port) {
        this->SetLastError("DB error - root port is null");
        IBDIAG_RETURN(NULL);
    }
    IBDIAG_RETURN(p_port);
}


// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::GetLocalPortState(OUT u_int8_t& state)
{
    IBDIAG_ENTER;
    IBPort *p_root_port = this->GetRootPort();
    if (!p_root_port) {
        this->SetLastError("DB error - failed to found root port");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    //check if logical state in init
    struct SMP_PortInfo *p_root_port_info = this->fabric_extended_info.getSMPPortInfo(p_root_port->createIndex);
    if (!p_root_port_info) {
        this->SetLastError("DB error - failed to found port info for root port");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    state = p_root_port_info->PortState;
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


int IBDiag::GetAllLocalPortGUIDs(OUT local_port_t local_ports_array[IBIS_MAX_LOCAL_PORTS],
            OUT u_int32_t *p_local_ports_num)
{
    IBDIAG_ENTER;
    if (this->ibis_obj.GetAllLocalPortGUIDs(local_ports_array, p_local_ports_num)) {
        this->SetLastError("Failed to get all local ports from ibis object, err=%s",
                this->ibis_obj.GetLastError());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::getLatestSupportedVersion(int page_number,
                                      unsigned int &latest_version)
{
    IBDIAG_ENTER;
    switch (page_number) {
        case VS_MLNX_CNTRS_PAGE0:
            latest_version = PAGE0_LATEST_VER;
            break;
        case VS_MLNX_CNTRS_PAGE1:
            latest_version = PAGE1_LATEST_VER;
            break;
        case VS_MLNX_CNTRS_PAGE255:
            latest_version = PAGE255_LATEST_VER;
            break;
        default:
            this->SetLastError("Failed to found latest version"  \
                               " for page %d",page_number);
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            break;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::DiscoverFabricFromFile(IN const string &csv_file)
{
    int rc = IBDIAG_SUCCESS_CODE;

    IBDiagFabric diag_fabric(discovered_fabric, fabric_extended_info, capability_module);

    try {
        if((rc = diag_fabric.UpdateFabric(csv_file))) {
            this->SetLastError("Failed to parse CSV file");
            IBDIAG_RETURN(rc);
        }
    } catch (exception &e) {
        this->SetLastError("Failed to parse CSV file, error: %s", e.what());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    } catch (...) {
        this->SetLastError("Failed to parse CSV file");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    }

    if((rc = BuildDirectRoutesDB()))
        IBDIAG_RETURN(rc);

    SCREEN_INFO_PRINT("Discovering ... " U32D_FMT " Nodes (" U32D_FMT " Switches & " U32D_FMT
                      " CAs) discovered.",
                      diag_fabric.getNodesFound(), diag_fabric.getSWFound(),
                      diag_fabric.getCAFound());

    IBDIAG_RETURN(rc);
}

IBSpecialPortType IBDiag::GetSpecialPortType(IBPort *p_port)
{
    if (!capability_module.IsSupportedSMPCapability(
            p_port->p_node, EnSMPCapIsSpecialPortsMarkingSupported))
        return IB_NOT_SPECIAL_PORT;

    struct SMP_MlnxExtPortInfo *p_mepi = fabric_extended_info.
            getSMPMlnxExtPortInfo(p_port->createIndex);

    if (p_mepi && p_mepi->IsSpecialPort)
       return (IBSpecialPortType)p_mepi->SpecialPortType;

    return IB_NOT_SPECIAL_PORT;
}

IBSpecialPortType IBDiag::GetSpecialCAPortType(IBNode *p_node)
{
    if (p_node->type == IB_SW_NODE)
        return IB_NOT_SPECIAL_PORT;

    // go over all ports of this node:
    for (u_int32_t i = 1; i <= p_node->numPorts; ++i) {
        IBPort *p_port = p_node->getPort((phys_port_t)i);
        if (!p_port || p_port->get_internal_state() !=  IB_PORT_STATE_ACTIVE)
            continue;

        if (!p_port->p_remotePort ||
            p_port->p_remotePort->get_internal_state() !=  IB_PORT_STATE_ACTIVE)
            continue;

        return GetSpecialPortType(p_port->p_remotePort);
    }

    return IB_NOT_SPECIAL_PORT;
}

int IBDiag::CheckRailOptimizedTopology(PCI_LeafSwithcesInfoVec &leafSwitchInfoVec, vec_pport &excludedCards)
{
    IBDIAG_ENTER;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {

        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"
                               " for key = %s\n", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }

        if (p_curr_node->type != IB_SW_NODE)
            continue;

        PCI_AddressToNodesMap pciAddresses;
        int nonSpecialCANum = 0;

        if (this->GetPCIRelatedData(*p_curr_node, pciAddresses, excludedCards, nonSpecialCANum))
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);

        if (!nonSpecialCANum)
            continue;

        int size = static_cast<int>(pciAddresses.size());
        if (size > 0 && size != this->GetRailFlow()) {
            leafSwitchInfoVec.push_back(PCI_LeafSwitchInfo());

            PCI_LeafSwitchInfo& back = leafSwitchInfoVec.back();
            back.p_switch = p_curr_node;
            back.pciAddressMap.swap(pciAddresses);
        }
    }//end nodes loop

    IBDIAG_RETURN(0);
}

int IBDiag::GetPCIRelatedData(IBNode &curr_swicth_node, PCI_AddressToNodesMap &pciAddresses,
                            vec_pport &excludedCards, int &nonSpecialCANum)
{
    IBDIAG_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
            "collecting PCI addresses for the switch: %s GUID=" U64H_FMT "\n",
            curr_swicth_node.name.c_str(), curr_swicth_node.guid_get());

    for (phys_port_t i = 1; i <= curr_swicth_node.numPorts; ++i) {
        IBPort *p_curr_port = curr_swicth_node.getPort(i);
        //Skip ports that are down or have undefined state
        //or are not in the scope of the fabric.
        if (!p_curr_port || !p_curr_port->is_data_worthy())
            continue;

        IBPort *p_remote_port = p_curr_port->p_remotePort;
        if (!p_remote_port)
           continue;

        IBNode *p_remote_node = p_remote_port->p_node;
        if (!p_remote_node) {
            this->SetLastError("DB error - the port with GUID=" U64H_FMT
                               "has null node\n", p_remote_port->guid_get());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }

        if (p_remote_node->type != IB_CA_NODE)
            continue;

        if (p_remote_node->isSpecialNode())
           continue;

        nonSpecialCANum++;

        if (this->ShouldFilterNode(p_remote_node->description))
            continue;

        if (p_remote_node->isSDM()) {
            excludedCards.push_back(p_remote_port);

            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                    "adding port: %s to exclude list\n", p_remote_port->getExtendedName().c_str());
            continue;
        }

        const PCI_AddressVec &vec = p_remote_node->getPCIaddresses();
        for (PCI_AddressVec::const_iterator it = vec.begin();
             it != vec.end(); it++) {
            vector_p_node &nodes =  pciAddresses[*it];
            nodes.push_back(p_remote_node);

            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                    "adding node %s GUID=" U64H_FMT " PCI address:  %x.%x.%x\n",
                    p_remote_node->name.c_str(), p_remote_node->guid_get(),
                    it->bus, it->device, it->local_port);
        }
    }

    IBDIAG_RETURN(0);
}

bool IBDiag::ShouldFilterNode(const string &description)
{
    IBDIAG_ENTER;

    if (!this->p_rail_regex)
        IBDIAG_RETURN(false);

    rexMatch *p_rexRes = p_rail_regex->apply(description.c_str());
    if (p_rexRes) {
        delete p_rexRes;
        IBDIAG_RETURN(false);
    }

    IBDIAG_RETURN(true);
}

bool IBDiag::SetRailRegex(const string &val)
{
    IBDIAG_ENTER;

    //p_rail_regex will be deleted at IBDiag's destructor if created
    p_rail_regex = new regExp(val.c_str());
    if (!p_rail_regex->valid())
       IBDIAG_RETURN(false);

    IBDIAG_RETURN(true);
}

int IBDiag::ParseSMDBFile()
{
    IBDIAG_ENTER;

    static int rc = -1;

    // This method should be called once at most.
    // In case of return code is equal to -1, it means that the function
    // wasn't called yet.
    // If we try to call this function more than once, the return code
    // will be the last return code (not -1) of this function.
    if (rc != -1)
        IBDIAG_RETURN(rc);

    rc = IBDIAG_SUCCESS_CODE;

    try {
        if((rc = this->ibdiag_smdb.ParseSMDB(this->smdb_path))) {
            this->SetLastError("Failed to parse SMDB file - %s", this->smdb_path.c_str());
            IBDIAG_RETURN(rc);
        }
    } catch (exception &e) {
        this->SetLastError("Failed to parse SMDB file - %s, error: %s", this->smdb_path.c_str(),
                           e.what());
        rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
        IBDIAG_RETURN(rc);
    } catch (...) {
        this->SetLastError("Failed to parse SMDB file - %s", this->smdb_path.c_str());
        rc = IBDIAG_ERR_CODE_PARSE_FILE_FAILED;
        IBDIAG_RETURN(rc);
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::WriteIbisStatFile(const std::string file_name)
{
    IBDIAG_ENTER;

    ofstream sout;
    int rc = this->OpenFile("Ibis Stat",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    this->DumpIbisStat(sout);

    this->CloseFile(sout);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::LoadSymbol(void *p_lib_handle, const char *name, void **symbol,
                       list_p_fabric_general_err &export_data_errors)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    char *err;

    // Since the value of the symbol could actually be NULL, the correct way to test dlsym() for an
    // error is to call dlerror() to clear any old error, then call dlsym(), and then call dlerror()
    // again.
    dlerror();
    *symbol = dlsym(p_lib_handle, name);
    if ((err = dlerror()) != NULL) {
        export_data_errors.push_back(new ExportDataErr(
            NULL, NULL, "Failed to load symbol %s: %s", name, err));
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

int IBDiag::InitExportAPI(const string &export_lib_path)
{
    IBDIAG_ENTER;

    int rc = 0;
    list_p_fabric_general_err export_data_errors;

    if (this->export_lib_handle) {
        ERR_PRINT("Export library is already loaded\n");
        IBDIAG_RETURN(1);
    }

    this->export_lib_handle = dlopen(export_lib_path.c_str(), RTLD_LOCAL | RTLD_LAZY);
    if (!this->export_lib_handle) {
        ERR_PRINT("Failed to load library %s: %s\n", export_lib_path.c_str(), dlerror());
        IBDIAG_RETURN(1);
    }

    if ((rc = this->LoadSymbol(this->export_lib_handle, "export_get_api_version",
                              (void **)&this->pf_export_get_api_version, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->export_lib_handle, "export_open_session",
                              (void **)&this->pf_export_open_session, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->export_lib_handle, "export_close_session",
                              (void **)&this->pf_export_close_session, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->export_lib_handle, "export_data_node",
                              (void **)&this->pf_export_data_node, export_data_errors)))
        goto exit;

    if ((rc = this->LoadSymbol(this->export_lib_handle, "export_data_port",
                              (void **)&this->pf_export_data_port, export_data_errors)))
        goto exit;

exit:
    if (rc) {
        for (list_p_fabric_general_err::iterator it = export_data_errors.begin();
                it != export_data_errors.end(); ++it) {
            ERR_PRINT("%s\n", (*it)->GetErrorLine().c_str());
            delete *it;
        }
        dlclose(this->export_lib_handle);
        this->export_lib_handle = NULL;
        this->pf_export_get_api_version = NULL;
        this->pf_export_open_session = NULL;
        this->pf_export_close_session = NULL;
        this->pf_export_data_node = NULL;
        this->pf_export_data_port = NULL;
    }

    IBDIAG_RETURN(rc);
}

export_session_handle_t IBDiag::ExportOpenSession(u_int64_t iteration)
{
    if (this->pf_export_open_session)
        return this->pf_export_open_session(iteration);

    return {};
}

void IBDiag::ExportCloseSession(export_session_handle_t session_handle, int rc)
{
    if (this->pf_export_close_session)
        this->pf_export_close_session(session_handle, rc);
}

int IBDiag::ExportData(export_session_handle_t session_handle,
                       list_p_fabric_general_err &export_data_errors)
{
    IBDIAG_ENTER;

    int rc2, rc = IBDIAG_SUCCESS_CODE;
    export_data_port_t export_data_port;
    export_data_node_t export_data_node;
    cable_record_data_t cable_record_data;

    for (map_str_pnode::iterator nI = this->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != this->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = nI->second;
        if (!p_curr_node || !p_curr_node->getInSubFabric())
            continue;

        CLEAR_STRUCT(export_data_node);

        export_data_node.node_guid = p_curr_node->guid_get();

        export_data_node.p_smp_node_info =
            this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
        export_data_node.p_smp_switch_info =
            this->fabric_extended_info.getSMPSwitchInfo(p_curr_node->createIndex);
        export_data_node.p_vendor_spec_general_info =
            this->fabric_extended_info.getVSGeneralInfo(p_curr_node->createIndex);
        export_data_node.p_smp_temp_sensing =
            this->fabric_extended_info.getSMPTempSensing(p_curr_node->createIndex);
        export_data_node.p_switch_network_info =
             this->fabric_extended_info.getVSSwitchNetworkInfo(p_curr_node->createIndex);

        // Congestion Control
        export_data_node.cc_congestion_info =
                this->fabric_extended_info.getCCEnhancedCongestionInfo(p_curr_node->createIndex);
        export_data_node.curr_switch_general_settings =
                this->fabric_extended_info.getCCSwitchGeneralSettings(p_curr_node->createIndex);

        rc2 = this->pf_export_data_node(session_handle, &export_data_node);
        if (rc2) {
            ExportDataErr *p_curr_fabric_err
                = new ExportDataErr(p_curr_node, NULL, "Failed to export node data: %d", rc2);
            if (!p_curr_fabric_err) {
                this->SetLastError("Failed to allocate ExportDataErr");
                rc = IBDIAG_ERR_CODE_NO_MEM;
                goto exit;
            }
            export_data_errors.push_back(p_curr_fabric_err);
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
        }

        for (phys_port_t pI = 0; pI <= p_curr_node->numPorts; ++pI) {
            IBPort *p_curr_port = p_curr_node->getPort(pI);
            if (!p_curr_port || !p_curr_port->getInSubFabric())
                continue;

            CLEAR_STRUCT(export_data_port);

            export_data_port.node_guid = p_curr_port->p_node->guid_get();
            export_data_port.port_guid = p_curr_port->guid_get();
            export_data_port.port_num = p_curr_port->num;

            if (p_curr_port->p_remotePort) {
                export_data_port.remote_node_guid = p_curr_port->p_remotePort->p_node->guid_get();
                export_data_port.remote_port_guid = p_curr_port->p_remotePort->guid_get();
                export_data_port.remote_port_num = p_curr_port->p_remotePort->num;
            }

            export_data_port.p_smp_port_info =
                this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
            export_data_port.p_smp_mlnx_ext_port_info =
                this->fabric_extended_info.getSMPMlnxExtPortInfo(p_curr_port->createIndex);

            export_data_port.p_pm_port_counters =
                this->fabric_extended_info.getPMPortCounters(p_curr_port->createIndex);
            export_data_port.p_pm_port_counters_extended =
                this->fabric_extended_info.getPMPortCountersExtended(p_curr_port->createIndex);
            export_data_port.p_pm_port_extended_speeds_counters =
                this->fabric_extended_info.getPMPortExtSpeedsCounters(p_curr_port->createIndex);
            export_data_port.p_pm_port_extended_speeds_rsfec_counters =
                this->fabric_extended_info.getPMPortExtSpeedsRSFECCounters(p_curr_port->createIndex);
            export_data_port.p_pm_port_calc_counters =
                this->fabric_extended_info.getPMPortCalcCounters(p_curr_port->createIndex);
            export_data_port.p_vendor_spec_port_llr_statistics =
                this->fabric_extended_info.getVSPortLLRStatistics(p_curr_port->createIndex);
            export_data_port.p_pm_port_rcv_error_details =
                this->fabric_extended_info.getPMPortRcvErrorDetails(p_curr_port->createIndex);
            export_data_port.p_pm_port_xmit_discard_details =
                this->fabric_extended_info.getPMPortXmitDiscardDetails(p_curr_port->createIndex);
            export_data_port.p_pm_port_samples_control =
                this->fabric_extended_info.getPMPortSamplesControl(p_curr_port->createIndex);

            // Congestion Control
            for (u_int8_t vl = 0; vl < IB_NUM_VL; ++vl)
                export_data_port.curr_port_profile_settings[vl] =
                    this->fabric_extended_info.getCCPortProfileSettings(p_curr_port->createIndex, vl);

            export_data_port.curr_sl_mapping_settings =
                    this->fabric_extended_info.getCCSLMappingSettings(p_curr_port->createIndex);
            export_data_port.curr_hca_general_settings =
                    this->fabric_extended_info.getCCHCAGeneralSettings(p_curr_port->createIndex);
            export_data_port.curr_hca_rp_parameters =
                    this->fabric_extended_info.getCCHCARPParameters(p_curr_port->createIndex);
            export_data_port.curr_hca_np_parameters =
                    this->fabric_extended_info.getCCHCANPParameters(p_curr_port->createIndex);
            export_data_port.curr_hca_statistics_query =
                    this->fabric_extended_info.getCCHCAStatisticsQuery(p_curr_port->createIndex);
            export_data_port.cc_hca_algo_config_sup =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigSup(p_curr_port->createIndex);

            for (int algo = 0; algo < MAX_CC_ALGO_SLOT; ++algo) {
                export_data_port.cc_hca_algo_config[algo] =
                    this->fabric_extended_info.getCC_HCA_AlgoConfig(p_curr_port->createIndex, algo);

                export_data_port.cc_hca_algo_config_params[algo] =
                    this->fabric_extended_info.getCC_HCA_AlgoConfigParams(p_curr_port->createIndex, algo);

                export_data_port.cc_hca_algo_counters[algo] =
                    this->fabric_extended_info.getCC_HCA_AlgoCounters(p_curr_port->createIndex, algo);
            }

            if (p_curr_port->p_combined_cable) {
                if (p_curr_port->p_combined_cable->ExportData(&cable_record_data))
                    export_data_port.p_cable_record_data = &cable_record_data;
            }

            rc2 = this->pf_export_data_port(session_handle, &export_data_port);
            if (rc2) {
                export_data_errors.push_back(new ExportDataErr(
                    p_curr_node, p_curr_port, "Failed to export port data: %d", rc2));
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
            }
        }
    }

exit:
    IBDIAG_RETURN(rc);
}
