/*
 * Copyright (c) 2004-2019 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <infiniband/ibdiag/ibdiag.h>

// reasons for bad links according to enum IbdiagBadDirectRouteReason_t in ibdiag_types.h
const char* bad_direct_route_reasons [IBDIAG_BAD_DR_MAX] = {
    "None",
    "Node info get failed",
    "Node duplicate GUID check failed",
    "Node description get failed",
    "Node inconsistency data error",
    "Bad Node Info data",
    "Port info get failed",
    "Port invalid lid",
    "Port inconsistency data error"
};


/****************************************************/
/******************private methods*******************/
/****************************************************/


/****************************************************/
/*******************public methods*******************/
/****************************************************/
//Return: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckDuplicatedNodeDescription(list_p_fabric_general_err& nodes_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    //check for duplication
    for (map_str_list_pnode::iterator lI = this->discovered_fabric.NodeByDesc.begin();
            lI != this->discovered_fabric.NodeByDesc.end(); ++lI) {

        // skip Aggregation Nodes
        if (GetSpecialCAPortType(lI->second.front()) == IB_SPECIAL_PORT_AN)
            continue;

        //found duplicated node description
        if ((*lI).second.size() > 1) {
            for (list_p_node::iterator nI = (*lI).second.begin(); nI != (*lI).second.end(); ++nI) {
                nodes_errors.push_back(new FabricErrNodeDuplicatedNodeDesc((*nI)));
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
        }
    }
    //clean map - do not need it anymore  - not free pointers to nodes (them we need)
    this->discovered_fabric.NodeByDesc.clear();
    IBDIAG_RETURN(rc);
}


//Return: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckLinks(list_p_fabric_general_err& links_errors, IBLinksInfo *p_ib_links_info)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    string bad_dr_msg, port_name;
    IBNode *p_node;
    phys_port_t port_num;
    IBPort *port;

    //check for bad links
    for (list_p_bad_direct_route::iterator it = this->bad_direct_routes.begin();
            it != this->bad_direct_routes.end(); ++it) {
        bad_dr_msg.clear();
        bad_dr_msg = bad_direct_route_reasons[(*it)->reason] + std::string(" ") + (*it)->message;

        //if it's related to port discovery errors, add port number
        if ((*it)->reason >= IBDIAG_BAD_DR_PORT_FIRST && (*it)->reason <= IBDIAG_BAD_DR_PORT_LAST) {
            std::ostringstream oss;
            oss << (*it)->port_num;
            bad_dr_msg += " (for port " + oss.str() + ")";
        }

        if ((*it)->direct_route->length > 1) {
            bad_dr_msg += " on node connected to ";
            port_name = "NA";
            //let's try to find a previous good node
            direct_route_t new_direct_route;
            CLEAR_STRUCT(new_direct_route);
            new_direct_route =  *((*it)->direct_route);
            new_direct_route.length = (u_int8_t)((*it)->direct_route->length - 1);
            IBNode *p_prev_node = this->GetNodeByDirectRoute(&new_direct_route);
            // previous node exists
            if (p_prev_node) {
                // connected on this port to the problematic node
                port_num = (*it)->direct_route->path.BYTE[(*it)->direct_route->length - 1];
                port = p_prev_node->getPort(port_num);

                if (port)
                    port_name = port->getName();
            }
        } else if ((*it)->direct_route->length == 1) {
            //it failed on root,
            //probably won't get here as we would fail discovery
            //and not continue to other stages
            p_node = this->root_node;
            bad_dr_msg += " on root node";
            //add port name if we can
            if (p_node) {
                port_num = (*it)->direct_route->path.BYTE[(*it)->direct_route->length - 1];
                port = p_node->getPort(port_num);
                if (port)
                    port_name = port->getName();
            }
        }
        bad_dr_msg += port_name;
        bad_dr_msg += " (DR:" + Ibis::ConvertDirPathToStr((*it)->direct_route) + ")";

        links_errors.push_back(new FabricErrDR(bad_dr_msg));
        rc = IBDIAG_ERR_CODE_CHECK_FAILED;
    }

    //check for not active links
    //mark all ports as not checked
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)                   //port not exists
            continue;
        p_curr_port->counter1 = 0;
    }

    //go over all ports
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)           //port not exists
            continue;
        IBPort *p_remote_port = p_curr_port->p_remotePort;
        if (!p_remote_port)         //port not connected
            continue;

        if ((p_curr_port->counter1) || (p_curr_port->p_remotePort->counter1))   //already checked
            continue;
        //mark as checked
        ++p_curr_port->counter1;
        ++p_curr_port->p_remotePort->counter1;

        // Count IB links
        if (p_ib_links_info) {
            p_ib_links_info->FillIBLinkWidthSpeedIndex(p_curr_port);
        }
        //check link logical state
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "(port=%s logical_state=%u)<-->(port=%s logical_state=%u)\n",
                p_curr_port->getName().c_str(), p_curr_port->get_internal_state(),
                p_remote_port->getName().c_str(), p_remote_port->get_internal_state());
        if ( (p_curr_port->get_internal_state() == IB_PORT_STATE_ACTIVE) &&
                (p_remote_port->get_internal_state() == IB_PORT_STATE_ACTIVE) )           //link is active
            continue;

        if (p_curr_port->get_internal_state() != p_remote_port->get_internal_state())
            links_errors.push_back(new FabricErrLinkLogicalStateWrong(p_curr_port, p_remote_port));
        else if (p_curr_port->get_internal_state() != IB_PORT_STATE_ACTIVE)
            links_errors.push_back(new FabricErrLinkLogicalStateNotActive(p_curr_port, p_remote_port));

        rc = IBDIAG_ERR_CODE_CHECK_FAILED;
    }
    IBDIAG_RETURN(rc);
}


//Return: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED
int IBDiag::CheckLids(list_p_fabric_general_err& lids_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    map_lid_list_p_port known_lids;

    //go over all nodes
    for (u_int32_t i = 0; i < this->fabric_extended_info.getNodesVectorSize(); ++i) {
        IBNode *p_curr_node = this->fabric_extended_info.getNodePtr(i);
        if (!p_curr_node)   //node not exists
            continue;

        u_int8_t start_port_idx;
        u_int8_t end_port_idx;
        if (p_curr_node->type == IB_SW_NODE) {
            start_port_idx = 0;
            end_port_idx = 0;   //only query manage port
        } else {   //CA
            start_port_idx = 1;
            end_port_idx = p_curr_node->numPorts;
        }

        // go over all ports of this node, if SW than query only manage port
        // use unsigned int index to prevent overwrap
        for (unsigned int port_num = start_port_idx; port_num <= end_port_idx; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)port_num);
            if (!p_curr_port)                                       //port not exist
                continue;
            //Skip ports that are down or have undefined state or
            //are not in the scope of the fabric.
            if ((port_num != 0) && !p_curr_port->is_data_worthy())
                continue;
            known_lids[p_curr_port->base_lid].push_back(p_curr_port);
        }
    }

    //check for duplicated / zero lids
    for (map_lid_list_p_port::iterator it = known_lids.begin();
            it != known_lids.end(); ++it) {
        if (it->first == 0) {   //check for zero lid
            for (list_p_port::iterator it2 = it->second.begin(); it2 != it->second.end(); ++it2)
                lids_errors.push_back(new FabricErrPortZeroLid((*it2)));

            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        } else {                //check for lid duplication
            if (it->second.size() > 1) {
                for (list_p_port::iterator it2 = it->second.begin(); it2 != it->second.end(); ++it2)
                    lids_errors.push_back(new FabricErrPortDuplicatedLid((*it2)));

                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }
        }
    }

    known_lids.clear();
    IBDIAG_RETURN(rc);
}


//Return: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED / ERR_CODE_DB_ERR
int IBDiag::CheckLinkWidth(list_p_fabric_general_err& width_errors, string expected_link_width_str)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    IBLinkWidth expected_link_width = IB_UNKNOWN_LINK_WIDTH;
    if (expected_link_width_str != "") {
        expected_link_width = char2width(expected_link_width_str.c_str());
        if (expected_link_width == IB_UNKNOWN_LINK_WIDTH) {
            this->SetLastError("Link width string is unknown");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
        }
    }

    //mark all ports as not checked
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)                   //port not exists
            continue;
        p_curr_port->counter1 = 0;
    }

    //check all ports
    char buffer[256];
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)           //port not exists
            continue;
        IBPort *p_remote_port = p_curr_port->p_remotePort;
        if (!p_remote_port)         //port not connected
            continue;
        SMP_PortInfo *p_curr_port_info = this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
        if (!p_curr_port_info) {
            stringstream ss;
            ss << "DB error - found connected port=" << p_curr_port->getName() << " without SMPPortInfo";

            width_errors.push_back(
                new FabricErrPortInfoFail(p_curr_port->p_node, p_curr_port->num, ss.str().c_str()));

            continue;
        }
        SMP_PortInfo *p_remote_port_info = this->fabric_extended_info.getSMPPortInfo(p_remote_port->createIndex);
        if (!p_remote_port_info) {
            stringstream ss;
            ss << "DB error - found connected port=" << p_remote_port->getName() << " without SMPPortInfo";

            width_errors.push_back(
                new FabricErrPortInfoFail(p_curr_port->p_node, p_curr_port->num, ss.str().c_str()));

            continue;
        }

        if ((p_curr_port->counter1) || (p_remote_port->counter1))   //already checked
            continue;
        //mark as checked
        ++p_curr_port->counter1;
        ++p_remote_port->counter1;


        /*
         * Step 1 - check same width on both ports
         */
        if (p_curr_port->get_internal_width() != p_remote_port->get_internal_width()) {
            width_errors.push_back(new FabricErrLinkDifferentWidth(p_curr_port, p_remote_port));
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            continue;
        }


        /*
         * Step 2 - check en <= sup
         */
        bool err_occured = false;

        // Step 2.1 - check port1
        if (!IsValidEnableWidth(p_curr_port_info->LinkWidthEn,
                p_curr_port_info->LinkWidthSup)) {
            snprintf(buffer, sizeof(buffer),
                     "enable_width=\"%s\" but supported_width=\"%s\"",
                     supwidth2char(p_curr_port_info->LinkWidthEn),
                     supwidth2char(p_curr_port_info->LinkWidthSup));

            width_errors.push_back(new FabricErrPortWrongConfig(p_curr_port, buffer));
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            //continue;     not continue because need to check other port as well
            err_occured = true;
        }

        // Step 2.2 - check port2
        if (!IsValidEnableWidth(p_remote_port_info->LinkWidthEn,
                p_remote_port_info->LinkWidthSup)) {
            snprintf(buffer, sizeof(buffer),
                     "enable_width=\"%s\" but supported_width=\"%s\"",
                     supwidth2char(p_remote_port_info->LinkWidthEn),
                     supwidth2char(p_remote_port_info->LinkWidthSup));

            width_errors.push_back(new FabricErrPortWrongConfig(p_remote_port, buffer));
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            //continue;     not continue because need to check other port as well
            err_occured = true;
        }
        if (err_occured)
            continue;   //not need next check in this case

        /*
         * Step 3 - check expected value
         */

        /* Step 3.1 - calculate final width if no specific value
           for checking is given from user */
        if (expected_link_width_str == "") {
            //get final speed
            expected_link_width = CalcFinalWidth(p_curr_port_info->LinkWidthEn,
                    p_remote_port_info->LinkWidthEn);
            if (expected_link_width == IB_UNKNOWN_LINK_WIDTH) {
                snprintf(buffer, sizeof(buffer),
                         "(enable_width1=\"%s\", enable_width2=\"%s\" therefore final width is unknown)",
                         supwidth2char(p_curr_port_info->LinkWidthEn),
                         supwidth2char(p_remote_port_info->LinkWidthEn));

                width_errors.push_back(new FabricErrLinkAutonegError(p_curr_port, p_remote_port, buffer));
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                continue;
            }
        }

        // Step 3.2 - check actual link width
        if (p_curr_port->get_internal_width() != expected_link_width) {
            if (expected_link_width_str == "") {
                snprintf(buffer, sizeof(buffer),
                         "(enable_width1=\"%s\", enable_width2=\"%s\" therefore final width should be %s)",
                         supwidth2char(p_curr_port_info->LinkWidthEn),
                         supwidth2char(p_remote_port_info->LinkWidthEn),
                         width2char(expected_link_width));
            } else
                memset(buffer, 0, sizeof(buffer));

            width_errors.push_back(
                    new FabricErrLinkUnexpectedWidth(p_curr_port, p_remote_port, buffer));

            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }
    IBDIAG_RETURN(rc);
}


//Return: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM / ERR_CODE_CHECK_FAILED / ERR_CODE_DB_ERR
int IBDiag::CheckLinkSpeed(list_p_fabric_general_err& speed_errors, string expected_link_speed_str)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;
    IBLinkSpeed expected_link_speed = IB_UNKNOWN_LINK_SPEED;
    if (expected_link_speed_str != "") {
        expected_link_speed = char2speed(expected_link_speed_str.c_str());
        if (expected_link_speed == IB_UNKNOWN_LINK_SPEED) {
            this->SetLastError("Link speed string is unknown");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
        }
    }

    //mark all ports as not checked
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)                   //port not exists
            continue;
        p_curr_port->counter1 = 0;
    }

    //check all ports
    char buffer[256];
    for (u_int32_t i = 0; i < this->fabric_extended_info.getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = this->fabric_extended_info.getPortPtr(i);
        if (!p_curr_port)           //port not exists
            continue;
        IBPort *p_remote_port = p_curr_port->p_remotePort;
        if (!p_remote_port)         //port not connected
            continue;
        SMP_PortInfo *p_curr_port_info = this->fabric_extended_info.getSMPPortInfo(p_curr_port->createIndex);
        if (!p_curr_port_info) {
            stringstream ss;
            ss << "DB error - found connected port=" << p_curr_port->getName() << " without SMPPortInfo";

            speed_errors.push_back(
                new FabricErrPortInfoFail(p_curr_port->p_node, p_curr_port->num, ss.str().c_str()));

            continue;
        }
        SMP_PortInfo *p_remote_port_info = this->fabric_extended_info.getSMPPortInfo(p_remote_port->createIndex);
        if (!p_remote_port_info) {
            stringstream ss;
            ss << "DB error - found connected port=" << p_remote_port->getName() << " without SMPPortInfo";

            speed_errors.push_back(
                new FabricErrPortInfoFail(p_curr_port->p_node, p_curr_port->num, ss.str().c_str()));

            continue;
        }
        SMP_MlnxExtPortInfo *p_curr_mlnx_ext_port_info =
                this->fabric_extended_info.getSMPMlnxExtPortInfo(p_curr_port->createIndex);
        SMP_MlnxExtPortInfo *p_remote_mlnx_ext_port_info =
                this->fabric_extended_info.getSMPMlnxExtPortInfo(p_remote_port->createIndex);

        if ((p_curr_port->counter1) || (p_remote_port->counter1))   //already checked
            continue;
        //mark as checked
        ++p_curr_port->counter1;
        ++p_remote_port->counter1;

        /*
         * Step 1 - check same speed on both ports
         */
        if (p_curr_port->get_internal_speed() != p_remote_port->get_internal_speed()) {
            speed_errors.push_back(new FabricErrLinkDifferentSpeed(p_curr_port, p_remote_port));
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            continue;
        }

        /*
         * Step 2 - find cap mask of both ports
         */
        u_int32_t cap_mask_curr_port            = 0;
        u_int32_t cap_mask_remote_port          = 0;

        // Step 2.1 - cap mask of current port
        struct SMP_PortInfo *p_tmp_port_info    = NULL;
        IBPort *p_tmp_port                      = NULL;
        if (p_curr_port->p_node->type == IB_SW_NODE) {
            p_tmp_port = p_curr_port->p_node->getPort(0);
            if (!p_tmp_port) {
                this->SetLastError("DB error - can not found manage port for switch=%s\n",
                        p_curr_port->p_node->getName().c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            p_tmp_port_info = this->fabric_extended_info.getSMPPortInfo(
                                    p_tmp_port->createIndex);
        } else
            p_tmp_port_info = p_curr_port_info;

        if (!p_tmp_port_info) {
            this->SetLastError("DB error - can not found port info for port =%s\n",
                    (p_curr_port->p_node->type == IB_SW_NODE) ? \
                            p_tmp_port->getName().c_str() : \
                            p_curr_port->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        cap_mask_curr_port = p_tmp_port_info->CapMsk;

        // Step 2.2 - cap mask of remote port
        p_tmp_port_info = NULL;
        p_tmp_port      = NULL;
        if (p_remote_port->p_node->type == IB_SW_NODE) {
            p_tmp_port = p_remote_port->p_node->getPort(0);
            if (!p_tmp_port) {
                this->SetLastError("DB error - can not found manage port for switch=%s\n",
                        p_remote_port->p_node->getName().c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            p_tmp_port_info = this->fabric_extended_info.getSMPPortInfo(
                              p_tmp_port->createIndex);
        } else
            p_tmp_port_info = p_remote_port_info;

        if (!p_tmp_port_info) {
            this->SetLastError("DB error - can not found port info for port=%s\n",
                    (p_remote_port->p_node->type == IB_SW_NODE) ?   \
                            p_tmp_port->getName().c_str() : \
                            p_remote_port->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        cap_mask_remote_port = p_tmp_port_info->CapMsk;

        /*
         * Step 3 - check en <= sup
         */
        bool err_occured = false;
        u_int32_t curr_port_speed_sup   = 0;
        u_int32_t curr_port_speed_en    = 0;

        //3.1 check current port
        //get total speeds: regular in first byte, extended in second,
        //mellanox in third
        GetTotalSpeeds(p_curr_port_info, p_curr_mlnx_ext_port_info,
                       cap_mask_curr_port,
                       &curr_port_speed_sup, &curr_port_speed_en, NULL);

        if (!IsValidEnableSpeed(curr_port_speed_en, curr_port_speed_sup)) {
            snprintf(buffer, sizeof(buffer),
                     "enable_speed=\"%s\" but supported_speed=\"%s\"",
                     supspeed2char(curr_port_speed_en).c_str(),
                     supspeed2char(curr_port_speed_sup).c_str());

            speed_errors.push_back(new FabricErrPortWrongConfig(p_curr_port, buffer));
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            //continue;     not continue because need to check other port as well
            err_occured = true;
        }

        //3.2 check the remote port
        u_int32_t remote_port_speed_sup = 0;
        u_int32_t remote_port_speed_en  = 0;

        GetTotalSpeeds(p_remote_port_info, p_remote_mlnx_ext_port_info,
                       cap_mask_remote_port,
                       &remote_port_speed_sup, &remote_port_speed_en, NULL);

        if (!IsValidEnableSpeed(remote_port_speed_en, remote_port_speed_sup)) {
            snprintf(buffer, sizeof(buffer),
                     "enable_speed=\"%s\" but supported_speed=\"%s\"",
                     supspeed2char(remote_port_speed_en).c_str(),
                     supspeed2char(remote_port_speed_sup).c_str());

            speed_errors.push_back(new FabricErrPortWrongConfig(p_remote_port, buffer));
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            err_occured = true;
        }

        if (err_occured)
            continue;   //not need next check in this case

        /*
         * Step 4 - check expected value
         * we already have enabled speeds for current port and its remote port
         * Step 4.1 Calculate final speed if no specific value
         * for checking is given from user
         */
        if (expected_link_speed_str == "") {
            expected_link_speed = CalcFinalSpeed(curr_port_speed_en,
                                                 remote_port_speed_en);
            if (expected_link_speed == IB_UNKNOWN_LINK_SPEED) {
                snprintf(buffer, sizeof(buffer),
                         "(enable_speed1=\"%s\", enable_speed2=\"%s\" "
                         "therefore final speed is unknown)",
                         supspeed2char(curr_port_speed_en).c_str(),
                         supspeed2char(remote_port_speed_en).c_str());

                speed_errors.push_back(new FabricErrLinkAutonegError(p_curr_port, p_remote_port, buffer));
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
                continue;
            }
        }

        // Step 4.2 - check actual link speed
        if (p_curr_port->get_internal_speed() != expected_link_speed) {
            if (expected_link_speed_str == "") {
                snprintf(buffer, sizeof(buffer),
                         "(enable_speed1=\"%s\", enable_speed2=\"%s\" therefore final speed should be %s)",
                         supspeed2char(curr_port_speed_en).c_str(),
                         supspeed2char(remote_port_speed_en).c_str(),
                         speed2char(expected_link_speed));
            } else
                memset(buffer, 0, sizeof(buffer));

            speed_errors.push_back(new FabricErrLinkUnexpectedSpeed(p_curr_port, p_remote_port, buffer));
            rc = IBDIAG_ERR_CODE_CHECK_FAILED;
        }
    }
    IBDIAG_RETURN(rc);
}


int IBDiag::CheckPortHierarchyInfo(list_p_fabric_general_err &hierarchy_errors)
{
    IBDIAG_ENTER;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_node = nI->second;
        if (!p_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (!p_node->should_support_port_hierarchy_info)
            continue;

        bool is_splitted = p_node->isSplitted();

        for (phys_port_t i = 1; i <= p_node->numPorts; ++i) {
            IBPort *p_port = p_node->getPort(i);
            if (!p_port || p_port->isSpecialPort())
                continue;

            // 1. Check that all switch and only active CA ports have hierarchy info
            if (!p_port->p_port_hierarchy_info) {
                if (p_node->type == IB_SW_NODE || p_port->get_internal_state() == IB_PORT_STATE_ACTIVE)
                    hierarchy_errors.push_back(new FabricErrPortHierarchyMissing(p_port));

                continue;
            }

            // 2. Cache is_splitted for this node, since p_node->isSplitted() doesn't work for HCAs
            if (p_port->p_port_hierarchy_info->m_split != -1)
                is_splitted = true;
        }

        for (phys_port_t i = 1; i <= p_node->numPorts; ++i) {
            IBPort *p_port = p_node->getPort(i);
            if (!p_port || p_port->isSpecialPort() || !p_port->p_port_hierarchy_info)
                continue;

            // 3. Check that correct port hierarchy info fields are present
            vector<string> missing_fields;
            vector<string> extra_fields;
            if (p_node->type == IB_SW_NODE) {
                // Check mandatory fields: ASIC, Cage, Port, Split
                if (p_port->p_port_hierarchy_info->m_asic == -1)
                    missing_fields.push_back("ASIC");
                if (p_port->p_port_hierarchy_info->m_cage == -1)
                    missing_fields.push_back("Cage");
                if (p_port->p_port_hierarchy_info->m_port == -1)
                    missing_fields.push_back("Port");
                if (is_splitted && p_port->p_port_hierarchy_info->m_split == -1)
                    missing_fields.push_back("Split");

                // Check extra fields: BDF, Type, Slot
                if (p_port->p_port_hierarchy_info->bus() != -1) {
                    extra_fields.push_back("Bus");
                    extra_fields.push_back("Device");
                    extra_fields.push_back("Function");
                }
                if (p_port->p_port_hierarchy_info->m_type != -1)
                    extra_fields.push_back("Type");
                if (p_port->p_port_hierarchy_info->m_slot_value != -1)
                    extra_fields.push_back("Slot");
            } else if (p_node->type == IB_CA_NODE) {
                // Check mandatory fields: Slot or BDF
                if (p_port->p_port_hierarchy_info->m_slot_type ==
                        PortHierarchyInfo::SLOT_TYPE_PHYSICAL) {
                    if (p_port->p_port_hierarchy_info->bus() != -1) {
                        extra_fields.push_back("Bus");
                        extra_fields.push_back("Device");
                        extra_fields.push_back("Function");
                    }
                } else {
                    if (p_port->p_port_hierarchy_info->bus() == -1) {
                        missing_fields.push_back("Bus");
                        missing_fields.push_back("Device");
                        missing_fields.push_back("Function");
                    }
                    if (p_port->p_port_hierarchy_info->m_slot_value != -1)
                        extra_fields.push_back("Slot (not physical)");
                }

                // Check mandatory fields: Cage, Port, Split
                if (p_port->p_port_hierarchy_info->m_cage == -1)
                    missing_fields.push_back("Cage");
                if (p_port->p_port_hierarchy_info->m_port == -1)
                    missing_fields.push_back("Port");
                if (is_splitted && p_port->p_port_hierarchy_info->m_split == -1)
                    missing_fields.push_back("Split");

                // Check extra fields: ASIC, Type
                if (p_port->p_port_hierarchy_info->m_asic != -1)
                    extra_fields.push_back("ASIC");
                if (p_port->p_port_hierarchy_info->m_type != -1)
                    extra_fields.push_back("Type");
            }

            if (!missing_fields.empty()) {
                hierarchy_errors.push_back(
                    new FabricErrPortHierarchyMissingFields(p_port, missing_fields));
                rc = IBDIAG_ERR_CODE_CHECK_FAILED;
            }

            if (!extra_fields.empty())
                hierarchy_errors.push_back(
                    new FabricErrPortHierarchyExtraFields(p_port, extra_fields));
        }
    }

    IBDIAG_RETURN(rc);
}
