/*
 * Copyright (c) 2004-2020 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <algorithm>
#include <infiniband/ibdiag/ibdiag.h>
#include <infiniband/ibdiag/ibdiag_types.h>
#include <infiniband/ibdiag/ibdiag_clbck.h>
#include <infiniband/ibdm/OutputControl.h>
#include "infiniband/ibutils/output_stream_primitives.h"

extern IBDiagClbck ibDiagClbck;

/****************************************************/
/******************private methods*******************/
/****************************************************/
inline bool IBDiag::IsBFSKnownPortGuid(IN u_int64_t guid)
{
    map_guid_list_p_direct_route::iterator nI = this->bfs_known_port_guids.find(guid);
    if (nI == this->bfs_known_port_guids.end())
            return false;
    return ((*nI).second.size() != 0);
}

inline bool IBDiag::IsBFSKnownNodeGuid(IN u_int64_t guid)
{
    map_guid_list_p_direct_route::iterator nI = this->bfs_known_node_guids.find(guid);
    if (nI == this->bfs_known_node_guids.end())
            return false;
    return ((*nI).second.size() != 0);
}

inline void IBDiag::MarkPortGuidAsBFSKnown(IN u_int64_t guid, IN direct_route_t *p_direct_route)
{
    this->bfs_known_port_guids[guid].push_back(p_direct_route);
}

inline void IBDiag::MarkNodeGuidAsBFSKnown(IN u_int64_t guid, IN direct_route_t *p_direct_route)
{
    this->bfs_known_node_guids[guid].push_back(p_direct_route);
}


inline void IBDiag::BFSPushPath(IN direct_route_t *p_direct_route)
{
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "BFSPushPath %s\n",
               Ibis::ConvertDirPathToStr(p_direct_route).c_str());
    this->bfs_list.push_back(p_direct_route);   //push to the end of the list
}

inline direct_route_t * IBDiag::BFSPopPath()
{
    direct_route_t *current = this->bfs_list.front();
    this->bfs_list.pop_front();                 //pop from the begin of the list
    return current;
}


inline void IBDiag::AddGoodPath(IN direct_route_t *p_direct_route)
{
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "AddGoodPath %s\n",
              Ibis::ConvertDirPathToStr(p_direct_route).c_str());

    this->good_direct_routes.push_back(p_direct_route);
}

inline int IBDiag::AddBadPath(IN IbdiagBadDirectRoute_t *bad_direct_route,
                               IN direct_route_t *direct_route)
{
    IbdiagBadDirectRoute_t *p_bad_direct_route = NULL;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "AddBadPath %s, fail reason \"%s\", port num: %d\n",
              Ibis::ConvertDirPathToStr(direct_route).c_str(),
               bad_direct_route_reasons[bad_direct_route->reason],
               bad_direct_route->port_num);

    p_bad_direct_route = new IbdiagBadDirectRoute_t;
    *p_bad_direct_route = *bad_direct_route;
    p_bad_direct_route->direct_route = direct_route;
    this->bad_direct_routes.push_back(p_bad_direct_route);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

inline void IBDiag::AddLoopPath(IN direct_route_t *p_direct_route)
{
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "AddLoopPath %s\n",
              Ibis::ConvertDirPathToStr(p_direct_route).c_str());

    this->loop_direct_routes.push_back(p_direct_route);
}


void IBDiag::PostDiscoverFabricProcess()
{
    IBDIAG_ENTER;
    if (this->ibdiag_discovery_status == DISCOVERY_DUPLICATED_GUIDS) {
        this->SetLastError("Duplicated GUIDs were found during discovery process");
        IBDIAG_RETURN_VOID;
    }
    this->ibdiag_discovery_status = DISCOVERY_SUCCESS;
    IBDIAG_RETURN_VOID;
}


// Returns: SUCCESS_CODE / ERR_CODE_FABRIC_ERROR / ERR_CODE_IBDM_ERR / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDiag::DiscoverFabricOpenCAPorts(IN IBNode *p_node,
        IN direct_route_t *p_direct_route,
        IN SMP_NodeInfo *p_node_info,
        IN bool is_root,
        OUT IbdiagBadDirectRoute_t *p_bad_direct_route_info,
        IN bool push_new_direct_route)
{
    IBDIAG_ENTER;
    SMP_PortInfo curr_port_info;
    bool to_send_port_info_mad = (is_root || this->send_port_info);

    //sending only to entry port because we can't exit from CA
    if (to_send_port_info_mad){
       if (this->ibis_obj.SMPPortInfoMadGetByDirect(p_direct_route,
                    p_node_info->LocalPortNum,
                    &curr_port_info)) {
            /*this->SetLastError("Failed to get port info for direct route %s (port=%u), err=%s",
                Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                p_node_info->LocalPortNum,
                this->ibis_obj.GetLastError());*/
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                    "Failed to get port info for direct route %s (port=%u), err=%s\n",
                    Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                    p_node_info->LocalPortNum,
                    this->ibis_obj.GetLastError());
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_PORT_INFO;
            p_bad_direct_route_info->port_num = p_node_info->LocalPortNum;
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }
    }
    else {
        CLEAR_STRUCT(curr_port_info);
        curr_port_info.LinkSpeedActv = IB_UNKNOWN_LINK_SPEED;
        curr_port_info.LinkWidthActv = IB_UNKNOWN_LINK_WIDTH;
        //has to be restored when a valid value is obtained
        curr_port_info.PortState = IB_PORT_STATE_ACTIVE;
        curr_port_info.PortPhyState = IB_PORT_PHYS_STATE_POLLING;

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                "CA direct route %s(port=%u) has to be filled in later\n",
                Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                p_node_info->LocalPortNum);
    }

    //set this port data to the node port
    IBLinkSpeed correct_speed = (IBLinkSpeed)curr_port_info.LinkSpeedActv;
    if ((curr_port_info.CapMsk & IBIS_IB_PORT_CAP_HAS_EXT_SPEEDS) &&
            curr_port_info.LinkSpeedExtActv) {
        /* port supports extended speeds and active extended speed
           is not zero ==> we have extended speed */
        correct_speed = extspeed2speed(curr_port_info.LinkSpeedExtActv);
    }

    int numLidsPerPort = 1 << curr_port_info.LMC;
    if ( curr_port_info.LID >= 0xc000 || ((int)curr_port_info.LID + numLidsPerPort) >= 0xc000) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s invalid lid:%u lmc:%u\n",
                p_node->name.c_str(),curr_port_info.LID, curr_port_info.LMC);

        p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INVALID_LID;
        p_bad_direct_route_info->port_num = p_node_info->LocalPortNum;

        errors.push_back(
                new FabricErrNodeInvalidLid(p_node, p_node_info->LocalPortNum,
                                            curr_port_info.LID, curr_port_info.LMC));

        if (!is_root)
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    IBPort *p_port = this->discovered_fabric.setNodePort(p_node,
            p_node_info->PortGUID,
            curr_port_info.LID,
            curr_port_info.LMC,
            p_node_info->LocalPortNum,
            (IBLinkWidth)curr_port_info.LinkWidthActv,
            correct_speed,
            (IBPortState)curr_port_info.PortState);
    if (!p_port) {
        this->SetLastError("Failed to set port data for port=%u of node=%s", p_node_info->LocalPortNum, p_node->getName().c_str());
        p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INCONSISTENT;
        p_bad_direct_route_info->port_num = p_node_info->LocalPortNum;
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }

    p_port->setPortInfoMadWasSent(to_send_port_info_mad);

    //save smp port info attribute in the fabric
    int rc = this->fabric_extended_info.addSMPPortInfo(p_port, curr_port_info);
    if (rc) {
        this->SetLastError("Failed to set smp_port_info for port %u of node in direct route %s, err=%s",
                p_port->num, Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                this->fabric_extended_info.GetLastError());
        p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INCONSISTENT;
        p_bad_direct_route_info->port_num = p_node_info->LocalPortNum;
        IBDIAG_RETURN(rc);
    }


    if (is_root == true) {  //if starting discovery from hca
        if (curr_port_info.PortPhyState != IB_PORT_PHYS_STATE_LINK_UP)    //not active
            IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

        if (push_new_direct_route) {
            //next bfs step
            direct_route_t *p_next_bfs_direct_route = new direct_route_t;
            *p_next_bfs_direct_route =  *p_direct_route;
            p_next_bfs_direct_route->path.BYTE[p_direct_route->length] = p_node_info->LocalPortNum;
            p_next_bfs_direct_route->length = (u_int8_t)(p_direct_route->length + 1); //TODO check length
            this->BFSPushPath(p_next_bfs_direct_route);
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_FABRIC_ERROR / ERR_CODE_IBDM_ERR / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDiag::DiscoverFabricOpenSWPorts(IN IBNode *p_node,
        IN direct_route_t *p_direct_route,
        IN SMP_NodeInfo *p_node_info,
        IN bool is_root,
        OUT IbdiagBadDirectRoute_t *p_bad_direct_route_info,
        IN bool push_new_direct_route)
{
    IBDIAG_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
               "DiscoverFabricOpenSWPorts SW GUID " U64H_FMT
               " Name: %s direct_route: %s\n",
               p_node_info->NodeGUID, p_node->name.c_str(),
               Ibis::ConvertDirPathToStr(p_direct_route).c_str());

    SMP_PortInfo curr_port_info;
    u_int16_t base_lid = 0;
    u_int32_t cap_mask = 0;
    u_int16_t cap_mask2 = 0;
    u_int8_t num_of_blocks = (u_int8_t)((p_node->numPorts / NUM_OF_PORT_STATES_IN_BLOCK) + 1);
    vector_smp_switch_port_state_table vec_switch_port_state_table(num_of_blocks);
    bool is_switch_port_state_table_supported = false;
    bool to_send_portinfo_mad = true;
    u_int8_t lmc = 0;
    //going over all ports of this node (SW)
    // use unsigned int index to prevent overwrap
    for (unsigned int i = 0; i <= p_node_info->NumPorts; ++i) {
        phys_port_t port_num = (phys_port_t)i;
        bool port_info_received = true;

        if (is_switch_port_state_table_supported && (port_num != 0)) {
            u_int8_t block_number = port_num / NUM_OF_PORT_STATES_IN_BLOCK;
            u_int8_t port_idx = (u_int8_t)(port_num % NUM_OF_PORT_STATES_IN_BLOCK);

                CLEAR_STRUCT(curr_port_info);
                curr_port_info.LinkSpeedActv = IB_UNKNOWN_LINK_SPEED;
                curr_port_info.LinkWidthActv = IB_UNKNOWN_LINK_WIDTH;
                curr_port_info.PortState = vec_switch_port_state_table[block_number].
                                                            PortStateBlock[port_idx];
                if (curr_port_info.PortState != IB_PORT_STATE_DOWN){
                    //PortPhyState will be taken later

                     IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                            "SW direct route %s (port=%u) port_state: %u has to be filled later\n",
                            Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                            port_num, curr_port_info.PortState);
                } else {
                    curr_port_info.PortPhyState = IB_PORT_PHYS_STATE_POLLING;

                    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                            "SW direct route %s (port=%u) port_state: %u has to be skiped\n",
                            Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                            port_num, curr_port_info.PortState);
                }

                to_send_portinfo_mad = false;
        }

        if (to_send_portinfo_mad) {
            //send port info mad to check if this port is active

            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "SW direct route %s (port=%u) will be sent",
                       Ibis::ConvertDirPathToStr(p_direct_route).c_str(), port_num);

            if (this->ibis_obj.SMPPortInfoMadGetByDirect(p_direct_route, port_num, &curr_port_info)) {
                /*this->SetLastError("Failed to get port info for direct route %s (port=%u), err=%s",
                        Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                        port_num,
                        this->ibis_obj.GetLastError());*/
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Failed to get port info for direct route %s (port=%u), err=%s\n",
                        Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                        port_num,
                        this->ibis_obj.GetLastError());

                p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_PORT_INFO;
                p_bad_direct_route_info->port_num = port_num;

                errors.push_back(
                        new FabricErrPortInfoFail(p_node, port_num, ibis_obj.GetLastError()));

                p_node->toIgnore=true;

                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s (%p) toIgnore:%d\n",
                        p_node->name.c_str(),p_node,p_node->toIgnore);


                /*
                 * If we didn't receive port info, we don't interrupt discovery. Instead we will
                 * add dummy ACTIVE port to the fabric. ibdiagnet will use this port to continue
                 * discovery and will re-try getting port info for that port later.
                 */
                port_info_received = false;
                CLEAR_STRUCT(curr_port_info);
            }
        }

        if (port_info_received) {
            if (port_num == 0)  {
                //manage port -
                //  we start loop from port_num = 0, therefore we will reach here first thing
                base_lid = curr_port_info.LID;
                cap_mask = curr_port_info.CapMsk;
                lmc = curr_port_info.LMC;
                cap_mask2 = curr_port_info.CapMsk2;

                if (IS_SUPPORT_SWITCH_PORT_STATE_TABLE(cap_mask2)) {
                    is_switch_port_state_table_supported = true;
                    for (u_int8_t block_number = 0; block_number < num_of_blocks; block_number++) {
                        CLEAR_STRUCT(vec_switch_port_state_table[block_number]);
                        if (this->ibis_obj.SMPSwitchPortStateTableMadGetByDirect(p_direct_route,
                                                        block_number,
                                                        &vec_switch_port_state_table[block_number])) {
                            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Failed to get Switch Port State Table for "
                                        "direct route %s (port=%u), err=%s\n",
                                        Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                                        port_num, this->ibis_obj.GetLastError());

                            is_switch_port_state_table_supported = false;
                            break;
                        }
                    }
                }
            }

            //add this port to discovered fabric or get ref to exist one
            IBLinkSpeed correct_speed = (IBLinkSpeed)curr_port_info.LinkSpeedActv;
            if ((cap_mask & IBIS_IB_PORT_CAP_HAS_EXT_SPEEDS) &&
                    curr_port_info.LinkSpeedExtActv) {
                /* port supports extended speeds and active extended speed
                is not zero ==> we have extended speed */
                correct_speed = extspeed2speed(curr_port_info.LinkSpeedExtActv);
            }

            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                    "SW direct route %s (port=%u) old correct speed: %d provided speed: %d\n",
                    Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                    port_num, correct_speed, curr_port_info.LinkSpeedActv);

            int numLidsPerPort = 1 << lmc;
            if ( base_lid >= 0xc000 || ((int)base_lid + numLidsPerPort) >= 0xc000) {
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node %s invalid lid:%u lmc:%u\n",
                        p_node->name.c_str(),base_lid, lmc);
                p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INVALID_LID;
                p_bad_direct_route_info->port_num = port_num;

                errors.push_back(
                        new FabricErrNodeInvalidLid(p_node, port_num, base_lid, lmc));
            }

            IBPort *p_port = this->discovered_fabric.setNodePort(p_node,
                    p_node_info->PortGUID,
                    base_lid,
                    lmc,
                    port_num,
                    (IBLinkWidth)curr_port_info.LinkWidthActv,
                    correct_speed,
                    (IBPortState)curr_port_info.PortState);
            if (!p_port) {
                this->SetLastError("Failed to store port data for port=%u of node=%s", p_node_info->LocalPortNum, p_node->getName().c_str());
                p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INCONSISTENT;
                p_bad_direct_route_info->port_num = port_num;
                IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
            }

            p_port->setPortInfoMadWasSent(to_send_portinfo_mad);

            //set cap mask; to be used later for filling port's speed.
            if (port_num == 0){
                p_port->setCapMask(cap_mask);
            }

            //save smp port info attribute in the fabric
            int rc = this->fabric_extended_info.addSMPPortInfo(p_port, curr_port_info);
            if (rc) {
                this->SetLastError("Failed to store smp_port_info for port %u of node in direct route %s, err=%s",
                        port_num, Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                        this->fabric_extended_info.GetLastError());
                p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INCONSISTENT;
                p_bad_direct_route_info->port_num = port_num;
                IBDIAG_RETURN(rc);
            }
        } else {
            IBPort *p_port = this->discovered_fabric.setNodePort(p_node,
                    p_node_info->PortGUID,
                    base_lid,
                    lmc,
                    port_num,
                    IB_UNKNOWN_LINK_WIDTH,
                    IB_UNKNOWN_LINK_SPEED,
                    IB_PORT_STATE_ACTIVE);
            if (!p_port) {
                this->SetLastError("Failed to store port data for port=%u of node=%s", p_node_info->LocalPortNum, p_node->getName().c_str());
                p_bad_direct_route_info->reason = IBDIAG_BAD_DR_PORT_INCONSISTENT;
                p_bad_direct_route_info->port_num = port_num;
                IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
            }
        }

        //manage port, can't go out from it
        if (port_num == 0)
            continue;
        //entry port, no sense to go out from it
        if ((is_root == false) && (port_num == p_node_info->LocalPortNum))
            continue;
        //not active
        if (curr_port_info.PortState == IB_PORT_STATE_DOWN)
            continue;

        if (push_new_direct_route) {
            //next bfs step
            direct_route_t *p_next_bfs_direct_route = new direct_route_t;
            *p_next_bfs_direct_route =  *p_direct_route;
            p_next_bfs_direct_route->path.BYTE[p_direct_route->length] = port_num;
            p_next_bfs_direct_route->length = (u_int8_t)(p_direct_route->length + 1); //TODO Fix type
            this->BFSPushPath(p_next_bfs_direct_route);
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_FABRIC_ERROR / ERR_CODE_IBDM_ERR / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDiag::DiscoverFabricBFSOpenPorts(IN direct_route_t * p_direct_route,
        IN IBNode *p_node,
        IN struct SMP_NodeInfo *p_node_info,
        IN bool is_visited_node,
        IN bool is_root,
        OUT IbdiagBadDirectRoute_t *p_bad_direct_route_info,
        IN bool push_new_direct_route)
{
    IBDIAG_ENTER;
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
               "DiscoverFabricBFSOpenPorts node GUID " U64H_FMT
               " Name: %s direct_route: %s\n",
               p_node_info->NodeGUID, p_node->name.c_str(),
               Ibis::ConvertDirPathToStr(p_direct_route).c_str());
    int rc;
    if ((p_node_info->NodeType == IB_SW_NODE) && (is_visited_node == false)) {  //not visited Switch
                                            //if visited than we already opened all it's ports
        rc = this->DiscoverFabricOpenSWPorts(p_node, p_direct_route, p_node_info, is_root, p_bad_direct_route_info,
                                             push_new_direct_route);
        if (rc)
            IBDIAG_RETURN(rc);
    } else if (p_node_info->NodeType == IB_CA_NODE ||
               p_node_info->NodeType == IB_RTR_NODE) { //for now RTR is like HCA
                                            //hca and therefore if visited already than
                                            //it means that now we are reaching it from it's other port
                                            //so we will open this port. In the first time we reached
                                            //it we didn't opened all the hca ports because we don't
                                            //know if we are in the same subnet
        rc = this->DiscoverFabricOpenCAPorts(p_node, p_direct_route, p_node_info, is_root, p_bad_direct_route_info,
                                             push_new_direct_route);
        if (rc)
            IBDIAG_RETURN(rc);
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


// Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_IBDM_ERR / ERR_CODE_TRY_TO_DISCONNECT_CONNECTED_PORT
int IBDiag::DiscoverFabricBFSCreateLink(IN direct_route_t * p_direct_route,
        IN IBPort * p_port)
{
    IBDIAG_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
               "DiscoverFabricBFSCreateLink  direct_route:%s"
               "  ports node name%s, lid:%d port number:%d\n",
               Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
               p_port->p_node->name.c_str(), p_port->base_lid, p_port->num);

    direct_route_t prev_direct_route = *p_direct_route;
    prev_direct_route.path.BYTE[p_direct_route->length - 1] = 0;
    --prev_direct_route.length;
    IBNode *p_prev_node = this->GetNodeByDirectRoute(&prev_direct_route);
    if (!p_prev_node) {
        this->SetLastError("DB error - can't find node for prev direct route = %s", Ibis::ConvertDirPathToStr(&prev_direct_route).c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
               "prev_node of direct_route:%s is %s\n",
               Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
               p_prev_node->name.c_str());

    IBPort* p_prev_port = p_prev_node->getPort(p_direct_route->path.BYTE[p_direct_route->length - 1]);
    if (!p_prev_port) {
        this->SetLastError("DB error - can't find port=%u for prev node=%s",
                p_direct_route->path.BYTE[p_direct_route->length - 1],
                p_prev_node->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    //check if not connected already
    if (p_prev_port->p_remotePort) {
        this->SetLastError("Failure - reached connected port when trying to connect it to another one"  \
                " port=%s is already connected to port=%s, therefore we can't connect it to port=%s\n",
                p_prev_port->getName().c_str(), p_prev_port->p_remotePort->getName().c_str(), p_port->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_TRY_TO_DISCONNECT_CONNECTED_PORT);
    }
    if (p_port->p_remotePort) {
        this->SetLastError("Failure - reached connected port when trying to connect it to another one"  \
                " port=%s is already connected to port=%s, therefore we can't connect it to port=%s\n",
                p_port->getName().c_str(), p_port->p_remotePort->getName().c_str(), p_prev_port->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_TRY_TO_DISCONNECT_CONNECTED_PORT);
    }

    if (this->discovered_fabric.makeLinkBetweenPorts(p_prev_port, p_port)) {
        this->SetLastError("Failed to create a link");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

// Returns: SUCCESS_CODE / ERR_CODE_FABRIC_ERROR / ERR_CODE_IBDM_ERR / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM
int IBDiag::DiscoverFabricBFSOpenNode(IN direct_route_t *p_direct_route,
        IN bool is_root,
        OUT IBNode **p_pnode,
        OUT struct SMP_NodeInfo *p_node_info,
        OUT bool *is_visited_node,
        IN  ProgressBar *p_progress_bar,
        OUT IbdiagBadDirectRoute_t *p_bad_direct_route_info,
        IN  bool send_node_info)
{
    IBDIAG_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
               "DiscoverFabricBFSOpenNode  direct_route:%s is_root:%d send_node_info: %d\n",
               Ibis::ConvertDirPathToStr(p_direct_route).c_str(), is_root, send_node_info);

    //NodeInfo may be already filled in.
    if (send_node_info) {
        if (this->ibis_obj.SMPNodeInfoMadGetByDirect(p_direct_route, p_node_info)) {
            /*this->SetLastError("Failed to get node information for direct route %s, err=%s",
                    Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                    this->ibis_obj.GetLastError());*/
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                    "Failed to get node information for direct route %s, err=%s\n",
                    Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                    this->ibis_obj.GetLastError());
                    p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_NODE_INFO;
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }
    }
    if (is_root)
        this->root_port_num = p_node_info->LocalPortNum;

    if (this->IsInvalidGUID(p_node_info->NodeGUID))
            this->errors.push_back(
                    new FabricInvalidNodeGuid(p_node_info->NodeGUID,
                                              Ibis::ConvertDirPathToStr(p_direct_route)));

    if (this->IsInvalidGUID(p_node_info->PortGUID))
            this->errors.push_back(
                    new FabricInvalidPortGuid(p_node_info->PortGUID,
                                              Ibis::ConvertDirPathToStr(p_direct_route)));

    //////////////////////////////////////////////////
    //Duplicated GUID algorithm
    //////////////////////////////////////////////////
    direct_route_t *p_old_direct_route;
    bool is_visited_port;
    /* is_visited_node is a income param.
     * we will use it latter to determine if we want to open a switch ports */
    if (this->check_duplicated_guids &&
        (this->check_switch_duplicated_guids || p_node_info->NodeType != IB_SW_NODE)) {
        bool duplicated_node_guid;
        bool duplicated_port_guid;
        int rc = this->IsDuplicatedGuids(p_direct_route, p_node_info,
                &duplicated_node_guid, &duplicated_port_guid,
                is_visited_node, &is_visited_port, &p_old_direct_route,
                p_bad_direct_route_info);
        if (rc) {
            if (p_bad_direct_route_info->reason == IBDIAG_BAD_DR_NONE) {
                p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_DUPLICATE_CHECK;
            }
            IBDIAG_RETURN(rc);
        }
        if (duplicated_node_guid || !*is_visited_node)
            this->MarkNodeGuidAsBFSKnown(p_node_info->NodeGUID, p_direct_route);
        if (duplicated_port_guid || !is_visited_port) {
            this->MarkPortGuidAsBFSKnown(p_node_info->PortGUID, p_direct_route);
        }

        IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, "\nDuplicated node GUID check - result=%s\n Duplicated port GUID check - result=%s\n Visited Node - result=%s\n",
                        duplicated_node_guid ? "True" : "False", duplicated_port_guid ? "True" : "False", *is_visited_node ? "True" : "False");
    } else {    //assume excellent fabric
        *is_visited_node = false;
        list_p_direct_route old_p_direct_routers_list =
                this->GetDirectRoutesByNodeGuid(p_node_info->NodeGUID);
        if (!old_p_direct_routers_list.size()) {
            this->MarkNodeGuidAsBFSKnown(p_node_info->NodeGUID, p_direct_route);
        } else {
            *is_visited_node = true;
            p_old_direct_route = old_p_direct_routers_list.front();
        }

        old_p_direct_routers_list =
                this->GetDirectRoutesByPortGuid(p_node_info->PortGUID);
        if (!old_p_direct_routers_list.size()) {
            this->MarkPortGuidAsBFSKnown(p_node_info->PortGUID, p_direct_route);
        }
    }

    SMP_NodeDesc curr_node_desc;
    if (*is_visited_node == false) {
        if (this->send_node_desc) {
            if (this->ibis_obj.SMPNodeDescMadGetByDirect(p_direct_route, &curr_node_desc)) {
                   /*this->SetLastError("Failed to get node description for direct route %s, err=%s",
                    Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                    this->ibis_obj.GetLastError());*/
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                          "Failed to get node description for direct route %s, err=%s\n",
                          Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                          this->ibis_obj.GetLastError());

                p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_NODE_DESCRIPTION;
                IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
            }
        }

        string additional_info;
        if (!IsValidNodeInfoData(p_node_info, additional_info)) {
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_BAD_NODE_INFO;
            p_bad_direct_route_info->message = additional_info;
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }

        //add this node to discovered fabric
        *p_pnode = this->discovered_fabric.makeNode((IBNodeType)p_node_info->NodeType,
                p_node_info->NumPorts,
                p_node_info->SystemImageGUID,
                p_node_info->NodeGUID,
                p_node_info->VendorID,
                p_node_info->DeviceID,
                p_node_info->revision,
                this->send_node_desc ? (char *)curr_node_desc.Byte: "",
                true);

        if (!*p_pnode) {
            int rc = 0;
            IbdiagBadDirectRouteReason_t reason = IBDIAG_BAD_DR_NONE;
            stringstream strStream;
            //if the flag "this->check_duplicated_guids" is turned on
            //then in a case of duplicated guids
            //the this->bfs_known_node_guids should keep a list of different "direct routes"
            map_guid_list_p_direct_route::iterator nI =
                            this->bfs_known_node_guids.find(p_node_info->NodeGUID);
            if (nI != this->bfs_known_node_guids.end() && nI->second.size() > 1) {
                strStream << "Duplicated GUIDs were found";

                for (list_p_direct_route::iterator dr_it = nI->second.begin();
                     dr_it != nI->second.end(); dr_it++)
                    strStream << "; Node GUID = "
                            << PTR(p_node_info->NodeGUID)
                            << " is duplicated at DR = "
                            << Ibis::ConvertDirPathToStr(*dr_it);
                rc = IBDIAG_ERR_CODE_DUPLICATED_GUID;
                reason = IBDIAG_BAD_DR_NODE_DUPLICATE_CHECK;
            } else {
                strStream << "Failed to store new node for direct route ="
                         << Ibis::ConvertDirPathToStr(p_direct_route)
                         << "; system GUID = "
                         << PTR(p_node_info->SystemImageGUID)
                         << ", GUID = "
                         << PTR(p_node_info->NodeGUID);
                rc = IBDIAG_ERR_CODE_IBDM_ERR;
                reason = IBDIAG_BAD_DR_NODE_INCONSISTENT;
            }

            this->SetLastError(strStream.str().c_str());
            p_bad_direct_route_info->reason = reason;
            IBDIAG_RETURN(rc);
        }
        p_progress_bar->push(*p_pnode);

        //save smp node info attribute in the fabric
        int rc = this->fabric_extended_info.addSMPNodeInfo(*p_pnode, *p_node_info);
        if (rc) {
            this->SetLastError("Failed to store smp_node_info for node in direct route %s, err=%s",
                     Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                     this->fabric_extended_info.GetLastError());
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_INCONSISTENT;
            IBDIAG_RETURN(rc);
        }
        p_progress_bar->complete(*p_pnode);
    } else {
        //get ref to exist one without consider all data given as args
        *p_pnode = this->GetNodeByDirectRoute(p_old_direct_route);
        //it's possible, if the node doesn't respond to 'node description' MAD,
        //then we won't create a node for it
        if (!*p_pnode) {
            p_bad_direct_route_info->reason = IBDIAG_BAD_DR_NODE_INCONSISTENT;
            //keep discovering fabric despite the fail
            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

// This function checks for valid node info data
bool IBDiag::IsValidNodeInfoData(IN struct SMP_NodeInfo *p_node_info,
                                 OUT string &additional_info)
{
    IBDIAG_ENTER;
    stringstream sstm;
    bool valid_node_info_data;

    //check that the number of port is valid
    if (!p_node_info)
        IBDIAG_RETURN(false);

    valid_node_info_data = true;
    if (p_node_info->NumPorts > IB_MAX_PHYS_NUM_PORTS ||
        p_node_info->NumPorts < IB_MIN_PHYS_NUM_PORTS) {
        valid_node_info_data = false;
        sstm << "Wrong number of ports " << (unsigned int)(p_node_info->NumPorts);
        additional_info = sstm.str();
    }
    IBDIAG_RETURN(valid_node_info_data);
}

u_int16_t IBDiag::GetPathNextNode(IN IBNode **p_node,
        IN lid_t dest_lid,
        INOUT direct_route_t **p_direct_route,
        IN direct_route_t *dr_path,
        IN struct SMP_NodeInfo &curr_node_info)
{
    struct SMP_LinearForwardingTable linear_forwarding_table;
    CLEAR_STRUCT(linear_forwarding_table);

    direct_route_t *new_p_direct_route = new direct_route_t;
    *new_p_direct_route = **p_direct_route;
    *p_direct_route = new_p_direct_route;

    //if ibdiagpath by direct route.
    if (dr_path->length > 0) {
        new_p_direct_route->path.BYTE[new_p_direct_route->length] =
            dr_path->path.BYTE[new_p_direct_route->length];
        new_p_direct_route->length++;
    } else {
        //must be the first local node or the source node
        if((*p_node)->type != IB_SW_NODE) {
            //if it is the first HCA/RTR node
            if (new_p_direct_route->length == 1) {
                new_p_direct_route->path.BYTE[new_p_direct_route->length] =
                    curr_node_info.LocalPortNum;
                new_p_direct_route->length++;
            //if we got to HCA source we need to got back to the switch
            } else {
                new_p_direct_route->path.BYTE[new_p_direct_route->length] = 0;
                new_p_direct_route->length--;
            }
        } else { //it is a switch
            if (this->ibis_obj.
                    SMPLinearForwardingTableGetByDirect(new_p_direct_route,
                        (lid_t)(dest_lid/IBIS_IB_MAD_SMP_LFT_NUM_BLOCKS),
                        &linear_forwarding_table,
                        NULL))
                IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
            u_int8_t next_port = linear_forwarding_table.Port[dest_lid % 64];
            new_p_direct_route->path.BYTE[new_p_direct_route->length] = next_port;
            new_p_direct_route->length++;
        }
    }
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

bool IBDiag::IsVirtualLidForNode(IBNode *p_node,
                                 lid_t lid,
                                 stringstream *pss)
{
    IBDIAG_ENTER;
    int rc = 0;
    char buffer[256];

    CLEAR_STRUCT(buffer);
    snprintf(buffer, sizeof(buffer), "-I- Looking lid=%d in node %s vlid list\n",
            lid, p_node->name.c_str());
    (*pss) << buffer;

    ibDiagClbck.Set(this, &fabric_extended_info, &errors, NULL);

    // Virtualization APIs get IBNode map as input instead of IBNode object.
    // Use IBNode map contains specific single IBNode object in this flow
    // for backward compatibility.
    map_str_pnode cur_node_map;
    cur_node_map[p_node->name] = p_node;
    //TODO chage to collecting virtual lid data in REAL asynchronous way.
    //First "for" on all ports and than "for" on all lids.
    rc = BuildVirtualizationBlock(&IBDiag::BuildVirtualizationInfoDB, cur_node_map, false, false);
    if(rc)
       IBDIAG_RETURN(rc);

    // second stage to get only the not down vport.
    rc = BuildVirtualizationBlock(&IBDiag::BuildVPortStateDB, cur_node_map, false, false);
    if(rc)
       IBDIAG_RETURN(rc);

    // third stage to get the VPort Info only for not down vports
    rc = BuildVirtualizationBlock(&IBDiag::BuildVPortInfoDB, cur_node_map, false, false);
    if(rc)
       IBDIAG_RETURN(rc);

    // fourth stage to get the VNode Info
    rc = BuildVirtualizationBlock(&IBDiag::BuildVNodeInfoDB, cur_node_map, false, false);
    if(rc)
       IBDIAG_RETURN(rc);

    // fifth stage to get the VNode Description
    rc = BuildVNodeDescriptionDB(p_node, false);
    if(rc != IBDIAG_SUCCESS_CODE) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,"Error during get on VNode" \
                   " Description on node %s skipping to next node\n",
                   p_node->name.c_str());
        IBDIAG_RETURN(rc);
    }

    rc = CheckAndSetVPortLid(errors);
    if(rc != IBDIAG_SUCCESS_CODE) {
        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                   "Error during get on CheckAndSetVPortLid" \
                   " Description on node %s skipping to next node\n",
                   p_node->name.c_str());
        IBDIAG_RETURN(rc);
    }

    for (unsigned int i = 1 ; i <=  p_node->numPorts ; i++) {
        IBPort *p_curr_port = p_node->getPort((phys_port_t)i);
        if (!p_curr_port || !p_curr_port->is_data_worthy()) //port not exist
            continue;

        map_vportnum_vport vports = p_curr_port->VPorts;
        for (map_vportnum_vport::iterator vpI = vports.begin();
                vpI != vports.end(); ++vpI) {
            IBVPort *p_vport = vpI->second;
            if(!p_vport)
                continue;
            if(p_vport->get_vlid() == lid) {
                CLEAR_STRUCT(buffer);
                snprintf(buffer, sizeof(buffer), "-I- Found vlid=%d in node %s\n",
                        lid, p_node->name.c_str());
                (*pss) << buffer;

                IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
            }
        }
    }

    IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
}

int IBDiag::DiscoverPath(IN u_int8_t max_hops, IN lid_t src_lid,
                         IN lid_t dest_lid, direct_route_t &dr_path,
                         OUT stringstream *pss)
{
    IBDIAG_ENTER;
    direct_route_t *p_current_direct_route = NULL;
    IbdiagBadDirectRoute_t bad_direct_route;
    struct SMP_NodeInfo curr_node_info;
    CLEAR_STRUCT(curr_node_info);
    bool is_visited_node;
    IBNode *p_node;
    ProgressBarDiscover progress_bar;
    int rc,rc2;
    u_int8_t current_hop = 0;
    lid_t target_lid;
    bool found_src = false;
    IBPort *p_port;
    vector <IBNode *> sub_nodes;
    vector <IBPort *> sub_ports;
    char buffer[256];

    //Allow sending NodeDesc MADs. This function prints discovered node's names.
    //And they are based on data provided by the MAD.
    this->send_node_desc = true;

    //Enables sending PortInfo MADs for HCA/Routers.
    //This function works with lids and we have to send PortInfo MADs for HCA/Routers.
    this->send_port_info = true;

    if (dr_path.length == 0 && (!src_lid || !dest_lid)) {
        this->SetLastError("Options src_lid and dest_lid or dr_path are"  \
                  " mandatory to run ibdiagapth");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
    }

    if (this->IsReady() == false) {
        this->SetLastError("IBDiag initialize wasn't done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NOT_READY);
    }
    if (max_hops > IBDIAG_MAX_HOPS) {
        this->SetLastError("MAX HOPS given is invalid, valid range is between 0 and 64");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
    }
    //clean previous data
    this->CleanUpInternalDB();

    //initial step
    p_current_direct_route = new direct_route_t;
    CLEAR_STRUCT(*p_current_direct_route);
    p_current_direct_route->length = 1;

    bad_direct_route.clear();

    rc = this->DiscoverFabricBFSOpenNode(p_current_direct_route, true,
                                         &p_node, &curr_node_info,
                                         &is_visited_node,
                                         &progress_bar,
                                         &bad_direct_route);
    if (rc) {
        this->SetLastError("Failed to open root node");
        rc2 = this->AddBadPath(&bad_direct_route, p_current_direct_route);
        if (rc2) {
            IBDIAG_RETURN(rc2);
        }
        if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
            this->SetLastError("Sanity check failed for root node discovery "
                               "failure reason");
        }
        IBDIAG_RETURN(rc);
    }
    this->root_node = p_node;
    rc = this->DiscoverFabricBFSOpenPorts(p_current_direct_route,
                                          p_node, &curr_node_info,
                                          false, true, &bad_direct_route,
                                          false);
    if (rc) {
        this->SetLastError("Failed to open ports of root node");
        rc2 = this->AddBadPath(&bad_direct_route, p_current_direct_route);
        if (rc2) {
            IBDIAG_RETURN(rc2);
        }
        if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
            this->SetLastError("Sanity check failed for root node ports"
                               " opening failure reason");
        }
        IBDIAG_RETURN(rc);
    }

    //first the target lid is the source lid and once the source is
    //found the target lid is the destination lid
    target_lid = src_lid;
    p_port = p_node->getPort(curr_node_info.LocalPortNum);
    if (!p_port) {
        this->SetLastError("DB error - failed to get port instance");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    /*if we run ibdiagpath using lid route and the root node is the source
     * or if we run ibdiagpath with direct route so the root node must be
     * the source*/
    if (p_port->is_lid_in_lmc_range(src_lid) || dr_path.length > 0) {
        found_src = true;
        target_lid = dest_lid;
        sub_nodes.push_back(p_node);
        sub_ports.push_back(p_port);
          (*pss) << TRAVERSE_SRC_TO_DEST;
    } else
          (*pss) << TRAVERSE_LOCAL_TO_SRC;

    rc = this->GetPathNextNode(&p_node,
                               (u_int32_t)target_lid, &p_current_direct_route,
                               &dr_path, curr_node_info);

    //discover path
    while (current_hop < max_hops) {

        IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, "Current BFS direct route = %s\n",
                   Ibis::ConvertDirPathToStr(p_current_direct_route).c_str());

        //open the node we reached to
        rc2 = this->DiscoverFabricBFSOpenNode(p_current_direct_route, false,
                &p_node, &curr_node_info, &is_visited_node,
                &progress_bar, &bad_direct_route);
        if (rc2) {
            this->SetLastError("Discovery failed for Lid %d when Quering LFT " \
                    "in Node %s", target_lid, p_node->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
        }
        //open the ports of this node
        rc2 = this->DiscoverFabricBFSOpenPorts(p_current_direct_route, p_node,
                                               &curr_node_info, is_visited_node,
                                               false, &bad_direct_route, false);
        if (rc2) {
            this->SetLastError("Sanity check failed for ports discovery"
                               " failure reason (DR %s) ",
                    (Ibis::ConvertDirPathToStr(p_current_direct_route)).c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
        }

        /*
         *  TO: p_port FROM: p_port->remote.
         *  p_port - current port, entry port to the current node.
         *  p_port->remote - previous port, exit port from previous node.
         *  When we get to the src, we may need to go back to already visited
         *  node. If the src is switch, then the new diect route holds the exit
         *  port from the switch. If it's HCA, then the new direct route is an
         *  old one without last entry, thus we get nodeinfo of visited node,
         *  on its old entry port. To get the right p_port, we take the exit
         *  port of that node, through which we got to current HCA, i.e current
         *  p_port->remote.
         */
        if (found_src && p_port->p_node->type != IB_SW_NODE &&
            p_port->p_remotePort)
            p_port = p_port->p_remotePort;
        else
            p_port = p_node->getPort(curr_node_info.LocalPortNum);

        if (!p_port) {
            this->SetLastError("DB error - failed to get port instance");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        /* Check that the link doesn't exist yet.
         * It is possible, that we already created the link while traversing
         * from local to source, so don't create it again when traversing
         * from source to destination.
         * This also covers case, where link is created between current node
         * and another already visited node from previous traversing
         * (example: running from dest to src, and then back).
         */
        if (!p_port->p_remotePort)
            rc2 = this->DiscoverFabricBFSCreateLink(p_current_direct_route, p_port);
        if (rc2)
            IBDIAG_RETURN(rc2);

        if (found_src) {
            sub_nodes.push_back(p_node);
            sub_ports.push_back(p_port);
            sub_ports.push_back(p_port->p_remotePort);
        }

        CLEAR_STRUCT(buffer);
        snprintf(buffer, sizeof(buffer), "-I- From: lid=%02d port guid=" U64H_FMT
                " dev=%d %s Port=%d\n",
                p_port->p_remotePort->is_lid_in_lmc_range(src_lid) ? src_lid : p_port->p_remotePort->base_lid,
                p_port->p_remotePort->guid_get(),
                p_port->p_remotePort->p_node->devId,
                p_port->p_remotePort->p_node->name.c_str(), p_port->p_remotePort->num);
        (*pss) << buffer;
        CLEAR_STRUCT(buffer);
        snprintf(buffer, sizeof(buffer), "-I- To : lid=%02d port guid=" U64H_FMT
                " dev=%d %s Port=%d\n",
                p_port->is_lid_in_lmc_range(target_lid) ? target_lid: p_port->base_lid,
                p_port->guid_get(), p_port->p_node->devId,
                p_port->p_node->name.c_str(), p_port->num);
        (*pss) << buffer;

        if (dr_path.length > 0 && p_current_direct_route->length == dr_path.length)
            break;
        else if(found_src) {
            if (p_port->is_lid_in_lmc_range(dest_lid))
                break;
            //if we already found source and we got to HCA that isn't the
            //distination it should be virtual lid or routing error from the switch LFT
            else if (p_node->type == IB_CA_NODE) {
                //check that this is not virtual node.
                if(IsVirtualLidForNode(p_node, dest_lid, pss))
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
                break;
            }
        } else if (p_port->is_lid_in_lmc_range(src_lid) ||
                   (p_port->base_lid != src_lid && p_node->type == IB_CA_NODE &&
                    !IsVirtualLidForNode(p_node, src_lid, pss))) {
                    found_src = true;
                    target_lid = dest_lid;
                    (*pss) << TRAVERSE_SRC_TO_DEST;
                    sub_nodes.push_back(p_node);
        }

        current_hop++;
        //////////////////////////////////////////////////
        //stop condition
        //////////////////////////////////////////////////
        if (current_hop >= max_hops) {
             IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER,"\nReached maximum hops=%u during"
                       " discovery", max_hops);
            FabricErrDiscovery *p_curr_fabric_err =
                new FabricErrDiscovery(
                        p_port->p_node,
                        max_hops);

            p_curr_fabric_err->SetLevel(EN_FABRIC_ERR_WARNING);
            errors.push_back(p_curr_fabric_err);

            goto finished_discovery;
        }

        rc = this->GetPathNextNode(&p_node,
                                   (u_int32_t)target_lid,
                                   &p_current_direct_route, &dr_path,
                                   curr_node_info);
    }       /* BFS while */

    this->ApplySubCluster(sub_nodes, sub_ports);

finished_discovery:
    this->PostDiscoverFabricProcess();
    IBDIAG_RETURN(rc);
}

/*
 * Remove all nodes that are in subCluster from Fabric::NodeByName and extended
 * fabric db.
 * Those are the two main places, that our stages check for nodes in fabric.
 * We don't delete the actual objects from db, since they are needed to find
 * nodes by direct route in function GetNodeByDirectRoute (otherwise there
 * would be "holes" in db).
 */
int IBDiag::ApplySubCluster(IN vector_p_node sub_nodes,
                            IN vector_p_port sub_ports)
{
    //for all nodes mark with "in_sub_fabric = true" only the ones in the path
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node) {
            cout << "-E-  found null node in nodes_vector " << endl;
            return 1;
        }
        //if the node isn't in sub_nodes list it's isnt in sub cluster.
        if(std::find(sub_nodes.begin(),
                    sub_nodes.end(),
                    p_node) == sub_nodes.end())
            p_node->setInSubFabric(false);
        else {
            uint8_t i = 0;
            if (p_node->type == IB_CA_NODE)
                i = 1;
            //set in_sub_cluster = false to unrelevant ports to the path
            for (;i <= p_node->numPorts; i++) {
                IBPort *p_port = p_node->getPort(i);
                if(!p_port) {
                    cout << "-E-  found null port in ports_vector " << endl;
                    return 1;
                }
                if(std::find(sub_ports.begin(),
                            sub_ports.end(),
                            p_port) == sub_ports.end())
                    p_port->setInSubFabric(false);
            }
        }
    }

    //remove all nodes that not in the path from NodeByName
    this->discovered_fabric.applySubCluster();

    //remove all node that not in the path from nodes_vector & ports_vector
    this->fabric_extended_info.applySubCluster();

    return 0;
}


/****************************************************/
/*******************public methods*******************/
/****************************************************/
class RoutesNodesListGuard {
    IBDiag::list_route_and_node_info& m_list;

    public:
        RoutesNodesListGuard(IBDiag::list_route_and_node_info & lst)
            : m_list(lst){}

        ~RoutesNodesListGuard() {
            for (auto & x : m_list)
                delete x.p_direct_route;
        }
};

// Returns: SUCCESS_CODE / ERR_CODE_FABRIC_ERROR / ERR_CODE_IBDM_ERR / ERR_CODE_INCORRECT_ARGS / ERR_CODE_NO_MEM / ERR_CODE_DB_ERR / ERR_CODE_TRY_TO_DISCONNECT_CONNECTED_PORT / ERR_CODE_INCORRECT_ARGS
int IBDiag::DiscoverFabric(IN u_int8_t max_hops)
{
    IBDIAG_ENTER;
    direct_route_t *p_bfs_direct_route = NULL;
    IbdiagBadDirectRoute_t bad_direct_route;
    struct SMP_NodeInfo curr_node_info;
    bool is_visited_node;
    IBNode *p_node;
    ProgressBarDiscover progress_bar;
    int rc,rc2, rc3;

    CLEAR_STRUCT(curr_node_info);

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Start DiscoverFabric max_hops:%d\n",
               max_hops);

    if (this->IsReady() == false) {
        this->SetLastError("IBDiag initialize wasn't done");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_NOT_READY);
    }
    if (max_hops > IBDIAG_MAX_HOPS) {
        this->SetLastError("MAX HOPS given is invalid, valid range is between 0 and 64");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_INCORRECT_ARGS);
    }

    //clean previous data
    this->CleanUpInternalDB();

    //initial step
    p_bfs_direct_route = new direct_route_t;
    CLEAR_STRUCT(*p_bfs_direct_route);
    p_bfs_direct_route->length = 1;

    bad_direct_route.clear();

    rc = this->DiscoverFabricBFSOpenNode(p_bfs_direct_route, true,
            &p_node, &curr_node_info, &is_visited_node,
            &progress_bar, &bad_direct_route);
    if (rc) {
        this->SetLastError("Failed to open root node");
        rc2 = this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
        if (rc2) {
            IBDIAG_RETURN(rc2);
        }
        if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
            this->SetLastError("Sanity check failed for root node discovery failure reason");
        }
        IBDIAG_RETURN(rc);
    }
    this->root_node = p_node;
    rc = this->DiscoverFabricBFSOpenPorts(p_bfs_direct_route, p_node, &curr_node_info, false, true, &bad_direct_route);
    if (rc) {
        this->SetLastError("Failed to open ports of root node");
        rc2 = this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
        if (rc2) {
            IBDIAG_RETURN(rc2);
        }
        if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
            this->SetLastError("Sanity check failed for root node ports opening failure reason");
        }
        IBDIAG_RETURN(rc);
    }
    this->AddGoodPath(p_bfs_direct_route);

    //bfs on all the fabric
    while (!this->bfs_list.empty()) {
        //list_p_direct_route level_routes_list;
        list_route_and_node_info level_routes_and_node_info_list;
        RoutesNodesListGuard guard(level_routes_and_node_info_list);

        if (this->GetAndValidateLevelRoutes(level_routes_and_node_info_list, max_hops)) {
            goto finished_discovery;
        }

        SendAndWaitForAllNodeInfo(level_routes_and_node_info_list);

        while(!level_routes_and_node_info_list.empty()) {
            DirectRouteAndNodeInfo entry = *level_routes_and_node_info_list.begin();
            level_routes_and_node_info_list.pop_front();
            p_bfs_direct_route = entry.p_direct_route;
            SMP_NodeInfo *p_node_info = &entry.node_info;

            //optimization - this direct route can be reached already from other side
            if (this->GetNodeByDirectRoute(p_bfs_direct_route)) {
                this->AddLoopPath(p_bfs_direct_route);
                continue;
            }

            if (!entry.is_filled) {
                bad_direct_route.reason = IBDIAG_BAD_DR_NODE_NODE_INFO;
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

                rc3 = this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
                if (rc3) {
                    IBDIAG_RETURN(rc3);
                }

                // clear for next time
                bad_direct_route.clear();

                //todo fill this in
                continue;
            }

            IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER, "Current BFS direct route = %s\n",
                    Ibis::ConvertDirPathToStr(p_bfs_direct_route).c_str());

            //open the node we reached to
            rc2 = this->DiscoverFabricBFSOpenNode(p_bfs_direct_route, false,
                                &p_node, p_node_info, &is_visited_node,
                                &progress_bar, &bad_direct_route, false);
            if (rc2 == IBDIAG_ERR_CODE_FABRIC_ERROR) {       //bad link, we will try to coninue
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                rc3 = this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
                if (rc3) {
                    IBDIAG_RETURN(rc3);
                }
                if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
                    this->SetLastError("Sanity check failed for node discovery failure reason (DR %s) ",
                                       (Ibis::ConvertDirPathToStr(p_bfs_direct_route)).c_str());
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
                }
                // clear for next time
                bad_direct_route.clear();
                continue;
            } else if (rc2) {    //error
                this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
                IBDIAG_RETURN(rc2);
            }

            //open the ports of this node
            rc2 = this->DiscoverFabricBFSOpenPorts(p_bfs_direct_route, p_node,
                                p_node_info, is_visited_node, false, &bad_direct_route);
            if (rc2 == IBDIAG_ERR_CODE_FABRIC_ERROR) {       //bad link, we will try to coninue
                rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
                rc3 = this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
                if (rc3) {
                    IBDIAG_RETURN(rc3);
                }
                if (bad_direct_route.reason == IBDIAG_BAD_DR_NONE) {
                    this->SetLastError("Sanity check failed for ports discovery failure reason (DR %s) ",
                                       (Ibis::ConvertDirPathToStr(p_bfs_direct_route)).c_str());
                    IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);
                }
                // clear for next time
                bad_direct_route.clear();
                continue;
            } else if (rc2) {    //error
                this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
                IBDIAG_RETURN(rc2);
            }

            //add link to this node - from previous port to enter port
            IBPort *p_port = p_node->getPort(p_node_info->LocalPortNum);
            if (!p_port) {
                this->SetLastError("DB error - failed to get port instance");
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            rc2 = this->DiscoverFabricBFSCreateLink(p_bfs_direct_route, p_port);
            if (rc2) {
                this->AddBadPath(&bad_direct_route, p_bfs_direct_route);
                IBDIAG_RETURN(rc2);
            }

            //mark good path or loop path if node already visited
            if (is_visited_node == false)
                this->AddGoodPath(p_bfs_direct_route);
            else
                this->AddLoopPath(p_bfs_direct_route);
        }    //for on levels
    }  /* BFS while */


finished_discovery:
    this->PostDiscoverFabricProcess();
    IBDIAG_RETURN(rc);
}

int IBDiag::SendNodeInfoMad(IN OUT IBDiag::NodeInfoSendData &node_info_send_data)
{
    IBDIAG_ENTER;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_data1 = (void*)&node_info_send_data;
    clbck_data.m_handle_data_func =
              &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPNodeInfoGetClbck>;

    //this function is also called from a callback.
    //So, we have to verify we are not at list's end
    if (node_info_send_data.position_itr != node_info_send_data.end_itr) {
        DirectRouteAndNodeInfo &entry = *(node_info_send_data.position_itr);
        direct_route_t *p_direct_route = entry.p_direct_route;

        clbck_data.m_data2 = (void*)(&entry);
        //call is done asynchroniously; so we advance iterrator before sending the MAD
        node_info_send_data.position_itr++;

        SMP_NodeInfo node_info;
        if (this->ibis_obj.SMPNodeInfoMadGetByDirect(p_direct_route, &node_info, &clbck_data)) {
                /*this->SetLastError("Failed to get node information for direct route %s, err=%s",
                            Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                            this->ibis_obj.GetLastError());*/
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                        "Failed to get node information for direct route %s, err=%s\n",
                        Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                        this->ibis_obj.GetLastError());

            IBDIAG_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


void IBDiag::SendAndWaitForAllNodeInfo(IN OUT list_route_and_node_info &level_routes_and_node_info_list)
{
    //we have to wait at any event untill all callbacks are done
    //no return on any errors
    IBDIAG_ENTER;

    //allocated on a stack and is passed to callbacks
    //but we wait in this function untill all callbacks are done.
    NodeInfoSendData  node_info_send_data(level_routes_and_node_info_list);

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Start sending: %lu NodeInfo MADs, pack size: %d, \n",
               level_routes_and_node_info_list.size(), this->max_node_info_mads_in_pack);

    ibDiagClbck.Set(this, NULL, NULL);

    //we send MADS utill we reach the list's end or we get to max possible number of MADS to be send
    //when a callback function is called we send one additional MAD.
    int sentMadsCount = 0;
    while (node_info_send_data.position_itr != node_info_send_data.end_itr &&
        sentMadsCount < this->max_node_info_mads_in_pack) {

        if (!this->SendNodeInfoMad(node_info_send_data)) {
            sentMadsCount++;
        }
    }

    this->ibis_obj.MadRecAll();
    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Waited for MADs\n");

    IBDIAG_RETURN_VOID;
}

int IBDiag::GetAndValidateLevelRoutes(OUT list_route_and_node_info &level_routes_and_node_info_list,
                                      IN u_int8_t max_hops)
{
    IBDIAG_ENTER;

    while (!this->bfs_list.empty()) {
        direct_route_t *p_direct_route = this->BFSPopPath();

        //////////////////////////////////////////////////
        //stop condition
        //////////////////////////////////////////////////
        if (p_direct_route->length > max_hops) {
            level_routes_and_node_info_list.clear();

            IBDIAG_LOG(TT_LOG_LEVEL_DISCOVER,
                   "Reached maximum hops=%u during discovery for the route: %s\n",
                    max_hops, Ibis::ConvertDirPathToStr(p_direct_route).c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_CHECK_FAILED);
        }

        DirectRouteAndNodeInfo entry;
        entry.p_direct_route = p_direct_route;
        level_routes_and_node_info_list.push_back(entry);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::PrintAllRoutes()
{
    IBDIAG_ENTER;
    list_p_direct_route::iterator it;
    list_p_bad_direct_route::iterator bad_dr_it;

    printf("Good Direct Routes:\n");
    for (it = this->good_direct_routes.begin(); it != this->good_direct_routes.end(); ++it) {
        printf("%s", Ibis::ConvertDirPathToStr(*it).c_str());
        printf("\t");
    }
    printf("\n\n");
    printf("Bad Direct Routes:\n");
    for (bad_dr_it = this->bad_direct_routes.begin(); bad_dr_it != this->bad_direct_routes.end(); ++bad_dr_it) {
        printf("%s", Ibis::ConvertDirPathToStr((*bad_dr_it)->direct_route).c_str());
        printf("\t");
    }
    printf("\n\n");
    printf("Loop Direct Routes:\n");
    for (it = this->loop_direct_routes.begin(); it != this->loop_direct_routes.end(); ++it) {
        printf("%s", Ibis::ConvertDirPathToStr(*it).c_str());
        printf("\t");
    }
    printf("\n\n");
/*
    this->discovered_fabric.dump(cout);
    printf("\n\n");
*/
    IBDIAG_RETURN_VOID;
}

void IBDiag::PrintAllDirectRoutes() // useful for debug
{
    map_guid_list_p_direct_route::iterator map_it;
    list_p_direct_route::iterator list_it;

    printf("NODES:\n");

    for (map_it = bfs_known_node_guids.begin(); map_it != bfs_known_node_guids.end(); ++map_it) {
        printf("GUID: " U64H_FMT ", DR: ", map_it->first);
        for (list_it = map_it->second.begin(); list_it != map_it->second.end(); ++list_it)
            printf("%s", Ibis::ConvertDirPathToStr(*list_it).c_str());
        printf("\n");
    }

    printf("\nPORTS:\n");

    for (map_it = bfs_known_port_guids.begin(); map_it != bfs_known_port_guids.end(); ++map_it) {
        printf("GUID: " U64H_FMT ", DR: ", map_it->first);
        for (list_it = map_it->second.begin(); list_it != map_it->second.end(); ++list_it)
            printf("%s", Ibis::ConvertDirPathToStr(*list_it).c_str());
        printf("\n");
    }
    printf("\n");
}

void IBDiag::GetGoodDirectRoutes(OUT list_string& good_direct_routes)
{
    string curr_direct_route;
    for (list_p_direct_route::iterator it = this->good_direct_routes.begin();
            it != this->good_direct_routes.end();
            ++it) {
        curr_direct_route = Ibis::ConvertDirPathToStr(*it);
        good_direct_routes.push_back(curr_direct_route);
    }
}

void IBDiag::GetLoopDirectRoutes(OUT list_string& loop_direct_routes)
{
    string curr_direct_route;
    for (list_p_direct_route::iterator it = this->loop_direct_routes.begin();
            it != this->loop_direct_routes.end();
            ++it) {
        curr_direct_route = Ibis::ConvertDirPathToStr(*it);
        loop_direct_routes.push_back(curr_direct_route);
    }
}

void IBDiag::GetBadDirectRoutes(OUT list_string& bad_direct_routes)
{
    string curr_direct_route;
    for (list_p_bad_direct_route::iterator it = this->bad_direct_routes.begin();
         it != this->bad_direct_routes.end();
         ++it) {
        curr_direct_route = Ibis::ConvertDirPathToStr((*it)->direct_route);
        bad_direct_routes.push_back(curr_direct_route);
    }
}


//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildSwitchInfoDB(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;
    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &this->fabric_extended_info, &retrieve_errors);

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
              &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPSwitchInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    // go over all nodes:
    struct SMP_SwitchInfo curr_switch_info;
    CLEAR_STRUCT(curr_switch_info);
    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map for key = %s",
                               (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //check if not switch
        if (p_curr_node->type != IB_SW_NODE)
            continue;

        //get direct route for this node
        direct_route_t *p_curr_direct_route = this->GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_curr_direct_route) {
            this->SetLastError("DB error - can't find direct route to node=%s",
                               p_curr_node->getName().c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        clbck_data.m_data1 = p_curr_node;
        //send switch info mad for this switch
        progress_bar.push(p_curr_node);
        if (this->ibis_obj.SMPSwitchInfoMadGetByDirect(p_curr_direct_route,
                                                       &curr_switch_info, &clbck_data)) {
            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                    "Failed to send SwithInfo Mad to the switch named: %s, route %s\n",
                    p_curr_node->name.c_str(), Ibis::ConvertDirPathToStr(p_curr_direct_route). c_str());
        }
    }

    this->ibis_obj.MadRecAll();

    if (ibDiagClbck.GetState())
        this->SetLastError(ibDiagClbck.GetLastError());
    else if (!retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}



//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR / ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildHierarchyInfoDB(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    ProgressBarPorts progress_bar;

    ibDiagClbck.Set(this, NULL, &retrieve_errors);

    //mark nodes clean of "non-supportiveness"
    ResetAppData();
    u_int32_t port_info_cap_mask = 0;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPHierarchyInfoGetClbck>;
    clbck_data.m_data1 = NULL; // IBPort *
    clbck_data.m_data2 = NULL; // Direct Route
    clbck_data.m_data3 = NULL; // HierarchyIndex
    clbck_data.m_data4 = &this->ibis_obj;
    clbck_data.m_p_progress_bar = &progress_bar;

    SMP_HierarchyInfo curr_port_hierarchy_info;
    CLEAR_STRUCT(curr_port_hierarchy_info);

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = nI->second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in Switches");
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (p_curr_node->isSpecialNode())
            continue;

        for (u_int32_t i = 0; i <= p_curr_node->numPorts; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);

            if (!p_curr_port || p_curr_port->isSpecialPort())
                continue;

            if (ReadPortInfoCapMask(p_curr_node, p_curr_port, port_info_cap_mask))
                continue;

            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node: %s port_info_cap_mask: 0x%x\n",
                    p_curr_node->name.c_str(), port_info_cap_mask);

            if (!IS_SUPPORT_HIERARCHY_INFO(port_info_cap_mask))
                continue;

            if (i != 0)
                p_curr_node->should_support_port_hierarchy_info = true;

            direct_route_t * p_direct_route =
                this->GetDirectRouteByPortGuid(p_curr_port->guid_get());

            if (!p_direct_route) {
                this->SetLastError("DB error - can't find direct route to node=%s",
                                   p_curr_node->getName().c_str());

                rc = IBDIAG_ERR_CODE_DB_ERR;
                goto exit;
            }

            clbck_data.m_data1 = p_curr_port;
            clbck_data.m_data2 = p_direct_route;
            // HierarchyIndex 0 is used to retrieve MaxActiveIndex and MaxSupportedIndex
            clbck_data.m_data3 = (uintptr_t)0;

            progress_bar.push(p_curr_port);

            this->ibis_obj.SMPHierarchyInfoMadGetByDirect(
                (direct_route_t *)clbck_data.m_data2,
                p_curr_port->num,
                (uint8_t)((uintptr_t)(clbck_data.m_data3) & 0xFF),
                &curr_port_hierarchy_info,
                &clbck_data);
        }

        if (ibDiagClbck.GetState())
            goto exit;
    }

exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of HierarchyInfo Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "Retrieve of HierarchyInfo Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DISCOVERY_NOT_SUCCESS / ERR_CODE_DB_ERR /
// ERR_CODE_NO_MEM / ERR_CODE_FABRIC_ERROR
int IBDiag::BuildPortInfoExtendedDB(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE, rc2;

    if (!this->IsDiscoveryDone())
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DISCOVERY_NOT_SUCCESS);

    //mark nodes clean of "non-supportiveness"
    ResetAppData();
    u_int32_t port_info_cap_mask = 0;
    u_int16_t port_info_cap_mask2 = 0;

    ibDiagClbck.Set(this, &fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
        &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPPortInfoExtendedGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName " \
                               "map for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        //get cap mask for this node
        bool to_read_cap = true;

        // skip on special nodes: AN/Router/GW
        if (p_curr_node->isSpecialNode())
            continue;

        // go over all ports of this node:
        //start from port 1, no need to read port 0
        for (u_int32_t i = 1; i <= p_curr_node->numPorts; ++i) {
            IBPort *p_curr_port = p_curr_node->getPort((phys_port_t)i);
            if (!p_curr_port)                       //port not exist
                continue;
            if (!p_curr_port->p_remotePort)         //port not connected
                continue;

            // skip on ports connected to special nodes: AN/Router/GW
            if (p_curr_port->isSpecialPort())
                continue;

            SMP_PortInfo *curr_port_info =
                this->fabric_extended_info.getSMPPortInfo(
                                            p_curr_port->createIndex);
            if (!curr_port_info) {
                this->SetLastError("DB error - found connected port=%s"
                                   " without SMPPortInfo",
                                   p_curr_port->getName().c_str());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }
            //don't query non active ports
            if (curr_port_info->PortPhyState != IB_PORT_PHYS_STATE_LINK_UP)
                continue;

            //if we already have FEC for this port, no need to send the MAD
            if (p_curr_port->get_fec_mode() != IB_FEC_NA)
                continue;

            //for switch read cap only on port 0, for others read on every port
            if (to_read_cap) {
                rc2 = ReadPortInfoCapMask(p_curr_node, p_curr_port,
                                port_info_cap_mask, &port_info_cap_mask2);
                if (rc2) {
                    rc = rc2;
                    goto exit;
                }
                IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "node: %s port_info_cap_mask: 0x%x port_info_cap_mask2: 0x%x\n",
                           (*nI).first.c_str(), port_info_cap_mask, port_info_cap_mask2);
                if (p_curr_node->type == IB_SW_NODE)
                    to_read_cap = false;
            }

            //is this device supports Port Info Extended Port Info mad
            if (!IS_SUPPORT_PORT_INFO_EXTENDED_FAILED(p_curr_node->appData1.val)) {
                if (IS_SUPPORT_PORT_INFO_CAP_MASK2(port_info_cap_mask) &&
                    IS_SUPPORT_PORT_INFO_EXTENDED(port_info_cap_mask2)) {
                    SMP_PortInfoExtended curr_port_info_ext;

                    //Send the MAD
                    clbck_data.m_data1 = p_curr_port;
                    direct_route_t * p_direct_route =
                      this->GetDirectRouteByNodeGuid(p_curr_port->p_node->guid_get());
                    if (!p_direct_route) {
                        this->SetLastError("DB error - can't find direct route to node=%s",
                                           p_curr_node->getName().c_str());

                        rc = IBDIAG_ERR_CODE_DB_ERR;
                        goto exit;
                    }

                    progress_bar.push(p_curr_port);
                    this->ibis_obj.SMPPortInfoExtMadGetByDirect(
                        p_direct_route,
                        p_curr_port->num,
                        &curr_port_info_ext,
                        &clbck_data);

                } else {
                    //don't count it as an error, this is an old fw
                    //that doesn't support yet
                    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG,
                               "The node %s doesn't support " \
                               "PortInfoExtended MAD\n",
                               p_curr_node->getName().c_str());

                    //mark we already checked for support
                    p_curr_node->appData1.val |= NOT_SUPPORT_PORT_INFO_EXTENDED;
                }
            }

            if(ibDiagClbck.GetState())
                goto exit;
        }
    }
exit:
    this->ibis_obj.MadRecAll();

    if (rc) {
        if (last_error.empty())
            this->SetLastError("Retrieve of PortInfoExtended Failed.");
        else
            IBDIAG_LOG(TT_LOG_LEVEL_ERROR, "Retrieve of PortInfoExtended Failed. \n");
    } else {
        rc = ibDiagClbck.GetState();
        if (rc)
            this->SetLastError(ibDiagClbck.GetLastError());
        else if (!retrieve_errors.empty())
            rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
    }

    IBDIAG_RETURN(rc);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::ParsePSLFile(const string & file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = discovered_fabric.parsePSLFile(file_name);

    char *buffer= ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    if(!root_node->isValidPSLInfo()) {
        this->SetLastError("Can not find path SL info for local node.");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    ibis_obj.SetPSLTable(root_node->PSL);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::ParseSADumpFile(const string & file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = discovered_fabric.parseSADumpFile(file_name);

    char *buffer= ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_IBDM_ERR
int IBDiag::DumpCapabilityMaskFile(const OutputControl::Identity& identity, string& output)
{
    IBDIAG_ENTER;

    ofstream sout;
    int rc = this->OpenFile("Capability Masks",
                            identity, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    ibdmClearInternalLog();

    rc = capability_module.DumpCapabilityMaskFile(sout);

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output"
                           " for capability masks output file");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    this->CloseFile(sout);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_IBDM_ERR
int IBDiag::DumpFullCapabilityMaskFile(const OutputControl::Identity& identity, string& output)
{
    IBDIAG_ENTER;

    ofstream sout;
    int rc = this->OpenFile("Full Capability Masks",
                            identity, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    ibdmClearInternalLog();

    rc = capability_module.DumpCapabilityMaskFile(sout);
    rc += capability_module.DumpGuid2Mask(sout, this->GetDiscoverFabricPtr());

    char * buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output"
                           " for capability masks output file");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    this->CloseFile(sout);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//for debug purposes
//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / ERR_CODE_IBDM_ERR
int IBDiag::DumpGuid2Mask(const string &file_name, string& output)
{
    IBDIAG_ENTER;

    ofstream sout;
    int rc = this->OpenFile("Guid to mask",
                            file_name, sout,
                            false, //no append
                            true); //add_header

    if (rc)
        IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    ibdmClearInternalLog();

    rc = capability_module.DumpGuid2Mask(sout, this->GetDiscoverFabricPtr());

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output"
                           " for capability masks output file");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    this->CloseFile(sout);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::DumpGeneralInfoSMP2CSV(CSVOut &csv_out)
{
    IBDIAG_ENTER;

    stringstream sout;

    SmpMask &smp_mask = capability_module.GetSmpMask();
    if (csv_out.DumpStart(SECTION_GENERAL_INFO_SMP))
        IBDIAG_RETURN_VOID;

    smp_mask.DumpCSVVSGeneralInfo(sout);
    csv_out.WriteBuf(sout.str());
    csv_out.DumpEnd(SECTION_GENERAL_INFO_SMP);

    IBDIAG_RETURN_VOID;
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::ParseCapabilityMaskFile(const char *file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = capability_module.ParseCapabilityMaskFile(file_name);

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output for capability use");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / IBDIAG_ERR_CODE_IBDM_ERR/ ERR_CODE_DB_ERR
int IBDiag::ParseNodeNameMapFile(const char* file_name, string& output)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    int rc = discovered_fabric.parseNodeNameMapFile(file_name);

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output"
                           " for node name map use");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR
int IBDiag::BuildCapabilityCache(string& output)
{
    IBDIAG_ENTER;

    bool is_smp_only_fw, is_gmp_only_fw;
    fw_version_obj_t fw;
    fw.major = fw.minor = fw.sub_minor = 0;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
            nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node "
                               "in NodeByName map for key = %s",
                               (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }
        //check if the node has a longest guid prefix matched in the capability
        //mask configuration
        u_int64_t curr_node_guid = p_curr_node->guid_get();
        u_int8_t prefix_len = 0;
        u_int64_t matched_guid = 0;
        query_or_mask_t qmask;
        capability_mask_t mask;

        mask.clear();

        //SMP
        bool smp_prefix_match =
            capability_module.IsLongestSMPPrefixMatch(curr_node_guid,
                                                      prefix_len,
                                                      matched_guid,
                                                      qmask);
        //first look if it matched by longest prefix match, query or mask
        //if mask is found, it will be added now, if query, then the mask will be added later
        //anyway don't check other configurations
        if (smp_prefix_match) {
            if (!qmask.to_query)
                //store the found mask in the cache
                capability_module.AddSMPCapabilityMask(curr_node_guid, qmask.mask);
        }
        //look for match by ven id + dev id that don't support query by mad
        else if (capability_module.IsSMPUnsupportedMadDevice(p_curr_node->vendId,
                                                             p_curr_node->devId,
                                                             mask))
            capability_module.AddSMPCapabilityMask(curr_node_guid, mask);
        //see if there a device configured only with fw = 0.0.0,
        //then the mask can be already added
        //no need to ask for device's fw in the fabric
        else {
            is_smp_only_fw = false;
            if (capability_module.GetSMPFwConfiguredMask(p_curr_node->vendId,
                                                         p_curr_node->devId,
                                                         fw,
                                                         mask,
                                                         &is_smp_only_fw) == IBDIAG_SUCCESS_CODE &&
                is_smp_only_fw)
                capability_module.AddSMPCapabilityMask(curr_node_guid, mask);
        }

        //GMP
        mask.clear();
        bool gmp_prefix_match = capability_module.IsLongestGMPPrefixMatch(
                                                    curr_node_guid,
                                                    prefix_len,
                                                    matched_guid,
                                                    qmask);

        if (gmp_prefix_match) {
            if (!qmask.to_query)
                //store the found mask in the cache
                capability_module.AddGMPCapabilityMask(curr_node_guid, qmask.mask);
        }
        else if (capability_module.IsGMPUnsupportedMadDevice(p_curr_node->vendId,
                                                             p_curr_node->devId,
                                                             mask))
            //look for match by ven id + dev id that don't support query by mad
            capability_module.AddGMPCapabilityMask(curr_node_guid, mask);
        //see if there a device configured only with fw = 0.0.0,
        //then the mask can be already added
        //no need to ask for device's fw in the fabric
        else {
            is_gmp_only_fw = false;
            if (capability_module.GetGMPFwConfiguredMask(p_curr_node->vendId,
                                                         p_curr_node->devId,
                                                         fw,
                                                         mask,
                                                         &is_gmp_only_fw) == IBDIAG_SUCCESS_CODE &&
                is_gmp_only_fw)
                capability_module.AddGMPCapabilityMask(curr_node_guid, mask);
        }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / IBDIAG_ERR_CODE_IBDM_ERR
int IBDiag::ParseScopePortGuidsFile(const string & file_name,
                                    string& output,
                                    bool include_in_scope,
                                    int &num_of_lines)
{
    IBDIAG_ENTER;

    ibdmClearInternalLog();

    //parse and apply scope
    int rc = discovered_fabric.parseScopePortGuidsFile(file_name,
                                                       include_in_scope,
                                                       num_of_lines);

    char *buffer = ibdmGetAndClearInternalLog();
    if (!buffer) {
        this->SetLastError("Failed to allocate buffer for ibdm output");
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }
    output += buffer;
    free(buffer);

    //apply scope marked in fabric to extended info
    this->fabric_extended_info.applySubCluster();

    if (rc)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}


int IBDiag::FillInPortInfo(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Filling in missing PortInfo data\n");
    int rc = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &this->fabric_extended_info, &retrieve_errors);

    ProgressBarPorts progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
            &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPPortInfoGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByName map"  \
                               " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        for (phys_port_t i = 0 ; i <=  p_curr_node->numPorts; i++) {
            IBPort *p_curr_port = p_curr_node->getPort(i);

            //Skip ports that are not active; or already filled with PortInfo
            if (!p_curr_port                                                ||
                p_curr_port->getPortInfoMadWasSent()                        ||
                p_curr_port->get_internal_state() == IB_UNKNOWN_PORT_STATE  ||
                p_curr_port->get_internal_state() == IB_PORT_STATE_DOWN){

                continue;
            }

            direct_route_t *p_direct_route = this->GetDirectRouteByNodeGuid(p_curr_node->guid_get());
            if (!p_direct_route){
                this->SetLastError("Failed to get direct rote for the node with GUID: "
                                U64H_FMT,
                                p_curr_node->guid_get());
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "direct_route: %s  port %u has to be filled in\n",
                       Ibis::ConvertDirPathToStr(p_direct_route).c_str(), i);

            clbck_data.m_data1 = p_curr_port;
            clbck_data.m_data2 = p_direct_route;
            SMP_PortInfo curr_port_info;
            progress_bar.push(p_curr_port);
            this->ibis_obj.SMPPortInfoMadGetByDirect(p_direct_route, i, &curr_port_info, &clbck_data);
        }
    }

    //wait for all
    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    else if (!retrieve_errors.empty())
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}

int IBDiag::FillInNodeDescription(list_p_fabric_general_err& retrieve_errors)
{
    IBDIAG_ENTER;

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Filling in missing NodeDescriptoon data\n");
    int rc = IBDIAG_SUCCESS_CODE;
    int error_state = IBDIAG_SUCCESS_CODE;

    ibDiagClbck.Set(this, &this->fabric_extended_info, &retrieve_errors);

    ProgressBarNodes progress_bar;

    clbck_data_t clbck_data;
    clbck_data.m_p_obj = &ibDiagClbck;
    clbck_data.m_handle_data_func =
              &forwardClbck<IBDiagClbck, &IBDiagClbck::SMPNodeDescGetClbck>;
    clbck_data.m_p_progress_bar = &progress_bar;

    for (map_guid_pnode::iterator nI = this->discovered_fabric.NodeByGuid.begin();
         nI != this->discovered_fabric.NodeByGuid.end() && error_state == IBDIAG_SUCCESS_CODE; ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->SetLastError("DB error - found null node in NodeByGuid map"  \
                               " for key = %016" PRIx64, (*nI).first);
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        direct_route_t *p_direct_route = this->GetDirectRouteByNodeGuid(p_curr_node->guid_get());
        if (!p_direct_route) {
            this->SetLastError("Failed to get direct rote for the node with GUID: "
                                U64H_FMT,
                                p_curr_node->guid_get());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "direct_route: %s  Node Description is going to be changed\n",
                       Ibis::ConvertDirPathToStr(p_direct_route).c_str());

        clbck_data.m_data1 = p_curr_node;
        //clbck_data.m_data2 = p_direct_route; //for debugging only
        SMP_NodeDesc node_description;
        progress_bar.push(p_curr_node);
        if (this->ibis_obj.SMPNodeDescMadGetByDirect(p_direct_route, &node_description, &clbck_data)){
           IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "Failed to get node description for direct route %s, err=%s\n",
                      Ibis::ConvertDirPathToStr(p_direct_route).c_str(),
                      this->ibis_obj.GetLastError());
        }
    }

    this->ibis_obj.MadRecAll();

    rc = ibDiagClbck.GetState();
    if (rc)
        this->SetLastError(ibDiagClbck.GetLastError());
    //we are not checking retrieve_errors beacuse we may recive already filled list
    //else if (!retrieve_errors.empty())
    //    rc = IBDIAG_ERR_CODE_FABRIC_ERROR;

    IBDIAG_RETURN(rc);
}


#define Node_GUID(p_node)  "\"" << nodetype2char_short(p_node->type) \
    << HEX(p_node->guid_get(), 16) << "\""

#define Port_Width_And_Speed(p_port)  width2char(p_port->get_internal_width()) \
    << speed2char_name(p_port->get_internal_speed())

#define NA_REMOTE_PEER(prefix)   " \"" prefix ".................\"[.]"
#define UP_PORT_WARNING_MESSAGE(port_type)  "# Warning: the " port_type "port is UP, " \
                                    "but remote peer is not responding"

int IBDiag::CreateIBNetDiscoverFile(const std::string& file_name, warnings_list& warnings)
{
    IBDIAG_ENTER;

    this->SetLastError("");

    ofstream  sout;
    int rc = this->OpenFile("IBNetDiscover", file_name, sout, false, true);
    if (rc)
       IBDIAG_RETURN(rc);

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    IBFabric::GetSwitchLabelPortNumExplanation(sout);

    rc = this->PrintSwitchesToIBNetDiscoverFile(sout, warnings);
    if (rc) {
       sout << endl << "### DumpSwitches error: " << this->GetLastError() << endl;
       IBDIAG_RETURN(rc);
    }

    rc = this->PrintHCAToIBNetDiscoverFile(sout, warnings);
    if (rc) {
       sout << endl << "### DumpHCA error: " << this->GetLastError() << endl;
       IBDIAG_RETURN(rc);
    }

    this->CloseFile(sout);

    //the ofstream's distructor closes an opened file
    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::PrintSwitchesToIBNetDiscoverFile(ostream &sout, warnings_list& warnings)
{
    IBDIAG_ENTER;

    for (set_pnode::iterator nI = this->discovered_fabric.Switches.begin();
         nI != this->discovered_fabric.Switches.end(); ++nI) {
        IBNode *p_node = *nI;
        if (!p_node) {
           this->SetLastError("DB error - found null node in Switches set");
           IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        int rc = this->PrintNodeInfo(p_node, sout, warnings);
        if (rc)
           return rc;

        rc = this->PrintSwitchNodePorts(p_node, sout, warnings);
        if (rc)
           return rc;

        sout << endl << endl;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::PrintSwitchNodePorts(IBNode *p_node, ostream &sout,
                                 warnings_list &warnings)
{
    IBDIAG_ENTER;

    for (phys_port_t i = 1 ; i <=  p_node->numPorts; i++) {
        IBPort *p_port = p_node->getPort(i);

        if (!p_port                                               ||
            p_port->get_internal_state() == IB_UNKNOWN_PORT_STATE ||
            p_port->get_internal_state() == IB_PORT_STATE_DOWN) {
            continue;
        }

        if (!p_port->p_remotePort) {
            sout << '#' << S_BRACKETS(DEC(i))
                 << R_BRACKETS(p_port->numAsString())
                 << NA_REMOTE_PEER("")
                 << "      " << UP_PORT_WARNING_MESSAGE("")
                 << endl;

            stringstream str_stream;
            str_stream << "The port "
                       << S_BRACKETS(DEC(i)) << R_BRACKETS(p_port->numAsString())
                       << " of the " << nodetype2char_capital(p_node->type)
                       << " guid=" << HEX(p_node->guid_get())
                       << " is UP, but remote peer is not responding";
            warnings.push_back(str_stream.str());
            continue;
        }

        sout << S_BRACKETS(DEC(i))
             << R_BRACKETS(p_port->numAsString()) << ' ';

        int rc = this->PrintRemoteNodeAndPortForSwitch(p_port, sout);
        if (rc)
           IBDIAG_RETURN(rc);

        sout << endl;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::PrintRemoteNodeAndPortForSwitch(IBPort *p_port, ostream &sout)
{
    IBDIAG_ENTER;

    IBPort *p_remote_port = p_port->p_remotePort;
    if (!p_remote_port) {
        this->SetLastError("The remote port connected to the port %s is NULL\n",
                p_port->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    IBNode *p_remote_node = p_remote_port->p_node;
    if (!p_remote_node) {
        this->SetLastError("The remote node connected to the port %s is NULL\n",
                   p_remote_port->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    sout << Node_GUID(p_remote_node)
         << S_BRACKETS(DEC(p_remote_port->num));

    if (p_remote_node->type != IB_SW_NODE) {
        sout << R_BRACKETS(HEX(p_remote_port->guid_get(), 0));
    }

    sout << "      # "
         << QUOTED(p_remote_node->description)
         << " lid " << DEC(p_remote_port->base_lid)
         << ' '
         << Port_Width_And_Speed(p_port);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::PrintHCAToIBNetDiscoverFile(ostream &sout, warnings_list &warnings)
{
    IBDIAG_ENTER;

    for (map_str_pnode::iterator nI = this->discovered_fabric.NodeByName.begin();
         nI != this->discovered_fabric.NodeByName.end(); ++nI) {
        IBNode *p_node = (*nI).second;
        if (!p_node) {
            this->SetLastError("DB error - found null node in NodeByName map"
                               " for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        if (p_node->type == IB_SW_NODE)
            continue;

        int rc = this->PrintNodeInfo(p_node, sout, warnings);
        if (rc)
           return rc;

        rc = this->PrintHCANodePorts(p_node, sout, warnings);
        if (rc)
           return rc;

        rc = this->PrintHCAVirtualPorts(p_node, sout, warnings);
        if (rc)
           return rc;

        sout <<  endl << endl;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::PrintHCAVirtualPorts(IBNode *p_node, ostream &sout, warnings_list &warnings)
{
    IBDIAG_ENTER;

    // use unsigned int index to prevent overwrap
    for (unsigned int i = 1; i <= p_node->numPorts; i++) {
        IBPort *p_curr_port = p_node->getPort((phys_port_t)i);
        if (!p_curr_port)
            continue;

        SMP_VirtualizationInfo *p_virtual_info =
        this->fabric_extended_info.getSMPVirtualizationInfo(p_curr_port->createIndex);
        if (!p_virtual_info || p_curr_port->VPorts.empty())
            continue;

        u_int16_t index_top = p_virtual_info->vport_index_top;
        if (!index_top)
            continue;

        sout << "vPorts TopIndex " << index_top << endl;

        map_vportnum_vport &vports = p_curr_port->VPorts;
        for (u_int16_t i = 1; i <= index_top; i++) {
            map_vportnum_vport::iterator vpI = vports.find(i);
            if (vpI == vports.end())
                continue;

            IBVPort *p_vport = vpI->second;
            if (!p_vport) {
                this->SetLastError("The virtual port associated "
                            "with index :%u is NULL\n", i);
                IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
            }

            IBVNode *p_vnode = p_vport->getVNodePtr();
            if (!p_vnode) {

                sout << '#' << S_BRACKETS(DEC(p_vport->getVPortNum()))
                     << R_BRACKETS(HEX(p_vport->guid_get()))
                     << NA_REMOTE_PEER("V-")
                     << ' ' << UP_PORT_WARNING_MESSAGE("virtual ")
                     << endl;

                stringstream str_stream;
                str_stream << "There is no virtual node to be connected to "
                              "the virtual port: " <<  p_vport->getName();
                warnings.push_back(str_stream.str());
                continue;
            }

            sout << S_BRACKETS(DEC(p_vport->getVPortNum()))
                 << R_BRACKETS(HEX(p_vport->guid_get()))
                 << " \"V-" << HEX(p_vnode->guid_get(), 16)
                 << "\""
                 << S_BRACKETS(DEC(p_vport->get_vlocal_port_num()))
                 << ' ';

            this->PrintVirtPortLidName(p_curr_port, p_vport, sout);

            sout << endl;
        }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

void IBDiag::PrintVirtPortLidName(IBPort *p_curr_port, IBVPort *p_vport, ostream &sout)
{
    IBDIAG_ENTER;

    IBVPort *p_vport2 = NULL;
    map_vportnum_vport::iterator itr = p_curr_port->VPorts.find(p_vport->get_lid_by_vport_index());
    if (itr != p_curr_port->VPorts.end())
        p_vport2 = itr->second;

    if (p_vport->get_vlid()) {
		sout << "# lid " << p_vport->get_vlid() << " (unique)";
	} else {
		if(p_vport->get_lid_by_vport_index() != 0 && p_vport2 != NULL && p_vport2->get_vlid())
			sout << "# lid " << p_vport2->get_vlid() << " (lidByIndex)";
		else
			sout << "# lid " <<  p_curr_port->base_lid << " (phy)";
	}

    IBDIAG_RETURN_VOID;
}

int IBDiag::PrintHCANodePorts(IBNode *p_node, ostream &sout, warnings_list &warnings)
{
    IBDIAG_ENTER;

    for (phys_port_t i = 0 ; i <=  p_node->numPorts; i++) {
         IBPort *p_port = p_node->getPort(i);

        if (!p_port                                               ||
            p_port->get_internal_state() == IB_UNKNOWN_PORT_STATE ||
            p_port->get_internal_state() == IB_PORT_STATE_DOWN) {
            continue;
        }

        IBPort  *p_remote_port = p_port->p_remotePort;
        if (!p_remote_port) {

            sout << '#' << S_BRACKETS(DEC(i))
                 << R_BRACKETS(HEX(p_port->guid_get()))
                 << NA_REMOTE_PEER("")
                 << ' ' << UP_PORT_WARNING_MESSAGE("") << endl;

            stringstream str_stream;
            str_stream << "The port "
                       << S_BRACKETS(DEC(i))
                       << " of the " << nodetype2char_capital(p_node->type)
                       << " guid=" << HEX(p_node->guid_get())
                       << " is UP, but remote peer is not responding";
            warnings.push_back(str_stream.str());
            continue;
        }

        sout << S_BRACKETS(DEC(i)) << R_BRACKETS(p_port->numAsString())
             << R_BRACKETS(HEX(p_port->guid_get(), 0)) << ' ';

        int rc = this->PrintRemoteNodeAndPortForHCA(p_port, sout);
        if (rc)
           IBDIAG_RETURN(rc);

        sout << endl;
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::PrintRemoteNodeAndPortForHCA(IBPort *p_port, ostream &sout)
{
    IBDIAG_ENTER;

    IBPort  *p_remote_port = p_port->p_remotePort;
    //has already been checked.
    if (!p_remote_port)
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    IBNode *p_remote_node = p_remote_port->p_node;
    if (!p_remote_node) {
        this->SetLastError("The remote node connected to the port %s is NULL\n",
                   p_port->getName().c_str());
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    sout << Node_GUID(p_remote_node)
         << S_BRACKETS(DEC(p_remote_port->num))
         << " #"
         << " lid " << DEC(p_port->base_lid)
         << " lmc " << DEC((uint16_t)p_port->lmc)
         << ' ' << QUOTED(p_remote_node->description)
         << " lid " << DEC(p_remote_port->base_lid)
         << ' '
         << Port_Width_And_Speed(p_port);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiag::PrintNodeInfo(IBNode *p_node, ostream &sout, warnings_list &warnings)
{
    IBDIAG_ENTER;

    IBPort *p_zero_port = NULL;
    const char *low_prefix = nodetype2char_low(p_node->type);
    const char *prefix = nodetype2char_capital(p_node->type);

    sout << "vendid=" << PTR(p_node->vendId, 0) <<  endl
         << "devid="  << PTR(p_node->devId)  <<  endl
         << "sysimgguid=" << PTR(p_node->system_guid_get(), 0) <<  endl;

    if (p_node->type != IB_SW_NODE) {
        sout << low_prefix  << "guid=" << PTR(p_node->guid_get(), 0) <<  endl;
    }
    else {
        p_zero_port = p_node->getPort(0);
        if (!p_zero_port) {
            this->SetLastError("The Zero port of the switch: %s guid=" U64H_FMT "is NULL\n",
                               p_node->getName().c_str(), p_node->guid_get());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        sout << low_prefix  << "guid=" << PTR(p_node->guid_get(), 0)
             << R_BRACKETS(HEX(p_zero_port->guid_get())) <<  endl;
    }

    sout << prefix << "  " << DEC((int)p_node->numPorts) << ' '
         << Node_GUID(p_node) << "      # " << QUOTED(p_node->description);

    if (p_node->type == IB_SW_NODE) {
        SMP_SwitchInfo *p_switch_info = this->fabric_extended_info.getSMPSwitchInfo(p_node->createIndex);
        if (!p_switch_info) {
            stringstream str_stream;
            str_stream << "Cannot identify 0 port type "
                       << "of the switch guid=" << HEX(p_node->guid_get())
                       << " no SwitchInfo is provided";
            warnings.push_back(str_stream.str());

            sout << "### Warning: cannot identify 0 port type";
        } else
           sout << (p_switch_info->ENP0 ? " enhanced port 0" : " base port 0");

        sout << " lid " << DEC(p_zero_port->base_lid)
             << " lmc " << DEC((uint16_t)p_zero_port->lmc);
    }

    sout << endl;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}
