/*
 * Copyright (c) 2017-2020 Mellanox Technologies LTD. All rights reserved.
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <infiniband/ibdiag/ibdiag_fabric.h>
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include <infiniband/ibutils/output_stream_primitives.h>

int IBDiagFabric::UpdateFabric(const string &csv_file)
{
    IBDiag::PrintFileTimestamp(csv_file, "DB CSV");

    int rc = IBDIAG_SUCCESS_CODE;
    CsvFileStream csv_file_stream(csv_file, m_csv_parser);

    SectionParser<NodeRecord>  node_section_parser;
    node_section_parser.Init(SECTION_NODES);
    rc = m_csv_parser.ParseAndHandleSection <NodeRecord, IBDiagFabric>(csv_file_stream,
            node_section_parser, this, &IBDiagFabric::CreateNode);
    if (rc) {
        if(!this->last_error.empty()) ERR_PRINT("%s", this->last_error.c_str());
        ERR_PRINT("Failed to parse " SECTION_NODES " section in csv file\n");
        return rc;
    }

    SectionParser<PortRecord>  port_section_parser;
    port_section_parser.Init(SECTION_PORTS);
    rc = m_csv_parser.ParseAndHandleSection <PortRecord, IBDiagFabric>(csv_file_stream,
            port_section_parser, this, &IBDiagFabric::CreatePort);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_PORTS " section in csv file\n");
        return rc;
    }
    rc = this->CreateDummyPorts();
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_PORTS " section in csv file\n");
        return rc;
    }

    SectionParser<SwitchRecord>  switch_section_parser;
    switch_section_parser.Init(SECTION_SWITCHES);
    rc = m_csv_parser.ParseAndHandleSection <SwitchRecord, IBDiagFabric>(csv_file_stream,
            switch_section_parser, this, &IBDiagFabric::CreateSwitch);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_SWITCHES " section in csv file\n");
        return rc;
    }

    SectionParser<LinkRecord>  link_section_parser;
    link_section_parser.Init(SECTION_LINKS);
    rc = m_csv_parser.ParseAndHandleSection <LinkRecord, IBDiagFabric>(csv_file_stream,
            link_section_parser, this, &IBDiagFabric::CreateLink);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_LINKS " section in csv file\n");
        return rc;
    }

    SectionParser<GeneralInfoGMPRecord> general_info_gmp_section_parser;
    general_info_gmp_section_parser.Init(SECTION_NODES_INFO);
    rc = m_csv_parser.ParseAndHandleSection <GeneralInfoGMPRecord, IBDiagFabric>(csv_file_stream,
            general_info_gmp_section_parser, this, &IBDiagFabric::CreateVSGeneralInfoGMP);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_NODES_INFO " section in csv file\n");
        return rc;
    }

    SectionParser<GeneralInfoSMPRecord> general_info_smp_section_parser;
    general_info_smp_section_parser.Init(SECTION_GENERAL_INFO_SMP);
    rc = m_csv_parser.ParseAndHandleSection <GeneralInfoSMPRecord, IBDiagFabric>(csv_file_stream,
            general_info_smp_section_parser, this, &IBDiagFabric::CreateVSGeneralInfoSMP);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_GENERAL_INFO_SMP " section in csv file\n");
        return rc;
    }

    SectionParser<ExtendedPortInfoRecord> extended_port_info_section_parser;
    extended_port_info_section_parser.Init(SECTION_EXTENDED_PORT_INFO);
    rc = m_csv_parser.ParseAndHandleSection <ExtendedPortInfoRecord, IBDiagFabric>(csv_file_stream,
            extended_port_info_section_parser, this, &IBDiagFabric::CreateExtendedPortInfo);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_EXTENDED_PORT_INFO " section in csv file\n");
        return rc;
    }

    SectionParser<PortInfoExtendedRecord> port_info_extended_section_parser;
    port_info_extended_section_parser.Init(SECTION_PORT_INFO_EXTENDED);
    rc = m_csv_parser.ParseAndHandleSection <PortInfoExtendedRecord, IBDiagFabric>(csv_file_stream,
            port_info_extended_section_parser, this, &IBDiagFabric::CreatePortInfoExtended);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_PORT_INFO_EXTENDED " section in csv file\n");
        return rc;
    }

    SectionParser<PortHierarchyInfoRecord> port_hierarchy_info_section_parser;
    port_hierarchy_info_section_parser.Init(SECTION_PORT_HIERARCHY_INFO);
    rc = m_csv_parser.ParseAndHandleSection <PortHierarchyInfoRecord, IBDiagFabric>(csv_file_stream,
            port_hierarchy_info_section_parser, this, &IBDiagFabric::CreatePortHierarchyInfo);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_PORT_HIERARCHY_INFO " section in csv file\n");
        return rc;
    }

    SectionParser<PhysicalHierarchyInfoRecord> physical_hierarchy_info_section_parser;
    physical_hierarchy_info_section_parser.Init(SECTION_PHYSICAL_HIERARCHY_INFO);
    rc = m_csv_parser.ParseAndHandleSection <PhysicalHierarchyInfoRecord, IBDiagFabric>(csv_file_stream,
            physical_hierarchy_info_section_parser, this, &IBDiagFabric::CreatePhysicalHierarchyInfo);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_PHYSICAL_HIERARCHY_INFO " section in csv file\n");
        return rc;
    }

    // AR Info
    SectionParser<ARInfoRecord> ar_info_section_parser;
    ar_info_section_parser.Init(SECTION_AR_INFO);
    rc = m_csv_parser.ParseAndHandleSection <ARInfoRecord, IBDiagFabric>(csv_file_stream,
            ar_info_section_parser, this, &IBDiagFabric::CreateARInfo);
    if (rc) {
        ERR_PRINT("Failed to parse " SECTION_AR_INFO " section in csv file\n");
        return rc;
    }

    return rc;
}

int IBDiagFabric::CreateNode(const NodeRecord &nodeRecord)
{
    int rc = 0;

    IBNode* p_node = discovered_fabric.makeNode((IBNodeType)nodeRecord.node_type,
                                                (phys_port_t)nodeRecord.num_ports,
                                                nodeRecord.system_image_guid,
                                                nodeRecord.node_guid,
                                                nodeRecord.vendor_id,
                                                nodeRecord.device_id,
                                                nodeRecord.revision,
                                                nodeRecord.node_description,
                                                true);

    if (!p_node) {
        stringstream ss;
        ss << "Cannot load from file IBNode GUID=" << HEX(nodeRecord.node_guid)
           << ", System GUID=" << HEX(nodeRecord.node_guid)
           << ", ports=" << DEC(nodeRecord.num_ports)
           << ", description '" <<  nodeRecord.node_description << "'" << endl;

        last_error = ss.str();

        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
    }

    //take care of progress bar
    ++nodes_found;
    if (nodeRecord.node_type == IB_CA_NODE)
       ++ca_found;
    else
       ++sw_found;
    ports_found += nodeRecord.num_ports;

    struct SMP_NodeInfo node_info;

    node_info.NumPorts = (phys_port_t)nodeRecord.num_ports;
    node_info.NodeType = nodeRecord.node_type;
    node_info.ClassVersion = nodeRecord.class_version;
    node_info.BaseVersion = nodeRecord.base_version;
    node_info.SystemImageGUID = nodeRecord.system_image_guid;
    node_info.NodeGUID = nodeRecord.node_guid;
    node_info.PortGUID = nodeRecord.port_guid;
    node_info.DeviceID = nodeRecord.device_id;
    node_info.PartitionCap = nodeRecord.partition_cap;
    node_info.revision = nodeRecord.revision;
    node_info.VendorID = nodeRecord.vendor_id;
    node_info.LocalPortNum = nodeRecord.local_port_num;

    rc = this->fabric_extended_info.addSMPNodeInfo(p_node, node_info);

    return rc;
}

int IBDiagFabric::CreatePort(const PortRecord &portRecord)
{
    int rc = 0;

    IBNode* p_node = discovered_fabric.NodeByGuid[portRecord.node_guid];
    IBLinkSpeed correct_speed = (IBLinkSpeed)(portRecord.link_speed_actv & 0xff);

    static u_int16_t base_lid; // static - to keep the value for all switch ports
    static u_int8_t lmc;
    static u_int32_t cap_mask;

    if (!(p_node->type == IB_SW_NODE && portRecord.port_num != 0)) {
        base_lid = portRecord.lid;
        lmc = portRecord.lmc;
        cap_mask = portRecord.cap_mask;
    }

    if ((cap_mask & IBIS_IB_PORT_CAP_HAS_EXT_SPEEDS) &&
            ((portRecord.link_speed_actv >> 8) & 0xff)) {
        /* port supports extended speeds and active extended speed
           is not zero ==> we have extended speed */
        correct_speed = extspeed2speed((u_int8_t)((portRecord.link_speed_actv >> 8) & 0xff));
    }

    IBPort *p_port = discovered_fabric.setNodePort(p_node,
                                                   portRecord.port_guid,
                                                   base_lid,
                                                   lmc,
                                                   portRecord.port_num,
                                                   (IBLinkWidth)portRecord.link_width_actv,
                                                   correct_speed,
                                                   (IBPortState)portRecord.port_state);

    if (!p_port)
        IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);

    struct SMP_PortInfo port_info;

    port_info.MKey = portRecord.m_key;
    port_info.GIDPrfx = portRecord.gid_prefix;
    port_info.MSMLID = portRecord.msm_lid;
    port_info.LID = portRecord.lid;
    port_info.CapMsk = portRecord.cap_mask;
    port_info.M_KeyLeasePeriod = portRecord.m_key_lease_period;
    port_info.DiagCode = portRecord.diag_code;
    port_info.LinkWidthActv = portRecord.link_width_actv;
    port_info.LinkWidthSup = portRecord.link_width_sup;
    port_info.LinkWidthEn = portRecord.link_width_en;
    port_info.LocalPortNum = portRecord.local_port_num;
    port_info.LinkSpeedEn = (u_int8_t)(portRecord.link_speed_en & 0xff);
    port_info.LinkSpeedActv = (u_int8_t)(portRecord.link_speed_actv & 0xff);
    port_info.LMC = portRecord.lmc;
    port_info.MKeyProtBits = portRecord.m_key_prot_bits;
    port_info.LinkDownDefState = portRecord.link_down_def_state;
    port_info.PortPhyState = portRecord.port_phy_state;
    port_info.PortState = portRecord.port_state;
    port_info.LinkSpeedSup = (u_int8_t)(portRecord.link_speed_sup & 0xff);
    port_info.VLArbHighCap = portRecord.vl_arbit_High_Cap;
    port_info.VLHighLimit = portRecord.vl_high_limit;
    port_info.InitType = portRecord.init_type;
    port_info.VLCap = portRecord.vl_cap;
    port_info.MSMSL = portRecord.msm_sl;
    port_info.NMTU = portRecord.nmtu;
    port_info.FilterRawOutb = portRecord.filter_raw_outbound;
    port_info.FilterRawInb = portRecord.filter_raw_inbound;
    port_info.PartEnfOutb = portRecord.part_enf_outbound;
    port_info.PartEnfInb = portRecord.part_enf_inbound;
    port_info.OpVLs = portRecord.op_VLs;
    port_info.HoQLife = portRecord.hoq_life;
    port_info.VLStallCnt = portRecord.vl_stall_cnt;
    port_info.MTUCap = portRecord.mtu_cap;
    port_info.InitTypeReply = portRecord.init_type_reply;
    port_info.VLArbLowCap = portRecord.vl_arbit_low_cap;
    port_info.PKeyViolations = portRecord.pkey_violations;
    port_info.MKeyViolations = portRecord.mkey_violations;
    port_info.SubnTmo = portRecord.subn_time_out;
    port_info.MulticastPKeyTrapSuppressionEnabled = portRecord.multicast_pkey_trap_suppression_enabled;
    port_info.ClientReregister = portRecord.client_reregister;
    port_info.GUIDCap = portRecord.guid_cap;
    port_info.QKeyViolations = portRecord.qkey_violations;
    port_info.MaxCreditHint = portRecord.max_credit_hint;
    port_info.OverrunErrs = portRecord.overrun_errs;
    port_info.LocalPhyError = portRecord.local_phy_error;
    port_info.RespTimeValue = portRecord.resp_time_value;
    port_info.LinkRoundTripLatency = portRecord.link_round_trip_latency;
    port_info.LinkSpeedExtEn = (u_int8_t)((portRecord.link_speed_en >> 8) & 0xff);
    port_info.LinkSpeedExtActv = (u_int8_t)((portRecord.link_speed_actv >> 8) & 0xff);
    port_info.LinkSpeedExtSup = (u_int8_t)((portRecord.link_speed_sup >> 8) & 0xff);
    port_info.CapMsk2 = portRecord.cap_mask_2;

    p_port->setPortInfoMadWasSent(true);

    rc = this->fabric_extended_info.addSMPPortInfo(p_port, port_info);

    return rc;
}

/*
 * We should create IBPorts and empty Port-Info data MADs also for
 * ports that weren't dump/MADs weren't sent for them.
*/
int IBDiagFabric::CreateDummyPorts()
{
    IBDIAG_ENTER;

    for (map_str_pnode::iterator it_node = this->discovered_fabric.NodeByName.begin();
             it_node != this->discovered_fabric.NodeByName.end(); ++it_node) {
         IBNode *p_curr_node = it_node->second;
         if (!p_curr_node)
             IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);

         struct SMP_NodeInfo *p_node_info =
                               this->fabric_extended_info.getSMPNodeInfo(p_curr_node->createIndex);
         if (!p_node_info)
             IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);

         lid_t base_lid = 0;
         uint8_t lmc = 0;
         if (p_curr_node->type == IB_SW_NODE) {
             IBPort *p_port_zero = p_curr_node->getPort(0);
             if (!p_port_zero){
                 ERR_PRINT("DB Error - found null port 0 for Node GUID: " U64H_FMT " \n",
                                p_curr_node->guid_get());
                 IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
             }

             base_lid = p_port_zero->base_lid;
             lmc = p_port_zero->lmc;
         }

         for (phys_port_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {
              IBPort *p_curr_port = p_curr_node->getPort(port_num);

              if (this->discovered_fabric.defAllPorts) {
                  if (!p_curr_port) {
                      ERR_PRINT("DB Error - found null port for Node GUID: " U64H_FMT " port num: %d\n",
                                p_curr_node->guid_get(), port_num);
                      IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
                  }

                  if (p_curr_port->guid_get())
                      continue;

                  //HCA only sets GUID, LID etc for its Active port
                  if (p_curr_node->type != IB_SW_NODE)
                      continue;
              } else
                  if (p_curr_port)
                      continue;

              struct SMP_PortInfo port_info;
              CLEAR_STRUCT(port_info);
              port_info.LinkSpeedActv = IB_UNKNOWN_LINK_SPEED;
              port_info.LinkWidthActv = IB_UNKNOWN_LINK_WIDTH;
              port_info.PortState = IB_PORT_STATE_DOWN;
              port_info.PortPhyState = IB_PORT_PHYS_STATE_POLLING;

              //this function resets base_lid and lmc for Zero port any time it is called
              //we have to provide correct values
              p_curr_port = this->discovered_fabric.setNodePort(p_curr_node,
                                                             p_node_info->PortGUID,
                                                             base_lid,
                                                             lmc,
                                                             (p_curr_node->type == IB_SW_NODE) ?
                                                               port_num : p_node_info->LocalPortNum,
                                                             (IBLinkWidth)port_info.LinkWidthActv,
                                                             (IBLinkSpeed)port_info.LinkSpeedActv,
                                                             (IBPortState)port_info.PortState);
              if (!p_curr_port)
                  IBDIAG_RETURN(IBDIAG_ERR_CODE_IBDM_ERR);
              p_curr_port->setPortInfoMadWasSent(false);

              int rc = this->fabric_extended_info.addSMPPortInfo(p_curr_port, port_info);
              if (rc)
                  IBDIAG_RETURN(rc);
         }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreateSwitch(const SwitchRecord &switchRecord)
{
    int rc = 0;

    IBNode* p_node = discovered_fabric.NodeByGuid[switchRecord.node_guid];

    struct SMP_SwitchInfo switch_info;

    switch_info.LinearFDBCap = switchRecord.linear_FDB_cap;
    switch_info.RandomFDBCap = switchRecord.random_FDB_cap;
    switch_info.MCastFDBCap = switchRecord.mcast_FDB_cap;
    switch_info.LinearFDBTop = switchRecord.linear_FDB_top;
    switch_info.DefPort = switchRecord.def_port;
    switch_info.DefMCastPriPort = switchRecord.def_mcast_pri_port;
    switch_info.DefMCastNotPriPort = switchRecord.def_mcast_not_pri_port;
    switch_info.LifeTimeValue = switchRecord.life_time_value;
    switch_info.PortStateChange = switchRecord.port_state_change;
    switch_info.OptimizedSLVLMapping = switchRecord.optimized_SLVL_mapping;
    switch_info.LidsPerPort = switchRecord.lids_per_port;
    switch_info.PartEnfCap = switchRecord.part_enf_cap;
    switch_info.InbEnfCap = switchRecord.inb_enf_cap;
    switch_info.OutbEnfCap = switchRecord.outb_enf_cap;
    switch_info.FilterRawInbCap = switchRecord.filter_raw_inb_cap;
    switch_info.FilterRawOutbCap = switchRecord.filter_raw_outb_cap;
    switch_info.ENP0 = switchRecord.en_port0;
    switch_info.MCastFDBTop = switchRecord.mcast_FDB_top;

    rc = fabric_extended_info.addSMPSwitchInfo(p_node, switch_info);

    IBDIAG_RETURN(rc);
}

int IBDiagFabric::CreateLink(const LinkRecord &linkRecord)
{
    IBDIAG_ENTER;

    IBNode *p_node1 = this->discovered_fabric.getNodeByGuid(linkRecord.node_guid1);
    IBNode *p_node2 = this->discovered_fabric.getNodeByGuid(linkRecord.node_guid2);
    if (!(p_node1 && p_node2)) {
        ERR_PRINT("DB error - found null node for Node %d GUID: " U64H_FMT " in csv file, section: "
                  SECTION_LINKS "\n", p_node1 ? 2 : 1,
                  p_node1 ? linkRecord.node_guid2 : linkRecord.node_guid1);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    IBPort *p_port1 = p_node1->getPort(linkRecord.port_num1);
    IBPort *p_port2 = p_node2->getPort(linkRecord.port_num2);
    if (!(p_port1 && p_port2)) {
        ERR_PRINT("DB error - found null port for Node %d GUID: " U64H_FMT " port num: " U32D_FMT
                  " in csv file, section: " SECTION_LINKS "\n", p_port1 ? 2 : 1,
                  p_port1 ? linkRecord.node_guid2 : linkRecord.node_guid1,
                  p_port1 ? linkRecord.port_num2 : linkRecord.port_num1);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    p_port1->p_remotePort = p_port2;
    p_port2->p_remotePort = p_port1;

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreateVSGeneralInfoSMP(const GeneralInfoSMPRecord &generalInfoSMPRecord)
{
    IBDIAG_ENTER;

    if (!this->discovered_fabric.getNodeByGuid(generalInfoSMPRecord.node_guid)) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_GENERAL_INFO_SMP "\n", generalInfoSMPRecord.node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    fw_version_obj_t  fw_info_smp;
    capability_mask_t capability_mask_smp;

    if ((generalInfoSMPRecord.fw_info_extended_major != "N/A") &&
        (generalInfoSMPRecord.fw_info_extended_minor != "N/A") &&
        (generalInfoSMPRecord.fw_info_extended_sub_minor != "N/A")) {
        CsvParser::Parse(generalInfoSMPRecord.fw_info_extended_major.c_str(), fw_info_smp.major, 16);
        CsvParser::Parse(generalInfoSMPRecord.fw_info_extended_minor.c_str(), fw_info_smp.minor, 16);
        CsvParser::Parse(generalInfoSMPRecord.fw_info_extended_sub_minor.c_str(),
                         fw_info_smp.sub_minor, 16);

        this->capability_module.AddSMPFw(generalInfoSMPRecord.node_guid, fw_info_smp);
    }

    bool is_capability_mask_available = true;
    for (int capability_field = 0; capability_field < NUM_CAPABILITY_FIELDS; capability_field++) {
         if (generalInfoSMPRecord.capability_mask_fields[capability_field] == "N/A") {
             is_capability_mask_available = false;
             break;
         }
         CsvParser::Parse(generalInfoSMPRecord.capability_mask_fields[capability_field].c_str(),
                          capability_mask_smp.mask[capability_field], 16);
    }

    if (is_capability_mask_available)
        this->capability_module.AddSMPCapabilityMask(generalInfoSMPRecord.node_guid,
                                                     capability_mask_smp);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreateVSGeneralInfoGMP(const GeneralInfoGMPRecord &generalInfoGMPRecord)
{
    IBDIAG_ENTER;

    IBNode *p_node = this->discovered_fabric.getNodeByGuid(generalInfoGMPRecord.node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_NODES_INFO "\n", generalInfoGMPRecord.node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    capability_mask_t capability_mask_gmp;
    bool is_capability_mask_available = true;
    for (int capability_field = 0; capability_field < NUM_CAPABILITY_FIELDS;
         capability_field++) {
         if (generalInfoGMPRecord.capability_mask_fields[capability_field] == "N/A") {
             is_capability_mask_available = false;
             break;
         }
         CsvParser::Parse(generalInfoGMPRecord.capability_mask_fields[capability_field].c_str(),
                          capability_mask_gmp.mask[capability_field], 16);
    }

    if (is_capability_mask_available)
        this->capability_module.AddGMPCapabilityMask(generalInfoGMPRecord.node_guid,
                                                     capability_mask_gmp);

    struct VendorSpec_GeneralInfo general_info_gmp;
    memset(&general_info_gmp, 0, sizeof(general_info_gmp));

    fw_version_obj_t  fw_info_gmp;
    if ((generalInfoGMPRecord.hw_info_device_id != "N/A") &&
        (generalInfoGMPRecord.hw_info_device_hw_revision != "N/A") &&
        (generalInfoGMPRecord.hw_info_up_time != "N/A") &&
        (generalInfoGMPRecord.fw_info_sub_minor != "N/A") &&
        (generalInfoGMPRecord.fw_info_minor != "N/A") &&
        (generalInfoGMPRecord.fw_info_major != "N/A") &&
        (generalInfoGMPRecord.fw_info_build_id != "N/A") &&
        (generalInfoGMPRecord.fw_info_year != "N/A") &&
        (generalInfoGMPRecord.fw_info_day != "N/A") &&
        (generalInfoGMPRecord.fw_info_month != "N/A") &&
        (generalInfoGMPRecord.fw_info_hour != "N/A") &&
        (generalInfoGMPRecord.fw_info_psid != "N/A") &&
        (generalInfoGMPRecord.fw_info_ini_file_version != "N/A") &&
        (generalInfoGMPRecord.fw_info_extended_major != "N/A") &&
        (generalInfoGMPRecord.fw_info_extended_minor != "N/A") &&
        (generalInfoGMPRecord.fw_info_extended_minor != "N/A") &&
        (generalInfoGMPRecord.sw_info_sub_minor != "N/A") &&
        (generalInfoGMPRecord.sw_info_minor != "N/A") &&
        (generalInfoGMPRecord.sw_info_major != "N/A")) {
            CsvParser::Parse(generalInfoGMPRecord.hw_info_device_id.c_str(),
                             general_info_gmp.HWInfo.DeviceID, 16);
            CsvParser::Parse(generalInfoGMPRecord.hw_info_device_hw_revision.c_str(),
                             general_info_gmp.HWInfo.DeviceHWRevision, 16);
            CsvParser::Parse(generalInfoGMPRecord.hw_info_up_time.c_str(),
                             general_info_gmp.HWInfo.UpTime, 16);
            CsvParser::Parse(generalInfoGMPRecord.fw_info_sub_minor.c_str(),
                             general_info_gmp.FWInfo.SubMinor, 16);
            CsvParser::Parse(generalInfoGMPRecord.fw_info_minor.c_str(),
                             general_info_gmp.FWInfo.Minor, 16);
            CsvParser::Parse(generalInfoGMPRecord.fw_info_major.c_str(),
                             general_info_gmp.FWInfo.Major, 16);
            CsvParser::Parse(generalInfoGMPRecord.fw_info_build_id.c_str(),
                             general_info_gmp.FWInfo.BuildID, 16);
            CsvParser::Parse(generalInfoGMPRecord.fw_info_year.c_str(),
                             general_info_gmp.FWInfo.Year, 16);
            CsvParser::Parse(generalInfoGMPRecord.fw_info_day.c_str(),
                             general_info_gmp.FWInfo.Day, 16);
            CsvParser::Parse(generalInfoGMPRecord.fw_info_month.c_str(),
                             general_info_gmp.FWInfo.Month, 16);
            CsvParser::Parse(generalInfoGMPRecord.fw_info_hour.c_str(),
                         general_info_gmp.FWInfo.Hour, 16);

            CLEAR_STRUCT(general_info_gmp.FWInfo.PSID.PSID);
            string psid = generalInfoGMPRecord.fw_info_psid;
            if (psid == "UNKNOWN")
                psid = "";
            strncpy((char *)general_info_gmp.FWInfo.PSID.PSID, psid.c_str(),
                    sizeof(general_info_gmp.FWInfo.PSID.PSID) - 1);

            CsvParser::Parse(generalInfoGMPRecord.fw_info_ini_file_version.c_str(),
                             general_info_gmp.FWInfo.INI_File_Version, 16);
            CsvParser::Parse(generalInfoGMPRecord.fw_info_extended_major.c_str(),
                             general_info_gmp.FWInfo.Extended_Major, 16);
            CsvParser::Parse(generalInfoGMPRecord.fw_info_extended_minor.c_str(),
                             general_info_gmp.FWInfo.Extended_Minor, 16);
            CsvParser::Parse(generalInfoGMPRecord.fw_info_extended_sub_minor.c_str(),
                             general_info_gmp.FWInfo.Extended_SubMinor, 16);
            CsvParser::Parse(generalInfoGMPRecord.sw_info_sub_minor.c_str(),
                             general_info_gmp.SWInfo.SubMinor, 16);
            CsvParser::Parse(generalInfoGMPRecord.sw_info_minor.c_str(),
                             general_info_gmp.SWInfo.Minor, 16);
            CsvParser::Parse(generalInfoGMPRecord.sw_info_major.c_str(),
                             general_info_gmp.SWInfo.Major, 16);

            if (is_capability_mask_available) {
                general_info_gmp.CapabilityMask.capability0 = capability_mask_gmp.mask[0];
                general_info_gmp.CapabilityMask.capability1 = capability_mask_gmp.mask[1];
                general_info_gmp.CapabilityMask.capability2 = capability_mask_gmp.mask[2];
                general_info_gmp.CapabilityMask.capability3 = capability_mask_gmp.mask[3];
            }

            fw_info_gmp = general_info_gmp;
            this->capability_module.AddGMPFw(generalInfoGMPRecord.node_guid, fw_info_gmp);

            int rc = this->fabric_extended_info.addVSGeneralInfo(p_node, general_info_gmp);
            if (rc) {
                ERR_PRINT("Failed to add VSGeneralInfo for node=%s, err=%u",
                          p_node->getName().c_str(), rc);
                IBDIAG_RETURN(rc);
            }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreateExtendedPortInfo(const ExtendedPortInfoRecord &extendedPortInfoRecord)
{
    IBDIAG_ENTER;

    IBNode *p_node = this->discovered_fabric.getNodeByGuid(extendedPortInfoRecord.node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_EXTENDED_PORT_INFO "\n", extendedPortInfoRecord.node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    IBPort *p_port = p_node->getPort((phys_port_t)extendedPortInfoRecord.port_num);
    if (!p_port) {
        ERR_PRINT("DB error - found null port for Node GUID " U64H_FMT " port num: " U8H_FMT
                  " in csv file, section: " SECTION_EXTENDED_PORT_INFO "\n",
                  extendedPortInfoRecord.node_guid, extendedPortInfoRecord.port_num);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    if (p_port->guid_get() != extendedPortInfoRecord.port_guid) {
        ERR_PRINT("DB error - Mismatch between Port %d GUID " U64H_FMT " in fabric to Port GUID "
                  U64H_FMT " in csv file, section: " SECTION_EXTENDED_PORT_INFO "\n",
                  p_port->num, p_port->guid_get(), extendedPortInfoRecord.port_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    struct SMP_MlnxExtPortInfo mepi;
    mepi.StateChangeEnable = extendedPortInfoRecord.state_change_enable;
    mepi.RouterLIDEn = extendedPortInfoRecord.router_lid_en;
    mepi.SHArPANEn = extendedPortInfoRecord.sharp_an_en;
    mepi.AME = extendedPortInfoRecord.ame;
    mepi.LinkSpeedSupported = extendedPortInfoRecord.link_speed_supported;
    mepi.LinkSpeedEnabled = extendedPortInfoRecord.link_speed_enabled;
    mepi.LinkSpeedActive = extendedPortInfoRecord.link_speed_active;
    mepi.ActiveRSFECParity = extendedPortInfoRecord.active_rsfec_parity;
    mepi.ActiveRSFECData = extendedPortInfoRecord.active_rsfec_data;
    mepi.CapabilityMask = extendedPortInfoRecord.capability_mask;
    mepi.FECModeActive = extendedPortInfoRecord.fec_mode_active;
    mepi.RetransMode = extendedPortInfoRecord.retrans_mode;
    mepi.FDR10FECModeSupported = extendedPortInfoRecord.fdr10_fec_mode_supported;
    mepi.FDR10FECModeEnabled = extendedPortInfoRecord.fdr10_fec_mode_enabled;
    mepi.FDRFECModeSupported = extendedPortInfoRecord.fdr_fec_mode_supported;
    mepi.FDRFECModeEnabled = extendedPortInfoRecord.fdr_fec_mode_enabled;
    mepi.EDR20FECModeSupported = extendedPortInfoRecord.edr20_fec_mode_supported;
    mepi.EDR20FECModeEnabled = extendedPortInfoRecord.edr20_fec_mode_enabled;
    mepi.EDRFECModeSupported = extendedPortInfoRecord.edr_fec_mode_supported;
    mepi.EDRFECModeEnabled = extendedPortInfoRecord.edr_fec_mode_enabled;
    mepi.FDR10RetranSupported = extendedPortInfoRecord.fdr10_retran_supported;
    mepi.FDR10RetranEnabled = extendedPortInfoRecord.fdr10_retran_enabled;
    mepi.FDRRetranSupported = extendedPortInfoRecord.fdr_retran_supported;
    mepi.FDRRetranEnabled = extendedPortInfoRecord.fdr_retran_enabled;
    mepi.EDR20RetranSupported = extendedPortInfoRecord.edr20_retran_supported;
    mepi.EDR20RetranEnabled = extendedPortInfoRecord.edr20_retran_enabled;
    mepi.EDRRetranSupported = extendedPortInfoRecord.edr_retran_supported;
    mepi.EDRRetranEnabled = extendedPortInfoRecord.edr_retran_enabled;
    mepi.IsSpecialPort = extendedPortInfoRecord.is_special_port;
    mepi.SpecialPortType = extendedPortInfoRecord.special_port_type;
    mepi.SpecialPortCapabilityMask = extendedPortInfoRecord.special_port_capability_mask;
    mepi.HDRFECModeSupported = extendedPortInfoRecord.hdr_fec_mode_supported;
    mepi.HDRFECModeEnabled = extendedPortInfoRecord.hdr_fec_mode_enabled;
    mepi.OOOSLMask = extendedPortInfoRecord.ooosl_mask;
    mepi.AdaptiveTimeoutSLMask = extendedPortInfoRecord.adaptive_timeout_sl_mask;
    mepi.NDRFECModeSupported = extendedPortInfoRecord.ndr_fec_mode_supported;
    mepi.NDRFECModeEnabled = extendedPortInfoRecord.ndr_fec_mode_enabled;

    // Maybe we need to change speed to MLNX speed
    IBLinkSpeed speed = (mepi.LinkSpeedActive ? \
                         mlnxspeed2speed(mepi.LinkSpeedActive) : \
                         p_port->get_common_speed());
    p_port->set_internal_speed(speed);

    // Check if can determine the FEC Mode
    if (EXT_FEC_MODE_CONTROL_AND_REPORT_SUPPORT(mepi.CapabilityMask))
            p_port->set_fec_mode((IBFECMode)mepi.FECModeActive);

    if (mepi.IsSpecialPort)
            p_port->setSpecialPortType((IBSpecialPortType)mepi.SpecialPortType);

    int rc = this->fabric_extended_info.addSMPMlnxExtPortInfo(p_port, mepi);
    if (rc) {
        ERR_PRINT("Failed to store vs extended port info for port %s err=%u\n",
                  p_port->getName().c_str(), rc);
        IBDIAG_RETURN(rc);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreatePortInfoExtended(const PortInfoExtendedRecord &portInfoExtendedRecord)
{
    IBDIAG_ENTER;

    IBNode *p_node = this->discovered_fabric.getNodeByGuid(portInfoExtendedRecord.node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_PORT_INFO_EXTENDED "\n", portInfoExtendedRecord.node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    IBPort *p_port = p_node->getPort((phys_port_t)portInfoExtendedRecord.port_num);
    if (!p_port) {
        ERR_PRINT("DB error - found null port for Node GUID " U64H_FMT " port num: " U8H_FMT
                  " in csv file, section: " SECTION_PORT_INFO_EXTENDED "\n",
                  portInfoExtendedRecord.node_guid, portInfoExtendedRecord.port_num);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    if (p_port->guid_get() != portInfoExtendedRecord.port_guid) {
        ERR_PRINT("DB error - Mismatch between Port %d GUID " U64H_FMT " in fabric to Port GUID "
                  U64H_FMT " in csv file, section: " SECTION_PORT_INFO_EXTENDED "\n",
                  p_port->num, p_port->guid_get(), portInfoExtendedRecord.port_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    struct SMP_PortInfoExtended port_info_extended = {0};
    port_info_extended.FECModeActive = portInfoExtendedRecord.fec_mode_active;
    port_info_extended.FDRFECModeSupported = portInfoExtendedRecord.fdr_fec_mode_supported;
    port_info_extended.FDRFECModeEnabled = portInfoExtendedRecord.fdr_fec_mode_enabled;
    port_info_extended.EDRFECModeSupported = portInfoExtendedRecord.edr_fec_mode_supported;
    port_info_extended.EDRFECModeEnabled = portInfoExtendedRecord.edr_fec_mode_enabled;
    port_info_extended.HDRFECModeSupported = portInfoExtendedRecord.hdr_fec_mode_supported;
    port_info_extended.HDRFECModeEnabled = portInfoExtendedRecord.hdr_fec_mode_enabled;
    port_info_extended.NDRFECModeSupported = portInfoExtendedRecord.ndr_fec_mode_supported;
    port_info_extended.NDRFECModeEnabled = portInfoExtendedRecord.ndr_fec_mode_enabled;
    port_info_extended.CapMsk = portInfoExtendedRecord.capability_mask;

    if (PORT_INFO_EXT_FEC_MODE_SUPPORT(port_info_extended.CapMsk)) {
            // convert from bit mask to real value
            IBFECMode fec_value = fec_mask2value(port_info_extended.FECModeActive);
            if (fec_value == IB_FEC_NA)
                ERR_PRINT("Wrong FECModeActive value: " U16H_FMT " in csv file, section: "
                          SECTION_PORT_INFO_EXTENDED "\n", port_info_extended.FECModeActive);
            p_port->set_fec_mode(fec_value);
    }

    int rc = this->fabric_extended_info.addSMPPortInfoExtended(p_port, port_info_extended);
    if (rc) {
        ERR_PRINT("Failed to store port info extended for port %s err=%u\n",
                  p_port->getName().c_str(), rc);
        IBDIAG_RETURN(rc);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreatePortHierarchyInfo(const PortHierarchyInfoRecord &portHierarchyInfoRecord)
{
    IBDIAG_ENTER;
    IBNode *p_node = this->discovered_fabric.getNodeByGuid(portHierarchyInfoRecord.m_node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                SECTION_PORT_HIERARCHY_INFO "\n", portHierarchyInfoRecord.m_node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    IBPort *p_port = p_node->getPort((phys_port_t)portHierarchyInfoRecord.m_port_num);
    if (!p_port) {
        ERR_PRINT("DB error - found null port for Node GUID " U64H_FMT " port num: " U8H_FMT
                  " in csv file, section: " SECTION_PORT_HIERARCHY_INFO "\n",
                  portHierarchyInfoRecord.m_node_guid, portHierarchyInfoRecord.m_port_num);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    if (p_port->guid_get() != portHierarchyInfoRecord.m_port_guid) {
        ERR_PRINT("DB error - Mismatch between Port %d GUID " U64H_FMT " in fabric to Port GUID "
                  U64H_FMT " in csv file, section: " SECTION_PORT_HIERARCHY_INFO "\n",
                  p_port->num, p_port->guid_get(), portHierarchyInfoRecord.m_port_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    p_port->p_node->should_support_port_hierarchy_info = true;
    p_port->p_port_hierarchy_info =
            new PortHierarchyInfo(
                    p_node->type,
                    {portHierarchyInfoRecord.m_bus, portHierarchyInfoRecord.m_device,
                        portHierarchyInfoRecord.m_function},
                    portHierarchyInfoRecord.m_type,
                    portHierarchyInfoRecord.m_slot_type,
                    portHierarchyInfoRecord.m_slot_value,
                    portHierarchyInfoRecord.m_asic,
                    portHierarchyInfoRecord.m_cage,
                    portHierarchyInfoRecord.m_port,
                    portHierarchyInfoRecord.m_split);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagFabric::CreatePhysicalHierarchyInfo(const PhysicalHierarchyInfoRecord &physicalHierarchyInfoRecord)
{
    IBDIAG_ENTER;

    IBNode *p_node = this->discovered_fabric.getNodeByGuid(physicalHierarchyInfoRecord.m_node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                SECTION_PHYSICAL_HIERARCHY_INFO "\n", physicalHierarchyInfoRecord.m_node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    p_node->p_physical_hierarchy_info =
            new PhysicalHierarchyInfo(
                    physicalHierarchyInfoRecord.m_campus_serial_num,
                    physicalHierarchyInfoRecord.m_room_serial_num,
                    physicalHierarchyInfoRecord.m_rack_serial_num,
                    physicalHierarchyInfoRecord.m_system_type,
                    physicalHierarchyInfoRecord.m_system_topu_num,
                    physicalHierarchyInfoRecord.m_board_type,
                    physicalHierarchyInfoRecord.m_board_slot_num,
                    physicalHierarchyInfoRecord.m_device_serial_num);

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

// AR Info
int IBDiagFabric::CreateARInfo(const ARInfoRecord &arInfoRecord)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    IBNode* p_node = this->discovered_fabric.getNodeByGuid(arInfoRecord.node_guid);
    if (!p_node) {
        ERR_PRINT("DB error - found null node for Node GUID " U64H_FMT " in csv file, section: "
                  SECTION_AR_INFO "\n", arInfoRecord.node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    adaptive_routing_info ar_info = {0};

    ar_info.e                       =  arInfoRecord.e;
    ar_info.is_arn_sup              =  arInfoRecord.is_arn_sup;
    ar_info.is_frn_sup              =  arInfoRecord.is_frn_sup;
    ar_info.is_fr_sup               =  arInfoRecord.is_fr_sup;
    ar_info.fr_enabled              =  arInfoRecord.fr_enabled;
    ar_info.rn_xmit_enabled         =  arInfoRecord.rn_xmit_enabled;
    ar_info.is_ar_trails_supported  =  arInfoRecord.is_ar_trails_supported;
    ar_info.sub_grps_active         =  arInfoRecord.sub_grps_active;
    ar_info.group_table_copy_sup    =  arInfoRecord.group_table_copy_sup;
    ar_info.direction_num_sup       =  arInfoRecord.direction_num_sup;
    ar_info.is4_mode                =  arInfoRecord.is4_mode;
    ar_info.glb_groups              =  arInfoRecord.glb_groups;
    ar_info.by_sl_cap               =  arInfoRecord.by_sl_cap;
    ar_info.by_sl_en                =  arInfoRecord.by_sl_en;
    ar_info.by_transp_cap           =  arInfoRecord.by_transp_cap;
    ar_info.dyn_cap_calc_sup        =  arInfoRecord.dyn_cap_calc_sup;
    ar_info.group_cap               =  arInfoRecord.group_cap;
    ar_info.group_top               =  arInfoRecord.group_top;
    ar_info.group_table_cap         =  arInfoRecord.group_table_cap;
    ar_info.string_width_cap        =  arInfoRecord.string_width_cap;
    ar_info.ar_version_cap          =  arInfoRecord.ar_version_cap;
    ar_info.rn_version_cap          =  arInfoRecord.rn_version_cap;
    ar_info.sub_grps_supported      =  arInfoRecord.sub_grps_supported;
    ar_info.enable_by_sl_mask       =  arInfoRecord.enable_by_sl_mask;
    ar_info.by_transport_disable    =  arInfoRecord.by_transport_disable;
    ar_info.ageing_time_value       =  arInfoRecord.ageing_time_value;
    ar_info.is_whbf_supported       =  arInfoRecord.is_whbf_supported;
    ar_info.whbf_en                 =  arInfoRecord.whbf_en;
    ar_info.is_hbf_supported        =  arInfoRecord.is_hbf_supported;
    ar_info.by_sl_hbf_en            =  arInfoRecord.by_sl_hbf_en;
    ar_info.enable_by_sl_mask_hbf   =  arInfoRecord.enable_by_sl_mask_hbf;
    ar_info.whbf_granularity        =  arInfoRecord.whbf_granularity;
    ar_info.is_symmetric_hash_supported = arInfoRecord.is_symmetric_hash_supported;
    ar_info.is_dceth_hash_supported     = arInfoRecord.is_dceth_hash_supported;
    ar_info.is_bth_dqp_hash_supported   = arInfoRecord.is_bth_dqp_hash_supported;
    ar_info.is_pfrn_supported           = arInfoRecord.is_pfrn_supported;
    ar_info.pfrn_enabled                = arInfoRecord.pfrn_enabled;

    rc = fabric_extended_info.addARInfo(p_node, ar_info);

    IBDIAG_RETURN(rc);
}


/*******************************
 *       SECTION RECORDS       *
 *******************************/

int NodeRecord::Init(vector < ParseFieldInfo <class NodeRecord> > &parse_section_info)
{
    int rc = 0;

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("NodeDesc", &NodeRecord::SetNodeDescription));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("NumPorts", &NodeRecord::SetNumPorts));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("NodeType", &NodeRecord::SetNodeType));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("ClassVersion", &NodeRecord::SetClassVersion));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("BaseVersion", &NodeRecord::SetBaseVersion));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("SystemImageGUID", &NodeRecord::SetSystemImageGUID));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("NodeGUID", &NodeRecord::SetNodeGUID));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("PortGUID", &NodeRecord::SetPortGUID));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("DeviceID", &NodeRecord::SetDeviceID));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("PartitionCap", &NodeRecord::SetPartitionCap));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("revision", &NodeRecord::SetRevision));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("VendorID", &NodeRecord::SetVendorID));

    parse_section_info.push_back(
            ParseFieldInfo <class NodeRecord>("LocalPortNum", &NodeRecord::SetLocalPortNum));

   return(rc);
}

int PortRecord::Init(vector < ParseFieldInfo <class PortRecord> > &parse_section_info)
{
    int rc = 0;

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("NodeGuid", &PortRecord::SetNodeGuid));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PortGuid", &PortRecord::SetPortGuid));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PortNum", &PortRecord::SetPortNum));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MKey", &PortRecord::SetMKey));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("GIDPrfx", &PortRecord::SetGIDPrfx));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MSMLID", &PortRecord::SetMSMLID));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LID", &PortRecord::SetLid));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("CapMsk", &PortRecord::SetCapMsk));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("M_KeyLeasePeriod", &PortRecord::SetM_KeyLeasePeriod));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("DiagCode", &PortRecord::SetDiagCode));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkWidthActv", &PortRecord::SetLinkWidthActv));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkWidthSup", &PortRecord::SetLinkWidthSup));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkWidthEn", &PortRecord::SetLinkWidthEn));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LocalPortNum", &PortRecord::SetLocalPortNum));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkSpeedEn", &PortRecord::SetLinkSpeedEn));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkSpeedActv", &PortRecord::SetLinkSpeedActv));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LMC", &PortRecord::SetLMC));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MKeyProtBits", &PortRecord::SetMKeyProtBits));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkDownDefState", &PortRecord::SetLinkDownDefState));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PortPhyState", &PortRecord::SetPortPhyState));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PortState", &PortRecord::SetPortState));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LinkSpeedSup", &PortRecord::SetLinkSpeedSup));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("VLArbHighCap", &PortRecord::SetVLArbHighCap));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("VLHighLimit", &PortRecord::SetVLHighLimit));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("InitType", &PortRecord::SetInitType));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("VLCap", &PortRecord::SetVLCap));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MSMSL", &PortRecord::SetMSMSL));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("NMTU", &PortRecord::SetNMTU));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("FilterRawOutb", &PortRecord::SetFilterRawOutb));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("FilterRawInb", &PortRecord::SetFilterRawInb));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PartEnfOutb", &PortRecord::SetPartEnfOutb));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PartEnfInb", &PortRecord::SetPartEnfInb));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("OpVLs", &PortRecord::SetOpVLs));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("HoQLife", &PortRecord::SetHoQLife));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("VLStallCnt", &PortRecord::SetVLStallCnt));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MTUCap", &PortRecord::SetMTUCap));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("InitTypeReply", &PortRecord::SetInitTypeReply));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("VLArbLowCap", &PortRecord::SetVLArbLowCap));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("PKeyViolations", &PortRecord::SetPKeyViolations));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MKeyViolations", &PortRecord::SetMKeyViolations));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("SubnTmo", &PortRecord::SetSubnTmo));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("ClientReregister", &PortRecord::SetClientReregister));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MulticastPKeyTrapSuppressionEnabled", &PortRecord::SetMulticastPKeyTrapSuppressionEnabled));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("GUIDCap", &PortRecord::SetGUIDCap));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("QKeyViolations", &PortRecord::SetQKeyViolations));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("MaxCreditHint", &PortRecord::SetMaxCreditHint));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("OverrunErrs", &PortRecord::SetOverrunErrs));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("LocalPhyError", &PortRecord::SetLocalPhyError));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("RespTimeValue", &PortRecord::SetRespTimeValue));

    parse_section_info.push_back(ParseFieldInfo <class PortRecord>
            ("LinkRoundTripLatency", &PortRecord::SetLinkRoundTripLatency));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("CapMsk2", &PortRecord::SetCapMsk2));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("FECActv", &PortRecord::SetFECActv));

    parse_section_info.push_back(
            ParseFieldInfo <class PortRecord>("RetransActv", &PortRecord::SetRetransActv));

   return(rc);
}

int SwitchRecord::Init(vector < ParseFieldInfo <class SwitchRecord> > &parse_section_info)
{
    int rc = 0;

    parse_section_info.push_back(
                ParseFieldInfo <class SwitchRecord>("NodeGUID", &SwitchRecord::SetNodeGuid));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("LinearFDBCap", &SwitchRecord::SetLinearFDBCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("RandomFDBCap", &SwitchRecord::SetRandomFDBCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("MCastFDBCap", &SwitchRecord::SetMCastFDBCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("LinearFDBTop", &SwitchRecord::SetLinearFDBTop));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("DefPort", &SwitchRecord::SetDefPort));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("DefMCastPriPort", &SwitchRecord::SetDefMCastPriPort));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("DefMCastNotPriPort", &SwitchRecord::SetDefMCastNotPriPort));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("LifeTimeValue", &SwitchRecord::SetLifeTimeValue));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("PortStateChange", &SwitchRecord::SetPortStateChange));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("OptimizedSLVLMapping", &SwitchRecord::SetOptimizedSLVLMapping));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("LidsPerPort", &SwitchRecord::SetLidsPerPort));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("PartEnfCap", &SwitchRecord::SetPartEnfCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("InbEnfCap", &SwitchRecord::SetInbEnfCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("OutbEnfCap", &SwitchRecord::SetOutbEnfCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("FilterRawInbCap", &SwitchRecord::SetFilterRawInbCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("FilterRawOutbCap", &SwitchRecord::SetFilterRawOutbCap));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("ENP0", &SwitchRecord::SetENP0));

    parse_section_info.push_back(
            ParseFieldInfo <class SwitchRecord>("MCastFDBTop", &SwitchRecord::SetMCastFDBTop));

   return(rc);
}

int LinkRecord::Init(vector < ParseFieldInfo <class LinkRecord> > &parse_section_info)
{
    int rc = 0;

    parse_section_info.push_back(
            ParseFieldInfo <class LinkRecord>("NodeGuid1", &LinkRecord::SetNodeGuid1));

   parse_section_info.push_back(
            ParseFieldInfo <class LinkRecord>("PortNum1", &LinkRecord::SetPortNum1));

   parse_section_info.push_back(
            ParseFieldInfo <class LinkRecord>("NodeGuid2", &LinkRecord::SetNodeGuid2));

   parse_section_info.push_back(
            ParseFieldInfo <class LinkRecord>("PortNum2", &LinkRecord::SetPortNum2));

   return rc;
}

int GeneralInfoSMPRecord::Init(vector < ParseFieldInfo
                                      <class GeneralInfoSMPRecord> > &parse_section_info)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoSMPRecord>
                                         ("NodeGUID", &GeneralInfoSMPRecord::SetNodeGUID));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoSMPRecord>
                                         ("FWInfo_Extended_Major",
                                          &GeneralInfoSMPRecord::SetFWInfoExtendedMajor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoSMPRecord>
                                         ("FWInfo_Extended_Minor",
                                          &GeneralInfoSMPRecord::SetFWInfoExtendedMinor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoSMPRecord>
                                         ("FWInfo_Extended_SubMinor",
                                          &GeneralInfoSMPRecord::SetFWInfoExtendedSubMinor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoSMPRecord>
                                         ("CapabilityMask_0",
                                          &GeneralInfoSMPRecord::SetCapabilityMaskField0));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoSMPRecord>
                                         ("CapabilityMask_1",
                                          &GeneralInfoSMPRecord::SetCapabilityMaskField1));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoSMPRecord>
                                         ("CapabilityMask_2",
                                          &GeneralInfoSMPRecord::SetCapabilityMaskField2));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoSMPRecord>
                                         ("CapabilityMask_3",
                                          &GeneralInfoSMPRecord::SetCapabilityMaskField3));

    IBDIAG_RETURN(rc);
}

int GeneralInfoGMPRecord::Init(vector < ParseFieldInfo
                                      <class GeneralInfoGMPRecord> > &parse_section_info)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("NodeGUID", &GeneralInfoGMPRecord::SetNodeGUID));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("HWInfo_DeviceID",
                                          &GeneralInfoGMPRecord::SetHWInfoDeviceID));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("HWInfo_DeviceHWRevision",
                                          &GeneralInfoGMPRecord::SetHWInfoDeviceHWRevision));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("HWInfo_UpTime",
                                          &GeneralInfoGMPRecord::SetHWInfoUpTime));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_SubMinor",
                                          &GeneralInfoGMPRecord::SetFWInfoSubMinor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_Minor",
                                          &GeneralInfoGMPRecord::SetFWInfoMinor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_Major",
                                          &GeneralInfoGMPRecord::SetFWInfoMajor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_BuildID",
                                          &GeneralInfoGMPRecord::SetFWInfoBuildID));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_Year",
                                          &GeneralInfoGMPRecord::SetFWInfoYear));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_Day",
                                          &GeneralInfoGMPRecord::SetFWInfoDay));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_Month",
                                          &GeneralInfoGMPRecord::SetFWInfoMonth));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_Hour",
                                          &GeneralInfoGMPRecord::SetFWInfoHour));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_PSID",
                                          &GeneralInfoGMPRecord::SetFWInfoPSID));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_INI_File_Version",
                                          &GeneralInfoGMPRecord::SetFWInfoINIFileVersion));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_Extended_Major",
                                          &GeneralInfoGMPRecord::SetFWInfoExtendedMajor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_Extended_Minor",
                                          &GeneralInfoGMPRecord::SetFWInfoExtendedMinor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("FWInfo_Extended_SubMinor",
                                          &GeneralInfoGMPRecord::SetFWInfoExtendedSubMinor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("SWInfo_SubMinor",
                                          &GeneralInfoGMPRecord::SetSWInfoSubMinor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("SWInfo_Minor",
                                          &GeneralInfoGMPRecord::SetSWInfoMinor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("SWInfo_Major",
                                          &GeneralInfoGMPRecord::SetSWInfoMajor));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("CapabilityMask_0",
                                          &GeneralInfoGMPRecord::SetCapabilityMaskField0));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("CapabilityMask_1",
                                          &GeneralInfoGMPRecord::SetCapabilityMaskField1));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("CapabilityMask_2",
                                          &GeneralInfoGMPRecord::SetCapabilityMaskField2));

    parse_section_info.push_back(ParseFieldInfo <class GeneralInfoGMPRecord>
                                         ("CapabilityMask_3",
                                          &GeneralInfoGMPRecord::SetCapabilityMaskField3));

    IBDIAG_RETURN(rc);
}

int ExtendedPortInfoRecord::Init(vector < ParseFieldInfo
                                      <class ExtendedPortInfoRecord> > &parse_section_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("NodeGuid", &ExtendedPortInfoRecord::SetNodeGuid));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("PortGuid", &ExtendedPortInfoRecord::SetPortGuid));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("PortNum", &ExtendedPortInfoRecord::SetPortNum));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("StateChangeEnable", &ExtendedPortInfoRecord::SetStateChangeEnable));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("SHArPANEn", &ExtendedPortInfoRecord::SetSharpAnEn));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("RouterLIDEn", &ExtendedPortInfoRecord::SetRouterLIDEn));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("AME", &ExtendedPortInfoRecord::SetAME));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("LinkSpeedSupported", &ExtendedPortInfoRecord::SetLinkSpeedSupported));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("LinkSpeedEnabled", &ExtendedPortInfoRecord::SetLinkSpeedEnabled));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("LinkSpeedActive", &ExtendedPortInfoRecord::SetLinkSpeedActive));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("ActiveRSFECParity", &ExtendedPortInfoRecord::SetActiveRSFECParity));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("ActiveRSFECData", &ExtendedPortInfoRecord::SetActiveRSFECData));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("CapabilityMask", &ExtendedPortInfoRecord::SetCapabilityMask));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("FECModeActive", &ExtendedPortInfoRecord::SetFECModeActive));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("RetransMode", &ExtendedPortInfoRecord::SetRetransMode));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("FDR10FECModeSupported", &ExtendedPortInfoRecord::SetFDR10FECModeSupported));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("FDR10FECModeEnabled", &ExtendedPortInfoRecord::SetFDR10FECModeEnabled));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("FDRFECModeSupported", &ExtendedPortInfoRecord::SetFDRFECModeSupported));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("FDRFECModeEnabled", &ExtendedPortInfoRecord::SetFDRFECModeEnabled));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("EDR20FECModeSupported", &ExtendedPortInfoRecord::SetEDR20FECModeSupported));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("EDR20FECModeEnabled", &ExtendedPortInfoRecord::SetEDR20FECModeEnabled));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("EDRFECModeSupported", &ExtendedPortInfoRecord::SetEDRFECModeSupported));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("EDRFECModeEnabled", &ExtendedPortInfoRecord::SetEDRFECModeEnabled));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("FDR10RetranSupported", &ExtendedPortInfoRecord::SetFDR10RetranSupported));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("FDR10RetranEnabled", &ExtendedPortInfoRecord::SetFDR10RetranEnabled));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("FDRRetranSupported", &ExtendedPortInfoRecord::SetFDRRetranSupported));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("FDRRetranEnabled", &ExtendedPortInfoRecord::SetFDRRetranEnabled));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("EDR20RetranSupported", &ExtendedPortInfoRecord::SetEDR20RetranSupported));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("EDR20RetranEnabled", &ExtendedPortInfoRecord::SetEDR20RetranEnabled));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("EDRRetranSupported", &ExtendedPortInfoRecord::SetEDRRetranSupported));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("EDRRetranEnabled", &ExtendedPortInfoRecord::SetEDRRetranEnabled));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("IsSpecialPort", &ExtendedPortInfoRecord::SetIsSpecialPort));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("SpecialPortType", &ExtendedPortInfoRecord::SetSpecialPortType));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("SpecialPortCapabilityMask", &ExtendedPortInfoRecord::SetSpecialPortCapabilityMask));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("HDRFECModeSupported", &ExtendedPortInfoRecord::SetHDRFECModeSupported));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("HDRFECModeEnabled", &ExtendedPortInfoRecord::SetHDRFECModeEnabled));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("OOOSLMask", &ExtendedPortInfoRecord::SetOOOSLMask));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("AdaptiveTimeoutSLMask", &ExtendedPortInfoRecord::SetAdaptiveTimeoutSLMask));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("NDRFECModeSupported", &ExtendedPortInfoRecord::SetNDRFECModeSupported));

    parse_section_info.push_back(ParseFieldInfo <class ExtendedPortInfoRecord>
             ("NDRFECModeEnabled", &ExtendedPortInfoRecord::SetNDRFECModeEnabled));


    IBDIAG_RETURN(rc);
}

int PortInfoExtendedRecord::Init(vector < ParseFieldInfo
                                      <class PortInfoExtendedRecord> > &parse_section_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("NodeGuid", &PortInfoExtendedRecord::SetNodeGuid));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("PortGuid", &PortInfoExtendedRecord::SetPortGuid));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("PortNum", &PortInfoExtendedRecord::SetPortNum));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("FECModeActive", &PortInfoExtendedRecord::SetFECModeActive));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("FDRFECModeSupported", &PortInfoExtendedRecord::SetFDRFECModeSupported));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("FDRFECModeEnabled", &PortInfoExtendedRecord::SetFDRFECModeEnabled));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("EDRFECModeSupported", &PortInfoExtendedRecord::SetEDRFECModeSupported));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("EDRFECModeEnabled", &PortInfoExtendedRecord::SetEDRFECModeEnabled));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("HDRFECModeSupported", &PortInfoExtendedRecord::SetHDRFECModeSupported));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("HDRFECModeEnabled", &PortInfoExtendedRecord::SetHDRFECModeEnabled));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("NDRFECModeSupported", &PortInfoExtendedRecord::SetNDRFECModeSupported));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("NDRFECModeEnabled", &PortInfoExtendedRecord::SetNDRFECModeEnabled));

    parse_section_info.push_back(ParseFieldInfo <class PortInfoExtendedRecord>
             ("CapabilityMask", &PortInfoExtendedRecord::SetCapabilityMask));

    IBDIAG_RETURN(rc);
}

int PortHierarchyInfoRecord::Init(vector < ParseFieldInfo
                                      <class PortHierarchyInfoRecord> > &parse_section_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("NodeGUID", &PortHierarchyInfoRecord::SetNodeGUID));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("PortGUID", &PortHierarchyInfoRecord::SetPortGUID));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("PortNum", &PortHierarchyInfoRecord::SetPortNum));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("Bus", &PortHierarchyInfoRecord::SetBus));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("Device", &PortHierarchyInfoRecord::SetDevice));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("Function", &PortHierarchyInfoRecord::SetFunction));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("Type", &PortHierarchyInfoRecord::SetType));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("SlotType", &PortHierarchyInfoRecord::SetSlotType));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("SlotValue", &PortHierarchyInfoRecord::SetSlotValue));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("ASIC", &PortHierarchyInfoRecord::SetASIC));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("Cage", &PortHierarchyInfoRecord::SetCage));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("Port", &PortHierarchyInfoRecord::SetPort));

    parse_section_info.push_back(ParseFieldInfo <class PortHierarchyInfoRecord>
             ("Split", &PortHierarchyInfoRecord::SetSplit));

    IBDIAG_RETURN(rc);
}

int PhysicalHierarchyInfoRecord::Init(vector < ParseFieldInfo
                                          <class PhysicalHierarchyInfoRecord> > &parse_section_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    parse_section_info.push_back(ParseFieldInfo <class PhysicalHierarchyInfoRecord>
             ("NodeGUID", &PhysicalHierarchyInfoRecord::SetNodeGUID));

    parse_section_info.push_back(ParseFieldInfo <class PhysicalHierarchyInfoRecord>
             ("CampusSerialNum", &PhysicalHierarchyInfoRecord::SetCampusSerialNum));

    parse_section_info.push_back(ParseFieldInfo <class PhysicalHierarchyInfoRecord>
             ("RoomSerialNum", &PhysicalHierarchyInfoRecord::SetRoomSerialNum));

    parse_section_info.push_back(ParseFieldInfo <class PhysicalHierarchyInfoRecord>
             ("RackSerialNum", &PhysicalHierarchyInfoRecord::SetRackSerialNum));

    parse_section_info.push_back(ParseFieldInfo <class PhysicalHierarchyInfoRecord>
             ("SystemType", &PhysicalHierarchyInfoRecord::SetSystemType));

    parse_section_info.push_back(ParseFieldInfo <class PhysicalHierarchyInfoRecord>
             ("SystemTopUNum", &PhysicalHierarchyInfoRecord::SetSystemTopUNum));

    parse_section_info.push_back(ParseFieldInfo <class PhysicalHierarchyInfoRecord>
             ("BoardType", &PhysicalHierarchyInfoRecord::SetBoardType));

    parse_section_info.push_back(ParseFieldInfo <class PhysicalHierarchyInfoRecord>
             ("BoardSlotNum", &PhysicalHierarchyInfoRecord::SetBoardSlotNum));

    parse_section_info.push_back(ParseFieldInfo <class PhysicalHierarchyInfoRecord>
             ("DeviceSerialNum", &PhysicalHierarchyInfoRecord::SetDeviceSerialNum));

    IBDIAG_RETURN(rc);
}

// AR Info
int ARInfoRecord::Init(vector < ParseFieldInfo <class ARInfoRecord> > &parse_section_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("NodeGUID", &ARInfoRecord::SetNodeGuid));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("e", &ARInfoRecord::Set_e));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("is_arn_sup", &ARInfoRecord::Set_is_arn_sup));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("is_frn_sup", &ARInfoRecord::Set_is_frn_sup));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("is_fr_sup", &ARInfoRecord::Set_is_fr_sup));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("fr_enabled", &ARInfoRecord::Set_fr_enabled));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("rn_xmit_enabled", &ARInfoRecord::Set_rn_xmit_enabled));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("is_ar_trails_supported", &ARInfoRecord::Set_is_ar_trails_supported));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("sub_grps_active", &ARInfoRecord::Set_sub_grps_active));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("group_table_copy_sup", &ARInfoRecord::Set_group_table_copy_sup));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("direction_num_sup", &ARInfoRecord::Set_direction_num_sup));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("is4_mode", &ARInfoRecord::Set_is4_mode));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("glb_groups", &ARInfoRecord::Set_glb_groups));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("by_sl_cap", &ARInfoRecord::Set_by_sl_cap));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("by_sl_en", &ARInfoRecord::Set_by_sl_en));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("by_transp_cap", &ARInfoRecord::Set_by_transp_cap));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("dyn_cap_calc_sup", &ARInfoRecord::Set_dyn_cap_calc_sup));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("group_cap", &ARInfoRecord::Set_group_cap));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("group_top", &ARInfoRecord::Set_group_top));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("group_table_cap", &ARInfoRecord::Set_group_table_cap));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("string_width_cap", &ARInfoRecord::Set_string_width_cap));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("ar_version_cap", &ARInfoRecord::Set_ar_version_cap));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("rn_version_cap", &ARInfoRecord::Set_rn_version_cap));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("sub_grps_supported", &ARInfoRecord::Set_sub_grps_supported));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("enable_by_sl_mask", &ARInfoRecord::Set_enable_by_sl_mask));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("by_transport_disable", &ARInfoRecord::Set_by_transport_disable));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("ageing_time_value", &ARInfoRecord::Set_ageing_time_value));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("is_whbf_supported", &ARInfoRecord::Set_is_whbf_supported));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("whbf_en", &ARInfoRecord::Set_whbf_en));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("is_hbf_supported", &ARInfoRecord::Set_is_hbf_supported));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("by_sl_hbf_en", &ARInfoRecord::Set_by_sl_hbf_en));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("enable_by_sl_mask_hbf", &ARInfoRecord::Set_enable_by_sl_mask_hbf));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("whbf_granularity", &ARInfoRecord::Set_whbf_granularity));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("is_symmetric_hash_supported", &ARInfoRecord::Set_is_symmetric_hash_supported));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                 ("is_dceth_hash_supported", &ARInfoRecord::Set_is_dceth_hash_supported));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                  ("is_bth_dqp_hash_supported", &ARInfoRecord::Set_is_bth_dqp_hash_supported));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                   ("is_pfrn_supported", &ARInfoRecord::Set_is_pfrn_supported));

    parse_section_info.push_back(ParseFieldInfo <class ARInfoRecord>
                   ("pfrn_enabled", &ARInfoRecord::Set_pfrn_enabled));

    IBDIAG_RETURN(rc);
}
