/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdarg.h>

#include "infiniband/ibdiag/ibdiag_fabric_errs.h"
#include "infiniband/ibutils/output_stream_primitives.h"
#include "infiniband/ibutils/trim.h"
#include <algorithm>

/*****************************************************/
/*****************************************************/
/*****************************************************/
void DumpCSVFabricErrorListTable(list_p_fabric_general_err& list_errors, CSVOut &csv_out,
                                 string name, EnFabricErrLevel_t type)
{
    IBDIAG_ENTER;
    if (list_errors.empty())
        IBDIAG_RETURN_VOID;

    for (unsigned i = 0; i < name.size(); ++i) {
        if (name[i] == ' ') {
            name[i] = '_';
            continue;
        }
        if ((name[i] >= 'a') && (name[i] <= 'z')) {
            name[i] = (char)((name[i] - 'a') + 'A');
        }
    }

    int status = 0;
    if (type == EN_FABRIC_ERR_WARNING)
        status = csv_out.DumpStart(string("WARNINGS_" + name).c_str());
    else
        status = csv_out.DumpStart(string("ERRORS_" + name).c_str());
    if (status)
        IBDIAG_RETURN_VOID;

    stringstream sstream;
    sstream << "Scope," << "NodeGUID," << "PortGUID," << "PortNumber," << "EventName," << "Summary" << endl;
    csv_out.WriteBuf(sstream.str());

    for (list_p_fabric_general_err::iterator it = list_errors.begin(); it != list_errors.end(); ++it) {
        sstream.str("");
        sstream << (*it)->GetCSVErrorLine() << endl;
        csv_out.WriteBuf(sstream.str());
    }

    if (type == EN_FABRIC_ERR_WARNING)
        csv_out.DumpEnd(string("WARNINGS_" + name).c_str());
    else
        csv_out.DumpEnd(string("ERRORS_" + name).c_str());

    IBDIAG_RETURN_VOID;
}


void CleanFabricErrorsList(list_p_fabric_general_err& list_errors)
{
    IBDIAG_ENTER;
    for (list_p_fabric_general_err::iterator it = list_errors.begin(); it != list_errors.end(); ++it)
        if (*it)
            delete *it;
    list_errors.clear();
    IBDIAG_RETURN_VOID;
}


string DescToCsvDesc(const string & desc)
{
    IBDIAG_ENTER

    string result = trim(desc);

    if (result.empty())
        IBDIAG_RETURN("NA");

    // change ',' into '-'
    size_t idx = result.find(',', 0);
    while (idx != string::npos) {
        result[idx] = '-';
        idx = result.find(',', idx + 1);
    }

    IBDIAG_RETURN(result);
}

//replaces all occurrences of the characters from the "from" string
//to characters in the "to" string.
string DescToCsvDesc(const string &desc, const string& from, const string& to)
{
    IBDIAG_ENTER

    string result = desc;
    string::const_iterator itrFrom = from.begin();
    string::const_iterator itrTo = to.begin();
    while (itrFrom != from.end() && itrTo != to.end()) {
        std::replace(result.begin(), result.end(), *itrFrom, *itrTo);
        itrFrom++;
        itrTo++;
    }

    std::replace(result.begin(), result.end(), ',', '-');
    result = trim(result);

    if (result.empty())
        IBDIAG_RETURN("NA");

    IBDIAG_RETURN(result);
}

void ResetAccumulatedErrors(list_p_fabric_general_err &list_errors)
{
    IBDIAG_ENTER

    std::map<int, FabricErrGeneral*> errorByLine;
    for (list_p_fabric_general_err::iterator it = list_errors.begin(); it != list_errors.end();) {
        FabricErrGeneral *p_error = *it;
        if (!p_error || !p_error->IsAccumulable()) {
            it++;
            continue;
        }

        std::pair<std::map<int, FabricErrGeneral*>::iterator, bool> entry =
                errorByLine.insert(std::pair<int, FabricErrGeneral*>(p_error->GetLine(), p_error));
        if (!entry.second) {
            entry.first->second->IncCount();

            delete p_error;
            it = list_errors.erase(it);
        }
        else
           it++;
    }

    IBDIAG_RETURN_VOID;
}

/*****************************************************/
/*****************************************************/
/*****************************************************/
string FabricErrDuplicatedNodeGuid::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->duplicated_guid,
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

FabricErrDuplicatedNodeGuid::FabricErrDuplicatedNodeGuid(IBNode *p_node, string direct_route_to_node_str, u_int64_t duplicated_guid) :
        FabricErrGuid(p_node, direct_route_to_node_str, duplicated_guid)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_DUPLICATED_NODE_GUID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Node GUID = " U64H_FMT " is duplicated at: ", this->duplicated_guid);
    this->description.assign(buffer);
    this->description += "Node = ";
    this->description += this->p_node->getName();
    if (this->p_node->orig_description.length()) {
        this->description += " (";
        this->description += this->p_node->orig_description;
        this->description += ")";
    }
    this->description += ", DR = ";
    this->description += this->direct_route_to_node_str;
    IBDIAG_RETURN_VOID;
}

string FabricErrDuplicatedPortGuid::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U64H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_node->guid_get(),
            this->duplicated_guid,
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

FabricErrDuplicatedPortGuid::FabricErrDuplicatedPortGuid(IBNode *p_node, string direct_route_to_node_str, u_int64_t duplicated_guid) :
        FabricErrGuid(p_node, direct_route_to_node_str, duplicated_guid)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_DUPLICATED_PORT_GUID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Port GUID = " U64H_FMT " is duplicated at: ", this->duplicated_guid);
    this->description.assign(buffer);
    this->description += "Node = ";
    this->description += this->p_node->getName();
    this->description += ", DR = ";
    this->description += this->direct_route_to_node_str;
    IBDIAG_RETURN_VOID;
}

//**********************************************************
//**********************************************************
//**********************************************************
FabricInvalidGuid::FabricInvalidGuid(uint64_t in_guid, const string &in_route,
            const string &in_descr, const string &in_type):
            FabricErrGeneral(),
            guid(in_guid), direct_route(in_route)
{
    IBDIAG_ENTER;
    this->scope = SCOPE_CLUSTER;
    this->err_desc = in_descr;

    stringstream strStream;
    strStream << "Invalid " << in_type << " GUID = " << PTR(this->guid)
              << ", DR = " << this->direct_route;

    this->description = strStream.str();
    IBDIAG_RETURN_VOID;
}

string FabricInvalidGuid::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    stringstream strStream;
    strStream << this->scope
              << ',' << PTR(this->guid)
              << ',' << this->err_desc
              << ",\"" << DescToCsvDesc(this->description) << '\"';
    IBDIAG_RETURN(strStream.str());
}


/*****************************************************/
/*****************************************************/
/*****************************************************/
FabricErrNotAllDevicesSupCap::FabricErrNotAllDevicesSupCap(string desc)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_NOT_ALL_NODES_SUP_CAP);
    this->description = "Not all devices support";
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID;
}

string FabricErrNotAllDevicesSupCap::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U8H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            0x0,            //NA
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrNotAllDevicesSupCap::GetErrorLine()
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(description);
}

/*****************************************************/
/*****************************************************/
/*****************************************************/
string FabricErrNode::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_node->guid_get(),
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrNode::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_node->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}


FabricErrNodeNotRespond::FabricErrNodeNotRespond(IBNode *p_node, string desc) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_NOT_RESPOND);
    this->description = "No response for MAD";
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID;
}


FabricErrNodeWrongConfig::FabricErrNodeWrongConfig(IBNode *p_node, string desc) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_WRONG_CONFIG);
    this->description = "Node is configured with wrong value.";
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID;
}


FabricErrNodeDuplicatedNodeDesc::FabricErrNodeDuplicatedNodeDesc(IBNode *p_node) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_NODE_DUPLICATED_NODE_DESC);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Node with GUID=" U64H_FMT " is configured with duplicated node description - %s",
            this->p_node->guid_get(),
            this->p_node->description.c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrNodeMlnxCountersPageVer::FabricErrNodeMlnxCountersPageVer(
        IBNode *p_node,
        unsigned int page,
        unsigned int curr_ver,
        unsigned int latest_ver) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_MLNX_CNTRS_WRONG_PAGE_VER);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "Node has Mlnx Counters Page" U32D_FMT " version " U32D_FMT   \
            ", while ibutils supports version " U32D_FMT,
            page, curr_ver, latest_ver);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrNodeWrongFWVer::FabricErrNodeWrongFWVer(IBNode *p_node, string desc) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_WRONG_FW_VER);
    this->description = desc;
    IBDIAG_RETURN_VOID;
}


FabricErrNodeNotSupportCap::FabricErrNodeNotSupportCap(IBNode *p_node, string desc) : FabricErrNode(p_node)
{
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_NOT_SUPPORT_CAP);
    this->description = desc;
    IBDIAG_RETURN_VOID;

}

FabricErrNodeInvalidLid::FabricErrNodeInvalidLid(IBNode *p_node, phys_port_t port, lid_t lid, uint8_t lmc) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_NODE_INVALID_LID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Configured with invalid lid=%u lmc=%u on port %u", lid, lmc, port);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

FabricErrSmpGmpFwMismatch::FabricErrSmpGmpFwMismatch(IBNode *p_node,
                              fw_version_obj_t &smp_fw,
                              fw_version_obj_t &gmp_fw) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_SMP_GMP_FW_MISMATCH);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Mismatch in reported firmware version. "
                    "SMP: %u.%u.%u, GMP: %u, %u, %u",
                    smp_fw.major, smp_fw.minor, smp_fw.sub_minor,
                    gmp_fw.major, gmp_fw.minor, gmp_fw.sub_minor);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

FabricErrSmpGmpCapMaskExist::FabricErrSmpGmpCapMaskExist(
    IBNode *p_node, bool is_smp, capability_mask_t &mask) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_NODE_SMP_GMP_CAP_MASK_EXIST);
    char buffer[1024];
    stringstream mask_str;

    mask_str << mask;
    snprintf(buffer, sizeof(buffer), "A %s capability mask already exists. Overriden with %s. ",
            is_smp ? "SMP" : "GMP", mask_str.str().c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}
/******************************************************/
/*****************************************************/
/*****************************************************/
string FabricErrSM::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    if (this->p_sm_obj)
        snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
                this->scope.c_str(),
                this->p_sm_obj->p_port->p_node->guid_get(),
                this->p_sm_obj->p_port->guid_get(),
                this->p_sm_obj->p_port->num,
                this->err_desc.c_str(),
                DescToCsvDesc(this->description).c_str());
    else
        snprintf(buffer, sizeof(buffer), STR_FMT "," U8H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
                this->scope.c_str(),
                0x0,            //NA
                0x0,            //NA
                0x0,            //NA
                this->err_desc.c_str(),
                DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrSM::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    if (this->p_sm_obj) {
        line = this->p_sm_obj->p_port->getExtendedName();
        line += " - ";
    }
    line += this->description;
    IBDIAG_RETURN(line);
}


FabricErrSMNotFound::FabricErrSMNotFound(sm_info_obj_t *p_sm_obj) : FabricErrSM(p_sm_obj)
{
    IBDIAG_ENTER;
    this->p_sm_obj = NULL;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_SM_NOT_FOUND);
    this->description = "Not found master subnet manager in fabric";
    IBDIAG_RETURN_VOID;
}


FabricErrSMUnknownState::FabricErrSMUnknownState(sm_info_obj_t *p_sm_obj) : FabricErrSM(p_sm_obj)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_SM_UNKNOWN_STATE);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Running subnet manager with unknown state=%u", this->p_sm_obj->smp_sm_info.SmState);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrSMManyExists::FabricErrSMManyExists(sm_info_obj_t *p_sm_obj) : FabricErrSM(p_sm_obj)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_SM_MANY_EXISTS);
    this->description = "Running duplicated master subnet manager";
    IBDIAG_RETURN_VOID;
}


FabricErrSMNotCorrect::FabricErrSMNotCorrect(sm_info_obj_t *p_sm_obj) : FabricErrSM(p_sm_obj)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_SM_NOT_CORRECT);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Master subnet manager that is running is not the correct one, subnet manager with priority=%u, GUID=" U64H_FMT " should be the master",
            this->p_sm_obj->smp_sm_info.Priority,
            this->p_sm_obj->smp_sm_info.GUID);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrPM::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    stringstream stream_ss;
    stream_ss << this->scope                            << ','
            << PTR(this->p_port->p_node->guid_get())    << ','
            << PTR(this->p_port->guid_get())            << ','
            <<  +this->p_port->num                      << ','
            << this->err_desc                           << ','
            << '\"' << DescToCsvDesc(this->description) << '\"';

    IBDIAG_RETURN(stream_ss.str());
}

string FabricErrPM::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getExtendedName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

FabricErrPMInvalidDelta::FabricErrPMInvalidDelta(IBPort *p_port,
                            const string &counters) : FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->dump_csv_only = true;
    this->level = EN_FABRIC_ERR_WARNING;
    this->scope = SCOPE_PORT;
    this->err_desc = FER_PM_NEGATIVE_DELTA_COUNTERS;
    this->description = "Negative delta values for PM counter(s):" + counters;
    IBDIAG_RETURN_VOID;
}

FabricErrPMCounterInvalidSize::FabricErrPMCounterInvalidSize(IBPort *p_port,
    const string& counter_name, u_int8_t real_size) : FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PM_COUNTER_INVALID_SIZE);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Invalid PM counter %s size. Counter real_size is %d",
            counter_name.c_str(), real_size);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

FabricErrPMCounterOverflow::FabricErrPMCounterOverflow(IBPort *p_port, string counter_name,
        u_int64_t overflow_value) : FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PM_COUNTER_OVERFLOW);
    char buffer1[512];
    snprintf(buffer1, sizeof(buffer1), U64D_FMT, overflow_value);
    char buffer2[1024];
    snprintf(buffer2, sizeof(buffer2), "%-35s : %-10s (overflow)", counter_name.c_str(), buffer1);
    this->description.assign(buffer2);
    IBDIAG_RETURN_VOID;
}

FabricErrPMBaseCalcCounterOverflow::FabricErrPMBaseCalcCounterOverflow(
        IBPort *p_port, string counter_name) : FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PM_BASE_COUNTER_OVERFLOW);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Base PM counters of %s counter are overflown,"
                   "  please run ibdiagnet -PC to reset.", counter_name.c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

FabricErrPMCounterExceedThreshold::FabricErrPMCounterExceedThreshold(IBPort *p_port, string counter_name,
        u_int64_t expected_value, u_int64_t actual_value) : FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PM_COUNTER_EXCEED_THRESH);
    char buffer1[512];
    snprintf(buffer1, sizeof(buffer1), U64D_FMT, actual_value);
    char buffer2[1024];
    snprintf(buffer2, sizeof(buffer2), "%-35s : %-10s (threshold=" U64D_FMT ")", counter_name.c_str(), buffer1, expected_value);
    this->description.assign(buffer2);
    IBDIAG_RETURN_VOID;
}


FabricErrPMCounterNotSupported::FabricErrPMCounterNotSupported(IBPort *p_port, string counter_name) :
        FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PM_COUNTER_NOT_SUPPORTED);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "%-35s : %-10s (Not Supported)", counter_name.c_str(), "");
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrPMErrCounterIncreased::FabricErrPMErrCounterIncreased(IBPort *p_port, string counter_name,
        u_int64_t expected_value, u_int64_t actual_value) : FabricErrPM(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PM_COUNTER_INCREASED);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "\"%s\" increased during the run "  \
            "(difference value=" U64D_FMT ",difference allowed threshold=" U64D_FMT ")",
            counter_name.c_str(),
            actual_value, expected_value);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrPMCountersAll::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string res = "";
    if (this->csv_err_line != "") {
        res += this->csv_err_line;
    }
    IBDIAG_RETURN(res);
}

string FabricErrPMCountersAll::GetErrorLine()
{
    IBDIAG_ENTER;
    char buffer[1024];
    string res = "";
    if (this->err_line != "") {
        snprintf(buffer, sizeof(buffer), "lid=" U16H_FMT " dev=" U32D_FMT " %s\n",
                this->p_port->base_lid,
                this->p_port->p_node->devId,
                this->p_port->getExtendedName().c_str());
        res += buffer;
        snprintf(buffer, sizeof(buffer), "%-35s : %-10s\n",
                "Performance Monitor counter", "Value");
        res += "        ";
        res += buffer;
        res += this->err_line;
    }
    IBDIAG_RETURN(res);
}


FabricErrPMCountersAll::FabricErrPMCountersAll(IBPort *p_port, list_p_pm_err& pm_errors) :
        FabricErrGeneral(), p_port(p_port), err_line(""), csv_err_line("")
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);                     //dummy one
    this->err_desc.assign("PM_COUNTERS_ALL");           //dummy one
    for (list_p_pm_err::iterator it = pm_errors.begin(); it != pm_errors.end(); ++it) {
        if (it != pm_errors.begin()) {
            this->err_line += "\n";
            this->csv_err_line += "\n";
        }
        this->err_line += "        ";
        this->err_line += (*it)->GetErrorLine();
        this->csv_err_line += (*it)->GetCSVErrorLine();
    }
    IBDIAG_RETURN_VOID;
}


/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrBER::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port->p_node->guid_get(),
            this->p_port->guid_get(),
            this->p_port->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrBER::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getExtendedName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

FabricErrFwBERExceedThreshold::FabricErrFwBERExceedThreshold(IBPort *p_port,
        double thresh, double value, IBBERType ber_type, string desc) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;

    scope.assign(SCOPE_PORT);
    err_desc.assign(FER_BER_EXCEED_THRESH);

    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "BER exceeds threshold - BER type: %s, FEC mode: %s, "
                                     "BER value = %e / threshold = %e %s",
                                        ber_type2char(ber_type),
                                        fec2char(p_port->get_fec_mode()),
                                        value,
                                        thresh,
                                        desc.c_str());
    description.assign(buffer);

    IBDIAG_RETURN_VOID;
}

FabricErrBERExceedThreshold::FabricErrBERExceedThreshold(IBPort *p_port,
        u_int64_t expected_value, long double actual_value) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_BER_EXCEED_THRESH);
    char buffer[1024];
    if (expected_value == OVERFLOW_VAL_64_BIT)
        snprintf(buffer, sizeof(buffer), "BER exceeds the threshold in port = %s"    \
                "(BER value=%Le, threshold=%e)",
                this->p_port->getName().c_str(),
                (1.0 / actual_value),
                0.0);
    else
        snprintf(buffer, sizeof(buffer), "BER exceeds the threshold in port = %s"    \
                "(BER value=%Le, threshold=%e)",
                this->p_port->getName().c_str(),
                (1.0 / actual_value),
                (1.0 / (double)expected_value));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrEffBERExceedThreshold::FabricErrEffBERExceedThreshold(IBPort *p_port,
        u_int64_t expected_value, long double actual_value) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_BER_EXCEED_THRESH);
    char buffer[1024];
    if (expected_value == OVERFLOW_VAL_64_BIT)
        snprintf(buffer, sizeof(buffer), "Effective BER exceeds the threshold in port = %s"    \
                "(BER value=%Le, threshold=%e) - "
                "Please upgrade FW to support Effective or Symbol BER",
                this->p_port->getName().c_str(),
                (1.0 / actual_value),
                0.0);
    else
        snprintf(buffer, sizeof(buffer), "Effective BER exceeds the threshold in port = %s"    \
                "(BER value=%Le, threshold=%e) - "
                "Please upgrade FW to support Effective or Symbol BER",
                this->p_port->getName().c_str(),
                (1.0 / actual_value),
                (1.0 / (double)expected_value));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrBERNoRcvData::FabricErrBERNoRcvData(IBPort *p_port) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_BER_NO_RCV_DATA);
    this->description = "No received data - can't calculate BER";
    IBDIAG_RETURN_VOID;
}


FabricErrBERIsZero::FabricErrBERIsZero(IBPort *p_port) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_BER_IS_ZERO);
    this->description = "BER value is ZERO";
    IBDIAG_RETURN_VOID;
}

FabricErrEffBERIsZero::FabricErrEffBERIsZero(IBPort *p_port) : FabricErrBER(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_BER_IS_ZERO);
    this->description = "Effective BER value is ZERO - " \
                        "Please upgrade FW to support Effective or Symbol BER";
    IBDIAG_RETURN_VOID;
}

/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrPort::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port->p_node->guid_get(),
            this->p_port->guid_get(),
            this->p_port->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrPort::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getExtendedName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}


FabricErrPortZeroLid::FabricErrPortZeroLid(IBPort *p_port) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_PORT_ZERO_LID);
    this->description = "Configured with ZERO lid";
    IBDIAG_RETURN_VOID;
}


FabricErrPortDuplicatedLid::FabricErrPortDuplicatedLid(IBPort *p_port) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_PORT_DUPLICATED_LID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Configured with duplicated lid=%u", p_port->base_lid);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

FabricErrPortNotRespond::FabricErrPortNotRespond(IBPort *p_port, string desc) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_NOT_RESPOND);
    this->description = "No response for MAD";
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID;
}


FabricErrPortWrongConfig::FabricErrPortWrongConfig(IBPort *p_port, string desc) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_WRONG_CONFIG);
    this->description = "Port is configured with wrong value";
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID;
}


FabricErrPortNotSupportCap::FabricErrPortNotSupportCap(IBPort *p_port, string desc) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_NOT_SUPPORT_CAP);
    this->description = desc;
    IBDIAG_RETURN_VOID;
}

FabricErrPortInvalidValue::FabricErrPortInvalidValue(IBPort *p_port, string desc) : FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_INVALID_VALUE);
    this->description = desc;
    IBDIAG_RETURN_VOID;
}

/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrDR::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line = "";
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U8H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            0x0,            //NA
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

FabricErrDR::FabricErrDR(string direct_route_str) :
        FabricErrGeneral(), direct_route_str(direct_route_str)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(FER_DIRECT_ROUTE);
    this->description = "Bad link was found. ";
    this->description += this->direct_route_str;
    IBDIAG_RETURN_VOID;
}


/******************************************************/
/******************************************************/
/******************************************************/
string FabricErrLink::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[2096];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port1->p_node->guid_get(),
            this->p_port1->guid_get(),
            this->p_port1->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    csv_line += "\n";
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port2->p_node->guid_get(),
            this->p_port2->guid_get(),
            this->p_port2->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line += string(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrLink::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = "Link: ";
    line += this->p_port1->getExtendedName();
    line += "<-->";
    line += this->p_port2->getExtendedName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}


FabricErrLinkLogicalStateWrong::FabricErrLinkLogicalStateWrong(IBPort *p_port1, IBPort *p_port2) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_STATE_WRONG);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Logical state is different in connected ports (port=%s state is %s and remote port=%s state is %s)",
            this->p_port1->getName().c_str(), portstate2char(this->p_port1->get_internal_state()),
            this->p_port2->getName().c_str(), portstate2char(this->p_port2->get_internal_state()));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrLinkLogicalStateNotActive::FabricErrLinkLogicalStateNotActive(IBPort *p_port1, IBPort *p_port2) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_STATE_NOT_ACTIVE);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Link logical state is %s", portstate2char(this->p_port1->get_internal_state()));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrLinkDifferentSpeed::FabricErrLinkDifferentSpeed(IBPort *p_port1, IBPort *p_port2) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_DIFFERENT_SPEED);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "Speed is different in connected ports (port=%s speed is %s and remote port=%s speed is %s)",
            this->p_port1->getName().c_str(), speed2char(this->p_port1->get_internal_speed()),
            this->p_port2->getName().c_str(), speed2char(this->p_port2->get_internal_speed()));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrLinkUnexpectedSpeed::FabricErrLinkUnexpectedSpeed(IBPort *p_port1, IBPort *p_port2, string desc) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_UNEXPECTED_SPEED);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
             "Unexpected actual link speed %s", speed2char(p_port1->get_internal_speed()));
    this->description.assign(buffer);
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID;
}


FabricErrLinkDifferentWidth::FabricErrLinkDifferentWidth(IBPort *p_port1, IBPort *p_port2) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_DIFFERENT_WIDTH);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
             "Width is different in connected ports (port=%s width is %s and remote port=%s width is %s)",
             this->p_port1->getName().c_str(), width2char(this->p_port1->get_internal_width()),
             this->p_port2->getName().c_str(), width2char(this->p_port2->get_internal_width()));
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}


FabricErrLinkUnexpectedWidth::FabricErrLinkUnexpectedWidth(IBPort *p_port1, IBPort *p_port2, string desc) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_UNEXPECTED_WIDTH);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
             "Unexpected width, actual link width is %s", width2char(p_port1->get_internal_width()));
    this->description.assign(buffer);
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID;
}


FabricErrLinkAutonegError::FabricErrLinkAutonegError(IBPort *p_port1, IBPort *p_port2, string desc) :
        FabricErrLink(p_port1, p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_LINK_LOGICAL_AUTONEG_ERR);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Autoneg should fail on this link");
    this->description.assign(buffer);
    if (desc != "") {
        this->description += " ";
        this->description += desc;
    }
    IBDIAG_RETURN_VOID;
}

/******************************************************/
/******************************************************/
/******************************************************/
FabricErrPKeyMismatch::FabricErrPKeyMismatch(IBPort *p_port1, IBPort *p_port2,
        string port1_pkey_str, string port2_pkey_str) :
	FabricErrGeneral(), p_port1(p_port1), p_port2(p_port2)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PKEY_MISMATCH);
    this->description = "Mismatching pkeys between ";

    this->description += this->p_port1->getName().c_str();
    if (port1_pkey_str != "") {
        this->description += " {";
        this->description += port1_pkey_str;
        this->description += "}";
    }

    this->description += " <--> ";

    this->description += this->p_port2->getName().c_str();
    if (port2_pkey_str != "") {
        this->description += " {";
        this->description += port2_pkey_str;
        this->description += "}";
    }

    IBDIAG_RETURN_VOID;
}

string FabricErrPKeyMismatch::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[2096];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port1->p_node->guid_get(),
            this->p_port1->guid_get(),
            this->p_port1->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    csv_line += "\n";
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port2->p_node->guid_get(),
            this->p_port2->guid_get(),
            this->p_port2->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line += string(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrPKeyMismatch::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->description;
    IBDIAG_RETURN(line);
}


/******************************************************/
/******************************************************/
/******************************************************/
FabricErrAGUID::FabricErrAGUID(IBPort *p_port1, string owner_name, u_int64_t guid, string guid_type):
        FabricErrGeneral(),
        p_port(p_port1), guid_owner_name(owner_name),
        duplicated_guid(guid), guid_type(guid_type)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_AGUID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Alias GUID " U64H_FMT " on port %s, used on %s as %s",
            this->duplicated_guid,
            this->p_port->getName().c_str(),
            this->guid_owner_name.c_str(),
            this->guid_type.c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

string FabricErrAGUID::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->duplicated_guid,
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrAGUID::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getExtendedName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

FabricErrAGUIDInvalidFirstEntry::FabricErrAGUIDInvalidFirstEntry(IBPort *port, u_int64_t first_alias_guid):
    FabricErrGeneral(),
    p_port(port),
    guid_zero_index(first_alias_guid)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_AGUID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Invalid guid ( " U64H_FMT " ) at the entry zero in guid info table of port %s. "
            "It must be equal to the port guid ( " U64H_FMT  " ).",
            this->guid_zero_index,
            this->p_port->getName().c_str(),
            this->p_port->guid_get());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

string FabricErrAGUIDInvalidFirstEntry::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];

    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->guid_zero_index,
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());

    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrAGUIDInvalidFirstEntry::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getExtendedName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

/******************************************************/
/******************************************************/
/******************************************************/
FabricErrPortInfoFail::FabricErrPortInfoFail(IBNode *p_node,
                                             unsigned int    port_num,
                                             const char *error_desc) :
	FabricErrGeneral(), m_p_node(p_node),m_port_num(port_num)
{
    IBDIAG_ENTER;
    char port_num_buff[3];
    snprintf(port_num_buff,2,"%u",m_port_num);
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_INFO_FAILED);
    this->description = "Port info failed on ";

    this->description += m_p_node->name.c_str();
    this->description += "  port number:";
    this->description += port_num_buff;
    this->description += "  with error:";
    this->description += error_desc;

    IBDIAG_RETURN_VOID;
}

string FabricErrPortInfoFail::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[2096];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U32D_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            scope.c_str(),
            m_p_node->guid_get(),
            m_port_num,
            err_desc.c_str(),
            DescToCsvDesc(description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrPortInfoFail::GetErrorLine() //TODO TAMIR change to const string &
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(description);
}

/******************************************************/
/******************************************************/
/******************************************************/
FabricErrDiscovery::FabricErrDiscovery(IBNode *p_node,
                                       uint8_t max_hops) :
	FabricErrGeneral(), p_node(p_node), max_hops(max_hops)
{
    IBDIAG_ENTER;
    char buffer[32];
    snprintf(buffer, sizeof(buffer), "%u", max_hops);
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_DISCOVERY_REACHED_MAX_HOP);
    this->description = "Discovery failed on ";
    this->description += p_node->name.c_str();
    this->description += "  when reached hop number:";
    this->description += buffer;
    IBDIAG_RETURN_VOID;
}

string FabricErrDiscovery::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[2096];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U32D_FMT "," "\"" STR_FMT "\"",
            scope.c_str(),
            p_node->guid_get(),
            max_hops,
            DescToCsvDesc(description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

FabricErrVPort::FabricErrVPort(IBVPort *p_port1, string owner_name,
                               u_int64_t guid, string guid_type):
    FabricErrGeneral(),
    p_port(p_port1), guid_owner_name(owner_name),
    duplicated_guid(guid), guid_type(guid_type)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_AGUID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "VPort GUID " U64H_FMT " on port %s, used on %s as %s",
            this->duplicated_guid,
            this->p_port->getName().c_str(),
            this->guid_owner_name.c_str(),
            this->guid_type.c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

string FabricErrVPort::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT
            "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->duplicated_guid,
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());
    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrDiscovery::GetErrorLine()
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(description);
}

string FabricErrVPort::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

FabricErrVPortGUIDInvalidFirstEntry::FabricErrVPortGUIDInvalidFirstEntry(
        IBPort *phys_port, IBVPort *port, u_int64_t first_alias_guid):
        FabricErrGeneral(),
        p_port(port),
        p_phys_port(phys_port),
        guid_at_zero_index(first_alias_guid)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_AGUID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Invalid guid ( " U64H_FMT " ) at the entry"
            " zero in guid info table of port %s. "
            "It must be equal to the port guid ( " U64H_FMT  " ).",
            this->guid_at_zero_index,
            this->p_port->getName().c_str(),
            this->p_phys_port->guid_get());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

string FabricErrVPortGUIDInvalidFirstEntry::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];

    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT
            "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->guid_at_zero_index,
            0x0,            //NA
            0x0,            //NA
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());

    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrVPortGUIDInvalidFirstEntry::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

FabricErrVPortIvalidTopIndex::FabricErrVPortIvalidTopIndex(IBPort *port, uint16_t cap,
        uint16_t top):
        FabricErrGeneral(),
        p_port(port),
        cap_idx(cap),
        top_idx(top)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_VIRT_INFO_INVALID_TOP);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "on port %s, top index %d can't be greater than "
            "capability index %d",
            this->p_port->getName().c_str(),
            this->top_idx,
            this->cap_idx);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

string FabricErrVPortIvalidTopIndex::GetCSVErrorLine()
{
   IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];

    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U64H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port->p_node->guid_get(),
            this->p_port->guid_get(),
            this->p_port->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());

    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrVPortIvalidTopIndex::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getExtendedName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

/********************************************/
/********************************************/
/********************************************/
string FabricErrVPortInvalid::GetCSVErrorLine()
{
   IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];

    snprintf(buffer, sizeof(buffer), STR_FMT "," U64H_FMT "," U64H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            this->p_port->p_node->guid_get(),
            this->p_port->guid_get(),
            this->p_port->num,
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());

    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrVPortInvalid::GetErrorLine()
{
    IBDIAG_ENTER;
    string line;
    line = this->p_port->getExtendedName();
    line += " - ";
    line += this->description;
    IBDIAG_RETURN(line);
}

FabricErrVLidZero::FabricErrVLidZero(IBPort *port, IBVPort *vport):
        FabricErrVPortInvalid(port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_VIRT_INFO_INVALID_VLID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
             "On vport %s, lid_required is 1 but vlid is 0",
             vport->getName().c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

FabricErrVPortInvalidLid::FabricErrVPortInvalidLid(IBPort *port, IBVPort *vport, lid_t lid):
        FabricErrVPortInvalid(port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_VIRT_INFO_INVALID_VLID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
             "Invalid LID on vport %s, vlid = %u",
             vport->getName().c_str(), lid);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

FabricErrInvalidIndexForVLid::FabricErrInvalidIndexForVLid(
        IBPort *port,
        IBVPort *vport,
        u_int16_t lid_by_vport_idx):
        FabricErrVPortInvalid(port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_VIRT_INFO_INVALID_VLID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
             "Invalid index num %d for vport %s, vport by index not found",
             lid_by_vport_idx,
             vport->getName().c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

FabricErrVlidForVlidByIndexIsZero::FabricErrVlidForVlidByIndexIsZero(
        IBPort *port, IBVPort *vport, IBVPort *vport_by_index,
        u_int16_t lid_by_vport_idx):
        FabricErrVPortInvalid(port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_VIRT_INFO_INVALID_VLID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Found vlid 0 in vport %s." \
                    "Found by vport %s with lid_required=0 and index num %d",
            vport_by_index->getName().c_str(),
            vport->getName().c_str(),
            lid_by_vport_idx);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

/******************************************************/
/******************************************************/
/******************************************************/
FabricErrCluster::FabricErrCluster(string err_desc, string desc)
{
    IBDIAG_ENTER;

    this->scope.assign(SCOPE_CLUSTER);
    this->err_desc.assign(err_desc);
    this->description.assign(desc);

    IBDIAG_RETURN_VOID;
}

string FabricErrCluster::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    string csv_line;
    char buffer[1024];

    snprintf(buffer, sizeof(buffer), STR_FMT "," U8H_FMT "," U8H_FMT "," U8H_FMT "," STR_FMT "," "\"" STR_FMT "\"",
            this->scope.c_str(),
            0, // node guid
            0, // port guid
            0, // port num
            this->err_desc.c_str(),
            DescToCsvDesc(this->description).c_str());

    csv_line.assign(buffer);
    IBDIAG_RETURN(csv_line);
}

string FabricErrCluster::GetErrorLine()
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(description);
}

SharpErrClassPortInfo::SharpErrClassPortInfo(string desc):
        FabricErrCluster(FER_SHARP_CLASS_PORT_INFO, desc) { }

SharpErrVersions::SharpErrVersions(string desc):
        FabricErrCluster(FER_SHARP_VERSIONING, desc) { }

SharpErrInvalidActiveVer::SharpErrInvalidActiveVer(IBNode *p_node) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_VERSIONING);
    this->description.assign("AN.active_class_ver is large than CPI.class_version");
    IBDIAG_RETURN_VOID;
}

SharpErrDiffVerMgmtAndSharp::SharpErrDiffVerMgmtAndSharp(IBNode *p_node, int class_ver, int sharp_ver) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_VERSIONING);

    stringstream sstream;
    sstream << "Different active_class_ver(" << class_ver
            << ") and active_sharp_ver(" << sharp_ver << ") on AN";
    this->description.assign(sstream.str());

    IBDIAG_RETURN_VOID;
}

SharpErrEdgeNodeNotFound::SharpErrEdgeNodeNotFound(
        IBNode *p_node, lid_t rlid) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_EDGE_NODE_NOT_FOUND);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "Node has no Edge remote lid: %d",
            rlid);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrParentTreeEdgeNotFound::SharpErrParentTreeEdgeNotFound(
        IBNode *p_node, lid_t rlid, u_int16_t tree_id) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_PARENT_TREE_EDGE_NOT_FOUND);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "Child Tree Node doesn't have QP to Parent Tree Edge remote lid: " U32D_FMT
            " for Tree ID: " U32D_FMT, rlid, tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrMismatchParentChildQPNumber::SharpErrMismatchParentChildQPNumber(
        IBNode *p_node, lid_t parent_lid, u_int32_t parent_qpn, u_int32_t parent_rqpn,
        lid_t lid_child, u_int32_t child_qpn, u_int32_t child_rqpn, u_int16_t tree_id) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_MISMATCH_CHILD_NODE_TO_PARENT_NODE);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "Mismatch: Parent Tree Node lid: " U32D_FMT "has QPN: " U32H_FMT " and RQPN: " U32H_FMT
            ", but Child Tree Node lid: " U32D_FMT "has QPN: " U32H_FMT " and RQPN: " U32H_FMT
            " for Tree ID: " U32H_FMT,
            parent_lid, parent_qpn, parent_rqpn, lid_child, child_qpn, child_rqpn, tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrMismatchParentChildQPConfig::SharpErrMismatchParentChildQPConfig(
        IBNode *p_node, lid_t lid_child, lid_t parent_lid, lid_t mismatch_parent_lid,
        u_int16_t tree_id) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_MISMATCH_CHILD_NODE_TO_PARENT_NODE);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "Mismatch: Parent Tree Node lid: " U32D_FMT " has Child Tree Node lid: " U32D_FMT ", "
            "but Child Tree Node lid: " U32D_FMT " has Parent Tree Node lid: " U32D_FMT
            " for Tree ID: " U32D_FMT,
            parent_lid, lid_child, lid_child, mismatch_parent_lid, tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrTreeNodeNotFound::SharpErrTreeNodeNotFound(
        IBNode *p_node, u_int16_t tree_id) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_TREE_NODE_NOT_FOUND);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "In Node %s, Tree Node was not found for treeID: %d",
            p_node->getName().c_str(), tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrDisconnectedTreeNode::SharpErrDisconnectedTreeNode(
        IBNode *p_node, u_int16_t tree_id) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_DISCONNECTED_TREE_NODE);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "In Node %s Tree Node is Disconnected for treeID: %d",
            p_node->getName().c_str(), tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrNodeTreeIDNotMatchGetRespondTreeID::SharpErrNodeTreeIDNotMatchGetRespondTreeID(
        IBNode *p_node, u_int16_t tree_id) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_TREE_ID_NOT_MATCH);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "In Node %s treeID not match get respond for treeID: %d",
            p_node->getName().c_str(), tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrRootTreeNodeAlreadyExistsForTreeID::SharpErrRootTreeNodeAlreadyExistsForTreeID(
        IBNode *p_node, u_int16_t tree_id) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_ROOT_ALREADY_EXISTS);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "In Node %s found root tree (parent qpn 0) which is already exists for treeID: %d",
            p_node->getName().c_str(), tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrDuplicatedQPNForAggNode::SharpErrDuplicatedQPNForAggNode(
        IBNode *p_node, u_int16_t tree_id, u_int16_t dup_tree_id,
        u_int32_t qpn) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_DUPLICATE_QPN_FOR_AN);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "In Node %s Duplicated QPN %d for TreeID: %d and TreeID: %d",
            p_node->getName().c_str(), qpn, tree_id, dup_tree_id);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrQPNotActive::SharpErrQPNotActive(
        IBNode *p_node, u_int32_t qpn, u_int8_t qp_state) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_QP_NOT_ACTIVE);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "In Node %s QPN %d state is not active. QP State: %d",
            p_node->getName().c_str(), qpn, qp_state);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrRQPNotValid::SharpErrRQPNotValid(
        IBNode *p_node, u_int32_t qpn, u_int32_t rqpn) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_RQP_NOT_VALID);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
            "In Node %s parent RQPN %d (as read from the parent QPCConfig)" \
            "is not the peer qpn. Peer QP: %d",
            p_node->getName().c_str(), rqpn, qpn);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrRemoteNodeDoesntExist::SharpErrRemoteNodeDoesntExist(
        IBNode *p_node) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_REMOTE_NODE_DOESNT_EXIST);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer), "Node %s is not connected to a remote switch", p_node->getName().c_str());
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrQPCPortNotZero::SharpErrQPCPortNotZero(
        IBNode *p_node, u_int8_t qpc_port, u_int8_t port_sel_supported, IBNode *p_remote_node,
        u_int8_t remote_qpc_port, u_int8_t remote_port_sel_supported) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_QPC_PORT_NOT_ZERO);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
             "In Node %s QPC port is not 0, actual QPC port: %d, port select supported: %d. " \
             "Remote Node: %s, QPC port: %d, port select supported: %d",
             p_node->getName().c_str(), qpc_port, port_sel_supported,
             p_remote_node->getName().c_str(), remote_qpc_port, remote_port_sel_supported);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

SharpErrQPCPortsNotConnected::SharpErrQPCPortsNotConnected(
        IBNode *p_node, u_int8_t qpc_port, IBNode *p_remote_node, u_int8_t remote_qpc_port) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_SHARP_QPC_PORTS_NOT_CONNECTED);
    char buffer[1024];
    snprintf(buffer, sizeof(buffer),
             "In Node %s QPC port %d is not physically connected to remote Node's %s QPC port %d",
             p_node->getName().c_str(), qpc_port, p_remote_node->getName().c_str(), remote_qpc_port);
    this->description.assign(buffer);
    IBDIAG_RETURN_VOID;
}

/******************************************************/
/******************************************************/
/******************************************************/
CableTemperatureErr::CableTemperatureErr(const IBPort *p_inPort, const string &message,
                                         const string& temp, const string &treshold):
                                         FabricErrGeneral(), p_port(p_inPort)
{
    IBDIAG_ENTER;

    this->level = EN_FABRIC_ERR_ERROR;
    this->scope = SCOPE_PORT;
    this->err_desc = FER_CABLE_TEMPERATURE_ERROR;

    stringstream  stream_ss;
    stream_ss << "Cable " << message
              << ", current temperature: " << temp << ", "
              << "threshold: " << treshold;

    this->description = stream_ss.str();

    IBDIAG_RETURN_VOID;
}

CableTemperatureErr::~CableTemperatureErr()
{
    IBDIAG_ENTER;

    IBDIAG_RETURN_VOID;
}

string CableTemperatureErr::GetErrorLine()
{
    IBDIAG_ENTER;

    stringstream stream_ss;
    stream_ss << this->p_port->getExtendedName()
              << " - "
              << this->description;

    IBDIAG_RETURN(stream_ss.str());
}

string CableTemperatureErr::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    stringstream stream_ss;
    stream_ss << this->scope                            << ','
            << PTR(this->p_port->p_node->guid_get())    << ','
            << PTR(this->p_port->guid_get())            << ','
            <<  +this->p_port->num                      << ','
            << this->err_desc                           << ','
            << '\"' << DescToCsvDesc(this->description) << '\"';

    IBDIAG_RETURN(stream_ss.str());
}

/******************************************************/
/******************************************************/
/******************************************************/
FabricErrPortHierarchyMissing::FabricErrPortHierarchyMissing(
        IBPort *p_port) :
        FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_HIERARCHY_MISSING);
    this->level = EN_FABRIC_ERR_WARNING;
    stringstream desc_ss;
    desc_ss << "In Node " << p_port->p_node->getName() << " Port " << +p_port->num
            << " hierarchy info is missing";
    this->description = desc_ss.str();
    IBDIAG_RETURN_VOID;
}

FabricErrPortHierarchyExtraFields::FabricErrPortHierarchyExtraFields(
        IBPort *p_port, vector<string> fields) :
        FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_HIERARCHY_EXTRA_FIELDS);
    this->level = EN_FABRIC_ERR_WARNING;
    stringstream desc_ss;
    desc_ss << "In Node " << p_port->p_node->getName() << " Port " << +p_port->num
            << " hierarchy info has extra field(s): ";
    for (size_t i = 0; i < fields.size(); ++i) {
        desc_ss << fields[i];
        if (i != fields.size() - 1)
            desc_ss << ", ";
    }
    this->description = desc_ss.str();
    IBDIAG_RETURN_VOID;
}

FabricErrPortHierarchyMissingFields::FabricErrPortHierarchyMissingFields(
        IBPort *p_port, vector<string> fields) :
        FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PORT_HIERARCHY_MISSING_FIELDS);
    stringstream desc_ss;
    desc_ss << "In Node " << p_port->p_node->getName() << " Port " << +p_port->num
            << " hierarchy info has missing field(s): ";
    for (size_t i = 0; i < fields.size(); ++i) {
        desc_ss << fields[i];
        if (i != fields.size() - 1)
            desc_ss << ", ";
    }
    this->description = desc_ss.str();
    IBDIAG_RETURN_VOID;
}

FabricErrHierarchyTemplateMismatch::FabricErrHierarchyTemplateMismatch(
        IBPort *p_port, u_int64_t template_guid, u_int8_t hierarchy_index) :
        FabricErrPort(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_HIERARCHY_TEMPLATE_MISMATCH);
    this->level = EN_FABRIC_ERR_WARNING;
    this->dump_csv_only = true;
    stringstream desc_ss;
    desc_ss << "In Node " << p_port->p_node->getName() << " Port " << +p_port->num
            << " got an unexpected Hierarchy Template GUID " << template_guid
            << ", Hierarchy Index " << +hierarchy_index;
    this->description = desc_ss.str();
    IBDIAG_RETURN_VOID;
}


FabricErrWHBFConfiguration::FabricErrWHBFConfiguration(
        IBNode *p_node) :
        FabricErrNode(p_node)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_WHBF_WRONG_CONFIGURATIONT);
    stringstream desc_ss;
    desc_ss << "In Node " << p_node->getName() << " WHBF is enabled but HBF is disabled";
    this->description = desc_ss.str();
    IBDIAG_RETURN_VOID;
}

/******************************************************/
/***            PFRN (Class 0xC) Errors             ***/
/******************************************************/
pFRNErrPartiallySupported::pFRNErrPartiallySupported(string desc):
        FabricErrCluster(FER_PFRN_PARTIALLY_SUPPORTED, desc) { }

pFRNErrDiffTrapLIDs::pFRNErrDiffTrapLIDs(string desc):
        FabricErrCluster(FER_PFRN_DIFFERENT_TRAP_LIDS, desc) { }

pFRNErrTrapLIDNotSM::pFRNErrTrapLIDNotSM(string desc):
        FabricErrCluster(FER_PFRN_TRAP_LID_NOT_TO_SM, desc) { }

pFRNErrNeighborNotExist::pFRNErrNeighborNotExist(
        IBNode *p_node, u_int32_t record) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;

    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_PFRN_NEIGHBOR_NOT_EXIST);

    stringstream desc;
    desc << "Neighbor connected to port " << record << " on node "  << p_node->getName()
         << " is not exist in the fabric";
    this->description = desc.str();

    IBDIAG_RETURN_VOID;
}

pFRNErrNeighborNotSwitch::pFRNErrNeighborNotSwitch(
        IBNode *p_node, u_int32_t record) : FabricErrNode(p_node)
{
    IBDIAG_ENTER;

    this->scope.assign(SCOPE_NODE);
    this->err_desc.assign(FER_PFRN_NEIGHBOR_NOT_SWITCH);

    stringstream desc;
    desc << "Neighbor connected to port " << record << " on node "  << p_node->getName()
         << " is not a switch";
    this->description = desc.str();

    IBDIAG_RETURN_VOID;
}

DifferentARGroupsIDForDLIDErr::DifferentARGroupsIDForDLIDErr(string desc):
        FabricErrCluster(FER_DIFFERENT_AR_GROUPS_ID_FOR_DLID, desc) { }

pFRNReceivedErrorNotZeroErr::pFRNReceivedErrorNotZeroErr(IBPort *p_port, u_int32_t pfrn_received_error) :
        FabricErrPort(p_port)
{
    IBDIAG_ENTER;

    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_PFRN_RECEIVED_ERROR_NOT_ZERO);

    stringstream desc;
    desc << "pfrn_received_error is not zero, value = " << pfrn_received_error;
    this->description = desc.str();

    IBDIAG_RETURN_VOID;
}

string FLIDError::GetCSVErrorLine()
{
    IBDIAG_ENTER;

    stringstream ss;
    ss << SCOPE_CLUSTER         << ','
       << "0x0,0x0,0x0"         << ','
       << FER_FLID_VALIDATION   << ','
       << '\"' << DescToCsvDesc(this->error, "\n", " ") << '\"';

    IBDIAG_RETURN(ss.str());
}

string FLIDError::GetErrorLine()
{
    IBDIAG_ENTER;

    while (!this->error.empty() && (this->error.back() == '\n' || this->error.back() == '\r'))
	    this->error.pop_back();

    IBDIAG_RETURN(this->error);
}

FLIDPortError::FLIDPortError(const IBPort *p_port, const string &message): FabricErrPort(p_port)
{
    IBDIAG_ENTER;

    this->scope = SCOPE_PORT;
    this->err_desc = FER_FLID_VALIDATION;
    this->description = message;

    this->SetLevel(EN_FABRIC_ERR_WARNING);

    IBDIAG_RETURN_VOID;
}

string NullPtrError::GetCSVErrorLine()
{
    IBDIAG_ENTER;

    stringstream ss;
    ss << SCOPE_CLUSTER         << ','
       << "0x0,0x0,0x0"         << ','
       << FER_INTERNAL_DB       << ','
       << '\"' << DescToCsvDesc(this->GetErrorLine()) << '\"';

    IBDIAG_RETURN(ss.str());
}

string NullPtrError::GetErrorLine()
{
    IBDIAG_ENTER;

    stringstream ss;
    ss << "Internal DB error of type=" << line
       << " was detected " << count << " times";

    IBDIAG_RETURN(ss.str());
}

FLIDNodeError::FLIDNodeError(const IBNode *p_node, const string &message): FabricErrNode(p_node)
{
    IBDIAG_ENTER;

    this->scope = SCOPE_NODE;
    this->err_desc = FER_FLID_VALIDATION;
    this->description = message;

    this->SetLevel(EN_FABRIC_ERR_WARNING);

    IBDIAG_RETURN_VOID;
}

/***********************************************************************
 *                   CC Algo verification errors
 ***********************************************************************/
CC_AlgoCounterEnErr::CC_AlgoCounterEnErr(IBPort *p_port, list <int> &counters_en_algos) :
        FabricErrPort(p_port)
{
    IBDIAG_ENTER;

    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_CC_ALGO_COUNTER_EN_ERROR);

    stringstream desc;
    desc << "More than one algo counter is enabled on port. algos: ";

    for (list <int>::iterator it = counters_en_algos.begin();
            it != counters_en_algos.end(); it++)
        desc << *it << "  ";

    this->description = rtrim(desc.str());

    IBDIAG_RETURN_VOID;
}

CC_AlgoSLEnErr::CC_AlgoSLEnErr(IBPort *p_port, u_int8_t sl, list <int> &algos) :
        FabricErrPort(p_port)
{
    IBDIAG_ENTER;

    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_CC_ALGO_SL_EN_ERROR);

    stringstream desc;
    desc << "SL " << +sl << " is enabled on more than one algo. algos: ";

    for (list <int>::iterator it = algos.begin(); it != algos.end(); it++)
        desc << *it << "  ";

    this->description = rtrim(desc.str());

    IBDIAG_RETURN_VOID;
}

CC_AlgoParamsSLEnErr::CC_AlgoParamsSLEnErr(IBPort *p_port, u_int8_t sl, list <int> &algos) :
        FabricErrPort(p_port)
{
    IBDIAG_ENTER;

    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_CC_ALGO_PARAMS_SL_EN_ERROR);

    stringstream desc;
    desc << "SL " << +sl << " is enabled on more than one algo params. algos: ";

    for (list <int>::iterator it = algos.begin(); it != algos.end(); it++)
        desc << *it << "  ";

    this->description = rtrim(desc.str());

    IBDIAG_RETURN_VOID;
}

CC_AlgoParamRangeErr::CC_AlgoParamRangeErr(IBPort *p_port, const string &desc) :
        FabricErrPort(p_port)
{
    IBDIAG_ENTER;

    this->scope.assign(SCOPE_PORT);
    this->err_desc.assign(FER_CC_ALGO_PARAM_OUT_OF_RANGE);
    this->description = desc;

    IBDIAG_RETURN_VOID;
}

/******************************************************/
/******************************************************/
/******************************************************/
ExportDataErr::ExportDataErr(IBNode *p_node, IBPort *p_port, const char *fmt, ...)
    : p_node(p_node), p_port(p_port)
{
    IBDIAG_ENTER;
    this->scope.assign(SCOPE_SYSTEM);
    this->err_desc.assign(FER_EXPORT_DATA);

    char buffer[2048];
    va_list args;
    va_start(args, fmt);
    vsnprintf(buffer, sizeof(buffer), fmt, args);
    va_end(args);

    this->description = string(buffer);
    IBDIAG_RETURN_VOID;
}

string ExportDataErr::GetCSVErrorLine()
{
    IBDIAG_ENTER;
    stringstream ss;
    ss << this->scope                                   << ','
       << (this->p_node ? this->p_node->guid_get() : 0) << ','
       << (this->p_port ? this->p_port->guid_get() : 0) << ','
       << (this->p_port ? this->p_port->num : 0)        << ','
       << this->err_desc                                << ','
       << '\"' << DescToCsvDesc(this->description)      << '\"';
    IBDIAG_RETURN(ss.str());
}

string ExportDataErr::GetErrorLine()
{
    IBDIAG_ENTER;
    IBDIAG_RETURN(this->description);
}
