/*
 * Copyright (c) 2020-2020 Mellanox Technologies LTD. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <infiniband/ibdiag/ibdiag_smdb.h>
#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include <infiniband/ibutils/output_stream_primitives.h>

int IBDiagSMDB::ParseSMDB(const string &smdb)
{
    IBDIAG_ENTER;

    IBDiag::PrintFileTimestamp(smdb, "SMDB");

    int rc = IBDIAG_SUCCESS_CODE;
    CsvFileStream csv_file_stream(smdb, m_csv_parser);

    SectionParser<SMDBSMRecord> sm_section_parser;
    sm_section_parser.Init("SM");
    rc = m_csv_parser.ParseAndHandleSection <SMDBSMRecord, IBDiagSMDB>(csv_file_stream,
            sm_section_parser, this, &IBDiagSMDB::ParseSMSection);
    if (rc) {
        ERR_PRINT("Failed to parse SM section in SMDB file - %s\n", smdb.c_str());
        return rc;
    }

    SectionParser<SMDBSwitchRecord> switch_section_parser;
    switch_section_parser.Init("SWITCHES");
    rc = m_csv_parser.ParseAndHandleSection <SMDBSwitchRecord, IBDiagSMDB>(csv_file_stream,
            switch_section_parser, this, &IBDiagSMDB::ParseSwitchSection);
    if (rc) {
        ERR_PRINT("Failed to parse SWITCHES section in SMDB file - %s\n", smdb.c_str());
        return rc;
    }

    this->is_smdb_parsed = true;

    IBDIAG_RETURN(rc);
}

int IBDiagSMDB::Apply(IBFabric &discovered_fabric)
{
    IBDIAG_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (!this->IsSMDBParsed())
        IBDIAG_RETURN(rc);

    discovered_fabric.SetRoutingEngine(this->routing_engine);

    for (set_pnode::iterator it_switch = discovered_fabric.Switches.begin();
         it_switch != discovered_fabric.Switches.end(); it_switch++ ) {
         IBNode *p_switch = *it_switch;
         map_smdb_guid_2_switch_info::iterator it_smdb_switch_info =
                                         this->guid_2_switch_info.find(p_switch->guid_get());
         if (it_smdb_switch_info == this->guid_2_switch_info.end()) {
             WARN_PRINT("Switch GUID: " U64H_FMT
                        " in Fabric doesn't exist in Switch Info table from SMDB file\n",
                        p_switch->guid_get());
             rc = IBDIAG_ERR_CODE_DB_ERR;
         }
    }

    for (map_smdb_guid_2_switch_info::iterator it_smdb_switch_info =
                                               this->guid_2_switch_info.begin();
         it_smdb_switch_info != this->guid_2_switch_info.end();
         it_smdb_switch_info++) {
         u_int64_t guid = it_smdb_switch_info->first;
         IBNode *p_node = discovered_fabric.getNodeByGuid(guid);
         if (!p_node) {
             WARN_PRINT("Switch GUID: " U64H_FMT
                        " in Switch Info table from SMDB file doesn't exist in Fabric\n",
                        guid);
             rc = IBDIAG_ERR_CODE_DB_ERR;
             continue;
         }

         if (p_node->type != IB_SW_NODE) {
             WARN_PRINT("Node GUID: " U64H_FMT
                        " in Switch Info table from SMDB file is %s and not %s\n",
                        p_node->guid_get(),
                        nodetype2char(p_node->type),
                        nodetype2char(IB_SW_NODE));
             rc = IBDIAG_ERR_CODE_DB_ERR;
             continue;
         }

         p_node->rank = it_smdb_switch_info->second.rank;
    }

    if (rc)
        WARN_PRINT("Apply SMDB Switch Info data was finished with warnings\n");
    else
        INFO_PRINT("Apply SMDB Switch Info data was finished successfully\n");

    discovered_fabric.SetIsSMDBApplied(true);

    IBDIAG_RETURN(rc);
}

int IBDiagSMDB::ParseSMSection(const SMDBSMRecord &smdbSMRecord)
{
    IBDIAG_ENTER;

    this->routing_engine = char2routing_engine(smdbSMRecord.routing_engine.c_str());
    if (this->routing_engine == IB_ROUTING_UNKNOWN) {
        WARN_PRINT("Unknown Routing Engine: %s in SMDB file, section: SM\n",
                   smdbSMRecord.routing_engine.c_str());
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

int IBDiagSMDB::ParseSwitchSection(const SMDBSwitchRecord &smdbSwitchRecord)
{
    IBDIAG_ENTER;

    struct smdb_switch_Info sw_info = {smdbSwitchRecord.rank};
    pair<map_smdb_guid_2_switch_info::iterator,bool> ret;
    ret = guid_2_switch_info.insert (pair< u_int64_t, struct smdb_switch_Info >
                                         (smdbSwitchRecord.node_guid, sw_info));
    if (!ret.second) {
        ERR_PRINT("Multiple switch GUID: " U64H_FMT " in SMDB file, "
                  "section: SWITCHES\n", smdbSwitchRecord.node_guid);
        IBDIAG_RETURN(IBDIAG_ERR_CODE_PARSE_FILE_FAILED);
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

/*******************************
 *       SECTION RECORDS       *
 *******************************/

int SMDBSMRecord::Init(vector < ParseFieldInfo <class SMDBSMRecord> > &parse_section_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    parse_section_info.push_back(ParseFieldInfo <class SMDBSMRecord>
             ("RoutingEngine", &SMDBSMRecord::SetRoutingEngine));

    IBDIAG_RETURN(rc);
}

int SMDBSwitchRecord::Init(vector < ParseFieldInfo <class SMDBSwitchRecord> > &parse_section_info)
{
    IBDIAG_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;

    parse_section_info.push_back(ParseFieldInfo <class SMDBSwitchRecord>
             ("NodeGUID", &SMDBSwitchRecord::SetNodeGUID));

    parse_section_info.push_back(ParseFieldInfo <class SMDBSwitchRecord>
             ("Rank", &SMDBSwitchRecord::SetRank));

    IBDIAG_RETURN(rc);
}
