/*
 * Copyright (c) 2004-2021 Mellanox Technologies LTD. All rights reserved
 * Copyright (C) 2021-2022, NVIDIA CORPORATION & AFFILIATES. ALL RIGHTS RESERVED.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <stdio.h>
#include <math.h>
#include <signal.h>
#include <map>
#include <list>
#include <iostream>
#include <errno.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>

#include "infiniband/ibdm/TopoMatch.h"
#include "infiniband/ibdm/SysDef.h"
#include "infiniband/ibdm/SysLogger.h"
#include <infiniband/ibdm/CredLoops.h>
#include <infiniband/ibutils/release_containers.h>
#include "infiniband/ibutils/output_stream_primitives.h"
#include "infiniband/ibdiag/ibdiag_dfp.h"
#include "infiniband/ibdiag/ibdiag_fat_tree.h"
#include "infiniband/ibdiag/ibdiag_flid.h"
#include "infiniband/ibdm/cable/CombinedCable.h"
#include "infiniband/ibdiag/ibdiagnet/ibdiagnet_export_api.h"

#include "ibdiagnet_stages.h"
#include "ibdiagnet_types.h"


/****************************************************/
/* globals */
static bool g_vs_cap_gmp_run = false;

/****************************************************/
int check_if_local_link_is_down(IN IBDiag *p_ibdiag,
        OUT bool& result)
{
    IBDIAGNET_ENTER;
    u_int8_t local_port_state;
    if (p_ibdiag->GetLocalPortState(local_port_state)) {
        ERR_PRINT("%s\n", p_ibdiag->GetLastError());
        IBDIAGNET_RETURN(1);
    }

    switch (local_port_state) {
    case IB_PORT_STATE_DOWN: {
        result = true;
        break;
    }
    default: {
        result = false;
        break;
    }
    }
    IBDIAGNET_RETURN(0);
}


/******************************************************/
/////
//DiscoveryPath public methods:
/////
DiscoveryPathStage::DiscoveryPathStage(IBDiag *p_ibdiag) :
        DiscoveryStage(p_ibdiag), src_lid(0), dest_lid(0)
{
    IBDIAGNET_ENTER;
    memset(&dr_path, 0, sizeof(dr_path));
    IBDIAGNET_RETURN_VOID;
}

int DiscoveryPathStage::DiscoverFabric()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    stringstream ss;

    rc = p_ibdiag->DiscoverPath(this->max_hops,
                                this->src_lid, this->dest_lid, this->dr_path,
                                &ss);
    if (rc)
        IBDIAGNET_RETURN(rc);

    PRINT("%s\n",ss.str().c_str());

    IBDIAGNET_RETURN(0);
}

int DiscoveryPathStage::Prepare()
{
    IBDIAGNET_ENTER;
    DiscoveryStage::Prepare();
    if (!this->to_check_dup_guids) {        //skipped duplicated guids detection
        this->p_ibdiag->GetRefToCheckDupGuids() = false;
    }
    IBDIAGNET_RETURN(0);
}

/******************************************************/
/////
//DiscoveryFabric private methods:
/////
int DiscoveryFabricStage::PostDiscover()
{
    IBDIAGNET_ENTER;
    string parser_output;
    int rc;

    string scope_file = this->scope_port_guid_file;
    u_int32_t errors = this->num_errors;
    u_int32_t warnings = this->num_warnings;

    int scope_lines = 0;
    int exclude_scope_lines = 0;

    //work on guids that are in a scope
    if (!scope_file.empty() &&
        access(scope_file.c_str(), F_OK) != -1) {
        PRINT("\n");
        IBDiag::PrintFileTimestamp(scope_file, "Scope");
        rc = p_ibdiag->ParseScopePortGuidsFile(
                            scope_file.c_str(),
                            parser_output,
                            true,
                            scope_lines);

        LOG_PRINT("%s\n", parser_output.c_str());

        //if critical errors found, just exit
        if (rc) {
            ERR_PRINT("Failed to parse Scope file (%s).\n",
                      scope_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
        //parse number of errors, warnings
        this->ParseIBDMOutput(parser_output);
        if (this->num_errors > errors) {
            ERR_PRINT("Parsing the scope file finished with errors.\n");
        } else if (this->num_warnings > warnings) {
            WARN_PRINT("Parsing the scope file finished with warnings.\n");
        }
    }
    string exclude_scope_file = this->exclude_scope_port_guid_file;
    //init for exclude scope
    parser_output.clear();
    errors = this->num_errors;
    warnings = this->num_warnings;

    //work on guids that are excluded from scope
    if (!exclude_scope_file.empty() &&
        access(exclude_scope_file.c_str(), F_OK) != -1) {
        PRINT("\n");
        IBDiag::PrintFileTimestamp(exclude_scope_file, "Exclude Scope");
        rc = p_ibdiag->ParseScopePortGuidsFile(
                            exclude_scope_file.c_str(),
                            parser_output,
                            false, //not to include, but to exclude
                            exclude_scope_lines);
        LOG_PRINT("%s\n", parser_output.c_str());

        //if critical errors found, just exit
        if (rc) {
            ERR_PRINT("Failed to parse Exclude Scope file (%s).\n",
                      exclude_scope_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
        //parse number of errors, warnings
        this->ParseIBDMOutput(parser_output);
        if (this->num_errors > errors) {
            ERR_PRINT("Parsing the exclude scope file finished"
                      " with errors.\n");
        } else if (this->num_warnings > warnings) {
            WARN_PRINT("Parsing the exclude scope file finished"
                       " with warnings.\n");
        }

    }

    size_t fabric_switches_total = 0;
    size_t fabric_switches_fully = 0;
    size_t fabric_switches_partial = 0;
    size_t fabric_switches_ports = 0;

    size_t fabric_cas_total = 0;
    size_t fabric_cas_fully = 0;
    size_t fabric_cas_partial = 0;
    size_t fabric_cas_ports = 0;

    IBFabric * p_fabric = this->p_ibdiag->GetDiscoverFabricPtr();

    for (map_str_pnode::iterator nI = p_fabric->NodeByName.begin();
            nI != p_fabric->NodeByName.end(); ++nI) {
        IBNode *p_node = nI->second;
        if (!p_node)
            continue;

        if (!p_node->getInSubFabric())
            continue;

        bool all_in_scope     = true;
        size_t  port_count    = 0;

        for (phys_port_t i = 1; i <= p_node->numPorts; ++i) {
            IBPort *p_port = p_node->getPort(i);
            if (!p_port)
                continue;

            if (!p_port->getInSubFabric()) {
                all_in_scope = false;
                continue;
            }

            ++port_count;
        }

        if (p_node->type == IB_SW_NODE)
            fabric_switches_total++;
        else if (p_node->type == IB_CA_NODE)
            fabric_cas_total++;

        if (port_count) {
            if (p_node->type == IB_SW_NODE) {

                fabric_switches_ports += port_count;
                if (all_in_scope)
                    ++fabric_switches_fully;
                else
                    ++fabric_switches_partial;

            } else if (p_node->type == IB_CA_NODE) {

                fabric_cas_ports += port_count;
                if (all_in_scope)
                    ++fabric_cas_fully;
                else
                    ++fabric_cas_partial;

            }
        }
    }

    char buff[1024];
    if (!scope_file.empty() || !exclude_scope_file.empty())
        snprintf(buff, sizeof(buff), "In scope [total/fully/partial/ports] switches [%lu/%lu/%lu/%lu], "
                    "CAs [%lu/%lu/%lu/%lu]; scope lines %d, exclude scope lines %d\n",
                    fabric_switches_total, fabric_switches_fully, fabric_switches_partial, fabric_switches_ports,
                    fabric_cas_total, fabric_cas_fully, fabric_cas_partial, fabric_cas_ports,
                    scope_lines, exclude_scope_lines);
    else
        snprintf(buff, sizeof(buff), "No scope files. Total switches/ports [%lu/%lu], CAs/ports [%lu/%lu]\n",
                    fabric_switches_total, fabric_switches_ports, fabric_cas_total, fabric_cas_ports);

    PRINT("\n-I- %s", buff);

    if (SysLogger::instance().get_enabled())
        SysLogger::instance().syslog(buff);

    IBDIAGNET_RETURN(0);
}

/////
//DiscoveryFabric public methods:
/////
DiscoveryFabricStage::DiscoveryFabricStage(IBDiag *p_ibdiag) :
        DiscoveryStage(p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int DiscoveryFabricStage::DiscoverFabric()
{

    IBDIAGNET_ENTER;

    int rc = 0;

    if (load_from_file) {
        rc = p_ibdiag->DiscoverFabricFromFile(load_from_file_path);
    } else {
        rc = p_ibdiag->DiscoverFabric(this->max_hops);
    }
    IBDIAGNET_RETURN(rc);
}

int DiscoveryFabricStage::Prepare()
{
    IBDIAGNET_ENTER;
    DiscoveryStage::Prepare();
    if (!this->to_check_dup_guids) {        //skipped duplicated guids detection
        WARN_PRINT("Advanced %s skipped (Only Basic detection will be performed)\n",
                CHECK_NAME_DUPLICATED_GUIDS);
        ++this->num_warnings;
        this->p_ibdiag->GetRefToCheckDupGuids() = false;
    }
    IBDIAGNET_RETURN(0);
}

/******************************************************/
/////
//Discovery private methods:
/////
void DiscoveryStage::PrintDuplicatedGuidsDetectErrors()
{
    IBDIAGNET_ENTER;
    if (!this->p_ibdiag->GetRefToDupGuidsDetectionErrors().size())
        IBDIAGNET_RETURN_VOID;
    LOG_PRINT("\n");
    LOG_WARN_PRINT("Duplicated GUIDs detection finished with some warnings:\n");
    for (list_string::iterator eI = this->p_ibdiag->GetRefToDupGuidsDetectionErrors().begin();
            eI != this->p_ibdiag->GetRefToDupGuidsDetectionErrors().end(); ++eI) {
        //will be printed to both screen + log
        LOG_WARN_PRINT("%s", (*eI).c_str());
        ++this->num_warnings;
    }
    LOG_PRINT("\n");
    IBDIAGNET_RETURN_VOID;
}

/////
//Discovery public methods:
/////
DiscoveryStage::DiscoveryStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_DISCOVERY, p_ibdiag),
        device_name(""), port_num(0), port_guid(0),
        max_hops(IBDIAG_MAX_HOPS), to_check_dup_guids(true), to_check_dup_node_desc(true),
        to_check_vs_cap_smp(true), to_check_vs_cap_gmp(true),
        initiliazed_with_device(false), initiliazed_with_guid(false), no_mepi(false),
        output_capability_file(""), output_full_capability_file(""),
        input_capability_file(""), show_ports_data_db2(true), dump_lst_with_lmc(false),
        node_name_map_file(""), p_mkey_manager(NULL), load_from_file(false), num_of_switches(0),
        num_of_channel_adapters(0), num_of_channel_adapters_sdm(0), num_of_routers(0),
        num_of_aggregation_nodes(0), check_switch_duplicated_guids(false)
{
    IBDIAGNET_ENTER;
    //check if we have environment variable with path
    //to capability mask configuration file
    char *capability_file = getenv(CAPABILITY_FILE_ENV);

    if (capability_file)
        input_capability_file = capability_file;

    char *map_file = getenv(MAP_FILE_ENV);

    //check if we have environment variable with path
    //to node name map file
    if (map_file)
        node_name_map_file = map_file;

    IBDIAGNET_RETURN_VOID;
}

int DiscoveryStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    //init - set port
    if (this->initiliazed_with_device && initiliazed_with_guid) {
        WARN_PRINT("Local GUID were given together with Device name / Port number, initializing only by GUID\n");
        ++this->num_warnings;
    }
    if (no_mepi)
        this->p_ibdiag->SetNoMEPI();

    this->p_ibdiag->SetCheckSwitchDuplicatedGuids(this->check_switch_duplicated_guids);

    if (!load_from_file) { // in load_from_file we set the device and port later
        if (initiliazed_with_guid) {    //init by guid
            if (this->p_ibdiag->SetPort(this->port_guid)) {
                ERR_PRINT("Failed to initialize - %s\n", this->p_ibdiag->GetLastError());
                ++this->num_errors;
                IBDIAGNET_RETURN(1);
            }
        } else {    //init by device - this is also default
            if (this->p_ibdiag->SetPort(this->device_name.c_str(), this->port_num)) {
                ERR_PRINT("Failed to initialize - %s\n", this->p_ibdiag->GetLastError());
                ++this->num_errors;
                IBDIAGNET_RETURN(1);
            }
        }

        INFO_PRINT("Using local IB device %s:%d\n",
                   this->p_ibdiag->GetIbisPtr()->GetDevName().c_str(),
                   this->p_ibdiag->GetIbisPtr()->GetPortNum());
        SysLogger::instance().syslog("Using local IB device %s:%d\n",
                                     this->p_ibdiag->GetIbisPtr()->GetDevName().c_str(),
                                     this->p_ibdiag->GetIbisPtr()->GetPortNum());
    }

    if (!this->to_check_dup_node_desc) {    //skipped duplicated node description detection
        WARN_PRINT("%s skipped\n", CHECK_NAME_DUPLICATED_NODE_DESC);
        ++this->num_warnings;
    }
    if (!this->to_check_vs_cap_smp) {    //skipped Vendor Specific SMP capability mask inquiry
        WARN_PRINT("%s skipped\n", CHECK_VS_CAP_SMP);
        ++this->num_warnings;
    }
    if (!this->to_check_vs_cap_gmp) {    //skipped Vendor Specific GMP capability mask inquiry
        WARN_PRINT("%s skipped\n", CHECK_VS_CAP_GMP);
        ++this->num_warnings;
    }

    //try to read Node Name Map file
    if (!this->node_name_map_file.empty() &&
        access(this->node_name_map_file.c_str(), F_OK) != -1) {
        string parser_output = string("");
        IBDiag::PrintFileTimestamp(node_name_map_file, "Node Map");

        int rc = p_ibdiag->ParseNodeNameMapFile(node_name_map_file.c_str(),
                                                parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse Node Name Map file (%s).\n",
                      node_name_map_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }

        u_int32_t errors = this->num_errors;
        u_int32_t warnings = this->num_warnings;
        //parse number of errors, warnings
        this->ParseIBDMOutput(parser_output);
        if (this->num_errors > errors) {
            ERR_PRINT("Parsing the Node Name Map file finished " \
                      "with errors.\n");
        } else if (this->num_warnings > warnings) {
            WARN_PRINT("Parsing the Node Name Map file " \
                       "finished with warnings.\n");
        }
    }


    IBDIAGNET_RETURN(0);
}

int DiscoveryStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    INFO_PRINT("Start Fabric Discover\n");
    int rc2 = DiscoverFabric();
    SCREEN_PRINT("\n");

    if (!this->load_from_file && !p_ibdiag->IsNodeDescriptionSent()){
        list_p_fabric_general_err& pi_errors = p_ibdiag->GetErrors();
        INFO_PRINT("Fill NodeDesc data\n");
        int local_rc = p_ibdiag->FillInNodeDescription(pi_errors);
        SCREEN_PRINT("\n");
        if (local_rc){
            ERR_PRINT("Failed to fill NodeDesc data: %s\n", p_ibdiag->GetLastError());
            IBDIAGNET_RETURN(1);
        }
        else {
            INFO_PRINT("NodeDesc finished successfully \n");
        }
    }

    int rc = this->AnalyzeCheckResults(p_ibdiag->GetErrors(),
                CHECK_FABRIC_DISCOVER,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                false);

    if (rc) {
        /*
        PRINT("\n");
        ERR_PRINT("Failed to discover fabric, err=%s\n",
                this->p_ibdiag->GetLastError());
        ++this->num_errors;
        */
        if (this->to_check_dup_guids)   //maybe we failed because duplicated GUIDs detection had errors
            this->PrintDuplicatedGuidsDetectErrors();

        IBDIAGNET_RETURN(1);
    }

    if (!this->load_from_file){
        SCREEN_PRINT("\n");

        list_p_fabric_general_err pi_errors;
        INFO_PRINT("Fill PortInfo data\n");
        rc = p_ibdiag->FillInPortInfo(pi_errors);
        SCREEN_PRINT("\n");
        if (rc){
            this->AnalyzeCheckResults(pi_errors,
                    CHECK_PORT_INFO,
                    rc,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    false);

            ERR_PRINT("Failed to fill PortInfo data\n");
        }
        else {
            INFO_PRINT("PortInfo finished successfully\n");
        }
    }

    if (this->p_mkey_manager) {
        if (p_ibdiag->GetIbisPtr() && p_ibdiag->GetIbisPtr()->GetMKeyManager())
        {
            if (!p_mkey_manager->getConstMKey()) {
                CreateLid2GuidDB();
                p_mkey_manager->setLidToNodeGuidMap(lid2guid_map);
            }
        }
        else
            ERR_PRINT("can't get MKeyManager Ptr\n");
    }

    this->SetKeyPerPort(this->guid_2_vs_key, IBIS_VS_KEY, "VS Key");
    this->SetKeyPerPort(this->guid_2_cc_key, IBIS_CC_KEY, "CC Key");
    this->SetKeyPerPort(this->guid_2_m2n_key, IBIS_CLASS_C_KEY, "M2N Key");

    //this is essential part of discovery, we can't continue without it
    //(example, scope failed)
    if (PostDiscover())
        IBDIAGNET_RETURN(1);

    string parser_output;

    if(!psl_dump_file.empty()) {
        parser_output = "";
        INFO_PRINT("Load PSL file (%s).\n", psl_dump_file.c_str());

        rc = p_ibdiag->ParsePSLFile(psl_dump_file, parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse PSL file (%s).\n", psl_dump_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
    }

    //try to read capability mask configuration file
    if (!this->input_capability_file.empty() &&
        access(this->input_capability_file.c_str(), F_OK) != -1) {
        parser_output = "";

        IBDiag::PrintFileTimestamp(input_capability_file, "Cpability Mask Configuration");

        rc = p_ibdiag->ParseCapabilityMaskFile(
            this->input_capability_file.c_str(), parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse Capability mask configuration file (%s).\n",
                      this->input_capability_file.c_str());
            ++this->num_errors;
        }
    }

    //check if we need to write the capability masks file
    if (!this->output_capability_file.empty()) {

    	parser_output = "";

    	OutputControl::Identity identity(this->output_capability_file,
    	                                 OutputControl::OutputControl_Flag_UserFile);

        if (p_ibdiag->DumpCapabilityMaskFile(identity, parser_output)) {
            WARN_PRINT("Writing Capability Masks file to %s failed\n", this->output_capability_file.c_str());
            ++this->num_errors;
        }
        LOG_PRINT("%s\n", parser_output.c_str());
    }

    parser_output = "";
    //initialize capability mask caches
    rc2 = p_ibdiag->BuildCapabilityCache(parser_output);
    if (rc2) {
        LOG_PRINT("%s\n", parser_output.c_str());
        ERR_PRINT("Failed to init capability mask inner cache.\n");
        ++this->num_errors;
    }
    SCREEN_PRINT("\n");

    //SMP and GMP substages to get capability masks
    //GMP
    if (load_from_file) {
        g_vs_cap_gmp_run = true;
    } else {
        if (this->to_check_vs_cap_gmp) {
            bool can_send_mads_by_lid = false;
            if (check_if_can_send_mads_by_lid(this->p_ibdiag, can_send_mads_by_lid) ||
                !can_send_mads_by_lid) {
                INFO_PRINT("%s skipped\n", CHECK_VS_CAP_GMP);
                PRINT("\n");
            } else {
                list_p_fabric_general_err vs_cap_gmp_errors;
                INFO_PRINT("Build VS Capability GMP\n");
                int rc2 = p_ibdiag->BuildVsCapGmpDB(vs_cap_gmp_errors);
                SCREEN_PRINT("\n");
                this->AnalyzeCheckResults(vs_cap_gmp_errors,
                        CHECK_VS_CAP_GMP,
                        rc2,
                        IBDIAG_ERR_CODE_FABRIC_ERROR,
                        this->num_errors,
                        this->num_warnings,
                        true);

                g_vs_cap_gmp_run = true;
            }
        }
    }

    //SMP
    if (this->to_check_vs_cap_smp && (!load_from_file)) {
        list_p_fabric_general_err vs_cap_smp_errors;
        INFO_PRINT("Build VS Capability SMP\n");
        rc2 = p_ibdiag->BuildVsCapSmpDB(vs_cap_smp_errors);
        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(vs_cap_smp_errors,
                CHECK_VS_CAP_SMP,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }

    if (!load_from_file) {
        //now we have capability -> can run ExtendedPortInfo MADs
        //to determine extended mellanox speeds speed and LLR
        //note: even if we haven't ran the SMP and GMP sub stages,
        //we might have a capability from initial configuration
        list_p_fabric_general_err vs_ext_pi_errors;
        INFO_PRINT("Build VS Extended Port Info\n");
        rc2 = p_ibdiag->BuildExtendedPortInfo(vs_ext_pi_errors);
        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(vs_ext_pi_errors,
                CHECK_VS_EXTENDED_PORT_INFO,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }

    if (!load_from_file) {
        //collect port info extended info for FEC data
        list_p_fabric_general_err pi_ext_errors;
        INFO_PRINT("Build VS Port Info Extended\n");
        rc2 = p_ibdiag->BuildPortInfoExtendedDB(pi_ext_errors);
        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(pi_ext_errors,
                CHECK_PORT_INFO_EXTENDED,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }

    if (!load_from_file) {
        //bring switch info
        list_p_fabric_general_err switch_errors;
        INFO_PRINT("Build Switch Info\n");
        rc2 = this->p_ibdiag->BuildSwitchInfoDB(switch_errors);
        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(switch_errors,
                CHECK_NAME_SW_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings);
    }

    if (!load_from_file) {
        //bring hierarchy info
        list_p_fabric_general_err hierarchy_errors;
        INFO_PRINT("Build Hierarchy Info\n");
        rc2 = this->p_ibdiag->BuildHierarchyInfoDB(hierarchy_errors);
        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(hierarchy_errors,
                CHECK_NAME_HIERARCHY_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings);
    }

    if (!load_from_file) {
    //bring AR info
    list_p_fabric_general_err ar_errors;
    INFO_PRINT("Build AR Info\n");
    rc2 = this->p_ibdiag->BuildARInfoDB(ar_errors);

    SCREEN_PRINT("\n");
    this->AnalyzeCheckResults(ar_errors,
            CHECK_NAME_AR_INFO_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings);
    }

    //here we have correct speeds after VS ExtendePortInfo MADs were sent
    //dump db file
    if (this->p_ibdiag->DumpInternalDBCSVTable(*this->p_csv_out,
                                               this->show_ports_data_db2)) {
        ERR_PRINT("Writing discovery info to DB failed, err=%s\n",
                this->p_ibdiag->GetLastError());
        ++this->num_errors;
        IBDIAGNET_RETURN(1);    /* exiting - there are tools
                                    that need to parse this file */
    }

    //dump lst file
    if (this->p_ibdiag->WriteLSTFile(LST_FILE, this->dump_lst_with_lmc)) {
        WARN_PRINT("Writing lst file failed\n");
        ++this->num_errors;
    }

    //network dump file
    if (this->p_ibdiag->WriteNetDumpFile(NET_DUMP_FILE)) {
        WARN_PRINT("Writing network dump file failed\n");
        ++this->num_errors;
    }

    p_ibdiag->DumpGeneralInfoSMP2CSV(*this->p_csv_out);

    //dump the capability mask configurations after we are done with both stages
    //using it as a hidden option for debug
    if (!this->output_full_capability_file.empty()) {
        string parser_output;

        OutputControl::Identity identity(this->output_full_capability_file,
                                         OutputControl::OutputControl_Flag_UserFile);


        if (p_ibdiag->DumpFullCapabilityMaskFile(identity, parser_output))
        {
            WARN_PRINT("Writing Guid to Capability Masks file to %s failed\n",
                       this->output_full_capability_file.c_str());
            ++this->num_errors;
        }

        LOG_PRINT("%s\n", parser_output.c_str());
    }

    IBDIAGNET_RETURN(0);
}

int DiscoveryStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;
    list_p_fabric_general_err nodes_errors;

    //print duplicated GUIDs detection error if not skipped
    if (this->to_check_dup_guids)
        this->PrintDuplicatedGuidsDetectErrors();

    //duplicated GUIDs detection - we can find duplicated GUID even if
    //skipped in some cases (like duplicated port GUID)
    list_p_fabric_general_err guids_errors;
    int rc2 = this->p_ibdiag->CheckDuplicatedGUIDs(guids_errors);
    rc = this->AnalyzeCheckResults(guids_errors,
            CHECK_NAME_DUPLICATED_GUIDS,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)    //abort and exit
        IBDIAGNET_RETURN(1);

    //if needed than check for duplicated node description
    if (this->to_check_dup_node_desc) {
        rc2 = this->p_ibdiag->CheckDuplicatedNodeDescription(nodes_errors);
        if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)
            WARN_PRINT("Note: If you have unmanaged systems then duplication can occur\n");
        rc |= this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_DUPLICATED_NODE_DESC,
                rc2,
                IBDIAG_ERR_CODE_CHECK_FAILED,
                this->num_errors,
                this->num_warnings,
                true);
    }

    //check port hierarchy info
    list_p_fabric_general_err hierarchy_errors;
    rc2 = this->p_ibdiag->CheckPortHierarchyInfo(hierarchy_errors);
    rc |= this->AnalyzeCheckResults(hierarchy_errors,
            CHECK_NAME_PORT_HIERARCHY_INFO,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    //if the port is down than exit
    bool local_link_is_down;
    if (check_if_local_link_is_down(this->p_ibdiag, local_link_is_down))
        IBDIAGNET_RETURN(1);
    if (local_link_is_down) {
        ERR_PRINT("The Local link is in DOWN state\n");
        IBDIAGNET_RETURN(1);
    }

    this->p_ibdiag->GetIBDiagSMDB().Apply(*(this->p_ibdiag->GetDiscoverFabricPtr()));

    IBDIAGNET_RETURN(rc);
}

int DiscoveryStage::CleanResources()
{
    IBDIAGNET_ENTER;
    p_ibdiag->GetDiscoverFabricPtr()->CleanUpInternalDB();
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


//Returns: SUCCESS_CODE / ERR_CODE_DB_ERR / IBDIAG_ERR_CODE_IBDM_ERR
int DiscoveryStage::CountNodeTypes()
{
    IBDIAG_ENTER;

    this->num_of_switches = 0;
    this->num_of_channel_adapters = 0;
    this->num_of_channel_adapters_sdm = 0;
    this->num_of_routers = 0;
    this->num_of_aggregation_nodes = 0;

    // Get over all nodes
    for (map_str_pnode::iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
         nI != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {
         IBNode *p_curr_node = (*nI).second;
         if (!p_curr_node) {
            this->p_ibdiag->SetLastError("DB error - found null node in NodeByName map"
                               "for key = %s", (*nI).first.c_str());
            IBDIAG_RETURN(IBDIAG_ERR_CODE_DB_ERR);
         }

         if (p_curr_node->type == IB_SW_NODE) {
             this->num_of_switches++;
             continue;
         }

         if (p_curr_node->type == IB_RTR_NODE) {
             this->num_of_routers++;
             continue;
         }

        if (p_curr_node->type == IB_CA_NODE) {
             IBPort *p_port = p_curr_node->getPort(1);
             if (p_port && p_port->p_remotePort &&
                 p_port->p_remotePort->getSpecialPortType() == IB_SPECIAL_PORT_AN) {
                 this->num_of_aggregation_nodes++;
                 continue;
             }

             if (p_curr_node->isSDM()) {
                 ++this->num_of_channel_adapters_sdm;
                 continue;
             }

             this->num_of_channel_adapters++;
         }
    }

    IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);
}

bool DiscoveryStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    if (this->CountNodeTypes())
        IBDIAGNET_RETURN(1);

    if (num_of_channel_adapters_sdm)
        this->max_spaces_fabric_summary = MAX_SPACES_FABRIC_SUMMARY_LONG;

    char buffer[1024] = {0};
    string nodes_str = "Total Nodes";
    snprintf(buffer, sizeof(buffer), "%*.s: %lu", this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
             this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.size());
    summary_str << nodes_str << buffer << "\n";

    CLEAR_STRUCT(buffer);
    nodes_str = "IB Switches";
    snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
            this->num_of_switches);
    summary_str << nodes_str << buffer << "\n";

    CLEAR_STRUCT(buffer);
    nodes_str = "IB Channel Adapters";
    snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
             this->num_of_channel_adapters);
    summary_str << nodes_str << buffer << "\n";

    if (num_of_channel_adapters_sdm) {
        CLEAR_STRUCT(buffer);
        nodes_str = "IB Channel Adapters Socket-Direct";
        snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
                 this->num_of_channel_adapters_sdm);
        summary_str << nodes_str << buffer << "\n";
    }

    CLEAR_STRUCT(buffer);
    nodes_str = "IB Aggregation Nodes";
    snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
             this->num_of_aggregation_nodes);
    summary_str << nodes_str << buffer << "\n";

    CLEAR_STRUCT(buffer);
    nodes_str = "IB Routers";
    snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(nodes_str.length()), " ",
             this->num_of_routers);
    summary_str << nodes_str << buffer << "\n";

    u_int64_t num_ar_switches = 0;
    u_int64_t num_hbf_en_switches = 0;
    u_int64_t num_hbf_sup_switches = 0;

    this->p_ibdiag->GetAREnabledNum(num_ar_switches, num_hbf_en_switches, num_hbf_sup_switches);
    summary_str << "\nAdaptive Routing is enabled on " << num_ar_switches << " switches.\n";

    if (num_hbf_sup_switches)
        summary_str << "Hashed Based Forwarding is enabled on " << num_hbf_en_switches << " switches.\n";

    IBDIAGNET_RETURN(true);
}

void DiscoveryStage::CreateLid2GuidDB()
{
    IBDIAGNET_ENTER;
    uint64_t guid = 0;
    uint16_t lid = 0;
    IBPort * p_port = NULL;
    if (!p_ibdiag->GetDiscoverFabricPtr())
        IBDIAGNET_RETURN_VOID;

    for (map_guid_pport::iterator It = p_ibdiag->GetDiscoverFabricPtr()->PortByGuid.begin();
         It != p_ibdiag->GetDiscoverFabricPtr()->PortByGuid.end();
         It++) {
        guid = (*It).first;
        p_port = (*It).second;
        if (!p_port)
            continue;
        lid = p_port->base_lid;
        lid2guid_map.insert(pair<uint16_t, uint64_t>(lid, guid));
    }
    IBDIAGNET_RETURN_VOID;
}

void DiscoveryStage::SetKeyPerPort(map_guid_2_key_t &guid_2_key, IBISKeyType key_type,
                                   string key_type_str)
{
    IBDIAGNET_ENTER;

    for (map_guid_2_key_t::iterator it = guid_2_key.begin(); it != guid_2_key.end(); it++) {
         IBPort *p_port = this->p_ibdiag->GetDiscoverFabricPtr()->getPortByGuid(it->first);
         if (p_port) {
             lid_t range = (lid_t)(p_port->base_lid + pow(2, p_port->lmc));
             for (lid_t lid = p_port->base_lid; lid < range; lid++) {
                  this->p_ibdiag->GetIbisPtr()->GetKeyManager().SetKey(lid,
                                                                       key_type,
                                                                       it->second);
             }
         } else {
             WARN_PRINT("%s - Port GUID: " U64H_FMT " doesn't exist in fabric\n",
                        key_type_str.c_str(), it->first);
         }
    }

    IBDIAGNET_RETURN_VOID;
}

/******************************************************/
LidsStage::LidsStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_LIDS, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int LidsStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int LidsStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    //dummy one
    IBDIAGNET_RETURN(0);
}

int LidsStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err lids_errors;
    int rc2 = this->p_ibdiag->CheckLids(lids_errors);
    if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)
        found_lids_errors = true;
    rc = this->AnalyzeCheckResults(lids_errors,
            CHECK_NAME_LIDS,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
/////
//ToplogtMatchStage public methods:
/////
TopologyMatchStage::TopologyMatchStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_TOPOLOGY_MATCH, p_ibdiag),
        spec_fabric(),
        output_topo_file(""), input_topo_file(""), topo_ibnl_dir(""),
        use_smart_names(false)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
	//this->compare_nodes_names = 0;
    IBDIAGNET_RETURN_VOID;
}

int TopologyMatchStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int TopologyMatchStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    if (topo_ibnl_dir.empty()) {
        topo_ibnl_dir = generate_file_name(*this->p_base_path, IBNL_OUTPUT_DIR);
        INFO_PRINT("Topo ibnl dir is not given, use default: %s\n", topo_ibnl_dir.c_str());
    }

    //create a topology file if needed
    if (!this->output_topo_file.empty()) {

        //construct Systems - creates IBSystem ports and check IBNL definitions
        ibdmClearInternalLog();

        list< string > dirs;
        dirs.push_back(topo_ibnl_dir);
        theSysDefsCollection()->parseSysDefsFromDirs(dirs);

        this->p_ibdiag->GetDiscoverFabricPtr()->constructSystems();
        char *buffer= ibdmGetAndClearInternalLog();
        if (!buffer) {
            ERR_PRINT("Failed to allocate buffer for ibdm output\n");
            IBDIAGNET_RETURN(1);
        } else {
            u_int32_t errors = num_errors;
            u_int32_t warnings = num_warnings;
            string construct_res(buffer);
            this->ParseIBDMOutput(construct_res);

            if (num_errors > errors) {
                ERR_PRINT("Preparing topology for discovered fabric"
                          " finished with errors.\n");
            } else if (num_warnings > warnings) {
                WARN_PRINT("Preparing topology for discovered fabric"
                           " finished with warnings.\n");
            }
            LOG_PRINT("%s", buffer);
            free(buffer);
        }

        //try to create topo file and ibnl files
        INFO_PRINT("Creating topology file for discovered fabric - %s\n",
                this->output_topo_file.c_str());

        ibdmClearInternalLog();

        OutputControl::Identity identity(this->output_topo_file,
        		                         OutputControl::OutputControl_Flag_UserFile);

        OutputControl::Properties properties(identity);

        if (!this->p_ibdiag->GetDiscoverFabricPtr()->dumpTopology(identity))
            this->p_ibdiag->AddGeneratedFile("Topology",  properties.path());
        else
        	this->p_ibdiag->AddGeneratedFile("Topology",  "Write file failure.");

        buffer = ibdmGetAndClearInternalLog();
        if (!buffer) {
            ERR_PRINT("Failed to allocate buffer for ibdm output\n");
            IBDIAGNET_RETURN(1);
        }
        PRINT("%s", buffer);
        free(buffer);
    }

    if (!this->input_topo_file.empty()) {

         //parsing the topology file
        ibdmClearInternalLog();

        //reparse topo_ibnl_dir in case aditional ibnls were added
        list< string > dirs;
        dirs.push_back(topo_ibnl_dir);
        theSysDefsCollection()->parseSysDefsFromDirs(dirs);

        IBDiag::PrintFileTimestamp(this->input_topo_file, "Topology File");
        int rc = GetSpecFabricPtr()->parseTopology(input_topo_file);
        char *buffer = ibdmGetAndClearInternalLog();
        if (!buffer) {
            ERR_PRINT("Failed to allocate buffer for ibdm output\n");
            IBDIAGNET_RETURN(1);
        }
        PRINT("%s", buffer);
        free(buffer);
        if (rc)
            ERR_PRINT("Parse topology file: %s finished with errors\n",
                      input_topo_file.c_str());
    }

    IBDIAGNET_RETURN(0);
}

int TopologyMatchStage::RunCheck()
{
    IBDIAGNET_ENTER;
    if (this->input_topo_file.empty())
        IBDIAGNET_RETURN(0);

    INFO_PRINT("Topology Matching\n");
    //get root port
    IBPort *p_root_port = this->p_ibdiag->GetRootPort();
    if (!p_root_port) {
        ERR_PRINT("Topology Matching failed, err=%s\n", this->p_ibdiag->GetLastError());
        IBDIAGNET_RETURN(1);
    }

    //match discovered fabric with spec fabric
    char *messages;
    TopoMatchOptions opts;

    /*
    if (TopoMatchFabrics(&this->spec_fabric,        // The specification fabric
            this->p_ibdiag->GetDiscoverFabricPtr(), // The discovered fabric
            p_root_port->p_node->name.c_str(),
                                                    // The name of the node to be the anchor point - in the spec fabric,
            p_root_port->num,                       // The port number of the anchor port - in the spec fabric
            p_root_port->guid_get(),                // Guid of the anchor port - in the discovered fabric
            &messages,
			this->compare_nodes_names)) {
           */
    //opts.mapFileName = mapFileName;
    opts.matchCAsByName = true; // match non-SW by name
    opts.matchSWsByName = true; // match SW by name
    opts.smartSWsName = this->use_smart_names; // use search switches by description if by name failed
    //opts.csvFileName = csvFileName;
    opts.outMapFileName = generate_file_name(*this->p_base_path, "topodiff.names");

    if (TopoMatchFabricsFromEdge(&this->spec_fabric,        // The specification fabric
            this->p_ibdiag->GetDiscoverFabricPtr(), // The discovered fabric
            opts, &messages)) {
        WARN_PRINT("Found mismatches between the topology defined in %s and the discovered fabric.\n",
                this->input_topo_file.c_str());
        if (messages) {
            string topo_result = messages;
            this->ParseIBDMOutput(topo_result);
            LOG_PRINT("%s\n", topo_result.c_str());
        }
    } else {
        INFO_PRINT("The topology defined in %s perfectly matches the discovered fabric\n",
                this->input_topo_file.c_str());
    }

    PRINT("\n");
    free(messages);
    IBDIAGNET_RETURN(0);
}


/******************************************************/
LinksStage::LinksStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_LINKS, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

bool LinksStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    IBLinkWidth width;
    IBLinkSpeed speed;
    string width_str;
    string speed_str;
    char buffer[1024] = {0};

    string links_message = "Total number of links";
    snprintf(buffer, sizeof(buffer), "%*.s: " U32D_FMT, this->max_spaces_fabric_summary - int(links_message.length()), " ",
             this->ib_links_info.GetNumOfIBLinks());
    summary_str << links_message << buffer << "\n";

    links_message = "Links at ";
    vec_vec_uint32 &ib_link_width_speed_matrix = this->ib_links_info.GetIBLinkWidthSpeedMatrix();

    for (int width_index = 1; width_index < IB_LINK_WIDTH_END_INDEX; width_index++) {
         width = width_index2width((IBLinkWidthIndex)width_index);
         width_str = width2char(width);
         for (int speed_index = 1; speed_index < IB_LINK_SPEED_END_INDEX; speed_index++) {
              if (!ib_link_width_speed_matrix[width_index][speed_index])
                  continue;
              speed = speed_index2speed((IBLinkSpeedIndex)speed_index);
              speed_str = speed2char(speed);

              CLEAR_STRUCT(buffer);
              string width_speed_str = width_str + speed_str;
              snprintf(buffer, sizeof(buffer), "%s%*.s: " U32D_FMT, width_speed_str.c_str(),
                       this->max_spaces_fabric_summary -
                       int(links_message.length() + width_speed_str.length()), " ",
                       ib_link_width_speed_matrix[width_index][speed_index]);

              summary_str << links_message << buffer << "\n";
         }
    }

    IBDIAGNET_RETURN(true);
}

int LinksStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int LinksStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    //dummy one
    IBDIAGNET_RETURN(0);
}

int LinksStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err links_errors;
    int rc2 = this->p_ibdiag->CheckLinks(links_errors, &this->ib_links_info);
    if (rc2 == IBDIAG_ERR_CODE_CHECK_FAILED)
        found_links_errors = true;
    rc = this->AnalyzeCheckResults(links_errors,
            CHECK_NAME_LINKS,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
SMStage::SMStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_SM, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

bool SMStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    string p_sm_masters_list_str = "";
    string p_sm_standby_list_str = "";

    list_p_sm_info_obj &sm_info_list = this->p_ibdiag->GetIBDMExtendedInfoPtr()->
                                             getSMPSMInfoListRef();

    char buffer[1024] = {0};
    //collect info
    for (list_p_sm_info_obj::iterator it = sm_info_list.begin(); it != sm_info_list.end(); ++it) {
         if (!(((*it)->smp_sm_info.SmState) == IBIS_IB_SM_STATE_MASTER) &&
             !(((*it)->smp_sm_info.SmState) == IBIS_IB_SM_STATE_STANDBY)) {
             continue;
         }

         CLEAR_STRUCT(buffer);
         snprintf(buffer, sizeof(buffer), "Port=" U32D_FMT " LID=%u" " GUID=" U64H_FMT \
                  " devid=" U32D_FMT " Priority:" U32D_FMT " Node_Type=%s Node_Description=%s",
                  (*it)->p_port->num,
                  (*it)->p_port->base_lid,
                  (*it)->p_port->guid_get(),
                  (*it)->p_port->p_node->devId,
                  (*it)->smp_sm_info.Priority,
                  nodetype2char((*it)->p_port->p_node->type),
                  (*it)->p_port->p_node->description.c_str());

         if (((*it)->smp_sm_info.SmState) == IBIS_IB_SM_STATE_MASTER) {
             p_sm_masters_list_str += "Master SM: " + string(buffer) + "\n";
         }

         if (((*it)->smp_sm_info.SmState) == IBIS_IB_SM_STATE_STANDBY) {
             p_sm_standby_list_str += "Standby SM: " + string(buffer) + "\n";
         }
    }

    if (p_sm_masters_list_str == "")
        p_sm_masters_list_str = "Master SM  : No Master SM\n";

    if (p_sm_standby_list_str == "")
        p_sm_standby_list_str = "Standby SM : No Standby SM\n";

    summary_str << p_sm_masters_list_str << p_sm_standby_list_str;

    IBDIAGNET_RETURN(true);
}

int SMStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int SMStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err sm_errors;
    int rc2 = this->p_ibdiag->BuildSMInfoDB(sm_errors);
    rc = this->AnalyzeCheckResults(sm_errors,
            CHECK_NAME_SM_INFO_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (!rc) {
        if (this->p_ibdiag->DumpSMInfoCSVTable(*this->p_csv_out)) {
            ERR_PRINT("Writing SM info to DB failed\n");
            ++this->num_errors;
        }
        if (this->p_ibdiag->WriteSMFile(SM_FILE)) {
            ERR_PRINT("Writing Subnet Manager file failed.\n");
            ++this->num_errors;
        }
    }
    IBDIAGNET_RETURN(rc);
}

int SMStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc;

    list_p_fabric_general_err sm_errors;
    int rc2 = this->p_ibdiag->CheckSMInfo(sm_errors);
    rc = this->AnalyzeCheckResults(sm_errors,
            CHECK_NAME_SM_INFO,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
/////
//pm private methods:
/////
static bool pm_caught_signal = false;

#ifdef __WIN__
    int signals_array[] = {SIGINT};
#else
    int signals_array[] = {SIGINT, SIGQUIT, SIGABRT, SIGPIPE, SIGTERM};
#endif


void pm_stage_sig_handler(int signum)
{
    IBDIAGNET_ENTER;
    // no need to set handler again because we want the default handler
    // from here

    //mark we caught the signal
    pm_caught_signal = true;

    SCREEN_PRINT("\n");
    if (signum == SIGINT) {
        INFO_PRINT("Received User ^C (%s). Stopping\n", strsignal(signum));
    } else {
        INFO_PRINT("Received signal %d (%s). Stopping\n", signum, strsignal(signum));
    }
    IBDIAGNET_RETURN_VOID;
}

void PMStage::Sleep(u_int32_t seconds)
{
    IBDIAGNET_ENTER;
    u_int32_t seconds_left = seconds;
    INFO_PRINT("Going to sleep for %u seconds until next counters sample\n", seconds);
    while (!pm_caught_signal && seconds_left) {
        SCREEN_INFO_PRINT("Time left to sleep ... %u seconds.\r", seconds_left);
        fflush(stdout);
        sleep(1);
        --seconds_left;
    }
    SCREEN_PRINT("\n");
    IBDIAGNET_RETURN_VOID;
}

/////
//pm public methods:
/////
PMStage::PMStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_PM, p_ibdiag),
        to_reset_counters(false),
        to_calc_ber(false),
        m_check_counters_bitset(0x00),
        en_per_lane_cnts(false),
        en_per_slvl_cntrs(false),
        ber_threshold(DEFAULT_BER_THRESHOLD),
        pause_time_between_samples(DEFAULT_SECONDS_BETWEEN_COUNTERS_SAMPLES),
        diff_time_between_samples(0),
        can_send_mads_by_lid(false),
        total_pm_symbol_errors(0)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int PMStage::CleanResources()
{
    IBDIAGNET_ENTER;

    this->total_pm_symbol_errors = 0;

    this->p_ibdiag->GetIBDMExtendedInfoPtr()->
           CleanPMInfoObjVector(this->first_sample_pm_info_obj_vector);
    this->p_ibdiag->GetIBDMExtendedInfoPtr()->
           CleanPMInfoObjVector(this->p_ibdiag->GetIBDMExtendedInfoPtr()->getPMInfoObjVectorRef());

    for (vec_slvl_cntrs::iterator it = m_slvl_cntrs_vec.begin();
         it != m_slvl_cntrs_vec.end(); it++) {

        if (*it)
            release_container_data((*it)->m_set_port_data_update);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

PMStage::~PMStage()
{
    IBDIAGNET_ENTER;

    this->CleanResources();
    release_container_data(m_slvl_cntrs_vec);

    IBDIAGNET_RETURN_VOID;
}

int PMStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag, this->can_send_mads_by_lid))
        IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

void PMStage::PreparePerSlVlCounters()
{

    IBDIAGNET_ENTER;
    //VS Class
    CountersPerSLVL * port_rcv_data_vl = new PortRcvDataVL();
    m_slvl_cntrs_vec.push_back(port_rcv_data_vl);

    CountersPerSLVL * port_xmit_data_vl = new PortXmitDataVL();
    m_slvl_cntrs_vec.push_back(port_xmit_data_vl);

    CountersPerSLVL * port_rcv_ext_data_vl = new PortRcvDataVLExt();
    m_slvl_cntrs_vec.push_back(port_rcv_ext_data_vl);

    CountersPerSLVL * port_xmit_ext_data_vl = new PortXmitDataVLExt();
    m_slvl_cntrs_vec.push_back(port_xmit_ext_data_vl);

    CountersPerSLVL * port_rcv_pkt_vl = new PortRcvPktVL();
    m_slvl_cntrs_vec.push_back(port_rcv_pkt_vl);

    CountersPerSLVL * port_xmit_pkt_vl = new PortXmitPktVL();
    m_slvl_cntrs_vec.push_back(port_xmit_pkt_vl);

    CountersPerSLVL * port_rcv_ext_pkt_vl = new PortRcvPktVLExt();
    m_slvl_cntrs_vec.push_back(port_rcv_ext_pkt_vl);

    CountersPerSLVL * port_xmit_ext_pkt_vl = new PortXmitPktVLExt();
    m_slvl_cntrs_vec.push_back(port_xmit_ext_pkt_vl);

    CountersPerSLVL * port_xmit_ext_wait_vl = new PortXmitWaitVLExt();
    m_slvl_cntrs_vec.push_back(port_xmit_ext_wait_vl);

    //PM Class
    CountersPerSLVL * port_xmit_data_sl = new PortXmitDataSL();
    m_slvl_cntrs_vec.push_back(port_xmit_data_sl);

    CountersPerSLVL * port_rcv_data_sl = new PortRcvDataSL();
    m_slvl_cntrs_vec.push_back(port_rcv_data_sl);

    CountersPerSLVL * port_xmit_data_sl_ext = new PortXmitDataSLExt();
    m_slvl_cntrs_vec.push_back(port_xmit_data_sl_ext);

    CountersPerSLVL * port_rcv_data_sl_ext = new PortRcvDataSLExt();
    m_slvl_cntrs_vec.push_back(port_rcv_data_sl_ext);

    CountersPerSLVL* port_vl_xmit_flow_ctl_update_errors = new PortVLXmitFlowCtlUpdateErrors();
    m_slvl_cntrs_vec.push_back(port_vl_xmit_flow_ctl_update_errors);

    /*

    //
    // Counters was replaced by PortXmitWaitVLExt from VS specs
    //
    CountersPerSLVL* port_vl_xmit_wait_vl = new PortXmitWaitVL();
    m_slvl_cntrs_vec.push_back(port_vl_xmit_wait_vl);

    */

    IBDIAGNET_RETURN_VOID;
}


void PMStage::CountTotalPMSymbolError(u_int64_t pm_threshold)
{
    IBDIAG_ENTER;

    IBDMExtendedInfo* fabric_extended_info = this->p_ibdiag->GetIBDMExtendedInfoPtr();
    for (u_int32_t i = 0; i < fabric_extended_info->getPortsVectorSize(); ++i) {
        IBPort *p_curr_port = fabric_extended_info->getPortPtr(i);
        if (!p_curr_port)               //port not exist
            continue;

        struct PM_PortCountersExtended *p_curr_port_counters_extended =
                                                 fabric_extended_info->getPMPortCountersExtended(i);
        if (!p_curr_port_counters_extended)      //info not exist
            continue;

        if (p_curr_port_counters_extended->SymbolErrorCounter > pm_threshold)
            this->total_pm_symbol_errors++;
    }

    IBDIAG_RETURN_VOID;
}


bool PMStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    // check if threshold was given to symbol_error_counter_extended counter
    map_str_uint64::iterator it = this->counters_to_threshold_map.
            find(string("symbol_error_counter_extended"));

    if (it == this->counters_to_threshold_map.end())
        IBDIAGNET_RETURN(false);

    this->CountTotalPMSymbolError(it->second);

    if (this->total_pm_symbol_errors == 0)
        IBDIAGNET_RETURN(false);

    summary_str << "High symbol errors reported by "
                << this->total_pm_symbol_errors
                << " ports"
                << endl;

    IBDIAGNET_RETURN(true);
}


int PMStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    if (!this->can_send_mads_by_lid) {
        if (this->to_reset_counters)
            INFO_PRINT("%s skipped\n", CHECK_NAME_PM_RESET);
        INFO_PRINT("%s skipped\n", CHECK_NAME_PM_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err ports_errors;
    //if need to reset
    if (this->to_reset_counters) {
        int rc2  = this->p_ibdiag->ResetPortCounters(ports_errors,
                                                     m_check_counters_bitset);
        SCREEN_PRINT("\n\n");
        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_PM_RESET,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (rc)
            IBDIAGNET_RETURN(rc);

        // Reset RN counters
        rc2 = this->p_ibdiag->ClearRNCounters(ports_errors);

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_ROUTE_RN_COUNTERS_CLEAR,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (rc)
            IBDIAGNET_RETURN(rc);

        // Reset HBF counters
        rc2 = this->p_ibdiag->ClearHBFCounters(ports_errors);

        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_HBF_COUNTERS_CLEAR,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    //retrieve all counters - first sample
    struct timeval start_measure;
    if ((rc = gettimeofday(&start_measure, NULL)) != 0) {
        IBDIAG_LOG(TT_LOG_LEVEL_INFO,
                   "Failed to get gettimeofday before first sample.");
        ERR_PRINT("Failed to retrieve first sample time\n");
        ++this->num_errors;
        IBDIAGNET_RETURN(1);
    }

    int rc2 = this->p_ibdiag->BuildPortCountersDB(ports_errors,
                                                  m_check_counters_bitset);
    PRINT("\n");
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_PM_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc)
        IBDIAGNET_RETURN(rc);

    // Build RN counters
    rc2 = this->p_ibdiag->BuildRNCounters(ports_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_ROUTE_RN_COUNTERS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc)
        IBDIAGNET_RETURN(rc);

    // Build HBF counters
    rc2 = this->p_ibdiag->BuildHBFCounters(ports_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_HBF_COUNTERS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc)
        IBDIAGNET_RETURN(rc);

    if (this->pause_time_between_samples != 0) {   //need to take another sample
        //copy counters vector of first sample
        this->p_ibdiag->CopyPMInfoObjVector(this->first_sample_pm_info_obj_vector);

        //set a signal handler for kill signals
        for (unsigned int i = 0; i < (sizeof(signals_array) / sizeof(signals_array[0])); ++i) {
            signal(signals_array[i], pm_stage_sig_handler);
        }
        //pausing
        this->Sleep(this->pause_time_between_samples);
        //set default handlers for kill signals
        for (unsigned int i = 0; i < (sizeof(signals_array) / sizeof(signals_array[0])); ++i) {
            signal(signals_array[i], SIG_DFL);
        }

        //retrieve all counters - second sample
        struct timeval end_measure;
        if ((rc = gettimeofday(&end_measure, NULL)) != 0) {
            IBDIAG_LOG(TT_LOG_LEVEL_INFO,
                       "Failed to get gettimeofday before second sample.");
            ERR_PRINT("Failed to retrieve second sample time\n");
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }

        int rc2 = this->p_ibdiag->BuildPortCountersDB(ports_errors,
                                                      m_check_counters_bitset);
        SCREEN_PRINT("\n\n");

        long long elapsed_mic_sec = (end_measure.tv_sec - start_measure.tv_sec)*1000000LL +
                                    (end_measure.tv_usec - start_measure.tv_usec);

        this->diff_time_between_samples = (double)elapsed_mic_sec / (double)1000000;

        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_PM_RETRIEVE2,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    if (this->en_per_slvl_cntrs) {
        this->PreparePerSlVlCounters();
        if (this->to_reset_counters) {
            for (vec_slvl_cntrs::iterator it = m_slvl_cntrs_vec.begin();
                    it != m_slvl_cntrs_vec.end(); it++) {
                CountersPerSLVL * cntrs_per_slvl = *it;
                int rc2 = this->p_ibdiag->BuildDBOrResetSLVLCntrs(
                        ports_errors,
                        true,//clear slvl counters
                        cntrs_per_slvl);

                SCREEN_PRINT("\n");
                rc = this->AnalyzeCheckResults(ports_errors,
                        cntrs_per_slvl->GetCntrHeader(),
                        rc2,
                        IBDIAG_ERR_CODE_FABRIC_ERROR,
                        this->num_errors,
                        this->num_warnings);
                ports_errors.clear();
            }
        }

        for (vec_slvl_cntrs::iterator it = m_slvl_cntrs_vec.begin();
             it != m_slvl_cntrs_vec.end(); it++) {
            CountersPerSLVL * cntrs_per_slvl = *it;
            int rc2 = this->p_ibdiag->BuildDBOrResetSLVLCntrs(
                    ports_errors,
                    false,//build slvl counters DB
                    cntrs_per_slvl);
            SCREEN_PRINT("\n");
            rc = this->AnalyzeCheckResults(ports_errors,
                    cntrs_per_slvl->GetCntrHeader(),
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings);
            ports_errors.clear();
        }
    }

    IBDIAGNET_RETURN(rc);
}

int PMStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("Ports counter Checks skipped\n");
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }
    list_p_fabric_general_err ports_errors;

    //calculate counters calculation
    int rc2 = this->p_ibdiag->CalcCounters(this->first_sample_pm_info_obj_vector,
                                           this->diff_time_between_samples,
                                           ports_errors);

    list_p_fabric_general_err pm_threshold_overflow_warnings;
    // Errors accumulated in the CalcCounters will be analyzed with CheckAllPMValues later
    //counters value check
    rc2 |= this->p_ibdiag->CheckAllPMValues(ports_errors,
                                            pm_threshold_overflow_warnings,
                                            this->counters_to_threshold_map,
                                            m_check_counters_bitset);

    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_PM_VALUE,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->AnalyzeCheckResults(pm_threshold_overflow_warnings,
            CHECK_NAME_OVERFLOW_PM_VALUE,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (this->p_ibdiag->DumpPMPortSamplesControlCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Writing Port Samples Control to DB CSV failed\n");
        ++this->num_errors;
    }

    // now when we have the calculated counters, we can print to the files
    if (this->p_ibdiag->DumpPortCountersCSVTable(*this->p_csv_out, m_check_counters_bitset)) {
        ERR_PRINT("Writing ports counters to DB failed\n");
        ++this->num_errors;
    }

    if (this->en_per_slvl_cntrs) {
        if (this->p_ibdiag->DumpPerSLVLPortCntrsCSVTable(*this->p_csv_out,
                                                         this->m_slvl_cntrs_vec)) {
            ERR_PRINT("Write SLVL Counters PM Data to DB failed\n");
            ++this->num_errors;
        }
    }

    if (this->p_ibdiag->DumpHBFCountersCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Write HBF counters to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WritePMFile(PM_FILE,
                                    m_check_counters_bitset,
                                    en_per_lane_cnts)) {
        ERR_PRINT("Writing ports counters file failed\n");
        ++this->num_errors;
    }

    list_p_fabric_general_err pfrn_errors;

    if (this->p_ibdiag->DumpRNCountersToCSV(*this->p_csv_out, pfrn_errors)) {
        ERR_PRINT("Writing RN counters to DB failed\n");
        ++this->num_errors;
    }

    // check for bad values in pfrn_received_error counter
    this->AnalyzeCheckResults(pfrn_errors,
            CHECK_NAME_PFRN_RECEIVED_ERROR,
            0,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings,
            true);

    if (this->p_ibdiag->WriteRNCounters_2_File(RNC_FILE_2)) {
        ERR_PRINT("Writing RN counters 2 file failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WriteRNCountersFile(RNC_FILE)) {
        ERR_PRINT("Writing RN counters file failed\n");
        ++this->num_errors;
    }

    //abort all other checks - we don't have another sample
    if (this->pause_time_between_samples == 0) {
        INFO_PRINT("Ports counters Difference Check will be skipped - pause time is zero\n");
        IBDIAGNET_RETURN(rc);
    }

    //no increase check
    rc2 = this->p_ibdiag->CheckCountersDiff(this->first_sample_pm_info_obj_vector,
            ports_errors);

    rc = this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_PM_DIFF,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    if (rc)
        IBDIAGNET_RETURN(rc);

    rc = this->p_ibdiag->DumpPortCountersDeltaCSVTable(*this->p_csv_out,
                                    this->first_sample_pm_info_obj_vector,
                                    m_check_counters_bitset,
                                    ports_errors);
    if (!ports_errors.empty())
        INFO_PRINT("Negative values were found while delta for port counters was calculated\n");

    this->AnalyzeCheckResults(ports_errors,
            CHECK_NAME_PM_DELTA,
            rc,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    //BER check - only if needed
    if (this->to_calc_ber) {
        rc2 = this->p_ibdiag->CalcBERErrors(this->first_sample_pm_info_obj_vector,
                (u_int64_t)(this->ber_threshold),
                this->diff_time_between_samples,
                ports_errors,
                *this->p_csv_out);
        rc = this->AnalyzeCheckResults(ports_errors,
                CHECK_NAME_BER,
                rc2,
                IBDIAG_ERR_CODE_CHECK_FAILED,
                this->num_errors,
                this->num_warnings);

        rc2 = this->p_ibdiag->CalcPhyTest(this->first_sample_pm_info_obj_vector,
                                          this->diff_time_between_samples,
                                          *this->p_csv_out);
    }
    IBDIAGNET_RETURN(rc);
}


/******************************************************/

/////
//RoutingStage public methods:
/////
RoutingStage::RoutingStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_ROUTING, p_ibdiag),
        retrieved_switch_info_ok(false), retrieved_uc_fdbs_ok(false),
        retrieved_mc_fdbs_ok(false), retrieved_ext_switch_info_ok(false),
        retrieved_plft_data_ok(false), retrieved_ar_data_ok(false),
        retrieved_rn_data_ok(false), ar_enabled(false), is_fat_tree(false), use_vlr(false),
        use_vs_route(true), check_mcast_credloop(false), check_sl(false),
        dump_only(false), dump_only_skip_routing_tables(false),
        static_ca2ca(false), default_sl(0), num_ar_switches(0), num_hbf_switches(0)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
    IBDIAGNET_RETURN_VOID;
}

int RoutingStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

void RoutingStage::DumpHBFTables()
{
    IBDIAGNET_ENTER;

    if (this->p_ibdiag->DumpCSV_HBFTable(*this->p_csv_out)) {
        ERR_PRINT("Writing of HBF table failed with: %s\n", p_ibdiag->GetLastError());
        ++this->num_errors;
    }

    IBDIAGNET_RETURN_VOID;
}

int RoutingStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = IBDIAG_SUCCESS_CODE;
    int rc2 = IBDIAG_SUCCESS_CODE;
    int rc3 = IBDIAG_SUCCESS_CODE;
    list_p_fabric_general_err nodes_errors;

    string parser_output;
    if(!sa_dump_file.empty() && check_mcast_credloop) {
        INFO_PRINT("Load SA Dump file (%s).\n", sa_dump_file.c_str());

        //To check vport and aguids in the sadump file,
        //need to call VirtualizationStage & AGuidStage before this function
        rc = p_ibdiag->ParseSADumpFile(sa_dump_file, parser_output);
        LOG_PRINT("%s\n", parser_output.c_str());

        if (rc) {
            ERR_PRINT("Failed to parse SA Dump file (%s).\n", sa_dump_file.c_str());
            ++this->num_errors;
            IBDIAGNET_RETURN(1);
        }
    }

    if (use_vs_route) {
        rc2  = this->p_ibdiag->RetrieveExtendedSwitchInfo(nodes_errors);
        if (rc2 == IBDIAG_SUCCESS_CODE)
            this->retrieved_ext_switch_info_ok = true;
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_EXT_SW_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (!rc)
            if (this->p_ibdiag->WriteVL2VLFile(VL2VL_FILE)) {
                ERR_PRINT("Writing VL2VL file failed\n");
                ++this->num_errors;
            }

        //bring PLFT Data
        unsigned int supportedDev;
        rc2  = this->p_ibdiag->RetrievePLFTData(nodes_errors, supportedDev);
        if (rc2 == IBDIAG_SUCCESS_CODE){
            this->retrieved_plft_data_ok = true;
            INFO_PRINT("PLFT is enabled on %d switches.\n",supportedDev);
        }
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_PLFT_DATA_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (!rc)
            if (this->p_ibdiag->WritePLFTFile(PLFT_FILE)) {
                ERR_PRINT("Writing PLFT file failed\n");
                ++this->num_errors;
            }

        //bring AR Data
        AdditionalRoutingDataMap routing_data_map;
        AdditionalRoutingDataMap* p_routing_data_map = NULL;
        p_routing_data_map = &routing_data_map;

        rc2  = this->p_ibdiag->RetrieveARData(nodes_errors,
                                              supportedDev,
                                              p_routing_data_map,
                                              this->dump_only_skip_routing_tables);
        //includes switches with either HBF or AR enabled
        this->num_ar_switches = (u_int32_t)supportedDev;
        INFO_PRINT("Adaptive Routing is enabled on %d switches.\n", supportedDev);
        if (rc2 == IBDIAG_SUCCESS_CODE){
            this->retrieved_ar_data_ok = true;
            this->ar_enabled = (supportedDev > 0);
        }
        int ar_result_rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_AR_DATA_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);


        //we have to retrive HBF data before the FAR file is written
        supportedDev = 0;
        rc2  = this->p_ibdiag->RetrieveHBFData(nodes_errors,
                                              supportedDev);
        INFO_PRINT("Hashed Based Forwarding is enabled on %d switches.\n", supportedDev);
        this->num_hbf_switches = (u_int32_t)supportedDev;
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_HBF_DATA_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        this->DumpHBFTables();

        if (!this->dump_only_skip_routing_tables) {
            if (!ar_result_rc) {
                if (this->p_ibdiag->WriteARFile(AR_FILE)) {
                    ERR_PRINT("Writing AR file failed\n");
                    ++this->num_errors;
                }
            }

            if (this->p_ibdiag->WriteFARFile(&routing_data_map, FAR_FILE, false)) {
                ERR_PRINT("Writing FAR file failed\n");
                ++this->num_errors;
            }

            if (this->p_ibdiag->WriteFARFile(&routing_data_map, FAR_FLID_FILE, true)) {
                ERR_PRINT("Writing FAR FLID file failed\n");
                ++this->num_errors;
            }
        }

        INFO_PRINT("Routing Notification is supported on %d switches.\n",
                AdditionalRoutingData::sw_supports_rn_count);

        if (AdditionalRoutingData::sw_supports_rn_count == 0)
            PRINT("\n");

        if (AdditionalRoutingData::sw_supports_rn_count != 0) {

            rc2 = this->p_ibdiag->RetrieveRNData(nodes_errors, &routing_data_map);

            if (rc2 == IBDIAG_SUCCESS_CODE)
                this->retrieved_rn_data_ok = true;

            this->AnalyzeCheckResults(nodes_errors,
                    CHECK_NAME_ROUTE_RN_DATA_RETRIEVE,
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    true);

            rc2 = this->p_ibdiag->WriteRNFile(nodes_errors, &routing_data_map, RN_FILE);

            if (rc2)
                ERR_PRINT("Writing RN file failed\n");

            this->AnalyzeCheckResults(nodes_errors,
                    CHECK_NAME_ROUTE_DUMP_RN_DATA,
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    true); // only_warning?
        }

        // Build pFRN Data
        INFO_PRINT("Build pFRN Data\n");
        rc2 = this->p_ibdiag->Retrieve_pFRNData(nodes_errors);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_PFRN_DATA_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        INFO_PRINT("Build N2N Class Port Info\n");
        rc2 = this->p_ibdiag->Retrieve_N2NClassPortInfo(nodes_errors);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_N2N_CPI_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        INFO_PRINT("Build N2N Neighbors Info\n");
        rc2 = this->p_ibdiag->Retrieve_NeighborsInfo(nodes_errors);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_NEIGHBORS_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        INFO_PRINT("Build N2N Key Info\n");
        rc2 = this->p_ibdiag->Retrieve_N2NKeyInfo(nodes_errors);

        SCREEN_PRINT("\n");
        this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_N2N_KEY_INFO_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }

    if (this->dump_only_skip_routing_tables) {
        LOG_INFO_PRINT("Requested to skip routing tables (LFTs) retrieving.\n");
    } else {

        //bring unicast FDBs info
        INFO_PRINT("Build Unicast FDBS Info\n");
        rc2  = this->p_ibdiag->RetrieveUCFDBSInfo(nodes_errors);
        if (rc2 == IBDIAG_SUCCESS_CODE)
            this->retrieved_uc_fdbs_ok = true;
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_UCFDBS_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (!rc)
            if (this->p_ibdiag->WriteUCFDBSFile(UC_FDBS_FILE)) {
                ERR_PRINT("Writing unicast FDBS file failed\n");
                ++this->num_errors;
            }

        if (!this->retrieved_uc_fdbs_ok)    //abort
            IBDIAGNET_RETURN(rc);

        //bring multicast FDBs info
        INFO_PRINT("Build Multicast FDBS Info\n");
        rc2  = this->p_ibdiag->RetrieveMCFDBSInfo(nodes_errors);
        if (rc2 == IBDIAG_SUCCESS_CODE)
            this->retrieved_mc_fdbs_ok = true;
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(nodes_errors,
                CHECK_NAME_ROUTE_MCASTDBS_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (!rc)
            if (this->p_ibdiag->WriteMCFDBSFile(MC_FDBS_FILE)) {
                ERR_PRINT("Writing multicast FDBS file failed\n");
                ++this->num_errors;
            }

        //Write SLVL File
        if (use_vlr || use_vs_route || check_mcast_credloop || check_sl) {
            INFO_PRINT("Build SLVL Table\n");
            rc2  = this->p_ibdiag->WriteSLVLFile(SLVL_FILE,
                    nodes_errors);
            SCREEN_PRINT("\n");
            rc = this->AnalyzeCheckResults(nodes_errors,
                    CHECK_NAME_ROUTE_DUMP_SLVL_TABLE,
                    rc2,
                    IBDIAG_ERR_CODE_FABRIC_ERROR,
                    this->num_errors,
                    this->num_warnings,
                    true);

        }

        // bring VL Arbitration Table info
        INFO_PRINT("Build VL Arbitration Table\n");
        rc2 = this->p_ibdiag->RetrieveVLArbitrationTable(nodes_errors);

        SCREEN_PRINT("\n");
        rc3 = this->AnalyzeCheckResults(nodes_errors,
                CHECK_VL_ARBITRATION_TABLE_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }

    if (rc)
        IBDIAGNET_RETURN(rc);
    if (rc3)
        IBDIAGNET_RETURN(rc3);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int RoutingStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    if (this->dump_only || this->dump_only_skip_routing_tables) {
        LOG_INFO_PRINT("Requested to skip Routing checks\n");
        IBDIAGNET_RETURN(rc);
    }

    if ((!this->retrieved_uc_fdbs_ok) || (!this->retrieved_mc_fdbs_ok)) {
        ERR_PRINT("Skipping Routing checks because not all information retrieved successfully\n");
        ++this->num_errors;
        IBDIAGNET_RETURN(rc);
    }

    this->p_ibdiag->SetDefaultSL(default_sl);

    if (use_vs_route) {

        if ((!this->retrieved_ext_switch_info_ok) ||
            (!this->retrieved_plft_data_ok) ||
            (!this->retrieved_ar_data_ok) ) {
            ERR_PRINT("Skipping Adaptive Routing checks because" \
                      " not all information retrieved successfully\n");
            ++this->num_errors;
            this->ar_enabled = false;
        }
    }

    LOG_PRINT("\n\n");
    string routing_output;

    if (check_sl) {
        LOG_PRINT("###################################\n");
        LOG_INFO_PRINT("SL2VL table check:\n");
        LOG_PRINT("###################################\n");
        this->p_ibdiag->CheckSL2VLTables(routing_output);
        this->ParseIBDMOutput(routing_output);
        LOG_PRINT("%s\n", routing_output.c_str());
    }

    LOG_PRINT("###################################\n");
    LOG_INFO_PRINT("Fabric Qualities Report:\n");
    LOG_PRINT("###################################\n");
    this->p_ibdiag->ReportFabricQualities(routing_output, this->p_base_path->c_str(),
                                          ar_enabled, static_ca2ca);
    this->ParseIBDMOutput(routing_output);
    LOG_PRINT("%s\n", routing_output.c_str());

    if (use_vs_route && this->ar_enabled) {

        LOG_PRINT("###################################\n");
        LOG_INFO_PRINT("Fabric AR Connectivity Report:\n");
        LOG_PRINT("###################################\n");
        this->p_ibdiag->ReportFabricARConnectivity(routing_output);
        this->ParseIBDMOutput(routing_output);
        LOG_PRINT("%s\n", routing_output.c_str());

        // AR validation - validate that AR routing tables are configured correctly.
        LOG_PRINT("###################################\n");
        LOG_INFO_PRINT("AR Validation Report:\n");
        LOG_PRINT("###################################\n");
        this->p_ibdiag->ReportFabricARValidation(routing_output);
        this->ParseIBDMOutput(routing_output);
        LOG_PRINT("%s\n", routing_output.c_str());
    }

    LOG_PRINT("###################################\n");
    LOG_INFO_PRINT("Credit Loops Report:\n");
    LOG_PRINT("###################################\n");
    if(check_mcast_credloop)
        CredLoopMode(0,1);

    this->p_ibdiag->ReportCreditLoops(routing_output, this->is_fat_tree,
                                      this->ar_enabled);
    this->ParseIBDMOutput(routing_output);
    LOG_PRINT("%s\n", routing_output.c_str());
    LOG_PRINT("\n\n");

    // Dump VL Arbitration Table to CSV
    this->p_ibdiag->DumpVLArbitrationCSVTable(*this->p_csv_out);

    this->p_ibdiag->Dump_pFRNConfigToCSV(*this->p_csv_out);
    this->p_ibdiag->Dump_N2NClassPortInfoToCSV(*this->p_csv_out);
    this->p_ibdiag->Dump_NeighborsInfoToCSV(*this->p_csv_out);
    this->p_ibdiag->Dump_N2NKeyInfoToCSV(*this->p_csv_out);

    // pFRN validations
    list_p_fabric_general_err pfrn_errors;
    int rc2 = 0;

    rc2 = this->p_ibdiag->pFRNSupportAndTrapsValidation(pfrn_errors);
    this->AnalyzeCheckResults(pfrn_errors,
            CHECK_NAME_PFRN_SUPPORT_AND_TRAPS,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    rc2 = this->p_ibdiag->pFRNNeighborsValidation(pfrn_errors);
    this->AnalyzeCheckResults(pfrn_errors,
            CHECK_NAME_PFRN_NEIGHBORS,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    rc2 = this->p_ibdiag->ARGroupsUniformValidation(pfrn_errors);
    this->AnalyzeCheckResults(pfrn_errors,
            CHECK_NAME_AR_GROUPS_UNIFORM,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    IBDIAGNET_RETURN(rc);
}


/******************************************************/
NodesInfoStage::NodesInfoStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_NODES_INFO, p_ibdiag),
        can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int NodesInfoStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag, this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int NodesInfoStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_NODES_INFO_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    if (!g_vs_cap_gmp_run) {
        INFO_PRINT("%s wasn't ran. Skipping %s\n",
                   CHECK_VS_CAP_GMP, CHECK_NAME_NODES_INFO_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err nodes_errors;

    if (this->p_ibdiag->DumpNodesInfoCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Writing Nodes Info to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WriteNodesInfoFile(NODES_INFO_FILE)) {
        ERR_PRINT("Writing Nodes Info file failed\n");
        ++this->num_errors;
    }

    IBDIAGNET_RETURN(rc);
}

int NodesInfoStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_NODES_INFO_FW);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }
    if (!g_vs_cap_gmp_run) {
        INFO_PRINT("%s wasn't ran. Skipping %s\n",
                   CHECK_VS_CAP_GMP, CHECK_NAME_NODES_INFO_FW);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err fw_errors;
    int rc2 = this->p_ibdiag->CheckFWVersion(fw_errors);
    rc = this->AnalyzeCheckResults(fw_errors,
            CHECK_NAME_NODES_INFO_FW,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
Speed_WidthStage::Speed_WidthStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_SPEED_WIDTH, p_ibdiag),
        expected_link_speed(""), expected_link_width("")
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int Speed_WidthStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int Speed_WidthStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    list_p_fabric_general_err links_errors;

    //check link speed
    if (this->expected_link_speed != "") {
        INFO_PRINT("Link Speed Check (Expected value given = %s)\n", this->expected_link_speed.c_str());
    } else {
        INFO_PRINT("Link Speed Check (Compare to supported link speed)\n");
    }
    int rc2 = this->p_ibdiag->CheckLinkSpeed(links_errors, this->expected_link_speed);
    rc = this->AnalyzeCheckResults(links_errors,
            CHECK_NAME_LINK_SPEED,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    //check link width
    if (this->expected_link_width != "") {
        INFO_PRINT("Link Width Check (Expected value given = %s)\n", this->expected_link_width.c_str());
    } else {
        INFO_PRINT("Link Width Check (Compare to supported link width)\n");
    }
    rc2 = this->p_ibdiag->CheckLinkWidth(links_errors, this->expected_link_width);
    rc = this->AnalyzeCheckResults(links_errors,
            CHECK_NAME_LINK_WIDTH,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
PKeyStage::PKeyStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_PKEY, p_ibdiag),
        can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

PKeyStage::~PKeyStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int PKeyStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int PKeyStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err pkey_errors;
    int rc2 = this->p_ibdiag->BuildPartitionKeysDB(pkey_errors);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(pkey_errors,
            CHECK_NAME_PKEY_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (!rc) {
        if (this->p_ibdiag->DumpPartitionKeysCSVTable(*this->p_csv_out)) {
            ERR_PRINT("Writing partition key info to DB failed\n");
            ++this->num_errors;
        }

        if (this->p_ibdiag->WritePKeyFile(PKEY_FILE)) {
            ERR_PRINT("Writing partition keys info file failed\n");
            ++this->num_errors;
        }
    }

    IBDIAGNET_RETURN(rc);
}

int PKeyStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err pkey_errors;
    int rc2 = this->p_ibdiag->CheckPartitionKeys(pkey_errors);
    rc = this->AnalyzeCheckResults(pkey_errors,
        STAGE_NAME_PKEY,
        rc2,
        IBDIAG_ERR_CODE_CHECK_FAILED,
        this->num_errors,
        this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
AGUIDStage::AGUIDStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_AGUID, p_ibdiag),
        can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
    IBDIAGNET_RETURN_VOID;
}

AGUIDStage::~AGUIDStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int AGUIDStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag, this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int AGUIDStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_AGUID_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err aguid_errors;
    int rc2 = this->p_ibdiag->BuildAliasGuidsDB(aguid_errors);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(aguid_errors,
            CHECK_NAME_AGUID_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (!rc) {
        if (this->p_ibdiag->DumpAliasGUIDCSVTable(*this->p_csv_out)) {
            ERR_PRINT("Writing alias gudis info to DB failed\n");
            ++this->num_errors;
        }

        if (this->p_ibdiag->WriteAliasGUIDFile(AGUID_FILE)) {
            ERR_PRINT("Writing alias gudis info file failed\n");
            ++this->num_errors;
        }
    }
    IBDIAGNET_RETURN(rc);
}

int AGUIDStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_AGUID);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err aguid_errors;
    int rc2 = this->p_ibdiag->CheckDuplicatedAliasGuids(aguid_errors);
    rc = this->AnalyzeCheckResults(aguid_errors,
            STAGE_NAME_AGUID,
            rc2,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);
    IBDIAGNET_RETURN(rc);
}


/******************************************************/
MlnxCntrsStage::MlnxCntrsStage(IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_MLNX_CNTRS, p_ibdiag),
    can_send_mads_by_lid(false), to_reset_counters(false)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default
    IBDIAGNET_RETURN_VOID;
}

MlnxCntrsStage::~MlnxCntrsStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int MlnxCntrsStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag,
        this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int MlnxCntrsStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_MLNX_CNTRS_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err mlnx_cntrs_errors;
    //if need to reset
    if (this->to_reset_counters) {
        int rc2  = this->p_ibdiag->ResetDiagnosticCounters(mlnx_cntrs_errors);
        SCREEN_PRINT("\n");
        rc = this->AnalyzeCheckResults(mlnx_cntrs_errors,
                CHECK_NAME_MLNX_CNTRS_RESET,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
    }
    if (rc)
        IBDIAGNET_RETURN(rc);

    int rc2 = this->p_ibdiag->BuildVsDiagnosticCounters(
            mlnx_cntrs_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(mlnx_cntrs_errors,
            CHECK_NAME_MLNX_CNTRS_RETRIEVE,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(rc);
}

int MlnxCntrsStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    // now when we have the calculated counters, we can print to the files
    if (this->p_ibdiag->DumpDiagnosticCountersCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Writing ports counters to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WriteMlnxCntrsFile(MLNX_CNTRS_FILE)) {
        ERR_PRINT("Writing Mellanox diagnostic counters file failed\n");
        ++this->num_errors;
    }

    IBDIAGNET_RETURN(rc);
}

//**********************************************
VirtualizationStage::VirtualizationStage(IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_VIRTUALIZATION, p_ibdiag),
    can_send_mads_by_lid(false)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

VirtualizationStage::~VirtualizationStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int VirtualizationStage::CleanResources()
{
    IBDIAGNET_ENTER;

    for (map_guid_pport::iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->PortByGuid.begin();
         nI != this->p_ibdiag->GetDiscoverFabricPtr()->PortByGuid.end(); ++nI) {
         IBPort *p_curr_port = (*nI).second;
         if (!p_curr_port) {
            this->p_ibdiag->SetLastError("DB error - found null port in PortByGuid map"
                                         "for GUID = " U64H_FMT, (*nI).first);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
         }
         p_curr_port->CleanVPorts();
    }
    this->p_ibdiag->GetIBDMExtendedInfoPtr()->CleanVPortDB();
    this->p_ibdiag->GetDiscoverFabricPtr()->ResetVPortIndex();
    this->p_ibdiag->GetDiscoverFabricPtr()->CleanVNodes();
    this->p_ibdiag->GetIBDMExtendedInfoPtr()->CleanVNodeDB();
    this->p_ibdiag->GetDiscoverFabricPtr()->ResetVNodeIndex();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int VirtualizationStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag,
        this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);
    IBDIAGNET_RETURN(0);
}

int VirtualizationStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    if (!this->can_send_mads_by_lid) {
        INFO_PRINT("%s skipped\n", CHECK_NAME_VIRTUALIZATION_RETRIEVE);
        PRINT("\n");
        IBDIAGNET_RETURN(rc);
    }

    list_p_fabric_general_err vport_errors;
    int rc2 = p_ibdiag->BuildVirtualizationDB(vport_errors);
    PRINT("\n");
    rc = this->AnalyzeCheckResults(vport_errors,
                                   STAGE_NAME_VIRTUALIZATION,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings,
                                   true);


    IBDIAGNET_RETURN(rc);
}

int VirtualizationStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err vport_errors;
    this->p_ibdiag->CheckAndSetVPortLid(vport_errors);
    rc = this->AnalyzeCheckResults(vport_errors,
            CHECK_NAME_VIRTUAL_DISCOVERY_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    this->p_ibdiag->CheckVPortDuplicatedGuids(vport_errors);
    rc = this->AnalyzeCheckResults(vport_errors,
            CHECK_NAME_VIRTUAL_DISCOVERY_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_CHECK_FAILED,
            this->num_errors,
            this->num_warnings);

    if (this->p_ibdiag->DumpCSVVNodesTable(*this->p_csv_out)) {
        ERR_PRINT("Writing VNodes to DB failed\n");
        ++this->num_errors;
    }
    if (this->p_ibdiag->DumpCSVVPortsTable(*this->p_csv_out)) {
        ERR_PRINT("Writing VPorts to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpCSVVPortsGUIDInfoTable(*this->p_csv_out)) {
        ERR_PRINT("Writing VPorts GUID Info to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WriteVPortsFile(VPORTS_FILE)) {
        ERR_PRINT("Writing vports info file failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->WriteVPortsPKeyTblFile(VPORTS_PKEY_FILE)) {
        ERR_PRINT("Writing vports pkey table file failed\n");
        ++this->num_errors;
    }

    IBDIAGNET_RETURN(rc);
}

TemperatureSensing::TemperatureSensing (IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_TEMPSENSING, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

TemperatureSensing::~TemperatureSensing()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int TemperatureSensing::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int TemperatureSensing::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err temp_sensing_errors;

    int rc2 = this->p_ibdiag->BuildSMPTempSensing(
            temp_sensing_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(temp_sensing_errors,
            STAGE_NAME_TEMPSENSING,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(rc);
}

int TemperatureSensing::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    if (this->p_ibdiag->DumpTempSensingCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Writing ports counters to DB failed\n");
        ++this->num_errors;
    }

    IBDIAGNET_RETURN(rc);
}

/*************************************************************************************************/
QoSStage::QoSStage (IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_QOS, p_ibdiag)
{
    IBDIAGNET_ENTER;
    //mark this stage as skipped by default
    MarkStageAsSkipped();
    IBDIAGNET_RETURN_VOID;
}

QoSStage::~QoSStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int QoSStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int QoSStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    list_p_fabric_general_err qos_config_sl_errors;
    list_p_fabric_general_err vports_qos_config_sl_errors;
    int rc = this->p_ibdiag->BuildSMPQoSConfigSL(qos_config_sl_errors,
                                                 false);

    SCREEN_PRINT("\n");
    int rc2 = this->AnalyzeCheckResults(qos_config_sl_errors,
            CHECK_NAME_QOS_CONFIG_SL_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    rc = this->p_ibdiag->BuildSMPQoSConfigSL(vports_qos_config_sl_errors,
                                             true);

    SCREEN_PRINT("\n");
    int rc3 = this->AnalyzeCheckResults(vports_qos_config_sl_errors,
            CHECK_NAME_VPORTS_QOS_CONFIG_SL_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);

    if (rc2)
        IBDIAGNET_RETURN(rc2);
    if (rc3)
        IBDIAGNET_RETURN(rc3);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int QoSStage::RunCheck()
{
    IBDIAGNET_ENTER;

    int rc1 = p_ibdiag->DumpQoSConfigSLCSVTable(*this->p_csv_out);
    if (rc1) {
        ERR_PRINT("Writing QoS config SL to csv file failed\n");
        ++this->num_errors;
    }

    int rc2 = p_ibdiag->DumpVPortsQoSConfigSLCSVTable(*this->p_csv_out);
    if (rc2) {
        ERR_PRINT("Writing VPorts QoS config SL to csv file failed\n");
        ++this->num_errors;
    }

    if (rc1)
        IBDIAGNET_RETURN(rc1);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


RoutersStage::RoutersStage(IBDiag *p_ibdiag) :
        Stage(STAGE_NAME_ROUTERS, p_ibdiag), p_FLIDsManager(NULL)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

RoutersStage::~RoutersStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

const char* RoutersStage::GetLastError()
{
    IBDIAGNET_ENTER;

    if (this->p_FLIDsManager)
        IBDIAGNET_RETURN(this->p_FLIDsManager->GetLastError());

    IBDIAGNET_RETURN(Stage::GetLastError());
}

int RoutersStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(0);
}

int RoutersStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;
    list_p_fabric_general_err errors;

    int rc2 = this->p_ibdiag->BuildRouterInfoDB(errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(errors,
                                   CHECK_ROUTERS_INFO_DB,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings);
    errors.clear();
    rc2 = this->p_ibdiag->BuildRouterTable(errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(errors,
                                   CHECK_ROUTERS_TABLE,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings);

    SCREEN_PRINT("\n");
    rc2 = this->p_ibdiag->RetrieveAdjSubnetsFLIDData(errors);
    rc = this->AnalyzeCheckResults(errors,
                                   CHECK_RETRIVE_ADJ_SUBNETS_FLID,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings);

    SCREEN_PRINT("\n");
    rc2 = this->p_ibdiag->RetrieveRouterFLIDTableData(errors);
    rc = this->AnalyzeCheckResults(errors,
                                   CHECK_RETRIVE_ROUTRS_FLID,
                                   rc2,
                                   IBDIAG_ERR_CODE_FABRIC_ERROR,
                                   this->num_errors,
                                   this->num_warnings);
    IBDIAGNET_RETURN(rc);
}

int RoutersStage::RunCheck()
{
    IBDIAGNET_ENTER;
    if (this->p_ibdiag->DumpRoutersInfoCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Writing routers info to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpRoutersAdjSiteLocalSubnetCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Writing routers adjacent site local subnets table to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpRoutersNextHopCSVTable(*this->p_csv_out)) {
        ERR_PRINT("Writing routers next hop table to DB failed\n");
        ++this->num_errors;
    }

    if (this->p_ibdiag->DumpRoutersAdjSubnetsFLIDData(*this->p_csv_out)) {
        ERR_PRINT("Writing routers next hop table to DB failed\n");
        ++this->num_errors;
    }

    this->CheckFLIDs();

    IBDIAGNET_RETURN(0);
}

void RoutersStage::CheckFLIDs()
{
    IBDIAGNET_ENTER;

    p_FLIDsManager = new FLIDsManager(this->p_ibdiag);

    list_p_fabric_general_err errorsList;
    int rc = p_FLIDsManager->CheckLocalSubnet(errorsList);
    this->AnalyzeCheckResults(errorsList,
            CHECK_LOCAL_SUBNET_FLID,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    if (!p_FLIDsManager->IsApplicable()) {
        INFO_PRINT("Skipping FLID verification\n");

        delete this->p_FLIDsManager;
        //last error will be taken from ibdiag
        this->p_FLIDsManager = NULL;
        IBDIAGNET_RETURN_VOID;
    }

    //if network has any router we should have information about FLIDs
    rc = p_FLIDsManager->CheckAdjacentSubnets(errorsList);
    this->AnalyzeCheckResults(errorsList,
            CHECK_ADJ_SUBNETS_FLID,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);


    rc = p_FLIDsManager->CheckHCAsAndSwitches(errorsList);
    this->AnalyzeCheckResults(errorsList,
            CHECK_HCA_SWITCHES_FLID,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    if (p_FLIDsManager->CollectEnabledFLIDs()) {
        ERR_PRINT("Failed to collect enabled FLID with: %s\n",
                                                p_FLIDsManager->GetLastError());
        this->num_errors++;
    }

    if (p_FLIDsManager->Dump(FLID_DUMP_FILE)) {
        ERR_PRINT("Failed to dump FLID data with: %s\n", p_FLIDsManager->GetLastError());
        this->num_errors++;
    }

    delete this->p_FLIDsManager;
    //last error will be taken from ibdiag
    this->p_FLIDsManager = NULL;

    IBDIAGNET_RETURN_VOID;
}

SharpStage::SharpStage (IBDiag *p_ibdiag):
    Stage(STAGE_NAME_SHARP, p_ibdiag), can_send_mads_by_lid(false),
    dump_counters(false), dump_hba_port_counters(false), clear_counters(false), version(0),
    p_sharp_mngr(NULL), sharp_versions_counter(SHARP_CONTROL_VERSION_MAX, 0), num_of_llts(0),
    num_of_sats(0), num_of_active_jobs(0)
{
    IBDIAGNET_ENTER;
    this->MarkStageAsSkipped(); //this stage will be skipped by default

    IBDIAGNET_RETURN_VOID;
}

SharpStage::~SharpStage()
{
    IBDIAGNET_ENTER;
    delete p_sharp_mngr;
    IBDIAGNET_RETURN_VOID;
}

void SharpStage::SetAMKeyPerPort()
{
    IBDIAGNET_ENTER;

    for (map_guid_2_key_t::iterator it = this->guid_2_am_key.begin();
         it != this->guid_2_am_key.end(); it++) {
         IBPort *p_port = this->p_ibdiag->GetDiscoverFabricPtr()->getPortByGuid(it->first);
         if (p_port) {
             if (p_port->p_node && p_port->p_node->isSpecialNode())
                 p_port->am_key.SetKey(it->second);
             else
                 WARN_PRINT("GUID 2 AM key - Port GUID: " U64H_FMT " doesn't belong to "
                           "Aggregation node\n",  it->first);
         } else {
             WARN_PRINT("GUID 2 AM key - Port GUID: " U64H_FMT " doesn't exist in fabric\n",
                       it->first);
         }
    }

    IBDIAGNET_RETURN_VOID;
}

void SharpStage::CountSharpVersionsAndNumOfActiveJobs()
{
    IBDIAGNET_ENTER;

    const list_sharp_an &sharp_ans = this->p_sharp_mngr->GetSharpANList();
    for (list_sharp_an::const_iterator an_iter = sharp_ans.begin(); an_iter != sharp_ans.end();
         ++an_iter) {
         SharpAggNode *p_sharp_aggnode = *an_iter;
         if (!p_sharp_aggnode)
             continue;

         const AM_ANActiveJobs &an_active_jobs = p_sharp_aggnode->GetANActiveJobs();
         for (size_t dw = 0; dw < AN_ACTIVE_JOBS_DWORDS_NUM; ++dw) {
             u_int32_t dword = an_active_jobs.active_jobs[dw];

             for (size_t bit = 0; bit < 8 * sizeof(u_int32_t); ++bit) {
                 if (dword & 1)
                     this->num_of_active_jobs++;
                 dword >>= 1;
             }
         }

         AM_ANInfo cur_an = p_sharp_aggnode->GetANInfo();
         u_int16_t bit = 0x1;
         if (!(cur_an.active_sharp_version_bit_mask & bit))
             this->sharp_versions_counter[0]++;
         for (u_int16_t version_idx = 1; version_idx < SHARP_CONTROL_VERSION_MAX; version_idx++) {
              if (cur_an.active_sharp_version_bit_mask & (bit << version_idx))
                  this->sharp_versions_counter[version_idx]++;
         }
    }

    IBDIAGNET_RETURN_VOID
}

void SharpStage::CountSharpTreeTypes()
{
    IBDIAGNET_ENTER;

    const vector_sharp_root_node &sharp_root_nodes = this->p_sharp_mngr->GetSharpRootNodeVec();
    for (vector_sharp_root_node::size_type tree_idx = 0; tree_idx < sharp_root_nodes.size(); ++tree_idx) {
         SharpTree *root_node = sharp_root_nodes[tree_idx];
         if (!root_node || !root_node->GetRoot())
             continue;

         if (root_node->GetRoot()->GetTreeConfig().tree_mode == TREE_MODE_SAT)
             this->num_of_sats++;
         else
             this->num_of_llts++;
    }

    IBDIAGNET_RETURN_VOID
}

bool SharpStage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    for (u_int16_t version_idx = 0; version_idx < SHARP_CONTROL_VERSION_MAX; version_idx++) {
         summary_str << "Number of aggregation nodes version " << version_idx + 1 << ": "
                     << this->sharp_versions_counter[version_idx] << endl;
    }

    summary_str << "Number of LL aggregation trees: " << num_of_llts << endl;
    summary_str << "Number of streaming aggregation trees: " << num_of_sats << endl;
    summary_str << "Number of active jobs: " << num_of_active_jobs << endl;

    IBDIAGNET_RETURN(true);
}

int SharpStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    if (check_if_can_send_mads_by_lid(this->p_ibdiag,
        this->can_send_mads_by_lid))
            IBDIAGNET_RETURN(1);

    p_sharp_mngr = new SharpMngr(p_ibdiag, version);
    IBDIAGNET_RETURN(0);
}

int SharpStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    if (!p_ibdiag->GetDiscoverFabricPtr() ||
        !p_ibdiag->GetIbisPtr())
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);

    if (!this->guid_2_am_key.empty())
        this->SetAMKeyPerPort();

    list_p_fabric_general_err sharp_discovery_errors;

    int rc2 = p_sharp_mngr->BuildSharpConfigurationDB(sharp_discovery_errors);

    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(sharp_discovery_errors,
            STAGE_NAME_SHARP,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (dump_counters) {
        INFO_PRINT("Build AM PerfCounters DB\n");
        rc2 = p_sharp_mngr->BuildPerformanceCountersDB(sharp_discovery_errors, false);
        SCREEN_PRINT("\n");
        if (rc) {
            ERR_PRINT("Failed to build AM PerfCounters DB.\n");
            IBDIAG_RETURN(rc);
        }

        rc = this->AnalyzeCheckResults(sharp_discovery_errors,
                CHECK_NAME_SHARP_PM_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    if (dump_hba_port_counters) {
        INFO_PRINT("Build AM HBA Ports PerfCounters DB\n");
        rc2 = p_sharp_mngr->BuildPerformanceCountersDB(sharp_discovery_errors, true);
        SCREEN_PRINT("\n");
        if (rc) {
            ERR_PRINT("Failed to build AM HBA Ports PerfCounters DB.\n");
            IBDIAG_RETURN(rc);
        }

        rc = this->AnalyzeCheckResults(sharp_discovery_errors,
                CHECK_NAME_HBA_PORT_COUNTERS_RETRIEVE,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    if (clear_counters) {
        INFO_PRINT("Reset AM PerfCounters DB\n");
        rc2 = p_sharp_mngr->ResetPerformanceCounters(sharp_discovery_errors);
        SCREEN_PRINT("\n");
        if (rc) {
            ERR_PRINT("Failed to reset AM PerfCounters DB.\n");
            IBDIAG_RETURN(rc);
        }

        rc = this->AnalyzeCheckResults(sharp_discovery_errors,
                CHECK_NAME_SHARP_PM_RESET,
                rc2,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc)
            IBDIAGNET_RETURN(rc);
    }

    IBDIAGNET_RETURN(rc);
}

int SharpStage::RunCheck()
{
    IBDIAGNET_ENTER;
    int rc = 0;

    list_p_fabric_general_err sharp_discovery_errors;

    int rc2 = p_sharp_mngr->VerifyTrapsLids(sharp_discovery_errors);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(sharp_discovery_errors,
            CHECK_NAME_SHARP_VERIFY_TRAPS_LIDS,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc2 = p_sharp_mngr->VerifyVersions(sharp_discovery_errors);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(sharp_discovery_errors,
            CHECK_NAME_SHARP_VERIFY_VERSIONS,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc2 = p_sharp_mngr->ConnectTreeEdges(sharp_discovery_errors);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(sharp_discovery_errors,
            CHECK_NAME_SHARP_CONNECT_TREE_EDGES,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings);
    if (rc)
        IBDIAGNET_RETURN(rc);

    rc2 = p_sharp_mngr->CheckSharpTrees(sharp_discovery_errors);
    SCREEN_PRINT("\n");
    rc = this->AnalyzeCheckResults(sharp_discovery_errors,
            CHECK_NAME_SHARP_CHECK_TREES,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc)
        IBDIAGNET_RETURN(rc);

    if (p_sharp_mngr->WriteSharpFile(SHARP_FILE)) {
        ERR_PRINT("Write sharp data file failed\n");
        ++this->num_errors;
    }

    if (p_sharp_mngr->DumpSharpANInfoToCSV(*p_csv_out)) {
        ERR_PRINT("Dump sharp ANs Info DB to CSV file failed \n");
        ++this->num_errors;
    }

    if (p_sharp_mngr->WriteSharpANInfoFile(SHARP_AN_FILE)) {
        ERR_PRINT("Write sharp ANs file failed\n");
        ++this->num_errors;
    }

    if (dump_counters)
        p_sharp_mngr->DumpSharpPMCountersToCSV(*p_csv_out);

    if (dump_hba_port_counters)
        p_sharp_mngr->DumpSharpPMHBAPortCountersToCSV(*p_csv_out);

    this->CountSharpVersionsAndNumOfActiveJobs();
    this->CountSharpTreeTypes();

    IBDIAGNET_RETURN(rc);
}



/**********************************
 ***  Congestion Control Stage  ***
 **********************************/
CongestionControlStage::CongestionControlStage(IBDiag *p_ibdiag) :
    Stage(STAGE_NAME_CC, p_ibdiag), to_get_congestion_counters(false),
    to_clear_congestion_counters(false)
{
    IBDIAGNET_ENTER;
    //mark this stage as skipped by default
    MarkStageAsSkipped();
    IBDIAGNET_RETURN_VOID;
}

CongestionControlStage::~CongestionControlStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int CongestionControlStage::Prepare()
{
    IBDIAGNET_ENTER;

    HDR_PRINT("%s\n", this->stage_name.c_str());

    if (!this->ppcc_algo_path.empty()) {
        if (this->ppcc_algo_database.ParseDir(this->ppcc_algo_path))
            IBDIAGNET_RETURN(1);
    }

    IBDIAGNET_RETURN(0);
}

int CongestionControlStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;

    list_p_fabric_general_err congestion_control_errors;

    int rc = this->p_ibdiag->BuildEnhancedCC(congestion_control_errors);
    SCREEN_PRINT("\n");
    int rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_ENHANCED_CC_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->BuildCCSwithGeneralSettings(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_ENHANCED_CC_SWITCH_GENERAL_SETTINGS_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->BuildCCSwithConfig(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_ENHANCED_CC_SWITCH_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->BuildCCHCAGeneralSettings(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_ENHANCED_CC_HCA_GENERAL_SETTINGS_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->BuildCCHCAConfig(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_ENHANCED_CC_HCA_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    if (this->to_get_congestion_counters) {
        rc = this->p_ibdiag->BuildCCHCAStatisticsQuery(congestion_control_errors,
                                                       this->to_clear_congestion_counters);
        SCREEN_PRINT("\n");
        rc2 = this->AnalyzeCheckResults(congestion_control_errors,
                CHECK_NAME_ENHANCED_CC_HCA_COUNTERS_RETRIEVE,
                rc,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);

        if (rc2)
            IBDIAGNET_RETURN(rc2);
    }

    rc = this->p_ibdiag->Build_CC_HCA_AlgoConfigSup(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_CC_HCA_ALGO_CONFIG_SUP_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->Build_CC_HCA_AlgoConfig(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_CC_HCA_ALGO_CONFIG_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    rc = this->p_ibdiag->Build_CC_HCA_AlgoConfigParams(congestion_control_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(congestion_control_errors,
            CHECK_NAME_CC_HCA_ALGO_PARAMS_RETRIEVE,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    if (this->to_get_congestion_counters) {
        rc = this->p_ibdiag->Build_CC_HCA_AlgoCounters(congestion_control_errors,
                                                       this->to_clear_congestion_counters);
        SCREEN_PRINT("\n");

        rc2 = this->AnalyzeCheckResults(congestion_control_errors,
                CHECK_NAME_CC_HCA_ALGO_COUNTERS_RETRIEVE,
                rc,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc2)
            IBDIAGNET_RETURN(rc2);
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CongestionControlStage::RunCheck()
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    int rc2 = IBDIAG_SUCCESS_CODE;
    list_p_fabric_general_err cc_errors;

    this->p_ibdiag->DumpCCEnhancedInfoCSVTable(*this->p_csv_out);
    this->p_ibdiag->DumpCCSwitchGeneralSettingsCSVTable(*this->p_csv_out);
    this->p_ibdiag->DumpCCPortProfileSettingsCSVTable(*this->p_csv_out);
    this->p_ibdiag->DumpCCSLMappingSettingsCSVTable(*this->p_csv_out);
    this->p_ibdiag->DumpCCHCAGeneralSettingsCSVTable(*this->p_csv_out);
    this->p_ibdiag->DumpCCHCARPParametersCSVTable(*this->p_csv_out);
    this->p_ibdiag->DumpCCHCANPParametersCSVTable(*this->p_csv_out);

    // CC Algo
    rc = this->p_ibdiag->DumpCC_HCA_AlgoConfigSupCSVTable(*this->p_csv_out, cc_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(cc_errors,
            CHECK_NAME_CC_HCA_ALGO_CONFIG_SUP_VALUES,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    this->p_ibdiag->DumpCC_HCA_AlgoConfigCSVTable(*this->p_csv_out);

    rc = this->p_ibdiag->DumpCC_HCA_AlgoConfigParamsCSVTable(*this->p_csv_out, cc_errors);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(cc_errors,
            CHECK_NAME_CC_HCA_ALGO_PARAMS_VALUES,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    if (this->to_get_congestion_counters) {

        this->p_ibdiag->DumpCCHCAStatisticsQuery(*this->p_csv_out);

        rc = this->p_ibdiag->DumpCC_HCA_AlgoCountersCSVTable(*this->p_csv_out, cc_errors);
        SCREEN_PRINT("\n");
        rc2 = this->AnalyzeCheckResults(cc_errors,
                CHECK_NAME_CC_HCA_ALGO_COUNTERS_VALUES,
                rc,
                IBDIAG_ERR_CODE_FABRIC_ERROR,
                this->num_errors,
                this->num_warnings,
                true);
        if (rc2)
            IBDIAGNET_RETURN(rc2);
    }

    // PPCC file
    if (!this->ppcc_algo_path.empty()) {
        if (this->p_ibdiag->WriteCC_HCA_AlgoDataToFile(PPCC_FILE, this->ppcc_algo_database)) {
            ERR_PRINT("Writing CC HCA Algo Data fo file failed\n");
            ++this->num_errors;
        }
    }

    // CC Algo Validations
    rc = this->p_ibdiag->CC_HCA_AlgoValidations(cc_errors, this->ppcc_algo_database);
    SCREEN_PRINT("\n");
    rc2 = this->AnalyzeCheckResults(cc_errors,
            CHECK_NAME_CC_HCA_ALGO_VALIDATIONS,
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            true);
    if (rc2)
        IBDIAGNET_RETURN(rc2);

    IBDIAGNET_RETURN(rc);
}


CheckRailOptimizedTopologyStage::CheckRailOptimizedTopologyStage(IBDiag *p_ibdiag):Stage(STAGE_NAME_RAIL_OPT_TOPO_CHECK, p_ibdiag)
{
    IBDIAGNET_ENTER;
    this->MarkStage(false);
    IBDIAGNET_RETURN_VOID;
}

CheckRailOptimizedTopologyStage::~CheckRailOptimizedTopologyStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int CheckRailOptimizedTopologyStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckRailOptimizedTopologyStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckRailOptimizedTopologyStage::RunCheck()
{
    IBDIAGNET_ENTER;

    if (this->CheckRailOptimizedTopology())
        ERR_PRINT("%s has failed\n", CHECK_RAIL_OPTIMIZED_TOPOLOGY_RETRIVE);

    SCREEN_PRINT("\n");

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckRailOptimizedTopologyStage::CheckRailOptimizedTopology()
{
    IBDIAGNET_ENTER;

    if (!this->p_ibdiag->IsRailDataCollected()) {
       ERR_PRINT("Error: The PCIe data needed for performing the "
                 "test has not been properly collected. "
                 "Verify that the Phy plugin has been loaded\n");

       this->num_errors++;
       IBDIAGNET_RETURN(1);
    }

    vec_pport excludedCards;
    vector<PCI_LeafSwitchInfo> pciLeafSwitchInfoVec;
    int rc = this->p_ibdiag->CheckRailOptimizedTopology(pciLeafSwitchInfoVec, excludedCards);
    if (rc) {
       ERR_PRINT("Error: %s \n", p_ibdiag->GetLastError());
       IBDIAGNET_RETURN(rc);
    }

    rc= this->PrintReport(pciLeafSwitchInfoVec, excludedCards);
    if (rc)
       IBDIAGNET_RETURN(rc);

    IBDIAGNET_RETURN(0);
}

int CheckRailOptimizedTopologyStage::PrintReport(const vector<PCI_LeafSwitchInfo> &pciLeafSwitchInfoVec,
                                                vec_pport &excludedCards)
{
    IBDIAGNET_ENTER;

    int size = static_cast<int>(pciLeafSwitchInfoVec.size());
    if (!size && !excludedCards.size()) {
        INFO_PRINT("%s finished successfully\n", CHECK_RAIL_OPTIMIZED_TOPOLOGY_RETRIVE);
        IBDIAGNET_RETURN(0);
    }

    //we have dump file in case if any mismatch is found
    ofstream  sout;
    if (this->p_ibdiag->OpenFile("Rails Dump", string(RAILS_DUMP_FILE), sout, false, true))
        WARN_PRINT("The file: %s with rails validation details"
                   " will not be provided.\n", RAILS_DUMP_FILE);

    IBFabric::GetSwitchLabelPortNumExplanation(sout);

    //stream does nor crash on write attempt if it has no associated file
    int rc = this->ReportOnExcludedCards(excludedCards, sout);

    for (vector<PCI_LeafSwitchInfo>::const_iterator it = pciLeafSwitchInfoVec.begin();
         !rc && it != pciLeafSwitchInfoVec.end(); it++) {

        rc = this->ReportOnLeafSwicth(*it, sout);
    }

    this->p_ibdiag->CloseFile(sout);


    this->num_warnings += size;
    WARN_PRINT("%s ended with %d warnings\n", CHECK_RAIL_OPTIMIZED_TOPOLOGY_RETRIVE,
            this->num_warnings);
    INFO_PRINT("Rail Optimized Topology validation is usually applicable to the compute nodes.\n");
    INFO_PRINT("If detected mis-cabled nodes are not compute ones, please apply Rail Optimized " \
               "Topology check for specific set of nodes by invoking: " \
               "--rail_validation_opt regex='reg expression'\n");

    IBDIAGNET_RETURN(rc);
}

int CheckRailOptimizedTopologyStage::ReportOnExcludedCards(vec_pport &excludedCards, ostream &ofs)
{
    IBDIAGNET_ENTER;

    if (excludedCards.empty())
        IBDIAGNET_RETURN(0);

    this->num_warnings++;
    WARN_PRINT("%lu socket-direct cards excluded from rail-optimized validation\n",
            excludedCards.size());

    ofs << excludedCards.size()
        << " socket-direct cards excluded from rail-optimized validation"
        << endl;

    IBDIAGNET_RETURN(0);
}

#define IDX_TO_STR_RANGE    ('Z' - 'A' + 1)
static char* idx_to_str(int index, char * buffer)
{
    if (index >= IDX_TO_STR_RANGE)
        buffer = idx_to_str(index / IDX_TO_STR_RANGE - 1, buffer);

    *buffer++ = static_cast<char>('A' + (index % IDX_TO_STR_RANGE));
    *(buffer) = 0;
    return buffer;
}

#define PCIe_fmt "PCIe 0000:%02x:%02x.%x"
int CheckRailOptimizedTopologyStage::ReportOnLeafSwicth(const PCI_LeafSwitchInfo &leafSwitchInfo, ostream &ofs)
{
    IBDIAGNET_ENTER;

    const IBNode *p_switch = leafSwitchInfo.p_switch;
    if (!p_switch) {
        ERR_PRINT("\tError: Null pointer to swicth node\n");
        IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
    }

    this->PrintSwitchData(p_switch, ofs);

    if (leafSwitchInfo.pciAddressMap.empty()) {
        WARN_PRINT("\t\t has no PCI address on all connected HCAs\n");

        ofs << "\t\t has no PCI address on all connected HCAs" << endl;
    } else {
        int index = 0;
        for (PCI_AddressToNodesMap::const_iterator it = leafSwitchInfo.pciAddressMap.begin();
            it != leafSwitchInfo.pciAddressMap.end(); it++) {

            const PCI_Address& address = it->first;
            const vector_p_node &nodes = it->second;
            if (nodes.empty()) {
                ERR_PRINT("\tError: Empty nodes list was provided "
                          "for the " PCIe_fmt "\n",
                          address.bus, address.device, address.local_port);
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
            }

            int rc = this->PrintRailData(address, nodes, index, ofs);
            if (rc)
               IBDIAGNET_RETURN(rc);

            index++;
        }
    }

    IBDIAGNET_RETURN(0);
}

int CheckRailOptimizedTopologyStage::PrintRailData(const PCI_Address& address,
                                                   const vector_p_node &nodes,
                                                   int index,
                                                   ostream &ofs)
{
    IBDIAGNET_ENTER;

    char buffer[32] = {'\0'};
    idx_to_str(index, buffer);
    const size_t size = nodes.size();

    for (size_t i = 0; i < size; i++)
    {
        IBNode *p_node = const_cast<IBNode*>(nodes[i]);
        if (!p_node) {
            ERR_PRINT("\tError: Null pointer found in HCAs "
                      "for the " PCIe_fmt "\n",
                      address.bus, address.device, address.local_port);
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }

        IBPort *p_switch_port = NULL;
        IBPort *p_node_port = NULL;
        if (this->GetSwitchAndNodePorts(p_node, p_switch_port, p_node_port)) {
            ERR_PRINT("\tError: "
                      " HCA GUID=" U64H_FMT " is not connected to any switch\n",
                      p_node->guid_get());
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_FABRIC_ERROR);
        }

        if (i == 0) {
            WARN_PRINT("\t\t rail %s (" PCIe_fmt "): "
                    "%lu ports ==> "
                    "%s <--> %s%s\n",
                    buffer,
                    address.bus, address.device, address.local_port,
                    size,
                    p_node_port->getExtendedName().c_str(),
                    p_switch_port->getExtendedName().c_str(), (size == 1) ? "" : ", ...");

            ofs << "    rail " << buffer
                << "(0000:" << HEX(address.bus, 2)
                << '.' << HEX(address.device, 2)
                <<  '.' << HEX(address.local_port) << "): "
                << size << " ports:" << endl
                << "                            " << p_node_port->getExtendedName()
                << " <--> " << p_switch_port->getExtendedName().c_str() << endl;
        } else
            ofs << "                            "
                << p_node_port->getExtendedName()
                << " <--> " << p_switch_port->getExtendedName().c_str() << endl;
    }//end for

    IBDIAGNET_RETURN(0);
}

void CheckRailOptimizedTopologyStage::PrintSwitchData(const IBNode *p_swicth,
                                                      ostream &ofs)
{
    IBDIAGNET_ENTER;

    WARN_PRINT("\tNode rail connectivity mismatch on the switch: \"%s\" GUID=" U64H_FMT "\n",
            p_swicth->description.c_str(), p_swicth->guid_get());

    ofs << endl << "Node rail connectivity mismatch on the switch: "
        << p_swicth->description
        << " GUID=" << PTR(p_swicth->guid_get()) << endl;

    IBDIAGNET_RETURN_VOID;
}

int CheckRailOptimizedTopologyStage::GetSwitchAndNodePorts(IBNode *p_node,
                                IBPort *&p_switch_port, IBPort *&p_node_port)
{
    IBDIAGNET_ENTER;

    int rc = 0;
    p_switch_port = NULL;
    p_node_port = NULL;

    for (phys_port_t i = 0; i <= p_node->numPorts; ++i) {
        IBPort *p_curr_port = p_node->getPort(i);
        //Skip ports that are down or have undefined state
        //or are not in the scope of the fabric.
        if (!p_curr_port || !p_curr_port->is_data_worthy())
            continue;

        IBPort *p_remote_port = p_curr_port->p_remotePort;
        if (p_remote_port) {
           p_switch_port = p_remote_port;
           p_node_port = p_curr_port;
           break;
        }
    }

    if (!p_switch_port || !p_node_port)
        rc = 1;

    IBDIAGNET_RETURN(rc);
}

PostReportsStage::PostReportsStage(IBDiag *p_ibdiag):Stage(STAGE_NAME_POST_REPORTS, p_ibdiag)
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

PostReportsStage::~PostReportsStage()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN_VOID;
}

int PostReportsStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int PostReportsStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int PostReportsStage::RunCheck()
{
    IBDIAGNET_ENTER;

    warnings_list warnings;
    if (p_ibdiag->CreateIBNetDiscoverFile(IBNET_DISCOVER_FILE, warnings)) {
        this->num_errors++;

        ERR_PRINT("Error: %s\n", p_ibdiag->GetLastError());
        ERR_PRINT("Writing of IBNetdDscover file failed\n");
    } else if (warnings.empty())
        INFO_PRINT("Writing of IBNetdDscover file finished successfully\n");
    else {
        this->num_warnings += static_cast<u_int32_t>(warnings.size());

        //we do not print any warning on screen and the log file.
        WARN_PRINT("Writing of IBNetdDscover file ended with warnings\n");
    }

    SCREEN_PRINT("\n");

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

void PostReportsStage::PrintWarnings(const warnings_list& warnings) const
{
    IBDIAGNET_ENTER;

    for (warnings_list::const_iterator itr = warnings.begin();
         itr != warnings.end(); itr++) {
        const string& ref = *itr;
        WARN_PRINT("%s\n", ref.c_str());
    }

    IBDIAGNET_RETURN_VOID;
}

CheckDFPTopologyStage::CheckDFPTopologyStage(IBDiag *p_ibdiag):Stage(STAGE_NAME_DFP_TOPO_CHECK, p_ibdiag),
                                        maxCAsOnSpine(-1), is_roots_from_file(false)
{
    IBDIAGNET_ENTER;
    this->MarkStage(false);
    IBDIAGNET_RETURN_VOID;

}

int CheckDFPTopologyStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckDFPTopologyStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckDFPTopologyStage::RunCheck()
{
    IBDIAGNET_ENTER;

    if (this->BuildAndValidate())
        ERR_PRINT("DFP validation has failed\n");
    else if (!this->num_warnings && !this->num_errors)
             INFO_PRINT("DFP validation finished successfully\n");
    else if (this->num_errors)
        WARN_PRINT("DFP validation finished with errors\n");
    else if (this->num_warnings && !this->num_errors)
        WARN_PRINT("DFP validation finished with warnings\n");

    SCREEN_PRINT("\n");

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckDFPTopologyStage::DiscoverDFPRootsBySMDB()
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    for (set_pnode::iterator it = this->p_ibdiag->GetDiscoverFabricPtr()->Switches.begin();
          it != this->p_ibdiag->GetDiscoverFabricPtr()->Switches.end();
          it++) {
         IBNode *p_switch = *it;
         if (p_switch->rank == RANK_OF_SPINE)
             this->dfp_roots[p_switch->guid_get()] = p_switch;
    }

    if (this->dfp_roots.empty()) {
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
        ERR_PRINT("No DFP Root switches are founded in fabric\n");
    }

    IBDIAGNET_RETURN(rc);
}

int CheckDFPTopologyStage::DiscoverDFPRootsByMaxCAsOnSpine()
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    for (set_pnode::iterator it = this->p_ibdiag->GetDiscoverFabricPtr()->Switches.begin();
          it != this->p_ibdiag->GetDiscoverFabricPtr()->Switches.end();
          it++) {
         int num_of_cas = 0;
         IBNode *p_switch = *it;
         for (u_int32_t i = 1 ; i <=  p_switch->numPorts ; i++) {
              IBPort *p_curr_port = p_switch->getPort((phys_port_t)i);
              if (!p_curr_port)                       //port not exist
                  continue;
              //Skip ports that are down or have undefined state
              //or are not in the scope of the fabric.
              if (!p_curr_port->is_data_worthy())
                  continue;
              if (!p_curr_port->p_remotePort || !p_curr_port->p_remotePort->p_node)
                  continue;
              // Skip on special ports
              if (p_curr_port->isSpecialPort())
                  continue;

              IBNode *p_remote_node = p_curr_port->p_remotePort->p_node;
              if (p_remote_node->type == IB_CA_NODE)
                  num_of_cas++;
         }

         if (num_of_cas <= this->maxCAsOnSpine)
             this->dfp_roots[p_switch->guid_get()] = p_switch;
    }

    if (this->dfp_roots.empty()) {
        rc = IBDIAG_ERR_CODE_FABRIC_ERROR;
        ERR_PRINT("No DFP Root switches are founded in fabric\n");
    }

    IBDIAGNET_RETURN(rc);
}

int CheckDFPTopologyStage::BuildAndValidate()
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;

    if (this->IsRootFromFile())
        rc = this->DiscoverDFPRootsBySMDB();
    else
        rc = this->DiscoverDFPRootsByMaxCAsOnSpine();

    if (rc) {
        this->num_errors++;
        ERR_PRINT("Failed to obtain DFP Roots\n");
        IBDIAGNET_RETURN(rc);
    }

    DFPTopology topology(*this->p_ibdiag->GetDiscoverFabricPtr());
    rc = topology.Build(this->dfp_roots);
    if (rc) {
        this->num_errors++;
        ERR_PRINT("Failed to build DFP topology\n");
        IBDIAGNET_RETURN(rc);
    }

    rc = topology.Validate(this->num_warnings, this->num_errors);
    if (rc) {
       this->num_errors++;
       IBDIAGNET_RETURN(rc);
    }

    ofstream  sout;

    rc = p_ibdiag->OpenFile("DFPToplogy", string(DFP_DUMP_FILE), sout, false, true);

    if (rc) {
        this->num_errors++;
        ERR_PRINT("Failed to create DFP topolgy dump file with: %s\n", p_ibdiag->GetLastError());
        IBDIAGNET_RETURN(rc);
    }

    if (!sout.is_open())
        IBDIAG_RETURN(IBDIAG_SUCCESS_CODE);

    if (this->IsRootFromFile())
        sout << "List of DFP Root switches is loaded from the SMDB file" << endl;
    else
        sout << "DFP Root switches are discovered by parameter "
             << "\"maximum CAs\"" << ": "
             << this->maxCAsOnSpine
             << endl;

    if (topology.DumpToStream(sout)) {
        this->num_errors++;
        ERR_PRINT("Failed to dump DFP topology\n");
    }

    this->p_ibdiag->CloseFile(sout);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}


CheckFatTreeTopologyStage::CheckFatTreeTopologyStage(IBDiag *p_ibdiag):Stage(STAGE_NAME_FAT_TREE_TOPO_CHECK, p_ibdiag),
                                        retries(7), equalResults(2), p_rootsRegExp(NULL), rootsFromFile(false)
{
    IBDIAGNET_ENTER;
    this->MarkStage(false);
    IBDIAGNET_RETURN_VOID;
}

int CheckFatTreeTopologyStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckFatTreeTopologyStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int CheckFatTreeTopologyStage::RunCheck()
{
    IBDIAGNET_ENTER;
    //todo think of backup algorithm

    this->p_ibdiag->SetLastError("");
    ofstream  outStream;
    if (p_ibdiag->OpenFile("FTToplogy", string(FAT_TREE_DUMP_FILE), outStream, false, true)) {
        this->num_errors++;
        ERR_PRINT("Failed to create Fat-Tree topology dump file with: %s\n",
                p_ibdiag->GetLastError());
    }

    FTTopology ftTopology(*this->p_ibdiag->GetDiscoverFabricPtr(), outStream);

    list_p_fabric_general_err errorsList;
    int rc = IBDIAG_SUCCESS_CODE;

    if (this->rootsFromFile)
        rc =  ftTopology.Build(errorsList, this->lastError);
    else if (this->p_rootsRegExp)
        rc =  ftTopology.Build(errorsList, this->lastError, *this->p_rootsRegExp);
    else
        rc =  ftTopology.Build(errorsList, this->lastError, this->retries, this->equalResults);

    this->AnalyzeCheckResults(errorsList,
            "Fat-Tree topology detection",
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    if (rc) {
        this->p_ibdiag->CloseFile(outStream);
        IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
    }

    INFO_PRINT("%s\n", ftTopology.LevelsReport().c_str());

    rc =  ftTopology.Validate(errorsList, this->lastError);
    SCREEN_PRINT("\n");
    this->AnalyzeCheckResults(errorsList,
            "Fat-Tree topology validation",
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    if (ftTopology.GetInternalIssues())
        ERR_PRINT("For more errors see the dump file: ibdiagnet2.fat_tree\n\n");

    INFO_PRINT("Calculated Fat-Tree bisectional bandwidth: %.2f [Gb/s]\n",
             ftTopology.GetMinimalRatio());

    if (ftTopology.Dump()) {
        this->num_errors++;
        ERR_PRINT("Failed to dump Fat-Tree topology\n");
    }

    this->p_ibdiag->CloseFile(outStream);

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

bool CheckFatTreeTopologyStage::SetRootsRegex(const string &val)
{
    IBDIAG_ENTER;

    //p_rootsRegExp will be deleted at IBDiag's destructor if created
    p_rootsRegExp = new regExp(val.c_str());
    if (!p_rootsRegExp->valid())
        IBDIAG_RETURN(false);

    IBDIAG_LOG(TT_LOG_LEVEL_DEBUG, "FT roots regular expressioon: %s\n", val.c_str());

    IBDIAG_RETURN(true);
}

ReportCableInfoStage::ReportCableInfoStage(IBDiag *p_ibdiag):Stage(STAGE_NAME_REPORT_CABLE_INFO, p_ibdiag)
{
    IBDIAGNET_ENTER;

    this->MarkStage(false);
    IBDIAGNET_RETURN_VOID;
}

int ReportCableInfoStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::RunCheck()
{
    IBDIAGNET_ENTER;

    list_p_fabric_general_err errorsList;
    int rc =  this->GeneralReport(errorsList);
    this->AnalyzeCheckResults(errorsList,
            "Cable Report",
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);
 
    rc =  this->TemperatureErrorsReport(errorsList);
    this->AnalyzeCheckResults(errorsList,
            "Cable Temperature Report",
            rc,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings,
            false);

    this->ReleaseCableInfo();

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::TemperatureErrorsReport(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    for (map_str_pnode::iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->lastError =
                    "DB error - found null node in NodeByName map for key = " + (*nI).first;
            this->num_errors++;
            IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
        }

        for (phys_port_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort(port_num);
            if (!p_curr_port || !p_curr_port->p_combined_cable)
                continue;

            bool isDetected = false;

            int rc = this->CheckTemperature(p_curr_port->p_combined_cable->GetTemperatureAlarms(),
                                            "Temperature Alarm reported", *p_curr_port, errorsList, isDetected);
            if (rc)
                IBDIAGNET_RETURN(rc);

            if (isDetected)
                continue;

            rc = this->CheckTemperature(p_curr_port->p_combined_cable->GetTemperatureErrorsByTreshold(),
                                        "Temperature detected", *p_curr_port, errorsList, isDetected);
            if (rc)
                IBDIAGNET_RETURN(rc);
        }
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::CheckTemperature(int tempCode, const string &message, const IBPort &port,
                                            list_p_fabric_general_err &errorsList, bool &isDetected)
{
    IBDIAGNET_ENTER;

    isDetected = false;

    if (tempCode < 0) {
        this->lastError = "DB error - found CombinedCableInfo object with wrong data. Port: "
                        + port.getExtendedName();
        this->num_errors++;
                IBDIAGNET_RETURN(IBDIAG_ERR_CODE_DB_ERR);
    }

    if (tempCode & CABLE_HIGH_TEMP_MASK) {
        isDetected = true;

        stringstream ss;
        ss << "High " << message;
        errorsList.push_back(new CableTemperatureErr(&port, ss.str(),
                                            port.p_combined_cable->GetTemperatureStr(),
                                            port.p_combined_cable->GetHighTemperatureThresholdStr()));
    }

    if (tempCode & CABLE_LOW_TEMP_MASK) {
        isDetected = true;

        stringstream ss;
        ss << "Low " << message;
        errorsList.push_back(new CableTemperatureErr(&port, ss.str(),
                                            port.p_combined_cable->GetTemperatureStr(),
                                            port.p_combined_cable->GetLowTemperatureThresholdStr()));
    }

    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ReportCableInfoStage::GeneralReport(list_p_fabric_general_err &errorsList)
{
    IBDIAGNET_ENTER;

    int rc = IBDIAG_SUCCESS_CODE;
    ofstream sout;

    //todo verify the stage is activated
    this->p_csv_out->DumpStart(SECTION_COMBINED_CABLE_INFO);
    this->p_csv_out->WriteBuf(CombinedCableInfo::GetCSVHeader().c_str());

    this->InitFileStream(sout);

    if (!sout.is_open())
        IBFabric::GetSwitchLabelPortNumExplanation(sout);

    for (map_str_pnode::iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
            this->lastError =
                    "DB error - found null node in NodeByName map for key = " + (*nI).first;
            this->num_errors++;
            rc = IBDIAG_ERR_CODE_DB_ERR;
            break;
        }

        for (phys_port_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort(port_num);
            if (!p_curr_port || !p_curr_port->p_combined_cable)
                continue;

            this->ReportToCSV(*p_curr_port, *p_curr_port->p_combined_cable);

            if (sout.is_open())
                this->ReportToStream(sout, *p_curr_port, *p_curr_port->p_combined_cable);
        }
    }

    this->p_csv_out->DumpEnd(SECTION_COMBINED_CABLE_INFO);

    if (sout.is_open())
        this->p_ibdiag->CloseFile(sout);

    IBDIAGNET_RETURN(rc);
}

void ReportCableInfoStage::InitFileStream(ofstream &fileStream)
{
    IBDIAGNET_ENTER;

    if (this->p_ibdiag->OpenFile("Cables Information",
                                string(COMBINED_CABLE_DUMP_FILE), fileStream, false, true)) {
        ERR_PRINT("Failed to open Cables Information file. Reason: %s\n", p_ibdiag->GetLastError());
        this->num_errors++;
    }

    IBDIAGNET_RETURN_VOID;
}

void ReportCableInfoStage::ReportToCSV(const IBPort &port, const CombinedCableInfo &combined)
{
    IBDIAGNET_ENTER;

    stringstream sstream;

    sstream << PTR(port.p_node->guid_get())      << ','
           << PTR(port.guid_get())               << ','
           << +port.num                          << ',';

    combined.ToCSVStream(sstream);
    this->p_csv_out->WriteBuf(sstream.str());

    IBDIAGNET_RETURN_VOID;
}

void ReportCableInfoStage::ReportToStream(ostream &stream,
                                        const IBPort &port,
                                        const CombinedCableInfo &combined) const
{
    IBDIAGNET_ENTER;

    stream << "-------------------------------------------------------" << endl
           <<  "Port="      << (int)port.num                            << ' '
           <<  "Lid="       << PTR(port.base_lid, 4)                    << ' '
           <<  "GUID="      << PTR(port.guid_get())                     << ' '
           <<  "Port Name=" << port.getName()                           << endl
           << "-------------------------------------------------------" << endl;

    combined.ToFileStream(stream);
    stream << endl << endl;

    IBDIAGNET_RETURN_VOID;
}

void ReportCableInfoStage::ReleaseCableInfo()
{
    IBDIAGNET_ENTER;

    for (map_str_pnode::iterator nI = this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.begin();
            nI != this->p_ibdiag->GetDiscoverFabricPtr()->NodeByName.end(); ++nI) {
        IBNode *p_curr_node = (*nI).second;
        if (!p_curr_node) {
             ERR_PRINT("DB error - found null node in NodeByName map for key = %s\n",
                    (*nI).first.c_str());
            continue;
        }

        for (phys_port_t port_num = 1; port_num <= p_curr_node->numPorts; ++port_num) {
            IBPort *p_curr_port = p_curr_node->getPort(port_num);
            if (p_curr_port && p_curr_port->p_combined_cable) {
                delete p_curr_port->p_combined_cable;
                p_curr_port->p_combined_cable = NULL;
            }
        }
    }

    p_ibdiag->SetCableExported(false);

    IBDIAGNET_RETURN_VOID;
}

bool ReportCableInfoStage::IsSkipped()
{
    IBDIAGNET_ENTER;

    bool is_skipped = Stage::IsSkipped();
    if (!is_skipped) {
       IBDIAGNET_RETURN(false);
    }

    if (this->p_ibdiag->IsCableExported()) {
        this->MarkStageAsActive();
        IBDIAGNET_RETURN(false);
    }

    IBDIAGNET_RETURN(is_skipped);
}

ExportDataStage::ExportDataStage(IBDiag *p_ibdiag, list_p_symbols_plugin *p_loaded_plugins_list)
    : Stage(STAGE_NAME_EXPORT_DATA, p_ibdiag), p_loaded_plugins_list(p_loaded_plugins_list)
{
    IBDIAGNET_ENTER;

    this->MarkStageAsSkipped();

    IBDIAGNET_RETURN_VOID;
}

ExportDataStage::~ExportDataStage()
{
}

int ExportDataStage::Prepare()
{
    IBDIAGNET_ENTER;
    HDR_PRINT("%s\n", this->stage_name.c_str());
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ExportDataStage::RetrieveInfo()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

int ExportDataStage::RunCheck()
{
    IBDIAGNET_ENTER;

    int rc, rc2;
    list_p_fabric_general_err export_data_errors;
    export_session_handle_t session_handle;

    CLEAR_STRUCT(session_handle);

    session_handle = this->p_ibdiag->ExportOpenSession(this->p_ibdiag->curr_iteration);

    rc2 = this->p_ibdiag->ExportData(session_handle, export_data_errors);

    for (list_p_symbols_plugin::iterator it = this->p_loaded_plugins_list->begin();
            it != this->p_loaded_plugins_list->end(); ++it) {
        Plugin *p_plugin = (*it)->p_plugin_obj;
        if (!p_plugin->IsFinished())
            continue;

        rc2 |= p_plugin->ExportData(session_handle, export_data_errors);
    }

    this->p_ibdiag->ExportCloseSession(session_handle, rc2);

    rc = this->AnalyzeCheckResults(export_data_errors,
            CHECK_EXPORT_DATA,
            rc2,
            IBDIAG_ERR_CODE_FABRIC_ERROR,
            this->num_errors,
            this->num_warnings);

    IBDIAGNET_RETURN(rc);
}
