 /*
 * Copyright (c) 2021-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 * This software is available to you under the terms of the
 * OpenIB.org BSD license included below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <iomanip>
#include "infiniband/ibutils/output_stream_primitives.h"
#include "uphy_dump_engine.h"

namespace UPHY
{
    const DumpEngine::block_t * DumpEngine::add(const DataSet * dataset, const AccRegKey * key, const block_t * data)
    {
        if( dataset == nullptr )
            return nullptr;

        if( data == nullptr )
            return nullptr;

        const DataKey * x = dynamic_cast<const DataKey *>(key);

        if( x == nullptr )
            return nullptr; // TODO:: ERROR MESSAGE

        if( x->base_key() == nullptr )
            return nullptr; // TODO:: ERROR MESSAGE

        const block_t* & ref = m_db[ dataset ][ x->base_key() ][ x->address() ];

        if( ref != nullptr )
            return nullptr; // TODO:: ERROR MESSAGE

        ref = data;

        return data;
    }

    void DumpEngine::build_section_name(const DataSet * dataset, std::string & section) const
    {
        std::stringstream ss;

        ss << "UPHY_"
           << (dataset->type() == UPHY::DataSetType_DLN ? "DLN" : "CLN") << "_"
           << +dataset->version().major_version() << "_"
           << +dataset->version().minor_version()
           << "_INFO";

        section = ss.str();
    }

    void DumpEngine::to_csv(std::ostream & stream, const AccRegKey * key, const DataSet::variant_collection_t & variant) const
    {
        key->DumpKeyHeader(stream);

        for( auto & j : variant )
        {
            stream << ',' << j.reg().name();

            if( j.unit() == RegisterUnitFlags_RX )
                stream << "_RX";
            else
                if( j.unit() == RegisterUnitFlags_TX )
                    stream << "_TX";
        }

        stream << std::endl;
    }


    void DumpEngine::to_csv(std::ostream & stream, const AccRegKey * key, const block_map_t & blocks, const DataSet::variant_collection_t & variant) const
    {
        ostream_flags_saver saver(stream);

        auto    x = blocks.begin();
        size_t  i = 0;

        key->DumpKeyData(stream);

        stream << std::hex << std::setfill('0');

        bool add_comma = false;

        for( auto & z : variant )
        {
            if( add_comma )
                stream << ',';

            add_comma = true;

            while( x != blocks.end() && i == x->second->payload_size )
            {
                i = 0;
                ++x;
            }

            if( x != blocks.end() )
            {
                const peucg_page_data & entry = x->second->page_data[ i ];

                if( entry.address == z.reg().address() && entry.rxtx == ( z.unit() == RegisterUnitFlags_RX )  )
                {
                    stream << "0x" << std::setw(4) << entry.payload_data;
                    ++i;
                }
                else
                    stream << "N/A";
            }
            else
                stream << "N/A";
        }

        stream << std::endl;
    }


    void DumpEngine::to_csv(CSVOut & csv, const std::list<std::string> & dumps) const
    {
        for( auto & x : m_db )
        {
            std::string section;
            build_section_name(x.first, section);

            csv.DumpStart(section.c_str());

            //
            std::stringstream                       header;
            const DataSet::variant_collection_t &   variants = x.first->variants(dumps);

            to_csv(header, x.second.begin()->first, variants);
            csv.WriteBuf(header.str());

            //
            for( auto & z : x.second )
            {
                std::stringstream data;
                to_csv(data, z.first, z.second, variants);
                csv.WriteBuf(data.str());
            }

            csv.DumpEnd(section.c_str());
        }
    }
}


