/*
 * Copyright (c) 2004-2020 Mellanox Technologies LTD. All rights reserved
 * Copyright (c) 2021-2021 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */


#include <sys/stat.h>
#include <dirent.h>
#include <stdarg.h>
#include <string>
using namespace std;

#include <infiniband/ibdiag/ibdiagnet/ibdiagnet_plugins_ifc.h>
#include <infiniband/ibdm/OutputControl.h>

//globals
bool found_lids_errors = false;
bool found_links_errors = false;

/******************************************************/
// global handler for log file
static FILE *log_file_sout = NULL;


int contruct_log_file(const char *file_name)
{
    IBDIAGNET_ENTER;

    int rc = construct_secure_file(file_name, &log_file_sout);

    IBDIAGNET_RETURN(rc);
}

void dump_to_log_file(const char *fmt, ...)
{
    IBDIAGNET_ENTER;
    if (!log_file_sout)          //not init
        IBDIAGNET_RETURN_VOID;

    va_list args;
    va_start(args, fmt);
    vfprintf(log_file_sout, fmt, args);
    va_end(args);

    fflush(log_file_sout);
    IBDIAGNET_RETURN_VOID;
}

void destroy_log_file()
{
    IBDIAGNET_ENTER;
    if (!log_file_sout)          //not init
        IBDIAGNET_RETURN_VOID;

    fflush(log_file_sout);
    fclose(log_file_sout);
    log_file_sout = NULL;
    IBDIAGNET_RETURN_VOID;
}


/******************************************************/
/* common functions */
int copy_file(const char SRC[], const char DST[], string& errors)
{
    IBDIAGNET_ENTER;
    std::ifstream src;  // the source file
    std::ofstream dst;  // the destination file

    errors = "";
    src.open(SRC, std::ios::binary);    // open in binary to prevent jargon at the end of the buffer

    if (!src.is_open()) {
        errors += "Unable to open file ";
        errors += SRC;
        IBDIAGNET_RETURN(1);
    }

    string err_message;
    if (IBFabric::OpenFile(DST, dst, false, err_message, false, std::ios::binary)) {
        errors += "IBFabric::OpenFile failure.";
        errors += DST;
        IBDIAGNET_RETURN(1);
    }
    //dst.open(DST, std::ios::binary);    // same again, binary

    if (!dst.is_open()) {
        errors += "Unable to open file ";
        errors += DST;
        IBDIAGNET_RETURN(1);
    }

    dst << src.rdbuf ();    // copy the content
    dst.close();            // close destination file
    src.close();            // close source file
    IBDIAGNET_RETURN(0);   // file copied successfully
}


list_string get_dir_files(string dir, list_string files_types)
{
    IBDIAGNET_ENTER;
    DIR *dp;
    struct dirent *ep;
    list_string res;
    char *lastDot;

    dp = opendir (dir.c_str());
    if (dp != NULL) {
        while ((ep = readdir (dp))) {
            lastDot = strrchr(ep->d_name,'.');
            if (lastDot)
                for (list_string::iterator it = files_types.begin();
                        it != files_types.end(); ++it)
                    if (!strcmp(lastDot, (*it).c_str())) {
                        res.push_back(ep->d_name);
                        break;
                    }
        }
        closedir(dp);
    }
    return res;
}


int check_if_can_send_mads_by_lid(IN IBDiag *p_ibdiag,
        OUT bool& result)
{
    IBDIAGNET_ENTER;
    u_int8_t local_port_state;
    if (p_ibdiag->GetLocalPortState(local_port_state)) {
        ERR_PRINT("%s", p_ibdiag->GetLastError());
        IBDIAGNET_RETURN(1);
    }
    result = false;

    switch (local_port_state) {
    case IB_PORT_STATE_DOWN: {
        INFO_PRINT("The Local link is in DOWN state\n");
        IBDIAGNET_RETURN(0);
    }
    case IB_PORT_STATE_ARM: {
        INFO_PRINT("The Local link is in ARM state\n");
        IBDIAGNET_RETURN(0);
    }
    case IB_PORT_STATE_INIT: {
        INFO_PRINT("The Local link is in INIT state\n");
        IBDIAGNET_RETURN(0);
    }
    default: {
        result = true;
        break;
    }
    }

    if (found_lids_errors)
        INFO_PRINT("Lids Check failed, no response for some MADs can occurred\n");
    if (found_links_errors)
        INFO_PRINT("Links Check failed, no response for some MADs can occurred\n");
    IBDIAGNET_RETURN(0);
}

/****************************************************/
/********************class Stage*********************/
/****************************************************/

/*
 * private methods
 */

void Stage::AddGeneratedFileName(string desc, string name)
{
    IBDIAGNET_ENTER;
    char buffer[256];
    snprintf(buffer, sizeof(buffer), "-I- %-35s : %s\n", desc.c_str(), name.c_str());
    this->generated_files_name += buffer;
    IBDIAGNET_RETURN_VOID;
}

void Stage::ParseIBDMOutput(string& output)
{
    IBDIAGNET_ENTER;
    char sLine[4096];
    istringstream iss(output);
    while (iss.good()) {
        iss.getline(sLine, 4096);
        if (!strncmp(sLine, "-E-", 3))
            ++this->num_errors;
        else if (!strncmp(sLine, "-W-", 3))
            ++this->num_warnings;
    }
    IBDIAGNET_RETURN_VOID;
}

/*
 * public methods
 */

Stage::Stage(string name, IBDiag *p_ibdiag_obj) :
        stage_status(ACTIVE),
        p_ibdiag(p_ibdiag_obj),
        num_warnings(0), num_errors(0),
        stage_name(name),
        p_num_of_errors_to_screen(NULL), p_base_path(NULL), p_csv_out(NULL)
{
    IBDIAGNET_ENTER;

    CLEAR_STRUCT(m_perf_time_info);

    IBDIAGNET_RETURN_VOID;
}

u_int32_t Stage::max_spaces_fabric_summary = MAX_SPACES_FABRIC_SUMMARY_SHORT;

void Stage::MarkStage(bool is_active)
{
    if (is_active)
        this->stage_status = ACTIVE;
    else
        this->stage_status = SKIPPED;
}


string Stage::GetSummaryLine()
{
    IBDIAGNET_ENTER;
    char buffer[1024];
    string line = "";

    if (this->IsActive()) {
        snprintf(buffer, sizeof(buffer), "%-35s %-10s %-10s %s",
                 this->stage_name.c_str(),
                 "", "", "NA");
    } else if (this->IsFinished()) {
        snprintf(buffer, sizeof(buffer), "%-35s %-10d %-10d",
                 this->stage_name.c_str(),
                 this->num_warnings,
                 this->num_errors);
    } else {    //skipped
        snprintf(buffer, sizeof(buffer), "%-35s %-10s %-10s %s",
                 this->stage_name.c_str(),
                 "", "", "Skipped");
    }
    line.assign(buffer);
    IBDIAGNET_RETURN(line);
}


string Stage::GetFilesLine()
{
    IBDIAGNET_ENTER;
    IBDIAGNET_RETURN(this->generated_files_name);
}


bool Stage::GetFabricSummarySection(stringstream &summary_str)
{
    IBDIAGNET_ENTER;

    IBDIAGNET_RETURN(false);
}

int Stage::CleanResources()
{
    IBDIAGNET_ENTER;
    CLEAR_STRUCT(m_perf_time_info);
    IBDIAGNET_RETURN(IBDIAG_SUCCESS_CODE);
}

//Returns: 0 - success / 1 - failure
int Stage::AnalyzeCheckResults(list_p_fabric_general_err& errors_list,
        string check_name,
        int check_rc,
        int err_code_value_for_report,
        u_int32_t& num_errors,
        u_int32_t& num_warnings,
        bool only_warning)
{
    IBDIAGNET_ENTER;
    int rc = 0;

    if (check_rc == IBDIAG_SUCCESS_CODE && errors_list.empty()) {
        if (!p_ibdiag->GetIbisPtr()->IsFailed())
            INFO_PRINT("%s finished successfully\n", check_name.c_str());
    } else
        if (check_rc == err_code_value_for_report || check_rc == IBDIAG_SUCCESS_CODE) {
            ResetAccumulatedErrors(errors_list);

            list_p_fabric_general_err errors, warnings;
        	list_p_fabric_general_err::const_iterator it;

            for (it = errors_list.begin(); it != errors_list.end(); ++it)
                if (*it) {
                    if ((*it)->GetLevel() == EN_FABRIC_ERR_ERROR && !only_warning)
                        errors.push_back(*it);
                    else
                        warnings.push_back(*it);
                }

            // errors list is empty
            if (errors.empty())
                WARN_PRINT("%s finished with warnings\n", check_name.c_str());
            else
                ERR_PRINT("%s finished with errors\n", check_name.c_str());

            // output error to log/screen/db_csv
            if (!errors.empty()) {
            	DumpCSVFabricErrorListTable(errors, *this->p_csv_out, check_name, EN_FABRIC_ERR_ERROR);

            	u_int32_t num_err = 1;
            	bool dump_csv_only = false;
            	for (it = errors.begin(); it != errors.end(); ++it, ++num_err) {
                     if (num_err <= *this->p_num_of_errors_to_screen) {
                         ERR_PRINT("%s\n", (*it)->GetErrorLine().c_str());
                     } else {
                         if (!(*it)->GetDumpCSVOnly())
                             LOG_ERR_PRINT("%s\n", (*it)->GetErrorLine().c_str());
                         else
                             dump_csv_only = true;
                     }
                }

                if (dump_csv_only)
                    INFO_PRINT("All other errors can be found in " DB_FILE "\n");
            }

            // output warnings to log/screen/db_csv
            if (!warnings.empty()) {
                DumpCSVFabricErrorListTable(warnings, *this->p_csv_out, check_name, EN_FABRIC_ERR_WARNING);

                u_int32_t num_warn = 0;
                bool dump_csv_only = false;
                for (it = warnings.begin(); (num_warn < *this->p_num_of_errors_to_screen) &&
                         (it != warnings.end()); ++it) {

                     if ((*it)->GetDumpCSVOnly()) {
                        dump_csv_only = true;
                        continue;
                     }

                     WARN_PRINT("%s\n", (*it)->GetErrorLine().c_str());
                     num_warn++;
                }

                if (dump_csv_only || warnings.size() > *this->p_num_of_errors_to_screen)
                    INFO_PRINT("All %swarnings can be found in " DB_FILE "\n", num_warn ? "other " : "");
            }

            // update error & warnings counter
            num_errors   += (uint32_t) errors.size();
            num_warnings += (uint32_t) warnings.size();

        } else {
            ERR_PRINT("%s failed, err=%s\n", check_name.c_str(), this->GetLastError());
            ++num_errors;
            rc = 1;
        }

    if (p_ibdiag->GetIbisPtr()->IsFailed()) {
        ERR_PRINT("%s failed, MAD err=%s\n",
                check_name.c_str(), p_ibdiag->GetIbisPtr()->GetLastError());
        p_ibdiag->GetIbisPtr()->ClearLastError();
        ++num_errors;
    }

    CleanFabricErrorsList(errors_list);
    PRINT("\n");
    IBDIAGNET_RETURN(rc);
}


/****************************************************/
/*******************class Plugin*********************/
/****************************************************/

/*
 * private methods
 */

void Plugin::SetLastError(const char *fmt, ...)
{
    IBDIAGNET_ENTER;
    char buff[IBDIAG_ERR_SIZE] = {0};
    va_list args;

    va_start(args, fmt);
    vsnprintf(buff, IBDIAG_ERR_SIZE, fmt, args);
    va_end(args);
    this->last_error.assign(buff);
    if (this->last_error.length() == IBDIAG_ERR_SIZE)
        this->last_error.append("\n-E- Message truncated");
    IBDIAGNET_LOG(TT_LOG_LEVEL_ERROR, "-E- %s\n", this->last_error.c_str());
    IBDIAGNET_RETURN_VOID;
}


/*
 * public methods
 */
const char* Plugin::GetLastError()
{
    IBDIAGNET_ENTER;
    if (this->last_error != "")
        IBDIAGNET_RETURN(this->last_error.c_str());
    IBDIAGNET_RETURN("Unknown");
}

