/*
 * Copyright (c) 2022-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 *
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdio.h>

#include "ibis.h"

typedef struct pcap_hdr_s {
    u_int32_t magic_number;   /* magic number */
    u_int16_t version_major;  /* major version number */
    u_int16_t version_minor;  /* minor version number */
    int32_t  thiszone;        /* GMT to local correction */
    u_int32_t sigfigs;        /* accuracy of timestamps */
    u_int32_t snaplen;        /* max length of captured packets, in octets */
    u_int32_t network;        /* data link type */
} __attribute__((packed)) pcap_hdr_t;

typedef struct pcaprec_hdr_s {
    u_int32_t ts_sec;         /* timestamp seconds */
    u_int32_t ts_usec;        /* timestamp microseconds */
    u_int32_t incl_len;       /* number of octets of packet saved in file */
    u_int32_t orig_len;       /* actual length of packet */
} __attribute__((packed)) pcaprec_hdr_t;

typedef struct erf_hdr_s {
    u_int64_t timestamp;
    u_int8_t type;
    u_int8_t flags;
    u_int16_t be_rlen;
    u_int16_t lctr;
    u_int16_t be_wlen;
} __attribute__((packed)) erf_hdr_t;

void Ibis::SetPcapFilePath(const char *pcap_path)
{
    if (this->pcap_fp)
        fclose(this->pcap_fp);

    this->pcap_fp = NULL;

    if (construct_secure_file(pcap_path, &this->pcap_fp)) {
        IBIS_LOG(TT_LOG_LEVEL_ERROR, "Failed to open pcap file=%s\n", pcap_path);
        return;
    }

    pcap_hdr_t pcap_hdr;
    CLEAR_STRUCT(pcap_hdr);

    pcap_hdr.magic_number = 0xA1B2C3D4;
    pcap_hdr.version_major = 2;
    pcap_hdr.version_minor = 4;
    pcap_hdr.thiszone = 0;
    pcap_hdr.sigfigs = 0;
    pcap_hdr.snaplen = 0xFFFF;
    pcap_hdr.network = 197;

    fwrite(&pcap_hdr, sizeof(pcap_hdr_t), 1, this->pcap_fp);
}

void Ibis::PcapDumpMAD(bool msg_send_mad)
{
    if (!this->pcap_fp)
        return;

    void *umad = msg_send_mad ? this->p_umad_buffer_send : this->p_umad_buffer_recv;
    if (!umad)
        return;

    struct timespec ts;

    clock_gettime(CLOCK_REALTIME, &ts);

    ib_mad_addr_t *mad_addr = umad_get_mad_addr(umad);

    // Packet record header
    pcaprec_hdr_t pcaprec_hdr;
    CLEAR_STRUCT(pcaprec_hdr);

    pcaprec_hdr.ts_sec = ts.tv_sec;
    pcaprec_hdr.ts_usec = ts.tv_nsec / 1000;
    pcaprec_hdr.incl_len = sizeof(erf_hdr_t) + IB_LRH_SIZE + IB_BTH_SIZE + IB_DETH_SIZE +
        IBIS_IB_MAD_SIZE + ICRC_SIZE + 2;
    if (mad_addr->grh_present)
        pcaprec_hdr.incl_len += IB_GRH_SIZE;
    pcaprec_hdr.orig_len = pcaprec_hdr.incl_len;

    fwrite(&pcaprec_hdr, sizeof(pcaprec_hdr_t), 1, this->pcap_fp);

    // Extensible Record Format header
    erf_hdr_t erf_hdr;
    CLEAR_STRUCT(erf_hdr);

    erf_hdr.timestamp = ((u_int64_t)ts.tv_sec << 32) |
        ((((u_int64_t)ts.tv_nsec << 32) / 1000000000) & 0xFFFFFFFF);
    erf_hdr.type = 0x15;
    erf_hdr.flags = 0x4;
    erf_hdr.be_rlen = htobe16(pcaprec_hdr.incl_len);
    erf_hdr.lctr = 0;
    erf_hdr.be_wlen = htobe16(pcaprec_hdr.incl_len - sizeof(erf_hdr_t));

    fwrite(&erf_hdr, sizeof(erf_hdr_t), 1, this->pcap_fp);

    u_int8_t buff[IB_GRH_SIZE + 1];
    memset(buff, 0, IB_GRH_SIZE + 1);

    // LRH
    IB_LRH ib_lrh;
    CLEAR_STRUCT(ib_lrh);

    u_int16_t slid = 0xFFFF;
    u_int16_t dlid = 0xFFFF;

    if (this->umad_get_port_done) {
        if (mad_addr->lid != 0xFFFF) {
            if (msg_send_mad) {
                slid = this->umad_port.base_lid;
                dlid = be16toh(mad_addr->lid);
            } else {
                slid = be16toh(mad_addr->lid);
                dlid = this->umad_port.base_lid;
            }
        }
    }

    ib_lrh.DLID = dlid;
    ib_lrh.LNH = 0x2 | mad_addr->grh_present;
    ib_lrh.Sl = mad_addr->sl;
    ib_lrh.LVer = 0;
    ib_lrh.Vl = 15;
    ib_lrh.SLID = slid;
    ib_lrh.PktLen = (IB_LRH_SIZE + IB_BTH_SIZE + IB_DETH_SIZE + IBIS_IB_MAD_SIZE + ICRC_SIZE) / 4;
    if (mad_addr->grh_present)
        ib_lrh.PktLen += IB_GRH_SIZE / 4;

    IB_LRH_pack(&ib_lrh, buff);
    fwrite(buff, IB_LRH_SIZE, 1, this->pcap_fp);

    // GRH
    if (mad_addr->grh_present) {
        IB_GRH ib_grh;
        CLEAR_STRUCT(ib_grh);

        ib_grh.FlowLabel = be32toh(mad_addr->flow_label);
        ib_grh.TClass = mad_addr->traffic_class;
        ib_grh.IPVer = 6;
        ib_grh.HopLmt = mad_addr->hop_limit;
        ib_grh.NxtHdr = 0x1B;
        ib_grh.PayLen = IB_BTH_SIZE + IB_DETH_SIZE + IBIS_IB_MAD_SIZE + ICRC_SIZE;
        memset(&ib_grh.SGID, 0, sizeof(GID_Block_Element));
        memcpy(&ib_grh.DGID, mad_addr->gid, sizeof(GID_Block_Element));

        IB_GRH_pack(&ib_grh, buff);
        fwrite(buff, IB_GRH_SIZE, 1, this->pcap_fp);
    }

    // BTH
    IB_BTH ib_bth;
    CLEAR_STRUCT(ib_bth);

    u_int16_t pkey = 0;

    if (this->umad_get_port_done) {
        if (mad_addr->pkey_index < this->umad_port.pkeys_size)
            pkey = this->umad_port.pkeys[mad_addr->pkey_index];
    }

    ib_bth.P_Key.P_KeyBase = pkey;
    ib_bth.P_Key.Membership_Type = 1;
    ib_bth.TVer = 0;
    ib_bth.PadCnt = 0;
    ib_bth.MigReq = 1;
    ib_bth.SE = 0;
    ib_bth.OpCode = 0x64;
    ib_bth.DestQP = be32toh(mad_addr->qpn);
    ib_bth.PSN = 0x000000;
    ib_bth.AckReq = 0;

    IB_BTH_pack(&ib_bth, buff);
    fwrite(buff, IB_BTH_SIZE, 1, this->pcap_fp);

    // DETH
    IB_DETH ib_deth;
    CLEAR_STRUCT(ib_deth);

    ib_deth.Q_Key = be32toh(mad_addr->qkey);
    ib_deth.SrcQP = 0x000000;

    IB_DETH_pack(&ib_deth, buff);
    fwrite(buff, IB_DETH_SIZE, 1, this->pcap_fp);

    // MAD
    void *mad = umad_get_mad(umad);
    fwrite(mad, IBIS_IB_MAD_SIZE, 1, this->pcap_fp);

    // ICRC
    ICRC icrc;
    CLEAR_STRUCT(icrc);
    fwrite(&icrc, ICRC_SIZE, 1, this->pcap_fp);

    // VCRC
    VCRC vcrc;
    CLEAR_STRUCT(vcrc);
    fwrite(&vcrc.VCRC, 2, 1, this->pcap_fp);
}
